/*
    Program to align vectorised images and compute their similarity.
    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/


import java.io.File;
import java.io.FileWriter;

import profi.matching.Image;
import profi.matching.ProbabilisticAlignment;
import profi.segmentation.RegionImage;

public class Comparator {

	public static void main(String[] args) {
		setSeed(0);
		if (args.length >= 3 && args[0].equalsIgnoreCase("segment")) {
			segment(args[1], args[2]);
		} else if (args.length >= 3 && args[0].equalsIgnoreCase("prepare-fu")) {
			prepareFu(args[1], args[2]);
		} else if (args.length >= 3 && args[0].equalsIgnoreCase("prepare-york")) {
			prepareYork(args[1], args[2]);
		} else if (args.length >= 3 && args[0].equalsIgnoreCase("import_1")) {
			importData(args[1], args[2], false);
		} else if (args.length >= 3 && args[0].equalsIgnoreCase("import_n")) {
			importData(args[1], args[2], true);
		} else if (args.length >= 3 && args[0].equalsIgnoreCase("segmentandprepare")) {
			segmentAndPrepare(args[1], args[2]);
		} else if (args.length >= 4 && args[0].equalsIgnoreCase("compare")) {
			Image.Result res = computeSimilarity(
					args[1],
					args[2],
					args[3],
					parseBoolean(args.length > 4 ? args[4] : "true"),
					parseBoolean(args.length > 5 ? args[5] : "true"),
					parseBoolean(args.length > 6 ? args[6] : "false"));
			System.out.println("ccm: " + res.sim(0));
			System.out.println("cpm: " + res.sim(1));
		} else {
			if (!(args.length >= 1 && args[0].equalsIgnoreCase("-help"))) {
				System.out.println();
				System.out.println("Bad usage:\n");
				System.out.print("\"");
				for(int i = 0; i < args.length; i++) {
					System.out.print(args[i] + ((i+1 < args.length) ? " " : ""));
				}
				System.out.println("\"");
			}
			System.out.println("Type\n");
			System.out.println("segment <inputFile> <outputFile>");
			System.out.println("\tfor segmenting an image (png) and storing the shapes");
			System.out.println();
			System.out.println("prepare-fu <inputFile> <outputFile>");
			System.out.println("\tfor reading in shapes (xml) and generating the image description");
			System.out.println("\t*with* rearrangement and grouping of polylines");
			System.out.println();
			System.out.println("prepare-york <inputFile> <outputFile>");
			System.out.println("\tfor reading in shapes (xml) and generating the image description");
			System.out.println("\t*without* rearrangement and grouping of polylines");
			System.out.println();
			System.out.println("import_[1,n] <inputFile> <outputFile>");
			System.out.println("\tfor reading in shapes (xml) and generating a dummy image description");
			System.out.println("\t'1' results one complex");
			System.out.println("\t'n' results in a single complex for each polyline");
			System.out.println();
			System.out.println("segmentandprepare <inputFile> <outputFile>");
			System.out.println("\tfor segmenting an image (png),");
			System.out.println("\tgenerating and storing the image description");
			System.out.println();
			System.out.println("compare <file1> <file2> <outputFile> [<ccm>] [<cpm>] [<ppm>]");
			System.out.println("\tfor comparing two image descriptions");
			System.out.println("\toptional parameters 'ccm', 'cpm', 'ppm' may be");
			System.out.println("\t'true', 'false' or '1', '0' respectively");
			System.out.println("\tdefault is 'true' 'true' 'false'");
			System.out.println();
		}
	}
	
	/**
	 * Derandomizes the comparison by specifying a seed for the random generator.
	 * This also effects the parts of the preprocessing that use the comparison.
	 * 
	 * @param seed The seed for the (pseudo) random generator.
	 */
	public static void setSeed(long seed) {
		ProbabilisticAlignment.setSeed(seed);
	}

	/**
	 * Segments an image and stores the shapes in xml format.
	 * Input and output is via files in the local file system. 
	 * 
	 * @param inputFileName Pathname of the input image (png or what ever javax.imageio.ImageIO is able to read).
	 * @param outputFileName Pathname for output shape file.
	 */
	public static void segment(String inputFileName, String outputFileName) {
		try {
			RegionImage regImg = RegionImage.getRegionImage(inputFileName, true);
			FileWriter out = new FileWriter(outputFileName, false);
			regImg.writeData(out);				
	        out.flush();
	        out.close();
		} catch (Exception ex) {
			ex.printStackTrace();
		}	        
	}
	
	/**
	 * Creates an image description for a given shape file.
	 * Input and output is via files in the local file system.
	 * Currently not the whole specification for the shape files is supported!
	 * 
	 * @param inputFileName Pathname of the input shape file.
	 * @param outputFileName Pathname for output image description.
	 */
	public static void prepareFu(String inputFileName, String outputFileName) {
		try {
			RegionImage regImg = RegionImage.readData(new File(inputFileName));
			Image image = new Image();
			image.defineStructure(regImg);
	        FileWriter out = new FileWriter(outputFileName, false);
			image.writeData(out, -1);				
	        out.flush();
	        out.close();
		} catch (Exception ex) {
			ex.printStackTrace();
		}	        
	}

	/**
	 * Creates an image description for a given shape file without rearranging and grouping.
	 * Input and output is via files in the local file system.
	 * Currently not the whole specification for the shape files is supported!
	 * 
	 * @param inputFileName Pathname of the input shape file.
	 * @param outputFileName Pathname for output image description.
	 */
	public static void prepareYork(String inputFileName, String outputFileName) {
		try {
			RegionImage regImg = RegionImage.readData(new File(inputFileName));
			Image image = new Image();
			image.defineStructure(regImg, Image.CLASSIFICATION);
	        FileWriter out = new FileWriter(outputFileName, false);
			image.writeData(out, -1);				
	        out.flush();
	        out.close();
		} catch (Exception ex) {
			ex.printStackTrace();
		}	        
	}
	
	/**
	 * Creates a dummy image description for a given shape file.
	 * Input and output is via files in the local file system.
	 * Currently not the whole specification for the shape files is supported!
	 * 
	 * @param inputFileName Pathname of the input shape file.
	 * @param outputFileName Pathname for output image description.
	 */
	public static void importData(String inputFileName, String outputFileName, boolean one2one) {
		try {
			RegionImage regImg = RegionImage.readData(new File(inputFileName));
			Image image = new Image();
			image.defineStructure(regImg, one2one ? Image.NO_CLASSIFICATION : Image.ONE_COMPLEX);
	        FileWriter out = new FileWriter(outputFileName, false);
			image.writeData(out, -1);				
	        out.flush();
	        out.close();
		} catch (Exception ex) {
			ex.printStackTrace();
		}	        
	}

	/**
	 * Creates an image description for a given (png-)image.
	 * Input and output is via files in the local file system.
	 * The interim shape representation is not stored in a shape file.
	 * 
	 * @param inputFileName Pathname of the input image (png or what ever javax.imageio.ImageIO is able to read).
	 * @param outputFileName Pathname for output image description.
	 */
	public static void segmentAndPrepare(String inputFileName, String outputFileName) {
		try {
			RegionImage regImg = RegionImage.getRegionImage(inputFileName, true);
			Image image = new Image();
			image.defineStructure(regImg);
	        FileWriter out = new FileWriter(outputFileName, false);
			image.writeData(out, -1);				
	        out.flush();
	        out.close();
		} catch (Exception ex) {
			ex.printStackTrace();
		}	        
	}

	/**
	 * Computes the similarity for two image descriptions.
	 * 
	 * @param fileName1 Pathname of first image description.
	 * @param fileName2 Pathname of second image description.
	 * @param fileNameOut Pathname of result. If <code>fileNameOut == null </code> no result file is written.
	 * @param ccm Whether complete-complete-match is to be computed.
	 * @param cpm Whether complete-partial-match is to be computed.
	 * @param ppm Whether partial-partial-match is to be computed.
	 * @return Result of the comparison.
	 */
	public static Image.Result computeSimilarity(
			String fileName1,
			String fileName2,
			String fileNameOut,
			boolean ccm,
			boolean cpm,
			boolean ppm) {
		try {
			Image img1 = Image.readData_03(new File(fileName1));
			Image img2 = Image.readData_03(new File(fileName2));
			Image.Result res = img1.getSimilarity2(img2, ccm, cpm, ppm);
			if (fileNameOut != null) {
				FileWriter out = new FileWriter(fileNameOut, false);
				res.writeData(out);
				out.close();
			}
			return res;
		} catch (Exception ex) {
			//ex.printStackTrace();
			System.err.println("Exception: " + ex.getMessage());
		}
		return new Image.Result();		
	}
		
	/**
	 * Conversion between similarity and distance.
	 * 
	 * @param similarity The similarity.
	 * @return 1.0 - similarity.
	 */
	public static double similarity2distance(double similarity) {
		return 1.0 - similarity;
	}

	/**
	 * Conversion between similarity and distance.
	 * 
	 * @param distance The distance. 
	 * @return 1.0 - distance.
	 */
	public static double distance2similarity(double distance) {
		return 1.0 - distance;
	}
	
	private static boolean parseBoolean(String s) {
		return s.equalsIgnoreCase("true") || s.equalsIgnoreCase("1") || s.equalsIgnoreCase("y") || s.equalsIgnoreCase("yes"); 
	}
}
