// vim:set shiftwidth=4 softtabstop=4 expandtab cindent:

// :map <F2> :up<CR>:make<CR>
// :imap <F2> <Esc><F2>

/*
    This program converts (some instances of) svg files to the
    profi shape format.

    Copyright (C) 2009 Utrecht University

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Author: Geert-Jan Giezeman <geert@cs.uu.nl>
    Contact: <http://www.cs.uu.nl/>
*/

import java.io.IOException;
import java.io.FileWriter;
import org.apache.batik.dom.svg.SAXSVGDocumentFactory;
import org.apache.batik.dom.svg.SVGDOMImplementation;
import org.apache.batik.dom.util.DOMUtilities;
import org.w3c.dom.Document;
import org.apache.batik.util.XMLResourceDescriptor;
import org.w3c.dom.*;
import org.w3c.dom.svg.*;
import org.w3c.dom.svg.SVGRect;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collection;
import java.util.Arrays;
import org.apache.batik.ext.awt.geom.Cubic;

import org.apache.batik.bridge.UserAgent;
import org.apache.batik.bridge.UserAgentAdapter;
import org.apache.batik.bridge.DocumentLoader;
import org.apache.batik.bridge.BridgeContext;
import org.apache.batik.bridge.GVTBuilder;
import org.apache.batik.gvt.GraphicsNode;


class Visitor {
    public Visitor()
    {
        doc=null;
        svg=null;
    }

    public Document getDocument()
    {
        return doc;
    }

    public boolean contentWasSkipped()
    {
        return svgAttempt || symbolAttempt;
    }

    public String skippedMsg()
    {
        StringBuilder sbuf = new StringBuilder();
        if (svgAttempt) {
            sbuf.append("One or more 'svg' elements were skipped\n");
        }
        if (symbolAttempt) {
            sbuf.append("One or more 'symbol' elements were skipped\n");
        }
        return sbuf.toString();
    }

    SVGMatrix newTransform(SVGAnimatedTransformList atl, SVGMatrix mat)
    {
        SVGTransformList tl = atl.getBaseVal();
        SVGTransform tfm = tl.consolidate();
        if (tfm==null)
            return mat;
        return mat.multiply(tfm.getMatrix());
    }

    private static float flt(SVGAnimatedLength al)
    {
        return al.getBaseVal().getValue();
    }

    private Element createNewPl(float[] bbox)
    {
        assert(bbox.length==4);
        assert(bbox[0]<=bbox[2]);
        assert(bbox[1]<=bbox[3]);
        Element root = doc.getDocumentElement();
        Element node = doc.createElementNS(profiNS,"shape");
        root.appendChild(node);
        node.setAttribute("id", ""+next_id++);
        Element plNode = doc.createElementNS(profiNS,"pl");
        node.appendChild(plNode);
        Element bbNode = doc.createElementNS(profiNS,"apbb");
        node.appendChild(bbNode);
        bbNode.appendChild(doc.createTextNode(bbox[0]+" "+bbox[1]+" "+bbox[2]+" "+bbox[3]));
        return plNode;
    }

    private static float[] computeBbox(SVGPoint pts[])
    {
        float[] result = new float[4];
        if (pts.length == 0)
            return result;
        result[0] = result[2] = pts[0].getX();
        result[1] = result[3] = pts[0].getY();
        for (int i=1; i<pts.length; ++i) {
            float x = pts[i].getX();
            if (x<result[0]) {
                result[0] = x;
            } else if (x>result[2]) {
                result[2] = x;
            }
            float y = pts[i].getY();
            if (y<result[1]) {
                result[1] = y;
            } else if (y>result[3]) {
                result[3] = y;
            }
        }
        return result;
    }

    private SVGPoint[] getPoints(int n)
    {
        SVGPoint inPoints[] = new SVGPoint[n];
        for (int i=0; i<n; ++i)
            inPoints[i] = svg.createSVGPoint();
        return inPoints;
    }

    private static SVGPoint[] getTransformedPoints(
            Collection<SVGPoint> inPoints, SVGMatrix mat)
    {
        SVGPoint tfmPoints[] = new SVGPoint[inPoints.size()];
        int i=0;
        for (Iterator<SVGPoint> iter = inPoints.iterator(); iter.hasNext(); )
            tfmPoints[i++] = iter.next().matrixTransform(mat);
        return tfmPoints;
    }

    private static SVGPoint[] getTransformedPoints(SVGPointList inPoints,
            SVGMatrix mat)
    {
        int n = inPoints.getNumberOfItems();
        SVGPoint tfmPoints[] = new SVGPoint[n];
        for (int i=0; i<n; ++i)
            tfmPoints[i] = inPoints.getItem(i).matrixTransform(mat);
        return tfmPoints;
    }

    private void treatLine(SVGLineElement line, SVGMatrix mat)
    {
        if (line.hasAttribute("transform")) {
            mat = newTransform(line.getTransform(), mat);
        }
        float x1 = flt(line.getX1());
        float y1 = flt(line.getY1());
        float x2 = flt(line.getX2());
        float y2 = flt(line.getY2());
        SVGPoint inPoints[] = getPoints(2);
        inPoints[0].setX(x1);
        inPoints[0].setY(y1);
        inPoints[1].setX(x2);
        inPoints[1].setY(y2);
        SVGPoint tfmPoints[] = getTransformedPoints(Arrays.asList(inPoints),mat);
        String data = tfmPoints[0].getX()+" "+tfmPoints[0].getY()
            +" "+tfmPoints[1].getX()+" "+tfmPoints[1].getY();
        Element plNode = createNewPl(computeBbox(tfmPoints));
        plNode.setAttribute("cl", "n");
        plNode.appendChild(doc.createTextNode(data));
    }

    private void treatRect(SVGRectElement rect, SVGMatrix mat)
    {
        if (rect.hasAttribute("transform")) {
            mat = newTransform(rect.getTransform(), mat);
        }
        SVGPoint inPoints[] = getPoints(4);
        float x = flt(rect.getX());
        float y = flt(rect.getY());
        float w = flt(rect.getWidth());
        float h = flt(rect.getHeight());
        inPoints[0].setX(x);
        inPoints[0].setY(y);
        inPoints[1].setX(x+w);
        inPoints[1].setY(y);
        inPoints[2].setX(x+w);
        inPoints[2].setY(y+h);
        inPoints[3].setX(x);
        inPoints[3].setY(y+h);
        SVGPoint tfmPoints[] = getTransformedPoints(Arrays.asList(inPoints), mat);
        String data = tfmPoints[0].getX()+" "+tfmPoints[0].getY()
            +" "+tfmPoints[1].getX()+" "+tfmPoints[1].getY()
            +" "+tfmPoints[2].getX()+" "+tfmPoints[2].getY()
            +" "+tfmPoints[3].getX()+" "+tfmPoints[3].getY();
        Element plNode = createNewPl(computeBbox(tfmPoints));
        plNode.setAttribute("cl", "i");
        plNode.appendChild(doc.createTextNode(data));
    }

    private void treatPoly(SVGPointList inPoints, SVGMatrix mat,
            boolean closed)
    {
        SVGPoint tfmPoints[] = getTransformedPoints(inPoints, mat);
        StringBuilder sbuf = new StringBuilder();
        for (int i=0; i<tfmPoints.length; ++i) {
            sbuf.append(tfmPoints[i].getX()+" "+tfmPoints[i].getY()+" ");
        }
        Element plNode = createNewPl(computeBbox(tfmPoints));
        plNode.setAttribute("cl", closed?"i":"n");
        plNode.appendChild(doc.createTextNode(sbuf.toString()));
    }

    private void treatPolygon(SVGPolygonElement pgn, SVGMatrix mat)
    {
        if (pgn.hasAttribute("transform")) {
            mat = newTransform(pgn.getTransform(), mat);
        }
        treatPoly(pgn.getPoints(), mat, true);
    }

    private void treatPolyline(SVGPolylineElement pln, SVGMatrix mat)
    {
        if (pln.hasAttribute("transform")) {
            mat = newTransform(pln.getTransform(), mat);
        }
        treatPoly(pln.getPoints(), mat, false);
    }

    SVGPoint newPoint(float x, float y)
    {
        SVGPoint result = svg.createSVGPoint();
        result.setX(x);
        result.setY(y);
        return result;
    }

    SVGPoint newPoint(Point2D.Double pt)
    {
        SVGPoint result = svg.createSVGPoint();
        result.setX((float)pt.getX());
        result.setY((float)pt.getY());
        return result;
    }

    private void treatPathStretch(SVGPathSegList segs, int start, int beyond,
            SVGMatrix mat)
    {
        assert(start<beyond);
        assert(0<=start);
        assert(beyond<=segs.getNumberOfItems());
        assert(segs.getItem(start).getPathSegType() ==
                SVGPathSeg.PATHSEG_MOVETO_ABS);
        boolean closedPath = false;
        if (segs.getItem(beyond-1).getPathSegType() ==
                SVGPathSeg.PATHSEG_CLOSEPATH) {
            --beyond;
            closedPath = true;
        }
        ArrayList<SVGPoint> inPoints = new ArrayList<SVGPoint>();
        float curx, cury;
        SVGPathSegMovetoAbs startPt = (SVGPathSegMovetoAbs)segs.getItem(start);
        float curX = startPt.getX();
        float curY = startPt.getY();
        inPoints.add(newPoint(curX, curY));
        ++start;
        while (start != beyond) {
            switch(segs.getItem(start).getPathSegType()) {
                case SVGPathSeg.PATHSEG_MOVETO_ABS:
                case SVGPathSeg.PATHSEG_CLOSEPATH:
                    assert(false);
                    break;
                case SVGPathSeg.PATHSEG_LINETO_ABS:
                    SVGPathSegLinetoAbs pt = (SVGPathSegLinetoAbs)
                        segs.getItem(start);
                    curX = pt.getX(); 
                    curY =pt.getY();
                    inPoints.add(newPoint(curX, curY));
                    break;
                case SVGPathSeg.PATHSEG_CURVETO_CUBIC_ABS:
                    SVGPathSegCurvetoCubicAbs ct = (SVGPathSegCurvetoCubicAbs)
                        segs.getItem(start);
                    Cubic cub = new Cubic(curX, curY,
                            ct.getX1(), ct.getY1(),
                            ct.getX2(), ct.getY2(),
                            ct.getX(), ct.getY());
                    // TODO: decide on n based on scale, length and so on
                    int n = 10;
                    for (int i=1; i<n; ++i) {
                        inPoints.add(newPoint(cub.eval(1.0*i/n)));
                    }
                    curX = ct.getX();
                    curY = ct.getY();
                    inPoints.add(newPoint(curX, curY));
                    break;
                default:
                    assert(false);
            }
            ++start;
        }
        SVGPoint tfmPoints[] = getTransformedPoints(inPoints, mat);
        StringBuilder sbuf = new StringBuilder();
        for (int i=0; i<tfmPoints.length; ++i) {
            sbuf.append(tfmPoints[i].getX()+" "+tfmPoints[i].getY()+" ");
        }
        Element plNode = createNewPl(computeBbox(tfmPoints));
        plNode.setAttribute("cl", closedPath?"i":"n");
        plNode.appendChild(doc.createTextNode(sbuf.toString()));
    }

    private void treatPath(SVGPathElement path, SVGMatrix mat)
    {
        if (path.hasAttribute("transform")) {
            mat = newTransform(path.getTransform(), mat);
        }
        SVGPathSegList pathSegs = path.getNormalizedPathSegList();
        int n = pathSegs.getNumberOfItems();
        int start = 0;
        while (start<n) {
            int beyond = start;
            boolean mustCut = false;
            while (!mustCut && beyond<n) {
                switch(pathSegs.getItem(beyond).getPathSegType()) {
                    case SVGPathSeg.PATHSEG_MOVETO_ABS:
                        if (beyond == start)
                            ++beyond;
                        else
                            mustCut = true;
                        break;
                    case SVGPathSeg.PATHSEG_CLOSEPATH:
                        mustCut = true;
                        // fallthrough
                    case SVGPathSeg.PATHSEG_LINETO_ABS:
                    case SVGPathSeg.PATHSEG_CURVETO_CUBIC_ABS:
                        ++beyond;
                        break;
                    default:
                        assert(false);
                }
            }
            treatPathStretch(pathSegs, start, beyond, mat);
            start = beyond;
        }
    }

    private void treatSVG(SVGSVGElement svg, SVGMatrix mat)
    {
        svgAttempt = true;
        // TODO: implement this (with new viewport)
        //SVGRect vp = svg.getViewport();
        //System.out.println("Viewport: "+vp.getWidth()+"x"+vp.getHeight());
        //System.out.println("Width: "+flt(svg.getWidth())+
        //        " Height:"+flt(svg.getHeight()));
        // make new viewport
        //visitChildren(svg, mat);
    }

    private void treatG(SVGGElement g, SVGMatrix mat)
    {
        if (g.hasAttribute("transform")) {
            mat = newTransform(g.getTransform(), mat);
        }
        visitChildren(g, mat);
    }

    private void treatUse(SVGUseElement u, SVGMatrix mat)
    {
        float dx = 0.0f;
        float dy = 0.0f;
        if (u.hasAttribute("x")) {
            dx = flt(u.getX());
        }
        if (u.hasAttribute("y")) {
            dy = flt(u.getY());
        }
        mat = mat.translate(dx, dy);
        if (u.hasAttribute("transform")) {
            mat = newTransform(u.getTransform(), mat);
        }
        // getInstanceRoot is not implemented in Batik
        //SVGElementInstance elInst = u.getInstanceRoot();
        //SVGElement cor = elInst.getCorrespondingElement();
        String id = u.getHref().getBaseVal();
        Element cor=null;
        if (id.charAt(0) == '#')
            cor = svg.getElementById(id.substring(1));
        if (cor == null) {
            System.err.println("No element found with id "+id);
            return;
        }
        if (cor instanceof SVGSymbolElement) {
            symbolAttempt = true;
        } else if (cor instanceof SVGSVGElement) {
            svgAttempt = true;
        } else {
            visitElement(cor, mat);
        }
    }

    /*
    private void printElement(Element el)
    {
        System.out.println(indent+el.getNodeName());
        NamedNodeMap attributes = el.getAttributes();
        for (int i=0; i<attributes.getLength(); ++i) {
            Node attr = attributes.item(i);
            System.out.println(indent+" -"+attr.getNodeName()+
                    ": "+attr.getNodeValue());
        }
    }
    */

    private void visitChildren(Element el, SVGMatrix mat)
    {
        NodeList nodes = el.getChildNodes();
        for(int i=0; i<nodes.getLength(); ++i) {
            Node node = nodes.item(i);
            switch (node.getNodeType()) {
                case 1:
                    visitElement((Element)node, mat);
                    break;
            }
        }
    }

    private void visitElement(Element el, SVGMatrix mat)
    {
        //printElement(el);
        if (el instanceof SVGRectElement) {
            treatRect((SVGRectElement)el, mat);
        } else if (el instanceof SVGCircleElement) {
            // TODO
        } else if (el instanceof SVGEllipseElement) {
            // TODO
        } else if (el instanceof SVGLineElement) {
            treatLine((SVGLineElement)el, mat);
        } else if (el instanceof SVGPolylineElement) {
            treatPolyline((SVGPolylineElement)el, mat);
        } else if (el instanceof SVGPolygonElement) {
            treatPolygon((SVGPolygonElement)el, mat);
        } else if (el instanceof SVGPathElement) {
            treatPath((SVGPathElement)el, mat);
        } else if (el instanceof SVGSVGElement) {
            treatSVG((SVGSVGElement)el, mat);
        } else if (el instanceof SVGAnimatedPoints) {
        } else if (el instanceof SVGGElement) {
            treatG((SVGGElement)el, mat);
        } else if (el instanceof SVGUseElement) {
            treatUse((SVGUseElement)el, mat);
        } 

    }

    public void visit(SVGSVGElement inSvg)
    {
        assert inSvg != null;
        assert inSvg instanceof SVGSVGElement;
        svg = inSvg;
        SVGMatrix mat = svg.createSVGMatrix();
        int width = 600;
        int height=600;
        if (svg.hasAttribute("viewBox")) {
            SVGAnimatedRect viewBox = svg.getViewBox();
            SVGRect viewBox0 = viewBox.getBaseVal();
            mat = mat.translate(-viewBox0.getX(), -viewBox0.getY());
            float vbw = viewBox0.getWidth();
            if (vbw < 1.0f) vbw = 1.0f;
            float vbh = viewBox0.getHeight();
            if (vbh < 1.0f) vbh = 1.0f;
            if (vbw > vbh) {
                height = (int)(width*(vbh/vbw));
                if (height<1) height = 1;
                mat = mat.scale(width/vbw);
            } else {
                width = (int)(height*(vbw/vbh));
                if (width<1) width=1;
                mat = mat.scale(height/vbh);
            }
        } else {
            float vbw = 600f, vbh = 600f;
            if (svg.hasAttribute("width")) {
                vbw = flt(svg.getWidth());
            }
            if (svg.hasAttribute("height")) {
                vbh = flt(svg.getHeight());
            }
            if (vbw > vbh) {
                height = (int)(width*(vbh/vbw));
                if (height<1) height = 1;
                mat = mat.scale(width/vbw);
            } else {
                width = (int)(height*(vbw/vbh));
                if (width<1) width=1;
                mat = mat.scale(height/vbh);
            }
        }
        //SVGRect viewport = svg.getViewport();
        //System.out.println(viewport.toString());
        //System.out.println(mat.getA()+" "+mat.getB()+" "+mat.getC()+" "+mat.getD()+" "+mat.getE()+" "+mat.getF());
        next_id = 1;
        DOMImplementation impl = SVGDOMImplementation.getDOMImplementation();
        doc = impl.createDocument(profiNS, "shapes", null);
        Element root = doc.getDocumentElement();
        root.setAttribute("v-min","4");
        root.setAttribute("v-maj","1");
        root.setAttribute("xmlns",profiNS);
        Element node = doc.createElementNS(profiNS,"src-img");
        root.appendChild(node);
        node.setAttribute("width", ""+width);
        node.setAttribute("height", ""+height);
        visitChildren(svg, mat);
    }
    
    private Document doc;
    private int next_id = 1;
    static private final String profiNS = "http://www.cs.uu.nl/profi/shapes-1";
    private SVGSVGElement svg;
    private boolean svgAttempt = false;
    private boolean symbolAttempt = false;
};

/*
 *  Main class
 */

public class Svg2Profi {
    public static void main(String[] args) {
        if (args.length != 2) {
            System.err.println("Program expects two arguments: URL and outputfile");
            System.exit(1);
        }
        try {
            UserAgent userAgent = new UserAgentAdapter();
            DocumentLoader loader = new DocumentLoader(userAgent);
            BridgeContext  ctx = new BridgeContext(userAgent, loader);
            ctx.setDynamicState(BridgeContext.DYNAMIC);
            GVTBuilder builder   = new GVTBuilder();
            String parser = XMLResourceDescriptor.getXMLParserClassName();
            SAXSVGDocumentFactory f = new SAXSVGDocumentFactory(parser);
            String uri = args[0];
            SVGDocument svgDoc = f.createSVGDocument(uri);
            SVGSVGElement root = svgDoc.getRootElement();
            GraphicsNode rootGN    = builder.build(ctx, svgDoc);
            if (root!=null) {
                SVGRect vp = root.getViewport();
                //System.out.println("Viewport: "+vp.getWidth()+"x"+vp.getHeight());
                Visitor visitor = new Visitor();
                visitor.visit(root);
                FileWriter fw = new FileWriter(args[1]);
                try {
                    DOMUtilities.writeDocument(visitor.getDocument(), fw);
                    fw.write('\n');
                    if (visitor.contentWasSkipped()) {
                        System.out.println("Warning:");
                        System.out.print(visitor.skippedMsg());
                    }
                } finally {
                    fw.close();
                }
            }

        } catch (IOException ex) {
            System.err.println("Program was ended by an exception:");
            System.err.println(ex.getMessage());
            System.exit(1);
        }
    }
};
    
