/*
    This file is part of a program to align vectorised images and
    compute their similarity.
    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/

package profi.experiments;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;

import profi.matching.ProbabilisticAlignment;
import profi.matching.WeightedClustering;

import profi.model.Box;
import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;
import sun.security.action.GetLongAction;

public class MappingTest {
	
	public static class Parameters {
		
		// experiment
		public long seed;
		public double stdevRel;
		public double vertexRatio;
		public int numberOfExperiments;
		
		public int transformationType;
		
		// algorithm
		public int numberOfVotes;
		public double numberOfVotesRel;
		public int numberOfRounds;
		public int numberOfClusters;
		public double clustersThreshold;
		public double clusterRadiusRel;
		
		// result
		public double meanErrorRel;
		public double rmsErrorRel;
		public double ci95ErrorRel;
		public double ci98ErrorRel;
		public double maxErrorRel;
		
		public Parameters() {
			seed = 0;
			stdevRel = 0.0;
			vertexRatio = 1.0;
			numberOfExperiments = 1;
			
			transformationType = ProbabilisticAlignment.TRANSLATION_ROTATION_SCALE;
			
			numberOfVotes = 0;
			numberOfVotesRel = 0.0;
			numberOfRounds = 0;
			numberOfClusters = 0;		
			clustersThreshold = 0.0;	
			clusterRadiusRel = 0.0;	

			meanErrorRel = 0.0;
			rmsErrorRel = 0.0;
			ci95ErrorRel = 0.0;
			ci98ErrorRel = 0.0;
			maxErrorRel = 0.0;
		}
		
		public String toString(String prefix) {
			StringBuffer buff = new StringBuffer();
			buff.append(prefix); buff.append("Parameters:\n");
			
			buff.append(prefix); buff.append("seed=\t"); buff.append(seed); buff.append("\n");
			buff.append(prefix); buff.append("stdevRel=\t"); buff.append(stdevRel); buff.append("\n");
			buff.append(prefix); buff.append("vertexRatio=\t"); buff.append(vertexRatio); buff.append("\n");
			buff.append(prefix); buff.append("numberOfExperiments=\t"); buff.append(numberOfExperiments); buff.append("\n");
			
			buff.append(prefix); buff.append("transformationType=\t"); buff.append(ProbabilisticAlignment.TransformationType2String(transformationType)); buff.append("\n");
			
			buff.append(prefix); buff.append("numberOfVotes=\t"); buff.append(numberOfVotes); buff.append("\n");
			buff.append(prefix); buff.append("numberOfVotesRel=\t"); buff.append(numberOfVotesRel); buff.append("\n");
			buff.append(prefix); buff.append("numberOfRounds=\t"); buff.append(numberOfRounds); buff.append("\n");
			buff.append(prefix); buff.append("numberOfClusters=\t"); buff.append(numberOfClusters); buff.append("\n");
			buff.append(prefix); buff.append("clustersThreshold=\t"); buff.append(clustersThreshold); buff.append("\n");
			buff.append(prefix); buff.append("clusterRadiusRel=\t"); buff.append(clusterRadiusRel); buff.append("\n");
			return buff.toString();
		}
		
	}
	
	Random random;
		
	public MappingTest() {
		random = new Random();
	}
		
	public void testSeries(int x, int vr) {
		Parameters par = new Parameters();
		String version = "version 0002";
		
		int noex = 0;
		double stddev = 0.05;
		ShapeSet set = new ShapeSet();
		Polyline pl = new Polyline(true);
		if (x == 0) {
			noex = 100;
			pl.addPoint(new Point(0, 0));
			pl.addPoint(new Point(100, 0));
			pl.addPoint(new Point(100, 100));
			pl.addPoint(new Point(0, 100));
		} else if (x == 1) {
			noex = 100;
			pl.addPoint(new Point(0, 0));
			pl.addPoint(new Point(100, 0));
			pl.addPoint(new Point(50, 100));
			pl.addPoint(new Point(50, 50));
			pl.addPoint(new Point(0, 50));
		} else if (x == 2) {
			noex = 500;
			pl.addPoint(new Point(0, 0));
			pl.addPoint(new Point(20, 0));
			pl.addPoint(new Point(20, 10));
			pl.addPoint(new Point(30, 10));
			pl.addPoint(new Point(30, 20));
			pl.addPoint(new Point(40, 20));
			pl.addPoint(new Point(60, 0));
			pl.addPoint(new Point(80, 0));
			pl.addPoint(new Point(40, 40));
			pl.addPoint(new Point(20, 40));
			pl.addPoint(new Point(20, 30));
			pl.addPoint(new Point(10, 30));
			pl.addPoint(new Point(10, 20));
			pl.addPoint(new Point(0, 20));
		} else if (x == 3) {
			noex = 500;
			pl.addPoint(new Point(0, 0));
			pl.addPoint(new Point(20, 0));
			pl.addPoint(new Point(20, 10));
			pl.addPoint(new Point(30, 10));
			pl.addPoint(new Point(30, 20));
			pl.addPoint(new Point(40, 20));
			pl.addPoint(new Point(60, 0));
			pl.addPoint(new Point(80, 0));
			pl.addPoint(new Point(40, 40));			
			pl.addPoint(new Point(40, 60));
			pl.addPoint(new Point(35, 40));
			pl.addPoint(new Point(35, 60));
			pl.addPoint(new Point(30, 40));
			pl.addPoint(new Point(30, 60));
			pl.addPoint(new Point(25, 40));
			pl.addPoint(new Point(25, 60));			
			pl.addPoint(new Point(20, 40));
			pl.addPoint(new Point(20, 30));
			pl.addPoint(new Point(10, 30));
			pl.addPoint(new Point(10, 20));
			pl.addPoint(new Point(0, 20));
		} else if (x == 4) {
			noex = 2000;
			pl.addPoint(new Point(	100	,	0	));
			pl.addPoint(new Point(	98.48	,	17.36	));
			pl.addPoint(new Point(	93.97	,	34.2	));
			pl.addPoint(new Point(	86.6	,	50	));
			pl.addPoint(new Point(	76.6	,	64.28	));
			pl.addPoint(new Point(	64.28	,	76.6	));
			pl.addPoint(new Point(	50	,	86.6	));
			pl.addPoint(new Point(	34.2	,	93.97	));
			pl.addPoint(new Point(	17.36	,	98.48	));
			pl.addPoint(new Point(	0	,	100	));
			pl.addPoint(new Point(	-17.36	,	98.48	));
			pl.addPoint(new Point(	-34.2	,	93.97	));
			pl.addPoint(new Point(	-50	,	86.6	));
			pl.addPoint(new Point(	-64.28	,	76.6	));
			pl.addPoint(new Point(	-76.6	,	64.28	));
			pl.addPoint(new Point(	-86.6	,	50	));
			pl.addPoint(new Point(	-93.97	,	34.2	));
			pl.addPoint(new Point(	-98.48	,	17.36	));
			pl.addPoint(new Point(	0	,	0	));
			pl.addPoint(new Point(	-100	,	0	));
			pl.addPoint(new Point(	-98.48	,	-17.36	));
			pl.addPoint(new Point(	-93.97	,	-34.2	));
			pl.addPoint(new Point(	-86.6	,	-50	));
			pl.addPoint(new Point(	-76.6	,	-64.28	));
			pl.addPoint(new Point(	-64.28	,	-76.6	));
			pl.addPoint(new Point(	-50	,	-86.6	));
			pl.addPoint(new Point(	-34.2	,	-93.97	));
			pl.addPoint(new Point(	-17.36	,	-98.48	));
			pl.addPoint(new Point(	0	,	-100	));
			pl.addPoint(new Point(	17.36	,	-98.48	));
			pl.addPoint(new Point(	34.2	,	-93.97	));
			pl.addPoint(new Point(	50	,	-86.6	));
			pl.addPoint(new Point(	64.28	,	-76.6	));
			pl.addPoint(new Point(	76.6	,	-64.28	));
			pl.addPoint(new Point(	86.6	,	-50	));
			pl.addPoint(new Point(	93.97	,	-34.2	));
			pl.addPoint(new Point(	98.48	,	-17.36	));
		} else if (x == 5) {
			noex = 5000;
			pl.addPoint(new Point(	100	,	0	));
			pl.addPoint(new Point(	98.48	,	17.36	));
			pl.addPoint(new Point(	93.97	,	34.2	));
			pl.addPoint(new Point(	86.6	,	50	));
			pl.addPoint(new Point(	76.6	,	64.28	));
			pl.addPoint(new Point(	64.28	,	76.6	));
			pl.addPoint(new Point(	50	,	86.6	));
			pl.addPoint(new Point(	34.2	,	93.97	));
			pl.addPoint(new Point(	17.36	,	98.48	));
			pl.addPoint(new Point(	0	,	100	));
			pl.addPoint(new Point(	-17.36	,	98.48	));
			pl.addPoint(new Point(	-34.2	,	93.97	));
			pl.addPoint(new Point(	-50	,	86.6	));
			pl.addPoint(new Point(	-64.28	,	76.6	));
			pl.addPoint(new Point(	-76.6	,	64.28	));
			pl.addPoint(new Point(	-86.6	,	50	));
			pl.addPoint(new Point(	-93.97	,	34.2	));
			pl.addPoint(new Point(	-98.48	,	17.36	));
			pl.addPoint(new Point(	0	,	0	));
			pl.addPoint(new Point(	-100	,	0	));
			pl.addPoint(new Point(	-98.48	,	-17.36	));
			pl.addPoint(new Point(	-93.97	,	-34.2	));
			pl.addPoint(new Point(	-86.6	,	-50	));
			pl.addPoint(new Point(	-76.6	,	-64.28	));
			pl.addPoint(new Point(	-64.28	,	-76.6	));
			pl.addPoint(new Point(	-50	,	-86.6	));
			pl.addPoint(new Point(	-34.2	,	-93.97	));
			pl.addPoint(new Point(	-17.36	,	-98.48	));
			pl.addPoint(new Point(	0	,	-100	));
			pl.addPoint(new Point(	17.36	,	-98.48	));
			pl.addPoint(new Point(	34.2	,	-93.97	));
			pl.addPoint(new Point(	50	,	-86.6	));
			pl.addPoint(new Point(	64.28	,	-76.6	));
			pl.addPoint(new Point(	76.6	,	-64.28	));
			pl.addPoint(new Point(	86.6	,	-50	));
			pl.addPoint(new Point(	93.97	,	-34.2	));
			pl.addPoint(new Point(	98.48	,	-17.36	));
			set.add(pl);
			set.apply(SimilarityTransformation.scaling(0.5));
			set.apply(SimilarityTransformation.translation(35, 30));
			pl = new Polyline(true);
			pl.addPoint(new Point(0, 0));
			pl.addPoint(new Point(20, 0));
			pl.addPoint(new Point(20, 10));
			pl.addPoint(new Point(30, 10));
			pl.addPoint(new Point(30, 20));
			pl.addPoint(new Point(40, 20));
			pl.addPoint(new Point(60, 0));
			pl.addPoint(new Point(80, 0));
			pl.addPoint(new Point(40, 40));			
			pl.addPoint(new Point(40, 60));
			pl.addPoint(new Point(35, 40));
			pl.addPoint(new Point(35, 60));
			pl.addPoint(new Point(30, 40));
			pl.addPoint(new Point(30, 60));
			pl.addPoint(new Point(25, 40));
			pl.addPoint(new Point(25, 60));			
			pl.addPoint(new Point(20, 40));
			pl.addPoint(new Point(20, 30));
			pl.addPoint(new Point(10, 30));
			pl.addPoint(new Point(10, 20));
			pl.addPoint(new Point(0, 20));
		} else if (x == 6) {
			noex = 10000;
			stddev = 0.01;
			pl.addPoint(new Point(288.0, 465.0)); pl.addPoint(new Point(286.0, 378.0)); pl.addPoint(new Point(270.0, 395.0)); pl.addPoint(new Point(245.0, 403.0)); pl.addPoint(new Point(222.0, 399.0));
			pl.addPoint(new Point(202.0, 383.0)); pl.addPoint(new Point(195.0, 361.0)); pl.addPoint(new Point(196.0, 284.0)); pl.addPoint(new Point(179.0, 305.0)); pl.addPoint(new Point(157.0, 313.0)); pl.addPoint(new Point(130.0, 307.0)); 
			pl.addPoint(new Point(115.0, 294.0)); pl.addPoint(new Point(108.0, 276.0)); pl.addPoint(new Point(110.0, 195.0)); pl.addPoint(new Point(97.0, 213.0)); pl.addPoint(new Point(72.0, 224.0)); pl.addPoint(new Point(55.0, 224.0)); 
			pl.addPoint(new Point(22.0, 92.0)); pl.addPoint(new Point(73.0, 142.0)); pl.addPoint(new Point(90.0, 139.0)); pl.addPoint(new Point(110.0, 119.0)); pl.addPoint(new Point(111.0, 183.0)); pl.addPoint(new Point(161.0, 232.0)); pl.addPoint(new Point(198.0, 209.0)); pl.addPoint(new Point(198.0, 271.0)); pl.addPoint(new Point(248.0, 319.0)); pl.addPoint(new Point(264.0, 318.0)); pl.addPoint(new Point(286.0, 298.0)); pl.addPoint(new Point(286.0, 359.0)); pl.addPoint(new Point(324.0, 398.0)); pl.addPoint(new Point(348.0, 407.0)); pl.addPoint(new Point(374.0, 387.0)); pl.addPoint(new Point(368.0, 472.0));
			set.add(pl);
			pl = new Polyline(true);
			pl.addPoint(new Point(443.0, 394.0)); pl.addPoint(new Point(417.0, 401.0)); pl.addPoint(new Point(390.0, 393.0)); pl.addPoint(new Point(375.0, 375.0)); pl.addPoint(new Point(371.0, 283.0)); pl.addPoint(new Point(360.0, 299.0)); pl.addPoint(new Point(338.0, 310.0)); pl.addPoint(new Point(310.0, 308.0)); pl.addPoint(new Point(291.0, 292.0)); pl.addPoint(new Point(283.0, 194.0)); pl.addPoint(new Point(253.0, 222.0)); pl.addPoint(new Point(226.0, 221.0)); pl.addPoint(new Point(206.0, 208.0)); pl.addPoint(new Point(196.0, 189.0)); pl.addPoint(new Point(195.0, 105.0)); pl.addPoint(new Point(165.0, 133.0)); pl.addPoint(new Point(138.0, 132.0)); pl.addPoint(new Point(118.0, 118.0)); pl.addPoint(new Point(108.0, 90.0)); pl.addPoint(new Point(106.0, 16.0));
			pl.addPoint(new Point(122.0, 15.0)); pl.addPoint(new Point(162.0, 52.0)); pl.addPoint(new Point(198.0, 31.0)); pl.addPoint(new Point(196.0, 91.0)); pl.addPoint(new Point(233.0, 131.0)); pl.addPoint(new Point(256.0, 142.0)); pl.addPoint(new Point(284.0, 118.0)); pl.addPoint(new Point(284.0, 181.0)); pl.addPoint(new Point(292.0, 190.0)); pl.addPoint(new Point(336.0, 230.0)); pl.addPoint(new Point(371.0, 207.0)); pl.addPoint(new Point(372.0, 270.0)); pl.addPoint(new Point(420.0, 317.0)); pl.addPoint(new Point(438.0, 316.0)); pl.addPoint(new Point(461.0, 298.0)); pl.addPoint(new Point(459.0, 375.0));
			set.add(pl);
			pl = new Polyline(true);
			pl.addPoint(new Point(524.0, 305.0)); pl.addPoint(new Point(494.0, 308.0)); pl.addPoint(new Point(474.0, 300.0)); pl.addPoint(new Point(460.0, 280.0)); pl.addPoint(new Point(458.0, 199.0)); pl.addPoint(new Point(443.0, 214.0)); pl.addPoint(new Point(420.0, 222.0)); pl.addPoint(new Point(395.0, 218.0)); pl.addPoint(new Point(376.0, 202.0)); pl.addPoint(new Point(370.0, 179.0)); pl.addPoint(new Point(370.0, 89.0)); pl.addPoint(new Point(424.0, 139.0)); pl.addPoint(new Point(438.0, 135.0)); pl.addPoint(new Point(451.0, 119.0)); pl.addPoint(new Point(459.0, 116.0)); pl.addPoint(new Point(459.0, 179.0)); pl.addPoint(new Point(503.0, 225.0)); pl.addPoint(new Point(523.0, 227.0)); pl.addPoint(new Point(546.0, 207.0)); pl.addPoint(new Point(547.0, 273.0)); pl.addPoint(new Point(541.0, 290.0));
			set.add(pl);
			pl = new Polyline(true);
			pl.addPoint(new Point(605.0, 218.0)); pl.addPoint(new Point(578.0, 219.0)); pl.addPoint(new Point(560.0, 212.0)); pl.addPoint(new Point(547.0, 195.0)); pl.addPoint(new Point(543.0, 102.0)); pl.addPoint(new Point(528.0, 122.0)); pl.addPoint(new Point(505.0, 131.0)); pl.addPoint(new Point(480.0, 127.0)); pl.addPoint(new Point(462.0, 109.0)); pl.addPoint(new Point(456.0, 96.0)); pl.addPoint(new Point(455.0, 9.0));
			pl.addPoint(new Point(467.0, 9.0)); pl.addPoint(new Point(507.0, 48.0)); pl.addPoint(new Point(523.0, 45.0)); pl.addPoint(new Point(544.0, 25.0)); pl.addPoint(new Point(544.0, 89.0)); pl.addPoint(new Point(586.0, 133.0)); pl.addPoint(new Point(606.0, 138.0)); pl.addPoint(new Point(632.0, 116.0));
		} else {
			throw new RuntimeException();
		}
		
		set.add(pl);
		
		//set = change(set, 0.0, 2.0);				
		
		par.seed = 0;
		
		par.stdevRel = stddev;
		par.vertexRatio = vr;
		par.numberOfExperiments = 1000;
		
		par.transformationType = ProbabilisticAlignment.TRANSLATION_ROTATION_SCALE;
		
		par.numberOfVotes = noex;
		par.numberOfVotesRel = -1.0;
		par.numberOfRounds = 2;
		par.numberOfClusters = 100;
		par.clustersThreshold = 0.25;
		par.clusterRadiusRel = 0.1;

		String fileName = 
			"test_crr_" + 
			x + 
			((int) (par.stdevRel * 100)) + 
			((int) par.vertexRatio) + 
			".dat";
		
		System.out.println(fileName);

		StringBuffer buff = new StringBuffer();
		buff.append("#");
		buff.append(version);
		buff.append("\n");
		StringWriter sw = new StringWriter();
		try {
			set.writeData_03(sw, "#");
		} catch (IOException e1) {
		}
		buff.append(sw.toString());
		buff.append("#\n#numberOfPoints= ");
		buff.append(set.getNumberOfPoints());
		buff.append("\n#\n");
		buff.append(par.toString("#"));
		buff.append("#\n");
		buff.append("#\tcrr\tmeanRel\trmsRel\tci95Rel\tci98Rel\tmaxRel\n");	
		
		try {
	        FileWriter out = new FileWriter("/home/datsche/scholz/" + fileName, true);
	        out.write(buff.toString());
	        out.flush();
	        out.close();
	    } catch (IOException e) {
	    }
	    		
	    random.setSeed(par.seed);
	    ProbabilisticAlignment.setSeed(par.seed);
		for (int i = 0; i < 10; i++) {
			par.clusterRadiusRel = Math.pow(1.005, i)-1.0;
			test(set, par);

			buff.setLength(0);
			buff.append(i);
			buff.append("\t");
			buff.append(par.clusterRadiusRel);
			buff.append("\t");
			buff.append(par.meanErrorRel);
			buff.append("\t");
			buff.append(par.rmsErrorRel);
			buff.append("\t");
			buff.append(par.ci95ErrorRel);
			buff.append("\t");
			buff.append(par.ci98ErrorRel);
			buff.append("\t");
			buff.append(par.maxErrorRel);			
			System.out.println(buff.toString());
			buff.append("\n");
			
			try {
		        FileWriter out = new FileWriter("/home/datsche/scholz/" + fileName, true);
		        out.write(buff.toString());
		        out.flush();
		        out.close();
		    } catch (IOException e) {
		    }
		}
	}
	
	
	public void test(ShapeSet original, Parameters par) {
		
		ProbabilisticAlignment.mapCount = par.numberOfClusters;
		ProbabilisticAlignment.mapThreshold = par.clustersThreshold;
		WeightedClustering.clusterRadiusRel = par.clusterRadiusRel;
		
		Box box = original.getBoundingBox();
		double maxDim = box.getMaxDim();
		double stdev = maxDim * par.stdevRel;
		int size1 = (int) (original.getNumberOfPoints() * par.vertexRatio);
		Point[] points = new Point[4];
		points[0] = new Point(box.getMinX(), box.getMinY());
		points[1] = new Point(box.getMaxX(), box.getMinY());
		points[2] = new Point(box.getMaxX(), box.getMaxY());
		points[3] = new Point(box.getMinX(), box.getMaxY());
		SimilarityTransformation.setPoints(points);
				
		int    found  = 0;
		double dSum   = 0.0;
		double dSumSQ = 0.0;
		double[] ds = new double[par.numberOfExperiments];
		
		
		for (int noEx = 0; noEx < par.numberOfExperiments; noEx++) {

			SimilarityTransformation t = randomTransform(par.transformationType);

			Point p1 = new Point(points[0]);
			t.applyTo(p1);
			Point p2 = new Point(points[2]);
			t.applyTo(p2);
			double diagSQ = p1.distanceSQ(p2);
			double foundThreshold = diagSQ / 10 / 10;
			
			ShapeSet changed = change(original, stdev, 1.0);
			changed.apply(t);
			changed = change(changed, 0.0, par.vertexRatio);
			int size2 = changed.getNumberOfPoints();		 
			
			int steps = par.numberOfVotes;
			if (steps == 0) {
				steps = (int) (par.numberOfVotesRel * Math.sqrt(size1*size2));
			}
						
			Transformation[] maps = ProbabilisticAlignment.computeTransformations(
					change(original, 0.0, par.vertexRatio), changed, false, par.transformationType, par.numberOfRounds, steps, false);
						
			double minSQ = Double.MAX_VALUE;
			for (int i = 0; i < maps.length && i < par.numberOfClusters; i++) {
				SimilarityTransformation t2 = (SimilarityTransformation) maps[i];
				double distSQ = t.itemDistSQ(t2); 
				if (distSQ < minSQ) {
					minSQ = distSQ;
				}
			}
			ds[noEx] = Math.sqrt(minSQ/diagSQ);
			dSum   += ds[noEx];
			dSumSQ += minSQ/diagSQ;
			if (minSQ < foundThreshold) found++;
		}
		Arrays.sort(ds);
		par.meanErrorRel = dSum/par.numberOfExperiments;
		par.rmsErrorRel = Math.sqrt(dSumSQ/par.numberOfExperiments);
		par.ci95ErrorRel = ds[(int) ((par.numberOfExperiments-1)*0.95)];
		par.ci98ErrorRel = ds[(int) ((par.numberOfExperiments-1)*0.98)];
		par.maxErrorRel = ds[par.numberOfExperiments-1];
	}
	
	public SimilarityTransformation randomTransform(int transformationType) {
		double scale = Math.pow(2.0, random.nextGaussian()*2);
		double angle = random.nextDouble() * Math.PI * 2;
		double dx = random.nextGaussian()*100;
		double dy = random.nextGaussian()*100;		
		if (transformationType == ProbabilisticAlignment.TRANSLATION_ROTATION_SCALE) {
			return SimilarityTransformation.explicit(false, scale, angle, dx, dy);
		} else if (transformationType == ProbabilisticAlignment.TRANSLATION_ROTATION) {
			return SimilarityTransformation.explicit(false, 1.0, angle, dx, dy);
		} else if (transformationType == ProbabilisticAlignment.TRANSLATION) {
			return SimilarityTransformation.explicit(false, 1.0, 0.0, dx, dy);
		} else if (transformationType == ProbabilisticAlignment.TRANSLATION_SCALE) {
			return SimilarityTransformation.explicit(false, scale, 0.0, dx, dy);
		}		
		return null;
	}
	
	public ShapeSet change(ShapeSet shapeOriginal, double stdev, double vertexRatio) {
	
		int currentIdx = 0;
		double currentLength = 0.0;
		double nextLength = 0.0;
		double totalLength = shapeOriginal.getLength();
		
		int vertexCount = shapeOriginal.getNumberOfPoints() * (int) Math.max(0.0, vertexRatio - 1.0);
		double[] positions = new double[vertexCount];
		for (int i = 0; i < vertexCount; i++) {
			positions[i] = random.nextDouble() * totalLength;
		}
		Arrays.sort(positions);
		
		ShapeSet shapeChanged = new ShapeSet();
		Iterator plIt = shapeOriginal.getShapeIterator();
		while (plIt.hasNext()) {
			Polyline polylineOriginal = (Polyline) plIt.next();
			Polyline polylineChanged = new Polyline(polylineOriginal.isClosed());
			Iterator it = polylineOriginal.getVertexIterator();
			Point p1 = (Point) it.next();
			polylineChanged.addPoint(new Point(p1.x + random.nextGaussian()*stdev, p1.y + random.nextGaussian()*stdev));
			while (it.hasNext()) {
				Point p2 = (Point) it.next();
				nextLength += p1.distance(p2);
				while (currentIdx < vertexCount && nextLength > positions[currentIdx]) {
					Point extra = p1.intermediatePoint((positions[currentIdx]-currentLength)/(nextLength-currentLength), p2);
					polylineChanged.addPoint(new Point(extra.x + random.nextGaussian()*stdev, extra.y + random.nextGaussian()*stdev));
					currentIdx++;
				}
				if (it.hasNext() || !polylineOriginal.isClosed()) {
					Point n = new Point(p2.x + random.nextGaussian()*stdev, p2.y + random.nextGaussian()*stdev);
					polylineChanged.addPoint(n);
				}
				p1 = p2;
				currentLength = nextLength;
			}
			shapeChanged.add(polylineChanged);
		}
		return shapeChanged;
	}
	
	public Polyline randomPolyline(Box box, double length) {
		return null;
	}
	
}
