/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.util.Iterator;
import java.util.LinkedList;

import profi.model.Box;
import profi.model.Correlations;
import profi.model.Point;
import profi.model.ShapeSet;
import profi.model.Correlations.Correlation;
import profi.model.Polyline.WeightedLine;

public class AverageWeightedDistance {

	public static boolean newWeight = false;
	
	public static boolean finding = false;
		
	/**
	 * The Correlations between the two shapes.
	 */
	public static void computeCorrelations(ShapeSet set, Box bb, ShapeSet other) {
		
		double diam = bb.getMaxDim();
		
		LinkedList linesS = set.getLines();
		LinkedList linesT = other.getLines();
				
		Iterator it = linesS.iterator();
		
		while (it.hasNext()) {
			WeightedLine line = (WeightedLine) it.next();
			LinkedList lines = linesT;
			computeCorrelations(line, lines , diam);
		}
	}
	
	public static void computeCorrelations(ShapeSet set, Box bb) {
		
		double diam = bb.getMaxDim();
		
		LinkedList linesS = set.getLines();
				
		Iterator it = linesS.iterator();
		while (it.hasNext()) {
			computeCorrelations((WeightedLine) it.next(), linesS, diam, false);
		}
	}
	
	/**
	 * The Correlation between the two shapes.
	 */
	private static boolean computeCorrelations(WeightedLine l, LinkedList other, double diam, boolean extern) {

		double minCorrelation = 0.0;

		Point p1 = l.start;
		Point p2 = l.end;
		Point v = new Point(p2.x - p1.x, p2.y - p1.y);
		double length = v.distance();
		if (length < 1e-6) {
			return false;
		}
		
		Iterator it = other.iterator();
		
		LinkedList correlations = new LinkedList();
		Correlations bestSoFar = Correlations.createDummy(length);
		
		bestSoFar.maxValue = 0.0;
		bestSoFar.minValue = 0.0;
				
		while (it.hasNext()) {

			double max = -1e6;
			double min = 1.0;
			
			WeightedLine line = (WeightedLine) it.next();
			Point q1 = line.start;
			Point q2 = line.end;
						
			double v2x = q2.x - q1.x;
			double v2y = q2.y - q1.y;
			double length2sq = v2x*v2x + v2y*v2y;
			double slopeFactor =  Math.min((v.x * v2x + v.y * v2y)*(v.x * v2x + v.y * v2y) / length /length / length2sq, 1.0); // cos()^2
			slopeFactor = slopeFactor * slopeFactor; // cos()^4
			
			if (finding) {
				slopeFactor = slopeFactor * slopeFactor; // cos()^8
				//slopeFactor = slopeFactor * slopeFactor; // cos()^16
			}

					
			double lambda1 = ((q1.x - p1.x) * v.x + (q1.y - p1.y) * v.y) / v.distanceSQ();
			double lambda2 = ((q2.x - p1.x) * v.x + (q2.y - p1.y) * v.y) / v.distanceSQ();
			if (Math.abs(lambda1 - lambda2) < 1e-6 || length2sq < 1e-12 || slopeFactor < 0.05) {
				continue; // perpendicular anyways
			}
			
			if (lambda1 > lambda2) {
				Point tmpPoint = q1; q1 = q2; q2 = tmpPoint;
				double tmpLambda = lambda1; lambda1 = lambda2; lambda2 = tmpLambda;
			}
			double c1x = q1.x -(p1.x + lambda1*v.x);
			double c1y = q1.y -(p1.y + lambda1*v.y);
			double d1 = Math.max(c1x, c1y);
			double c2x = q2.x -(p1.x + lambda2*v.x);
			double c2y =  q2.y -(p1.y + lambda2*v.y);
			double d2 = Math.max(c2x, c2y);
						
			if (Correlation.computeValue(slopeFactor, d1/diam) < minCorrelation && Correlation.computeValue(slopeFactor, Math.abs(d2)/diam) < minCorrelation) {
				//System.out.print('+');
				continue;
			}
			//System.out.print('-');
			d1 = Math.sqrt(c1x*c1x+c1y*c1y);
			d2 = Math.sqrt(c2x*c2x+c2y*c2y);
			if (c1x * c2x + c1y * c2y < 0) {
				d2 *= -1.0;
			}
			
			Correlations newCorrelations = new Correlations();
			
			double d1prime = d1 - lambda1 * (d2-d1)/(lambda2-lambda1);
			double d2prime = d2 + (1-lambda2) * (d2-d1)/(lambda2-lambda1);

			if (lambda1 > 0.0) {	
				Correlation cp1 = new Correlation();
				cp1.isPoint = true;
				cp1.length = length;
				cp1.slopeFactor = slopeFactor;
				cp1.diameterRec = 1.0/diam;
				cp1.start = 0.0;
				cp1.end = Math.min(lambda1, 1.0);
				cp1.b = d1 + length*lambda1;
				cp1.m = -length;
				if (newWeight && !extern) {
					cp1.b = diam;
					cp1.m = 0;
				}
				cp1.update();
				newCorrelations.correlations.add(cp1);
				double c0s = cp1.getValue(cp1.start);
				double c0e = cp1.getValue(cp1.end);
				min = Math.min(min, c0s);
				min = Math.min(min, c0e);
				max = Math.max(max, c0s);
				max = Math.max(max, c0e);
			}

			if (lambda2 > 0.0 && lambda1 < 1.0) {
				boolean broken = false;
				if (d1prime - d2prime > 1e-6) {
					double br = d1prime / (d1prime - d2prime);
					if (br > lambda1 && br < lambda2 && br > 0.0 && br < 1.0) {
						Correlation cl1 = new Correlation();
						cl1.isPoint = false;
						cl1.length = length;
						cl1.slopeFactor = slopeFactor;
						cl1.diameterRec = 1.0/diam;
						cl1.start = Math.max(lambda1, 0.0);
						cl1.end = br;
						cl1.b = Math.abs(d1prime);
						cl1.m = -Math.abs(d2prime - d1prime);
						Correlation cl2 = new Correlation(cl1);
						cl2.start = br;
						cl2.end = Math.min(lambda2, 1.0);
						cl2.b = -cl1.b;
						cl2.m = -cl1.m;
						cl1.update();
						cl2.update();
						newCorrelations.correlations.add(cl1);
						newCorrelations.correlations.add(cl2);
						double cls = cl1.getValue(cl1.start);
						double clb = cl1.getValue(br);
						double cle = cl2.getValue(cl2.end);
						min = Math.min(min, cls);
						min = Math.min(min, clb);
						min = Math.min(min, cle);
						max = Math.max(max, cls);
						max = Math.max(max, clb);
						max = Math.max(max, cle);
						broken = true;
					}
				} 
				if (!broken) {
					Correlation cl = new Correlation();
					cl.isPoint = false;
					cl.length = length;
					cl.slopeFactor = slopeFactor;
					cl.diameterRec = 1.0/diam;
					cl.start = Math.max(lambda1, 0.0);
					cl.end = Math.min(lambda2, 1.0);
					cl.b = d1prime;
					cl.m = d2prime - d1prime;
					cl.update();
					newCorrelations.correlations.add(cl);
					double cls = cl.getValue(cl.start);
					double cle = cl.getValue(cl.end);
					min = Math.min(min, cls);
					min = Math.min(min, cle);
					max = Math.max(max, cls);
					max = Math.max(max, cle);
				}
			}

			if (lambda2 < 1.0) {
				Correlation cp2 = new Correlation();
				cp2.isPoint = true;
				cp2.length = length;
				cp2.slopeFactor = slopeFactor;
				cp2.diameterRec = 1.0/diam;
				cp2.start = Math.max(lambda2, 0.0);
				cp2.end = 1.0;
				cp2.b = Math.abs(d2) - length*lambda2;
				cp2.m = length;
				if (newWeight && !extern) {
					cp2.b = diam;
					cp2.m = 0;
				}
				cp2.update();
				newCorrelations.correlations.add(cp2);
				double c2s = cp2.getValue(cp2.start);
				double c2e = cp2.getValue(cp2.end);
				min = Math.min(min, c2s);
				min = Math.min(min, c2e);
				max = Math.max(max, c2s);
				max = Math.max(max, c2e);
			}
			newCorrelations.minValue = min;
			newCorrelations.maxValue = max;

			if (extern) {
				if (newCorrelations.maxValue < bestSoFar.minValue) {
					//nop
				} else if (newCorrelations.minValue > bestSoFar.minValue) {
					if (bestSoFar.maxValue < newCorrelations.minValue) {
						//nop
					} else {
						correlations.addFirst(bestSoFar);
					}
					bestSoFar = newCorrelations;
				} else {
					correlations.addLast(newCorrelations);
				}
			} else {
				correlations.addLast(newCorrelations);
			}
		}
		
		if (extern) {
			Iterator cIt = correlations.iterator();
			while (cIt.hasNext()) {
				bestSoFar.merge((Correlations) cIt.next());
			}
			l.externCorrelations = bestSoFar.correlations;
		} else {
			Correlations sum = Correlations.createDummy(length);
			Correlation dummy = (Correlation) sum.correlations.getFirst();
			dummy.length = length;
			Iterator cIt = correlations.iterator();
			while (cIt.hasNext()) {
				Correlations cor = (Correlations) cIt.next();
				sum.add(cor);
			}
			sum.repair();
			l.internCorrelations = sum.correlations;			
		}
		return true;
	}
	
	/**
	 * The Correlation between the two shapes.
	 */
	private static boolean computeCorrelations(WeightedLine l, LinkedList other, double diam) {

		Point p1 = l.start;
		Point p2 = l.end;
		Point v = new Point(p2.x - p1.x, p2.y - p1.y);
		double length = v.distance();
		if (length < 1e-6) {
			return false;
		}
		
		Iterator it = other.iterator();

		Correlation[] newCorrelations = new Correlation[4];
		LinkedList correlations = new LinkedList();
		Correlations bestSoFar = Correlations.createDummy(length);
		bestSoFar.maxValue = 0.0;
		bestSoFar.minValue = 0.0;
				
		while (it.hasNext()) {

			correlations.clear();
			newCorrelations[0] = newCorrelations[1] = newCorrelations[2] = newCorrelations[3] = null; 
						
			WeightedLine line = (WeightedLine) it.next();
			Point q1 = line.start;
			Point q2 = line.end;
						
			double v2x = q2.x - q1.x;
			double v2y = q2.y - q1.y;
			double length2sq = v2x*v2x + v2y*v2y;
			double slopeFactor =  Math.min((v.x * v2x + v.y * v2y)*(v.x * v2x + v.y * v2y) / length /length / length2sq, 1.0); // cos()^2
			slopeFactor = slopeFactor * slopeFactor; // cos()^4

			if (finding) {
				slopeFactor = slopeFactor * slopeFactor; // cos()^8
				//slopeFactor = slopeFactor * slopeFactor; // cos()^16
			}

					
			double lambda1 = ((q1.x - p1.x) * v.x + (q1.y - p1.y) * v.y) / v.distanceSQ();
			double lambda2 = ((q2.x - p1.x) * v.x + (q2.y - p1.y) * v.y) / v.distanceSQ();
			if (Double.isNaN(lambda1) || Double.isNaN(lambda2) || Math.abs(lambda1 - lambda2) < 1e-6 || length2sq < 1e-12 || slopeFactor < 0.05) {
				continue; // perpendicular anyways
			}
			
			if (lambda1 > lambda2) {
				Point tmpPoint = q1; q1 = q2; q2 = tmpPoint;
				double tmpLambda = lambda1; lambda1 = lambda2; lambda2 = tmpLambda;
			}
			double c1x = q1.x -(p1.x + lambda1*v.x);
			double c1y = q1.y -(p1.y + lambda1*v.y);
			double d1 = Math.max(c1x, c1y);
			double c2x = q2.x -(p1.x + lambda2*v.x);
			double c2y =  q2.y -(p1.y + lambda2*v.y);
			double d2 = Math.max(c2x, c2y);

			if (lambda1 > 1) {
				d1 = Math.max(d1, (lambda1-1)*length);
				d2 = Math.max(d2, (lambda2-1)*length);
			} else if (lambda2 < 0) {
				d1 = Math.max(d1, (-lambda1)*length);
				d2 = Math.max(d2, (-lambda2)*length);
			}
						
			if (Correlation.computeValue(slopeFactor, d1/diam) < bestSoFar.minValue && Correlation.computeValue(slopeFactor, Math.abs(d2)/diam) < bestSoFar.minValue) {
				//System.out.print('+');
				continue;
			}
			//System.out.print('-');
			d1 = Math.sqrt(c1x*c1x+c1y*c1y);
			d2 = Math.sqrt(c2x*c2x+c2y*c2y);
			if (c1x * c2x + c1y * c2y < 0) {
				d2 *= -1.0;
			}

			double d1prime = d1 - lambda1 * (d2-d1)/(lambda2-lambda1);
			double d2prime = d2 + (1-lambda2) * (d2-d1)/(lambda2-lambda1);
			
			int noC = 0;
			boolean broken = false;			
			
			if (lambda1 > 0.0) {	
				newCorrelations[noC] = new Correlation();
				newCorrelations[noC].isPoint = true;
				newCorrelations[noC].length = length;
				newCorrelations[noC].slopeFactor = slopeFactor;
				newCorrelations[noC].diameterRec = 1.0/diam;
				newCorrelations[noC].start = 0.0;
				newCorrelations[noC].end = Math.min(lambda1, 1.0);
				newCorrelations[noC].b = d1 + length*lambda1;
				newCorrelations[noC].m = -length;
				newCorrelations[noC].update();
				noC++;
			}

			if (lambda2 > 0.0 && lambda1 < 1.0) {
				if (d1prime - d2prime > 1e-6) {
					double br = d1prime / (d1prime - d2prime);
					if (br > lambda1 && br < lambda2 && br > 0.0 && br < 1.0) {
						newCorrelations[noC] = new Correlation();
						newCorrelations[noC].isPoint = false;
						newCorrelations[noC].length = length;
						newCorrelations[noC].slopeFactor = slopeFactor;
						newCorrelations[noC].diameterRec = 1.0/diam;
						newCorrelations[noC].start = Math.max(lambda1, 0.0);
						newCorrelations[noC].end = br;
						newCorrelations[noC].b = Math.abs(d1prime);
						newCorrelations[noC].m = -Math.abs(d2prime - d1prime);
						newCorrelations[noC].update();
						noC++;
						newCorrelations[noC] = new Correlation(newCorrelations[noC-1]);
						newCorrelations[noC].start = br;
						newCorrelations[noC].end = Math.min(lambda2, 1.0);
						newCorrelations[noC].b = -newCorrelations[noC].b;
						newCorrelations[noC].m = -newCorrelations[noC].m;
						newCorrelations[noC].update();
						noC++;
						broken = true;
					}
				} 
				if (!broken) {
					newCorrelations[noC] = new Correlation();
					newCorrelations[noC].isPoint = false;
					newCorrelations[noC].length = length;
					newCorrelations[noC].slopeFactor = slopeFactor;
					newCorrelations[noC].diameterRec = 1.0/diam;
					newCorrelations[noC].start = Math.max(lambda1, 0.0);
					newCorrelations[noC].end = Math.min(lambda2, 1.0);
					newCorrelations[noC].b = d1prime;
					newCorrelations[noC].m = d2prime - d1prime;
					newCorrelations[noC].update();
					noC++;
				}
			}

			if (lambda2 < 1.0) {
				newCorrelations[noC] = new Correlation();
				newCorrelations[noC].isPoint = true;
				newCorrelations[noC].length = length;
				newCorrelations[noC].slopeFactor = slopeFactor;
				newCorrelations[noC].diameterRec = 1.0/diam;
				newCorrelations[noC].start = Math.max(lambda2, 0.0);
				newCorrelations[noC].end = 1.0;
				newCorrelations[noC].b = Math.abs(d2) - length*lambda2;
				newCorrelations[noC].m = length;
				newCorrelations[noC].update();
			}
			bestSoFar.merge(newCorrelations);
		}
		l.externCorrelations = bestSoFar.correlations;
		return true;
	}	

}
