/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Writer;
import java.util.Vector;

import org.w3c.dom.Element;

import profi.model.AffineTransformation;
import profi.model.Point;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;


public class Bunch extends Figure {

	private static class Node {
		
		public Point p;
		public double w;
		
		public Node(Point p, double w) {
			this.p = p; 
			this.w = w;
		}
		
		public Node(Node n) {
			this.p = new Point(n.p); 
			this.w = n.w;
		}
	}
	
	private Vector nodes;
	private Vector layers;
	private double length;
	double[] standardVotes;
	
	public Bunch() {
		super(Figure.BUNCH);
		nodes = new Vector();
		layers = new Vector();
		this.standardVotes = new double[CLASSES];
	}
	
	public Bunch(Bunch other) {
		super(other);
		nodes = new Vector();
		for (int i = 0; i < other.nodes.size(); i++) {
			nodes.add(new Node((Node) other.nodes.get(i)));
		}
		layers = new Vector();
		for (int i = 0; i < other.layers.size(); i++) {
			layers.add(new Double(((Double) other.layers.get(i)).doubleValue()));
		}
		length = other.length;
		this.standardVotes = new double[other.standardVotes.length];
		for (int i = 0; i < standardVotes.length; i++) {
			this.standardVotes[i] = other.standardVotes[i];
		}
	}

	public void addSpecificData_01(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("BUNCH\n");
		int nn = nodes.size();
		buff.append(prefix); buff.append("nodes = "); buff.append(nn); buff.append("\n");
		for (int i = 0; i < nn; i++) {
			Node node = (Node) nodes.get(i);
			buff.append(prefix); buff.append("node.x = "); buff.append(node.p.x); buff.append("\n");
			buff.append(prefix); buff.append("node.y = "); buff.append(node.p.y); buff.append("\n");
			buff.append(prefix); buff.append("node.w = "); buff.append(node.w); buff.append("\n");			
		}
		int nl = layers.size();
		buff.append(prefix); buff.append("layers = "); buff.append(nl); buff.append("\n");
		for (int i = 0; i < nl; i++) {
			double l = ((Double) layers.get(i)).doubleValue();
			buff.append(prefix); buff.append("layer = "); buff.append(l); buff.append("\n");			
		}
		for (int i = 0; i < CLASSES; i++) {
			buff.append(prefix); buff.append("sv"); buff.append(i); buff.append(" = "); buff.append(standardVotes[i]); buff.append("\n");
		}
		writer.write(buff.toString());
	}
	
	public void addSpecificData_02(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<!-- BUNCH -->\n");
		int nn = nodes.size();
		buff.append(prefix); buff.append("<nodes>"); buff.append(nn); buff.append("</nodes>\n");
		for (int i = 0; i < nn; i++) {
			Node node = (Node) nodes.get(i);buff.append(prefix);
			buff.append("<node.x>"); buff.append(node.p.x); buff.append("</node.x>\n");
			buff.append("<node.y>"); buff.append(node.p.y); buff.append("</node.y>\n");
			buff.append("<node.w>"); buff.append(node.w); buff.append("</node.w>\n");			
		}
		int nl = layers.size();
		buff.append(prefix); buff.append("<layers>"); buff.append(nl); buff.append("</layers>\n");
		for (int i = 0; i < nl; i++) {
			double l = ((Double) layers.get(i)).doubleValue();
			buff.append(prefix); buff.append("<layer>"); buff.append(l); buff.append("</layer>\n");	
		}
		for (int i = 0; i < CLASSES; i++) {
			buff.append(prefix); buff.append("<sv number=\""); buff.append(i); buff.append("\">"); buff.append(standardVotes[i]); buff.append("</sv>\n");
		}
		writer.write(buff.toString());
	}
	
	public void readSpecificData_01(BufferedReader reader) throws IOException {
		String line;
		String[] parts;
		length = 0.0;
		
		line = reader.readLine().trim();
		parts = line.split(" = ");
		int nn = Integer.parseInt(parts[1]);
		for (int i = 0; i < nn; i++) {
			line = reader.readLine().trim();
			parts = line.split(" = ");
			double x = Double.parseDouble(parts[1]);
			line = reader.readLine().trim();
			parts = line.split(" = ");
			double y = Double.parseDouble(parts[1]);
			line = reader.readLine().trim();
			parts = line.split(" = ");
			double w = Double.parseDouble(parts[1]);
			nodes.add(new Node(new Point(x, y), w));
			if (i > 0) {
				Point pl = ((Node) nodes.get(nodes.size() - 2)).p;
				Point pc = ((Node) nodes.get(nodes.size() - 1)).p;
				length += pl.distance(pc);
			}
		}
		line = reader.readLine().trim();
		parts = line.split(" = ");
		int nl = Integer.parseInt(parts[1]);
		for (int i = 0; i < nl; i++) {
			line = reader.readLine().trim();
			parts = line.split(" = ");
			double l = Double.parseDouble(parts[1]);
			layers.add(new Double(l));
		}
		for (int i = 0; i < CLASSES; i++) {
			line = reader.readLine().trim();
			parts = line.split(" = ");
			standardVotes[i] = Double.parseDouble(parts[1]);
		}
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("FIGURE end"));
	}

	public void readSpecificData_02(BufferedReader reader) throws IOException {
		String line;
		String field;
		length = 0.0;
		
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		int nn = Integer.parseInt(field);
		for (int i = 0; i < nn; i++) {
			line = reader.readLine().trim();
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			double x = Double.parseDouble(field);
			line = reader.readLine().trim();
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			double y = Double.parseDouble(field);
			line = reader.readLine().trim();
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			double w = Double.parseDouble(field);
			nodes.add(new Node(new Point(x, y), w));
			if (i > 0) {
				Point pl = ((Node) nodes.get(nodes.size() - 2)).p;
				Point pc = ((Node) nodes.get(nodes.size() - 1)).p;
				length += pl.distance(pc);
			}
		}			
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		int nl = Integer.parseInt(field);
		for (int i = 0; i < nl; i++) {
			line = reader.readLine().trim();
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			double l = Double.parseDouble(field);
			layers.add(new Double(l));
		}
		for (int i = 0; i < CLASSES; i++) {
			line = reader.readLine().trim();
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			standardVotes[i] = Double.parseDouble(field);
		}
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("</FIGURE>"));
	}
	
	public ShapeSet getShapeSet() {
		ShapeSet set = new ShapeSet();
		//todo...
		return set;
	}

	public double getSize() {
		return length / 2;
	}

	public double getSalience() {
		return getSize();
	}

	public double getLength() {
		return length;
	}

	public Point getCenter() {
		int nn = nodes.size();
		Point p1 = ((Node) nodes.get(0)).p;
		Point p2 = ((Node) nodes.get(nn/2)).p;
		Point p3 = ((Node) nodes.get(nn-1)).p;
		return new Point((p1.x + p2.x + p3.x) / 3, (p1.y + p2.y + p3.y) / 3);
	}

	public Figure copyItem() {
		return new Bunch(this);
	}

	public boolean isCopy(Figure other) {
		if (other instanceof Bunch) return isTheSame((Bunch) other);
		else return false;
	}	

	public double sqDist(Figure o) {
		if (!(o instanceof Bunch)) {
			return Double.MAX_VALUE;
		}
		Bunch other = (Bunch) o;
		//todo...
		return 0;
	}
	
	public boolean isTheSame(Bunch other) {
		double maxRadius = Math.max(this.length, other.length) / 4;
		double maxSqDist = sqDist(other);
		return (maxSqDist < maxRadius * maxRadius / 5 / 5);
	}


	public void transfor(Transformation t) {
		if (t instanceof SimilarityTransformation) {
			apply((SimilarityTransformation) t);
		} else {
			apply((AffineTransformation) t);
		}		
	}
	
	public void apply(SimilarityTransformation sim) {
		//todo...
		this.length *= sim.getScale();
	}
	
	public void apply(AffineTransformation aff) {
		//todo...	
	}

	public double weightFactor() {
		return 1.5;
	}

	public boolean layering(Figure fig) {
		return false;
	}
	
	public void computeStandardVotes() {
		ShapeSet thisSet = getShapeSet();
		for(int i = 0; i < CLASSES; i++) {
			standardVotes[i] = StandardComparator.compare(thisSet, standardFigures[i].getShapeSet());
		}
	}
	
	public double maybeFrame() {
		return 0.0;
	}

	public double pithiness() {
		//todo...
		return 0.8;
	}

	public Figure copy() {
		return new Bunch(this);
	}

	public void addSpecificData_03(Writer writer, String prefix) throws IOException {
		// TODO Auto-generated method stub		
	}

	public String getTypeName() {
		return "bunch";
	}

	public void readSpecificData_03(Element element) {
		// TODO Auto-generated method stub		
	}

	public Object getItemSource() {
		// TODO Auto-generated method stub
		return null;
	}

	public void setItemSource(Object source) {
		// TODO Auto-generated method stub
		
	}
}
