/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Writer;
import java.util.StringTokenizer;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;

public class Circle extends Figure {

	static Circle createStandardFigure() {
		Circle circle = new Circle();
		circle.radius = 100;
		circle.center = new Point(0.0, 0.0);
		circle.angle = 0.0;
		circle.ratio = 1.0;
		return circle;
	}
	
	public Point center;
	public double radius;
	// for ellipses...
	public double angle;
	public double ratio;
	
	public Circle() {
		super(Figure.CIRCLE);
	}
	
	public Circle(Circle other) {
		super(other);
		this.center = new Point(other.center);
		this.radius = other.radius;
		this.angle = other.angle;
		this.ratio = other.ratio;
	}
	
	public Circle(Point center, double radius) {
		super(Figure.CIRCLE);
		this.center = new Point(center);
		this.radius = radius;
		this.angle = 0.0;
		this.ratio = 1.0;
	}

	public void addSpecificData_01(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("CIRCLE\n");
		buff.append(prefix); buff.append("center.x = "); buff.append(center.x); buff.append("\n");
		buff.append(prefix); buff.append("center.y = "); buff.append(center.y); buff.append("\n");
		buff.append(prefix); buff.append("radius = "); buff.append(radius); buff.append("\n");
		buff.append(prefix); buff.append("angle = "); buff.append(angle); buff.append("\n");
		buff.append(prefix); buff.append("ratio = "); buff.append(ratio); buff.append("\n");
		writer.write(buff.toString());
	}
	
	public void addSpecificData_02(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<!-- CIRCLE -->\n");
		buff.append(prefix); buff.append("<center.x>"); buff.append(center.x); buff.append("</center.x>\n");
		buff.append(prefix); buff.append("<center.y>"); buff.append(center.y); buff.append("</center.y>\n");
		buff.append(prefix); buff.append("<radius>"); buff.append(radius); buff.append("</radius>\n");
		buff.append(prefix); buff.append("<angle>"); buff.append(angle); buff.append("</angle>\n");
		buff.append(prefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		writer.write(buff.toString());
	}

	public void readSpecificData_01(BufferedReader reader) throws IOException {
		String line;
		String[] parts;
		center = new Point();
		
		line = reader.readLine().trim();
		parts = line.split(" = ");
		center.x = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		center.y = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		radius = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		angle = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		ratio = Double.parseDouble(parts[1]);
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("FIGURE end"));
	}
	
	public void readSpecificData_02(BufferedReader reader) throws IOException {
		String line;
		String field;
		center = new Point();
		
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		center.x = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		center.y = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		radius = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		angle = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		ratio = Double.parseDouble(field);
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("</FIGURE>"));
	}
	
	public ShapeSet getShapeSet() {
		ShapeSet set = new ShapeSet();
		int k = 30;
		for (int l = -1; l < layers.size(); l++) {
			double r = 1.0;
			if (l > -1) {
				r = ((Double) layers.get(l)).doubleValue();
			}
			r *= radius;
			double cos = Math.cos(angle);
			double sin = Math.sin(angle);
			Polyline poly = new Polyline(true);
			for (int i = 0; i < k; i++) {
				double x = r * Math.cos(i*Math.PI/k*2);
				double y = r * ratio * Math.sin(i*Math.PI/k*2);
				double xr = cos * x - sin * y;
				double yr = sin * x + cos * y;
				poly.addPoint(new Point(center.x + xr, center.y + yr));
			}
			set.add(poly);
		}
		return set;
	}
	
	public double getSize() {
		return radius + radius*ratio;
	}
	
	public double getSalience() {
		return getSize() * (0.5 + ratio);
	}
	
	public double getLength() {
		return radius * Math.PI * (1 + ratio);
	}
	
	public Point getCenter() {
		return center;
	}

	public Figure copy() {
		return new Circle(this);
	}

	public boolean isCopy(Figure other) {
		if (other instanceof Circle) return isTheSame((Circle) other);
		else return false;
	}

	public double sqDist(Figure o) {

		if (!(o instanceof Circle)) {
			return Double.MAX_VALUE;
		}
		Circle other = (Circle) o;
		double tcos = Math.cos(this.angle);
		double tsin = Math.sin(this.angle);
		double ocos = Math.cos(other.angle);
		double osin = Math.sin(other.angle);
		
		double maxDist = 0.0;
		for (int i = 0; i < 4; i++) {
			double a = i * Math.PI / 2;		
		
			double x1 = radius * Math.cos(a);
			double y1 = radius * ratio * Math.sin(a);
			double x1r = tcos * x1 - tsin * y1 + center.x;
			double y1r = tsin * x1 + tcos * y1 + center.y;
			
			double x2 = other.radius * Math.cos(this.angle - other.angle + a);
			double y2 = other.radius * other.ratio * Math.sin(this.angle - other.angle + a);
			double x2r = ocos * x2 - osin * y2 + other.center.x;
			double y2r = osin * x2 + ocos * y2 + other.center.y;

			double dist = (x1r - x2r)*(x1r - x2r) + (y1r - y2r)*(y1r - y2r);
			if (dist > maxDist) {
				maxDist = dist;
			}

			double x3 = other.radius * Math.cos(a);
			double y3 = other.radius * ratio * Math.sin(a);
			double x3r = ocos * x3 - osin * y3 + other.center.x;
			double y3r = osin * x3 + ocos * y3 + other.center.y;
			
			double x4 = radius * Math.cos(other.angle - this.angle + a);
			double y4 = radius * ratio * Math.sin(other.angle - this.angle + a);
			double x4r = tcos * x4 - tsin * y4 + this.center.x;
			double y4r = tsin * x4 + tcos * y4 + this.center.y;

			dist = (x3r - x4r)*(x3r - x4r) + (y3r - y4r)*(y3r - y4r);
			if (dist > maxDist) {
				maxDist = dist;
			}
		}
		return maxDist;
	}
	
	public boolean isTheSame(Circle other) {
		double minRadius = Math.min(this.radius, other.radius);
		double maxRadius = Math.max(this.radius, other.radius);
		double dr = maxRadius - minRadius;
		double centerDist = this.center.distance(other.center);
		if (dr > maxRadius/5 || centerDist > maxRadius/5) {
			return false;
		} else {
			double maxSqDist = sqDist(other);
			return (maxSqDist < maxRadius * maxRadius / 5 / 5);
		}
	}

	public void transfor(Transformation t) {
		if (t instanceof SimilarityTransformation) {
			apply((SimilarityTransformation) t);
		} else {
			System.out.println("t not supported.");
		}		
	}
	
	public void apply(SimilarityTransformation sim) {
		sim.applyTo(this.center);
		this.radius *= sim.getScale();
		this.angle += sim.getAngle();
	}

	public double weightFactor() {
		return 2.0;
	}

	public boolean layering(Figure fig) {
		if (fig instanceof Circle) {
			Circle other = (Circle) fig;
			double centerDist = this.center.distance(other.center);
			if (centerDist > other.radius/3*other.ratio || other.radius < this.radius * 0.5) {
				return false;
			}
			if (this.ratio / other.ratio > 1.5 || other.ratio / this.ratio > 1.5) {
				return false;
			}			
			int dAngle = (int) ((this.angle - other.angle) / Math.PI * 180);
			dAngle = dAngle % 180;
			dAngle += 180;
			dAngle = dAngle % 180;
			boolean angleOk = false;
			if (this.ratio > 0.9) {
				// nearly circle
				angleOk = true;				
			} else if (this.ratio > 0.75) {
				// elllise
				angleOk = dAngle < 20 || dAngle > 160;				
			} else {
				// ellipse
				angleOk = dAngle < 10 || dAngle > 170;
			}
			if (!angleOk) {
				return false;
			}
						
			double rNew = other.radius / this.radius * (1 + other.ratio / this.ratio) / 2;
			for (int l = -1; l < layers.size(); l++) {
				double r = 1.0;
				if (l > -1) {
					r = ((Double) layers.get(l)).doubleValue();
				}
				if(Math.abs(rNew-r) < 0.2) {
					return true;
				}
			}
			this.layers.add(new Double(rNew));
			return true;
		}
		return false;
	}

	public double maybeFrame() {
		return 1.0;
	}

	public double pithiness() {
		return 1-(1-ratio)*(1-ratio)*0.75;
	}

	public void addSpecificData_03(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<ellipse>\n");
		buff.append(pprefix); buff.append("<center>"); buff.append(center.x); buff.append(" "); buff.append(center.y); buff.append("</center>\n");
		buff.append(pprefix); buff.append("<radius>"); buff.append(radius); buff.append("</radius>\n");
		buff.append(pprefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		buff.append(pprefix); buff.append("<angle>"); buff.append(angle); buff.append("</angle>\n");
		buff.append(prefix); buff.append("</ellipse>\n");
		writer.write(buff.toString());		
	}

	public String getTypeName() {
		return "ellipse";
	}

	public void readSpecificData_03(Element element) {
		NodeList childNodes = element.getChildNodes();
		int nodeCount = 0;
		int stepCount = 0;
		while (stepCount <= 3) {
			Node node = childNodes.item(nodeCount);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element) node;
				switch (stepCount) {
				case 0:
					StringTokenizer tokens = new StringTokenizer(e.getTextContent());
					center = new Point(Double.parseDouble(tokens.nextToken()), Double.parseDouble(tokens.nextToken()));
					break;
				case 1:
					radius = Double.parseDouble(e.getTextContent());
					break;
				case 2:
					ratio = Double.parseDouble(e.getTextContent());
					break;
				case 3:
					angle = Double.parseDouble(e.getTextContent());
					break;
				}
				stepCount++;
			}
			nodeCount++;
		}
	}

	public Object getItemSource() {
		// TODO Auto-generated method stub
		return null;
	}

	public void setItemSource(Object source) {
		// TODO Auto-generated method stub
		
	}
}
