/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import profi.model.Box;
import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;

import Jama.Matrix;


public class CircleFinder {

	public static final int TRANSLATION = 0;
	public static final int TRANSLATION_SCALE = 1;
	public static final int TRANSLATION_ROTATION = 2;
	public static final int TRANSLATION_ROTATION_SCALE = 3;
	public static final int AFFINE = 4;
	public static final int AFFINE_DISTORTION_PENALIZED = 5;
			
	
	private static abstract class Calculator {

		protected static final double DIST_APPROX_FACTOR = (1+Math.sqrt(0.5))/2;
		
		public static class Sample {
			public Point p1, p2;
			
			public Sample(Point p1, Point p2) {
				this.p1 = p1;
				this.p2 = p2;
			}
		}
		
		protected Sample lastSample;
		protected LinkedList points;
		
		protected Calculator() {
			this.points = new LinkedList();
		}
		
		public abstract void reset();
		public abstract void consolidate();
		public abstract void add(Point p1, Point p2);
	}

	private static class Calculator01 {

		
		
		Box bb;
		LinkedList allPoints;
		
		public double abortionThreshold;

		private Point first;
		private Point last;
		
		private Point firstV, lastV;
		
		private Point c, cBest;
		private Point v, vBest;
		private double rBest;
		
		private int n, nBest;
		
		private double sumPx, sumPxBest;
		private double sumPy, sumPyBest;
		private double sumVx, sumVxBest;
		private double sumVy, sumVyBest;
		
		private double sumVTPx, sumVTPxBest;
		private double sumVTPy, sumVTPyBest;

		private int counter;
		
		Matrix m;
		Matrix b;

				
				
		public Calculator01(Box bb) {
			this.bb = bb;
			abortionThreshold = 0.05;
			allPoints = new LinkedList();
			c = new Point(); cBest = new Point();
			v = new Point();vBest = new Point();
			m = new Matrix(4, 4);
			b = new Matrix(4, 1);
			counter = 0;
		}
		
		public void reset() {
			allPoints.clear();
			n = nBest = 0;
			rBest = -1.0;
			sumPx = sumPxBest = 0.0;
			sumPy = sumPyBest = 0.0;
			sumVx = sumVxBest = 0.0;
			sumVy = sumVyBest = 0.0;
			sumVTPx = sumVTPxBest = 0.0;
			sumVTPy = sumVTPyBest = 0.0;
		}

		public void addFirst(Point p1, Point p2) {
			add(p1, p2);
			first = p1.intermediatePoint(0.5, p2);
			firstV = new Point(p2.x - p1.x, p2.y - p1.y);
			firstV.normalize();
		}

		public void add(Point p1, Point p2) {
			n++;
			Point p = p1.intermediatePoint(0.5, p2);
			Point v = new Point(p2.x - p1.x, p2.y - p1.y);
			v.normalize();
			lastV = v;
			last = p;
			allPoints.add(new Calculator.Sample(p, v));
			sumPx += p.x;
			sumPy += p.y;
			sumVx += v.x;
			sumVy += v.y;
			sumVTPx +=  v.x * p.x + v.y * p.y;
			sumVTPy += -v.y * p.x + v.x * p.y;
		}

		public void consolidate() {
			n = nBest;
			sumPx = sumPxBest;
			sumPy = sumPyBest;
			sumVx = sumVxBest;
			sumVy = sumVyBest;
			sumVTPx = sumVTPxBest;
			sumVTPy = sumVTPyBest;
			while (allPoints.size() > n) {
				allPoints.removeLast();
			}
		}
				
		public boolean computeCircle() {
			
			m.set(0, 0, n);
			m.set(1, 1, n);
			m.set(2, 2, n);
			m.set(3, 3, n);
			
			m.set(0, 2, -sumVx); m.set(0, 3,  sumVy);
			m.set(1, 2, -sumVy); m.set(1, 3, -sumVx);

			m.set(2, 0, -sumVx); m.set(2, 1, -sumVy);
			m.set(3, 0,  sumVy); m.set(3, 1, -sumVx);
			
			b.set(0, 0, sumPx);
			b.set(1, 0, sumPy);
			b.set(2, 0, -sumVTPx);
			b.set(3, 0, -sumVTPy);
			
			try  {
				Matrix x = m.solve(b);
				
				c.x = x.get(0, 0);
				c.y = x.get(1, 0);
				v.x = x.get(2, 0);
				v.y = x.get(3, 0);		
			} catch (Exception ex) {
		    	consolidate();
		    	return true;
			}
			double r  = Math.sqrt(v.x * v.x + v.y * v.y);
			double maxErr = 0.0;
			Iterator it = allPoints.iterator();
			while (it.hasNext()) {
				Calculator.Sample s = (Calculator.Sample) it.next();
				Point p = s.p1;
				double rc = Math.sqrt((c.x - p.x) * (c.x - p.x) + (c.y - p.y) * (c.y - p.y));
				maxErr = Math.max(maxErr, Math.abs(r-rc));
			}
			
			double rf = Math.sqrt((c.x - first.x) * (c.x - first.x) + (c.y - first.y) * (c.y - first.y));
		    double rl = Math.sqrt((c.x - last.x) * (c.x - last.x) + (c.y - last.y) * (c.y - last.y));
		    
		    double drf = Math.abs(rf - r)/r;
		    double drl = Math.abs(rl - r)/r;
		    double drm = Math.max(drf, drl);
		    //System.out.println(drf + " " + drl);
		    counter++;
		    //if (counter > 10) System.exit(0);
		    
		    if (maxErr/r > abortionThreshold) {
		    	consolidate();
		    } else {
		    	nBest = n;
		    	cBest.x = c.x;
		    	cBest.y = c.y;
		    	vBest.x = v.x;
		    	vBest.y = v.y;
		    	rBest = r;
		    	sumPxBest = sumPx;
		    	sumPyBest = sumPy;
		    	sumVxBest = sumVx;
		    	sumVyBest = sumVy;
		    	sumVTPxBest = sumVTPx;
		    	sumVTPyBest = sumVTPy;
		    	if (
		    			r > bb.getMaxDim()*0.75 ||
		    			cBest.x < bb.getMinX() ||
		    			cBest.x > bb.getMaxX() ||
		    			cBest.y < bb.getMinY() ||
		    			cBest.y > bb.getMaxY()) {
		    		return false;
		    	}
		    }
		    return true;
		}
		
		public Circle getCircle() {
			
			if (rBest <= 0.0) {
				return null;
			}
			
			//double eo = 0.0;
			//double er = 0.0;
			//int na = 0;
			
			double error = 0.0;
			int n = 0;
			Iterator it = allPoints.iterator();
			while (it.hasNext()) {
				Calculator.Sample s = (Calculator.Sample) it.next();
				Point p = s.p1;
				double e = Math.abs(rBest - Math.sqrt((cBest.x - p.x) * (cBest.x - p.x) + (cBest.y - p.y) * (cBest.y - p.y)));

				//double cxo = p.x + vBest.x * v.x - vBest.y * v.y;
				//double cyo = p.y + vBest.x * v.y + vBest.y * v.x;

				//double cxr = p.x + 0 * v.x - 160 * v.y;
				//double cyr = p.y + 0 * v.y + 160 * v.x;
				
				//eo += (cBest.x - cxo) * (cBest.x - cxo) + (cBest.y - cyo) * (cBest.y - cyo);
				//er += (320 - cxr) * (320 - cxr) + (290 - cyr) * (290 - cyr);
				//na++;

				if (e < abortionThreshold * rBest) {
					error += e*e;
					n++;
				}
			}
			error =  Math.sqrt(error / n) / rBest;
			
			
			//System.out.println(Math.sqrt(eo/na) + " " + Math.sqrt(er/na));
			
			Circle circle = new Circle(cBest, rBest);
			//TODO...
			circle.weightAbs = n / (1 + 5*error);
			circle.setItemWeight(n / (1 + 5*error));
			return circle;
		}
		
		public double getCenterSquaredError(Point p1, Point p2) {
			Point p = p1.intermediatePoint(0.5, p2);
			Point v = new Point(p2.x - p1.x, p2.y - p1.y);
			v.normalize();
			double cx = p.x + vBest.x * v.x - vBest.y * v.y;
			double cy = p.y + vBest.x * v.y + vBest.y * v.x;
			return ((cx - cBest.x) * (cx - cBest.x) + (cy - cBest.y) * (cy - cBest.y)) / (rBest * rBest); 
		}
	}
	
	
	
	
	private ShapeSet startShape;
	private Calculator.Sample[][] startPoints;
	int nStartPoints;
		
	Calculator01 calc;
	
	/**
	 * The clustering algorithm - as well as the clusters.
	 */
	private WeightedClustering mqlsClustering;
		
	private Box bb;

	/**
	 * Creates an algorithm object.
	 * 
	 * @param startShape 
	 * @param goalShape
	 * @param distance
	 * @param useScale Whether scaling is allowed or not.
	 */
	public CircleFinder(ShapeSet startShape) {
		bb = startShape.getBoundingBox();
		calc = new Calculator01(bb);
						
		this.startShape = startShape;
		startPoints = createArrays(startShape);

		nStartPoints = 0;
		for (int i = 0; i < startPoints.length; i++) {
			nStartPoints += startPoints[i].length;
		}
		
		this.mqlsClustering = new WeightedClustering(Figure.circle);
		
		double maxDim = startShape.getBoundingBox().getMaxDim();
		this.mqlsClustering.setDiam(maxDim / 50);
		this.mqlsClustering.setReference(Figure.circle, maxDim);
				
	}
			
	private Calculator.Sample[][] createArrays(ShapeSet lineSet) {
		List plList = ((ShapeSet) lineSet).getPlList();
		Polyline[] pls = new Polyline[plList.size()];
		Iterator plIt = plList.iterator();
		for (int i = 0; plIt.hasNext(); i++) {
			pls[i] = (Polyline) plIt.next();
		}
		Calculator.Sample[][] samples = new Calculator.Sample[pls.length][];
		for (int i = 0; i < samples.length; i++) {
			LinkedList pList = pls[i].getPoints();
			boolean pg = pls[i].isClosed();
			samples[i] = new Calculator.Sample[pList.size() - (pg ? 0 : 1)];
			Iterator pIt = pList.iterator();
			Point start = (Point) pIt.next();
			for (int j = 0; pIt.hasNext(); j++) {
				Point end = (Point) pIt.next();
				samples[i][j] = new Calculator.Sample(start, end);
				start = end;
			}
			if (pg) {
				samples[i][samples[i].length - 1] = new Calculator.Sample((Point) pList.getLast(), (Point) pList.getFirst());
			}
		}
		return samples;		
	}
	
	public void steps(int k) {
				
		//System.out.println("doing " + newSteps + " steps...");		
		
		for (int i = 0; i < k; i++) {
			doOneStep();
			break;
		}
		mqlsClustering.performClustering();
	}
	
	protected void doOneStep() {
		
		/*
		int startSeed = randomGenerator.nextInt(nStartPoints);
		int startDir  = randomGenerator.nextInt(2) * 2 - 1;
		int startPl = 0;
		for (int i = 0; i < this.startPoints.length; i++) {
			if (startSeed < this.startPoints[i].length) {
				startPl = i;
				break;
			} else {
				startSeed -= this.startPoints[i].length;
			}
		}
		int startFirstIdx = startSeed;
		if (startFirstIdx + startDir < 0 || startFirstIdx + startDir >= startPoints[startPl].length) {
			return;
		}
		*/
		/*
		startDir = 1;
		startPl = 0;
		startFirstIdx = (startPoints[startPl].length-1) * 2 / 3;
		//System.out.println(startFirstIdx);
		//startFirstIdx = randomGenerator.nextInt(startPoints[startPl].length-1);
		int k = startPoints[startPl].length;
		for (int i = 0; i < k; i++) {
		//	startPoints[startPl][i] = new Calculator.Sample(new Point(320 + 160 * Math.cos(Math.PI*2/k*i), 290 + 160 * Math.sin(Math.PI*2/k*i)),
		//			new Point(320 + 160 * Math.cos(Math.PI*2/k*(i+1)), 290 + 160 * Math.sin(Math.PI*2/k*(i+1))));
		}
		// */		
		for(int pls = 0; pls < startPoints.length; pls++) {
			for (int pos = 0; pos < startPoints[pls].length-1; pos++) {
		
				int startPl = pls;
				int startFirstIdx = pos;
				int startDir = 1;
						
				calc.reset();
				calc.addFirst(startPoints[startPl][startFirstIdx].p1, startPoints[startPl][startFirstIdx].p2);
				calc.add(startPoints[startPl][startFirstIdx + startDir].p1, startPoints[startPl][startFirstIdx + startDir].p2);
				if (startPoints[startPl].length > startFirstIdx + 2*startDir && 0 <= startFirstIdx + 2*startDir) {
					calc.add(startPoints[startPl][startFirstIdx + 2*startDir].p1, startPoints[startPl][startFirstIdx + 2*startDir].p2);
				}
				if (!calc.computeCircle()) {
					continue;
				}
												
				// going to both sides from the starting point...
				for (int pl = 0; pl < startPoints.length; pl++) {
					int plIdx = (pl + startPl) % startPoints.length; 
				
					for (int dir = 1; dir > -2; dir -= 2) {
										
						// initialisation for the travel...
						int sourceCurrent = (pl == 0) ? (dir == 1 ? startFirstIdx + 1 : startFirstIdx) : -1;
						
						boolean inside = 
							sourceCurrent + dir*startDir >= 0 && sourceCurrent + dir*startDir < startPoints[plIdx].length;
						while (inside) {
							int sourceNext = sourceCurrent + dir*startDir;
							Calculator.Sample current = startPoints[plIdx][sourceNext];
							double ce1 = calc.getCenterSquaredError(current.p1, current.p2);
							double ce2 = calc.getCenterSquaredError(current.p2, current.p1);
							if (Math.min(ce1, ce2) < 1.0) {
								if (ce1 <= ce2) {
									calc.add(current.p1, current.p2);
								} else {
									calc.add(current.p2, current.p1);
								}
								calc.computeCircle();
							}
							sourceCurrent = sourceNext;
							inside = 
								sourceCurrent + dir*startDir >= 0 && sourceCurrent + dir*startDir < startPoints[plIdx].length;
							
						} // inside
					} // dir
				} // pl
				Circle circle = calc.getCircle();
				//circle = new Circle(new Point(280, 265), 250);
		    	mqlsClustering.add(circle);
			}
		}
	}
	
	/**
	 * The maps that have <code>weight</code> at least <code>bestMap.weight * f</code>.
	 * 
	 * @param f The factor.
	 * @return The maps.
	 */
	public Figure[] getFigures(double f) {
		/*
		WeightedSimilarity[] newMaps = mqlsClustering.getMaps(1e-6);
		WeightedSimilarity[] oldMaps = refClustering.getMaps(1e-6);
		WeightedSimilarity[] tmpMaps = new WeightedSimilarity[Math.min(10, Math.min(newMaps.length, oldMaps.length)*2)];
		for (int i = 0; i < tmpMaps.length / 2; i++) {
			tmpMaps[i] = newMaps[i];
			tmpMaps[i + (tmpMaps.length / 2)] = oldMaps[i];
		}
		return tmpMaps;
		*/
		LinkedList li = mqlsClustering.getItems(f, 50, false); 
		Iterator it = li.iterator();
		Figure[] ret = new Figure[li.size()];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = (Figure) it.next();
		}		
		return ret;
	}
	
	public int getTotalMaps() {
		return mqlsClustering.numberOfPoints();
	}

	public synchronized void reset() {
		mqlsClustering.reset();
	}
	
	// getters and setters...
			
	public void setClusterDiam(double d) {
		if(mqlsClustering.getClusterDiam() != d) {
			reset();
		}
		mqlsClustering.setDiam(d);
	}

	public double getClusterDiam() {
		return mqlsClustering.getClusterDiam();
	}
}
