/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;


public class Complex extends Figure {

	private ShapeSet shapeSet;
	public Polyline ch;
	public Point center;
	private double radius;
	private double ratio;
	double[] standardVotes;
		
	public Complex() {
		super(Figure.COMPLEX);
		shapeSet = new ShapeSet();
		this.ratio = 1.0;
		this.standardVotes = new double[CLASSES];
	}

	public Complex(ShapeSet shapeSet) {
		super(Figure.COMPLEX);
		this.shapeSet = new ShapeSet(shapeSet);
		this.ch = Convex.convexHull(shapeSet);
		this.center = this.ch.getBoundingBox().getCenter();
		this.radius = this.ch.getDiameter() / 2;
		this.ratio = new DeformationFinder(this.ch).getRatio();
		this.standardVotes = new double[CLASSES];
	}
	
	public Complex(Complex other) {
		super(other);
		this.shapeSet = new ShapeSet(other.shapeSet);
		this.ch = new Polyline(other.ch);
		this.center = new Point(other.center);
		this.radius = other.radius;
		this.ratio = other.ratio;
		this.standardVotes = new double[other.standardVotes.length];
		for (int i = 0; i < standardVotes.length; i++) {
			this.standardVotes[i] = other.standardVotes[i];
		}
	}

	public void addSpecificData_01(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("COMPLEX\n");
		buff.append(prefix); buff.append("ch = "); buff.append(toString(ch)); buff.append("\n");
		buff.append(prefix); buff.append("center.x = "); buff.append(center.x); buff.append("\n");
		buff.append(prefix); buff.append("center.y = "); buff.append(center.y); buff.append("\n");
		buff.append(prefix); buff.append("radius = "); buff.append(radius); buff.append("\n");
		for (int i = 0; i < CLASSES; i++) {
			buff.append(prefix); buff.append("sv"); buff.append(i); buff.append(" = "); buff.append(standardVotes[i]); buff.append("\n");
		}
		
		Iterator it = shapeSet.getShapeIterator();
		while (it.hasNext()) {
			Polyline pl = (Polyline) it.next();		
			buff.append(prefix); buff.append("polyline = "); buff.append(toString(pl)); buff.append("\n");
		}
		writer.write(buff.toString());
	}

	public void addSpecificData_02(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<!-- COMPLEX -->\n");
		buff.append(prefix); buff.append(toXml(ch, true, false)); buff.append("\n");
		buff.append(prefix); buff.append("<center.x>"); buff.append(center.x); buff.append("</center.x>\n");
		buff.append(prefix); buff.append("<center.y>"); buff.append(center.y); buff.append("</center.y>\n");
		buff.append(prefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		buff.append(prefix); buff.append("<radius>"); buff.append(radius); buff.append("</radius>\n");
		for (int i = 0; i < CLASSES; i++) {
			buff.append(prefix); buff.append("<sv number=\""); buff.append(i); buff.append("\">"); buff.append(standardVotes[i]); buff.append("</sv>\n");
		}
		
		Iterator it = shapeSet.getShapeIterator();
		while (it.hasNext()) {
			Polyline pl = (Polyline) it.next();		
			buff.append(prefix); buff.append(toXml(pl, false, false)); buff.append("\n");
		}
		writer.write(buff.toString());
	}
	
	public void readSpecificData_01(BufferedReader reader) throws IOException {
		String line;
		String[] parts;
		
		line = reader.readLine().trim();
		parts = line.split(" = ");
		ch = (Polyline) toPolyLine(parts[1]);
		
		center = new Point();		
		line = reader.readLine().trim();
		parts = line.split(" = ");
		center.x = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		center.y = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		radius = Double.parseDouble(parts[1]);
		for (int i = 0; i < CLASSES; i++) {
			line = reader.readLine().trim();
			parts = line.split(" = ");
			standardVotes[i] = Double.parseDouble(parts[1]);
		}
		
		boolean end = false;
		do {
			line = reader.readLine().trim();
			end = line.equalsIgnoreCase("FIGURE end");
			if (!end) {
				parts = line.split(" = ");
				shapeSet.add(toPolyLine(parts[1]));
			}
		} while (!end);
		shapeSet.updateLength();
	}

	public void readSpecificData_02(BufferedReader reader) throws IOException {
		String line;
		String field;
		
		line = reader.readLine().trim();
		ch = readXml(line);
		
		center = new Point();		
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		center.x = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		center.y = Double.parseDouble(field);
		line = reader.readLine().trim();
		if (line.indexOf("ratio") != -1) {
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			ratio = Double.parseDouble(field);
			line = reader.readLine().trim();
		}
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		radius = Double.parseDouble(field);
		for (int i = 0; i < CLASSES; i++) {
			line = reader.readLine().trim();
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			standardVotes[i] = Double.parseDouble(field);
		}
		
		boolean end = false;
		do {
			line = reader.readLine().trim();
			end = line.equalsIgnoreCase("</FIGURE>");
			if (!end) {
				shapeSet.add(readXml(line));
			}
		} while (!end);
		shapeSet.updateLength();
	}
	
	public ShapeSet getShapeSet() {
		return shapeSet;
	}

	public double getSize() {
		return radius + radius*ratio;
	}

	public double getSalience() {
		return getSize() * (0.25 + 0.75*ratio);
	}

	public double getLength() {
		return Math.min(ch.getLength(), shapeSet.getLength());
	}

	public Point getCenter() {
		return center;
	}

	public Figure copy() {
		return new Complex(this);
	}

	public boolean isCopy(Figure other) {
		if (other instanceof Complex) return false;
		else return false;
	}	
	
	public void transform(SimilarityTransformation sim) {
		//TODO...
		//sim.applyTo(this.center);
		//this.radius *= sim.getScale();
	}

	public double sqDist(Figure other) {
		// TODO Auto-generated method stub
		return Double.MAX_VALUE;
	}

	public void transfor(Transformation t) {
		// TODO Auto-generated method stub
		
	}

	public void computeStandardVotes() {
		ShapeSet thisSet = getShapeSet();
		for(int i = 0; i < CLASSES; i++) {
			standardVotes[i] = StandardComparator.compare(thisSet, standardFigures[i].getShapeSet());
		}
	}

	public boolean layering(Figure fig) {
		return false;
	}

	public double maybeFrame() {
		return 0;
	}

	public double pithiness() {
		double sym = isSymmetric();
		double rot = isRotating();
		sym = sym*sym;
		rot = rot*rot;
		return (0.5 + 0.1*sym + 0.1*rot)-(1-ratio)*(1-ratio)*0.2;
	}

	public void addSpecificData_03(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<complex>\n");
		writer.write(buff.toString());
		shapeSet.writeData_03(writer, pprefix);
		ch.writeData_03(writer, pprefix, "convexHull");
		buff.setLength(0);
		buff.append(pprefix); buff.append("<radius>"); buff.append(radius); buff.append("</radius>\n");
		buff.append(pprefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		buff.append(pprefix);
		buff.append("<ratings>");
		for (int i = 0; i < CLASSES; i++) {
			buff.append(standardVotes[i]);
			buff.append(" ");
		}
		buff.deleteCharAt(buff.length()-1);
		buff.append("</ratings>\n");
		buff.append(prefix); buff.append("</complex>\n");
		writer.write(buff.toString());		
	}

	public String getTypeName() {
		return "complex";
	}

	public void readSpecificData_03(Element element) {
		// shape, convexHull, radius, ratio, ratings
		NodeList childNodes = element.getChildNodes();
		int nodeCount = 0;
		int stepCount = 0;
		while (stepCount <= 4) {
			Node node = childNodes.item(nodeCount);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element) node;
				switch (stepCount) {
				case 0:
					shapeSet = ShapeSet.readData_03(e);
					break;
				case 1:
					ch = Polyline.readData_03(e);
					break;
				case 2:
					radius = Double.parseDouble(e.getTextContent());
					break;
				case 3:
					ratio = Double.parseDouble(e.getTextContent());
					break;
				case 4:
					StringTokenizer tokens = new StringTokenizer(e.getTextContent());
					for (int i = 0; i < Figure.CLASSES; i++) {
						standardVotes[i] = Double.parseDouble(tokens.nextToken());
					}
					break;
				}
				stepCount++;
			}
			nodeCount++;
		}
	}

	public Object getItemSource() {
		// TODO Auto-generated method stub
		return null;
	}

	public void setItemSource(Object source) {
		// TODO Auto-generated method stub
		
	}
}
