/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.StringTokenizer;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import profi.model.Box;
import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;


public class Convex extends Figure {

	private static class PointComparator implements Comparator {

		public int compare(Object o1, Object o2) {
			Point p1 = (Point) o1;
			Point p2 = (Point) o2;
			if (p1.x < p2.x) return -1;
			else if (p1.x > p2.x) return 1;
			else if (p1.y < p2.y) return -1;
			else if (p1.y > p2.y) return 1;
			else return 0;
		}		
	}
	
	private static Comparator comparator = new PointComparator();

	/**
	 * Whether p0, p1 and p2 perform a left turn or not.
	 */
	public static boolean isLeft(Point p0, Point p1, Point p2) {
	    return ( (p1.x - p0.x) * (p2.y - p0.y) - (p2.x - p0.x) * (p1.y - p0.y) ) > 0.0;
	}

	/**
	 * The convex hull of a set of shapes in counterclockwise order.
	 */
	public static Polyline convexHull(ShapeSet s) {
		Polyline ch = new Polyline(true);
		
		LinkedList list = new LinkedList();
		Iterator plIt = s.getShapeIterator();
		while (plIt.hasNext()) {
			Polyline pl = (Polyline) plIt.next();
			list.addAll(pl.getPoints());
		}

		Iterator it = list.iterator();
		
		Point[] points = new Point[list.size()];
		int i = 0;
		while (it.hasNext()) {
			points[i] = (Point) it.next();
			i++;
		}
		Arrays.sort(points, comparator);
		
		list.clear();
		Point p0 = points[0];
		list.add(p0);
		for (i = 1; i < points.length; i++) {
			if (points[i].x != p0.x || points[i].y != p0.y) {
				p0 = points[i];
				list.add(p0);
			}
		}
		
		it = list.iterator();
		if (list.size() < 3) {
			while (it.hasNext()) {
				ch.addPoint((Point) it.next());
			}	
			return ch;
		}
		
		points = new Point[list.size()];
		i = 0;
		while (it.hasNext()) {
			points[i] = (Point) it.next();
			i++;
		}		
		
		LinkedList upper = new LinkedList();
		upper.addFirst(points[0]);
		upper.addFirst(points[1]);
		for (i = 2; i < points.length; i++) {
			Point p1 = (Point) upper.get(1);
			Point p2 = (Point) upper.getFirst();
			Point p3 = points[i];
			while (isLeft(p1, p2, p3)) {
				if (upper.size() > 2) {
					upper.removeFirst();
					p2 = p1;
					p1 = (Point) upper.get(1);
				} else {
					upper.removeFirst();
					break;
				}
			}
			upper.addFirst(p3);
		}
		ch.addPoints(upper);
		
		upper.clear();
		upper.addFirst(points[points.length-1]);
		upper.addFirst(points[points.length-2]);
		for (i = points.length-3; i >= 0; i--) {
			Point p1 = (Point) upper.get(1);
			Point p2 = (Point) upper.getFirst();
			Point p3 = points[i];
			while (isLeft(p1, p2, p3)) {
				if (upper.size() > 2) {
					upper.removeFirst();
					p2 = p1;
					p1 = (Point) upper.get(1);
				} else {
					upper.removeFirst();
					break;
				}
			}
			upper.addFirst(p3);
		}
		upper.removeFirst();
		upper.removeLast();
		ch.addPoints(upper);
			
		return ch;
	}
	
	public static Convex getConvexHull(ShapeSet s) {
		return new Convex(convexHull(s));
	}
		

	/**
	 * The boundary in counterclockwise order.
	 */
	private Polyline outline;
	private Point center;
	private Point layeringCenter;
	private double radius;
	public double ratio;
	double[] standardVotes;
		
	public Convex() {
		super(Figure.CONVEX);
		outline = new Polyline(true);
		layeringCenter = null;
		this.ratio = 1.0;
		this.standardVotes = new double[CLASSES];
	}

	public Convex(Polyline outline) {
		super(Figure.CONVEX);
		this.outline = new Polyline(outline);
		layeringCenter = null;
		this.center = this.outline.getBoundingBox().getCenter();
		this.radius = this.outline.getDiameter() / 2;
		this.ratio = new DeformationFinder(outline).getRatio();
		this.deformation = ratio;
		this.standardVotes = new double[CLASSES];
	}
	
	public Convex(Convex other) {
		super(other);
		this.layeringCenter = new Point(other.layeringCenter);
		this.outline = new Polyline(other.outline);
		this.center = new Point(other.center);
		this.radius = other.radius;
		this.ratio = other.ratio;
		this.standardVotes = new double[other.standardVotes.length];
		for (int i = 0; i < standardVotes.length; i++) {
			this.standardVotes[i] = other.standardVotes[i];
		}
	}
	
	public void addSpecificData_01(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("CONVEX\n");
		buff.append(prefix); buff.append("outline = "); buff.append(toString(outline)); buff.append("\n");
		buff.append(prefix); buff.append("center.x = "); buff.append(center.x); buff.append("\n");
		buff.append(prefix); buff.append("center.y = "); buff.append(center.y); buff.append("\n");
		buff.append(prefix); buff.append("radius = "); buff.append(radius); buff.append("\n");
		buff.append(prefix); buff.append("ratio = "); buff.append(ratio); buff.append("\n");
		if (this.layers.size() > 0) {
			Point c = this.layeringCenter;
			if (c == null) {
				c = getCenter();
			}
			buff.append(prefix); buff.append("layering = "); buff.append(true); buff.append("\n");
			buff.append(prefix); buff.append("lc.x = "); buff.append(c.x); buff.append("\n");
			buff.append(prefix); buff.append("lc.y = "); buff.append(c.y); buff.append("\n");
		} else {
			buff.append(prefix); buff.append("layering = "); buff.append(false); buff.append("\n");
		}
		for (int i = 0; i < CLASSES; i++) {
			buff.append(prefix); buff.append("sv"); buff.append(i); buff.append(" = "); buff.append(standardVotes[i]); buff.append("\n");
		}
		writer.write(buff.toString());
	}

	public void addSpecificData_02(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<!-- CONVEX -->\n");
		buff.append(prefix); buff.append(toXml(outline, true, false)); buff.append("\n");
		buff.append(prefix); buff.append("<center.x>"); buff.append(center.x); buff.append("</center.x>\n");
		buff.append(prefix); buff.append("<center.y>"); buff.append(center.y); buff.append("</center.y>\n");
		buff.append(prefix); buff.append("<radius>"); buff.append(radius); buff.append("</radius>\n");
		buff.append(prefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		if (this.layers.size() > 0) {
			Point c = this.layeringCenter;
			if (c == null) {
				c = getCenter();
			}
			buff.append(prefix); buff.append("<layering>"); buff.append(true); buff.append("</layering>\n");
			buff.append(prefix); buff.append("<lc.x>"); buff.append(c.x); buff.append("</lc.x>\n");
			buff.append(prefix); buff.append("<lc.y>"); buff.append(c.y); buff.append("</lc.y>\n");
		} else {
			buff.append(prefix); buff.append("<layering>"); buff.append(false); buff.append("</layering>\n");
		}
		for (int i = 0; i < CLASSES; i++) {
			buff.append(prefix); buff.append("<sv number=\""); buff.append(i); buff.append("\">"); buff.append(standardVotes[i]); buff.append("</sv>\n");
		}
		writer.write(buff.toString());
	}
	
	public void readSpecificData_01(BufferedReader reader) throws IOException {
		String line;
		String[] parts;
		
		line = reader.readLine().trim();
		parts = line.split(" = ");
		outline = toPolyLine(parts[1]);

		center = new Point();		
		line = reader.readLine().trim();
		parts = line.split(" = ");
		center.x = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		center.y = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		radius = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		ratio = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		boolean layering = Boolean.valueOf(parts[1]).booleanValue();
		if (layering) {
			layeringCenter = new Point();
			line = reader.readLine().trim();
			parts = line.split(" = ");
			layeringCenter.x = Double.parseDouble(parts[1]);
			line = reader.readLine().trim();
			parts = line.split(" = ");
			layeringCenter.y = Double.parseDouble(parts[1]);
		} else {
			layeringCenter = null;
		}
		for (int i = 0; i < CLASSES; i++) {
			line = reader.readLine().trim();
			parts = line.split(" = ");
			standardVotes[i] = Double.parseDouble(parts[1]);
		}
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("FIGURE end"));
	}


	public void readSpecificData_02(BufferedReader reader) throws IOException {
		String line;
		String field;
		
		line = reader.readLine().trim();
		outline = readXml(line);

		center = new Point();		
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		center.x = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		center.y = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		radius = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		ratio = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		boolean layering = Boolean.valueOf(field).booleanValue();
		if (layering) {
			layeringCenter = new Point();
			line = reader.readLine().trim();
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			layeringCenter.x = Double.parseDouble(field);
			line = reader.readLine().trim();
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			layeringCenter.y = Double.parseDouble(field);
		} else {
			layeringCenter = null;
		}
		for (int i = 0; i < CLASSES; i++) {
			line = reader.readLine().trim();
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			standardVotes[i] = Double.parseDouble(field);
		}
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("</FIGURE>"));
	}

	public ShapeSet getShapeSet() {
		ShapeSet set = new ShapeSet();
		set.add(outline);
		Point c = this.layeringCenter;
		if (c == null) {
			c = this.center;
		}
		for (int i = 0; i < this.layers.size(); i++) {
			Polyline scaled = new Polyline(outline);
			double s = ((Double) layers.get(i)).doubleValue();
			SimilarityTransformation t = new SimilarityTransformation(false, s, 0.0, (1-s)*c.x, (1-s)*c.y);
			scaled.apply(t);
			set.add(scaled);
		}
		return set;
	}

	public double getSize() {
		return radius + radius*ratio;
	}
	
	public double getSalience() {
		return getSize() * (0.25 + 0.75*ratio);
	}

	public double getLength() {
		return outline.getLength();
	}

	public Point getCenter() {
		return center;
	}
	
	public Figure copy() {
		return new Convex(this);
	}

	public boolean isCopy(Figure other) {
		if (other instanceof Convex) {
			return isTheSame((Convex) other);
		}
		return false;
	}	
	
	public boolean isTheSame(Convex other) {
		ShapeSet s1 = new ShapeSet();
		s1.add(this.outline);
		ShapeSet s2 = new ShapeSet();
		s2.add(other.outline);
		double r1 = StandardComparator.representedBy(s1, s2, false);
		double r2 = StandardComparator.representedBy(s2, s1, false);
		return (r1 + r2) / 2 > 0.975;
	}
	
	public void transform(SimilarityTransformation sim) {
		//TODO...
		//sim.applyTo(this.center);
		//this.radius *= sim.getScale();
	}

	public double sqDist(Figure other) {
		// TODO Auto-generated method stub
		return Double.MAX_VALUE;
	}

	public void transfor(Transformation t) {
		// TODO Auto-generated method stub
		
	}
	
	public void computeStandardVotes() {
		ShapeSet thisSet = getShapeSet();
		for(int i = 0; i < CLASSES; i++) {
			Figure fig = standardFigures[i];
			if (i == Figure.CIRCLE) {
				Circle c = new Circle((Circle) fig);
				c.ratio = ratio;
				fig = c;
			}
			if (i == Figure.SQUARE) {
				Square s = new Square((Square) fig);
				s.ratio = ratio;
				fig = s;
			}
			double sv = StandardComparator.compare(thisSet, fig.getShapeSet());
			standardVotes[i] = sv;
		}
	}
	
	public boolean isPointInside(Point p) {
		Iterator it = outline.getPointIterator();
		Point p0 = (Point) it.next();
		Point current = p0;
		boolean hasWork = true;
		while (hasWork) {
			Point next = null;
			if (it.hasNext()) {
				next = (Point) it.next();
			} else {
				next = p0;
				hasWork = false;
			}
			if (!isLeft(current, next, p)) {
				return false;
			}
			current = next;
		}		
		return true;
	}
	
	public boolean arePointsInside(Polyline pl) {
		Iterator it = pl.getPointIterator();
		while (it.hasNext()) {
			Point p = (Point) it.next();
			if (!isPointInside(p)) {
				return false;
			}
		}
		return true;
	}
	
	public Polyline getOutline() {
		return this.outline;
	}

	public boolean layering(Figure fig) {
		if (!(fig instanceof Convex)) {
			return false;
		}

		Convex other = (Convex) fig;
		
		Box tbb = outline.getBoundingBox();
		ShapeSet thisSet = new ShapeSet(new Polyline(outline));
		
		Polyline pg = other.outline;
		Box obb = pg.getBoundingBox();
		ShapeSet otherSet = new ShapeSet(new Polyline(pg));
		
		double s = (tbb.getWidth()*obb.getWidth() + tbb.getHeight()*obb.getHeight()) / (tbb.getWidth()*tbb.getWidth() + tbb.getHeight()*tbb.getHeight());
		if (s < 0.5) {
			return false;
		}
		SimilarityTransformation scaling = new SimilarityTransformation(false, s, 0.0, 0.0, 0.0);
		thisSet.apply(scaling);
		Box sbb = thisSet.getBoundingBox();
		
		double dx = obb.getCenter().x - sbb.getCenter().x;
		double dy = obb.getCenter().y - sbb.getCenter().y;
		SimilarityTransformation translation  = new SimilarityTransformation(false, 1.0, 0.0, dx, dy);
		thisSet.apply(translation);
				
		double res = StandardComparator.compare(thisSet, otherSet, false, false, false, false, false, false, false).value[0];
		if (res > 0.8) {
			boolean add = true;
			if (s > 0.95) {
				add = Math.sqrt(dx*dx+dy*dy) < tbb.getDiag()*0.1;
			} else {
				Point lc = new Point();
				lc.x = dx / (1-s);
				lc.y = dy / (1-s);
				if (!isPointInside(lc)) {
					return false;
				}
				if (layeringCenter == null) {
					layeringCenter = lc;
				}
				add = layeringCenter.distance(lc) < other.radius * 0.2;
			}
			
			if (add) {
				for (int l = -1; l < layers.size(); l++) {
					double sOld = 1.0;
					if (l > -1) {
						sOld = ((Double) layers.get(l)).doubleValue();
					}
					if(Math.abs(s-sOld) < 0.2) {
						return true;
					}
				}
				this.layers.add(new Double(s));
				return true;
			}			
		}
		return false;
	}

	public double maybeFrame() {
		double iS = isSymmetric();
		return iS*iS;
	}

	public double pithiness() {
		double rsym = isRotating();
		double ssym = isSymmetric();
		rsym = rsym*rsym;
		ssym = ssym*ssym;
		return (0.75 + 0.125 * rsym + 0.125 * ssym)-(1-ratio)*(1-ratio)*0.4;
	}
	
	public void inflate(double f) {
		double dp = this.radius*2 * f;
		Polyline pl = new Polyline(true);
		Iterator it = this.outline.getPointIterator();
		while (it.hasNext()) {
			Point p0 = (Point) it.next();
			pl.addPoint(new Point(p0.x + dp, p0.y + dp));
			pl.addPoint(new Point(p0.x - dp, p0.y + dp));
			pl.addPoint(new Point(p0.x - dp, p0.y - dp));
			pl.addPoint(new Point(p0.x + dp, p0.y - dp));
		}
		this.outline = convexHull(new ShapeSet(pl));
	}

	public void addSpecificData_03(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<convex>\n");
		writer.write(buff.toString());
		outline.writeData_03(writer, pprefix);
		buff.setLength(0);
		buff.append(pprefix); buff.append("<radius>"); buff.append(radius); buff.append("</radius>\n");
		buff.append(pprefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		buff.append(pprefix);
		buff.append("<ratings>");
		for (int i = 0; i < CLASSES; i++) {
			buff.append(standardVotes[i]);
			buff.append(" ");
		}
		buff.deleteCharAt(buff.length()-1);
		buff.append("</ratings>\n");
		buff.append(prefix); buff.append("</convex>\n");
		writer.write(buff.toString());		
	}

	public String getTypeName() {
		return "convex";
	}

	public void readSpecificData_03(Element element) {
		NodeList childNodes = element.getChildNodes();
		int nodeCount = 0;
		int stepCount = 0;
		while (stepCount <= 3) {
			Node node = childNodes.item(nodeCount);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element) node;
				switch (stepCount) {
				case 0:
					outline = Polyline.readData_03(e);
					break;
				case 1:
					radius = Double.parseDouble(e.getTextContent());
					break;
				case 2:
					ratio = Double.parseDouble(e.getTextContent());
					break;
				case 3:
					StringTokenizer tokens = new StringTokenizer(e.getTextContent());
					for (int i = 0; i < Figure.CLASSES; i++) {
						standardVotes[i] = Double.parseDouble(tokens.nextToken());
					}
					break;
				}
				stepCount++;
			}
			nodeCount++;
		}
	}

	public Object getItemSource() {
		// TODO Auto-generated method stub
		return null;
	}

	public void setItemSource(Object source) {
		// TODO Auto-generated method stub
		
	}
}
