/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;

import profi.model.AffineTransformation;
import profi.model.Box;
import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;

public class DeformationFinder {
	
	private Polyline deformed;
	private Polyline original;
		
	private double angle;	
	private double ratio;
	
	/**
	 * Creates an algorithm object.
	 * 
	 * @param startShape 
	 * @param goalShape
	 * @param distance
	 * @param useScale Whether scaling is allowed or not.
	 */
	public DeformationFinder(Polyline polygon) {
		this.original = new Polyline(polygon);
		this.deformed = new Polyline(polygon);
		int k = 18;
		SimilarityTransformation rot = SimilarityTransformation.rotation(0.5/k*Math.PI);
		ratio = Double.MAX_VALUE;
		double ratio2 = Double.MAX_VALUE;
		for (int i = 0; i <= k; i++) {
			Box bb = this.deformed.getBoundingBox();
			double w = Math.max(bb.getWidth(), 1e-3);
			double h = Math.max(bb.getHeight(), 1e-3);
			if (h*h/w < ratio2) {
				ratio2 = h*h/w;
				ratio = h/w;
				angle = i * Math.PI/2/k;
			} if (w*w/h < ratio2) {
				ratio2 = w*w/h;
				ratio = w/h;
				angle = (i * 0.5/k + 0.5)*Math.PI;
			}
			this.deformed.apply(rot);
		}
		rot = SimilarityTransformation.rotation(angle);
		this.deformed = new Polyline(polygon);
		this.deformed.apply(rot);
		AffineTransformation scale = new AffineTransformation(1.0, 0.0, 0.0, 1/ratio, 0.0, 0.0, null);
		this.deformed.apply(scale);
	}
	
	public double getRatio() {
		return ratio;
	}
	
	public Circle getEllipse(double mapThreshold) {
		if (ratio < 0.5) {
			return null;
		}
		ShapeSet ss = new ShapeSet(this.deformed);
		
		CircleFinder pc = new CircleFinder(new ShapeSet(this.deformed));
		pc.reset();
		pc.steps(1);
		Figure[] circles = pc.getFigures(mapThreshold);
		if (circles.length > 0) {
			Circle c1 = (Circle) circles[0];
			c1.ratio = ratio;
			c1.deformation = ratio;
			c1.center.y *= ratio;
			c1.angle = -angle;
			double cx = Math.cos(-angle) * c1.center.x - Math.sin(-angle) * c1.center.y;
			double cy = Math.sin(-angle) * c1.center.x + Math.cos(-angle) * c1.center.y;
			c1.center = new Point(cx, cy);		
	
			ShapeSet movable = c1.getShapeSet();
			Box bb2 = movable.getBoundingBox();
			AverageWeightedDistance.computeCorrelations(movable, bb2);
			movable.updateWeightedLength();
			AverageWeightedDistance.computeCorrelations(movable, bb2, ss);
			double c = movable.getWeight();
			c1.represented = c;
			c1.weightAbs = c;
			return c1;
		}
		pc = new CircleFinder(new ShapeSet(this.original));
		pc.reset();
		pc.steps(1);
		circles = pc.getFigures(mapThreshold);
		if (circles.length > 0) {
			Circle c2 = (Circle) circles[0];
			ShapeSet movable = c2.getShapeSet();
			Box bb2 = movable.getBoundingBox();
			AverageWeightedDistance.computeCorrelations(movable, bb2);
			movable.updateWeightedLength();
			AverageWeightedDistance.computeCorrelations(movable, bb2, ss);
			double c = movable.getWeight();
			c2.represented = c;
			c2.weightAbs = c;
			return c2;
		}
		return null;
	}
	
	public Square getRectangle() {
		if (ratio < 0.25) {
			return null;
		}

		LinkedList results = new LinkedList();
		
		ShapeSet ss = new ShapeSet(this.deformed);
		int size1 = Math.max(200, ss.getNumberOfPoints());
		
		ShapeSet movable = Figure.square.getShapeSet();
		Box bb2 = movable.getBoundingBox();
		AverageWeightedDistance.computeCorrelations(movable, bb2);
		movable.updateWeightedLength();
		int size2 = Math.max(200, movable.getNumberOfPoints());		
		
		Transformation[] maps = 	ProbabilisticAlignment.computeTransformations(
				movable, // movable
				ss, // fixed
				false, // flips
				ProbabilisticAlignment.TRANSLATION_SCALE, // type
				2, // rounds
				(int) (50 * Math.sqrt(size1*size2)), // samples
				false // force rotation
				);
									
		for (int l = 0; l < maps.length && l < 50; l++) {
			
			ShapeSet moved = new ShapeSet(movable);
			moved.apply(maps[l]);
			bb2 = moved.getBoundingBox();
			
			AverageWeightedDistance.computeCorrelations(moved, bb2, ss);
			//AverageWeightedDistance.computeCorrelations(s1_, bb1, moved, true);
			double c = moved.getWeight();
			if (c > 0.75) {
				SimilarityTransformation sm = (SimilarityTransformation) maps[l];
				sm.weight = c;
				results.add(sm);
			}
		}
		Figure.FigureWeightComparator comp = new Figure.FigureWeightComparator(Figure.FigureWeightComparator.DESCENDING);
		Figure[] returns = new Figure[results.size()];
		Iterator it = results.iterator();
		for (int i = 0; it.hasNext(); i++) {
			SimilarityTransformation r = (SimilarityTransformation) it.next();
			returns[i] = Figure.square.copy();
			returns[i].transfor(r);
			returns[i].represented = r.weight;
			returns[i].weightAbs = r.weight;
		}
		Arrays.sort(returns, comp);
		if (returns.length < 1) return null;
		Square  sq = (Square) returns[0];
		
		sq.ratio = ratio;
		sq.deformation = ratio;
		sq.center.y *= ratio;
		sq.angle = -angle;
		double cx = Math.cos(-angle) * sq.center.x - Math.sin(-angle) * sq.center.y;
		double cy = Math.sin(-angle) * sq.center.x + Math.cos(-angle) * sq.center.y;
		sq.center = new Point(cx, cy);
		return sq;
	}	
}
