/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.Transformation;


public abstract class Figure implements WeightedClustering.Item {

	public static final int CIRCLE		=  1;
	public static final int SQUARE		=  2;
	public static final int TRIANGLE	=  3;
	public static final int CLASSES     = TRIANGLE + 1;
	public static final int CONVEX		=  4;
	public static final int COMPLEX		=  5;
	public static final int BUNCH		=  6;
	
	public static final double[] ID_THRESHOLD = {0.70, 0.85, 0.9, 0.9, 1.0, 0.7};
	public static final double FRAME_THRESHOLD = 0.75;
	
	public static Figure[] standardFigures = createStandardFigures();
	public static Circle circle = Circle.createStandardFigure();
	public static Square square = Square.createStandardFigure();
	public static Triangle triangle = Triangle.createStandardFigure();
	
	public static class FigureLengthComparator implements Comparator {
		
		/**
		 * Constant indicating ascending order.
		 */
		public static final int ASCENDING = 1;

		/**
		 * Constant indicating descending order.
		 */
		public static final int DESCENDING = -1;
		
		/**
		 * The ordering direction - ascending or descending.
		 */
		private int dir;
		
		/**
		 * Default orders ascending.
		 */
		public FigureLengthComparator() {
			this(ASCENDING);
		}
		
		/**
		 * Comparator ordering ascending or descending.
		 * 
		 * @param dir The ordering direction - ascending or descending.
		 */
		public FigureLengthComparator(int dir) {
			this.dir = dir;
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Object o1, Object o2) {
			double w1 = ((Figure) o1).getLength();
			double w2 = ((Figure) o2).getLength();
			return w1 == w2 ? 0 :
				w1 < w2 ? -1 * this.dir : 1 * this.dir;
		}		
	}	
	
	public static class FigureWeightComparator implements Comparator {
		
		/**
		 * Constant indicating ascending order.
		 */
		public static final int ASCENDING = 1;

		/**
		 * Constant indicating descending order.
		 */
		public static final int DESCENDING = -1;
		
		/**
		 * The ordering direction - ascending or descending.
		 */
		private int dir;
		
		private boolean estimates;
		
		/**
		 * Default orders ascending.
		 */
		public FigureWeightComparator() {
			this(ASCENDING);
		}
		
		/**
		 * Comparator ordering ascending or descending.
		 * 
		 * @param dir The ordering direction - ascending or descending.
		 */
		public FigureWeightComparator(int dir) {
			this.dir = dir;
			this.estimates = false;
		}
		
		public FigureWeightComparator(int dir, boolean estimates) {
			this.dir = dir;
			this.estimates = estimates;
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Object o1, Object o2) {
			if (estimates) {
				double w1 = ((Figure) o1).estimateWeight();
				double w2 = ((Figure) o2).estimateWeight();
				return w1 == w2 ? 0 :
					w1 < w2 ? -1 * this.dir : 1 * this.dir;
			}
			double w1 = ((Figure) o1).weightAbs;
			double w2 = ((Figure) o2).weightAbs;
			return w1 == w2 ? 0 :
				w1 < w2 ? -1 * this.dir : 1 * this.dir;
		}		
	}
	
	public static class Pair {		
		public Figure source;
		public Figure target;
		public Relation rel_st;
		public Relation rel_ts;
		
		public double voteRelation(Pair other) {
			return this.rel_st.vote(other.rel_st);
		}		
	}
	
	public static class Peer {
		public Figure peer;
		public double vote;
		
		public Peer(Figure peer, double vote) {
			this.peer = peer;
			this.vote = vote;
		}
	}
	
	public static double similarity(Figure f1, Figure f2) {	
		return StandardComparator.compare(f1.getShapeSet(), f2.getShapeSet());
	}

	public static Figure readData_01(BufferedReader reader) throws IOException {
		Figure fig;
		String line;
		String[] parts;
		line = reader.readLine().trim();
		parts = line.split(" = ");
		int idx = Integer.parseInt(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		int type = Integer.parseInt(parts[1]);
		
		switch (type) {
			case CIRCLE:
				fig = new Circle();
				break;
			case SQUARE:
				fig = new Square();
				break;
			case CONVEX:
				fig = new Convex();
				break;
			case COMPLEX:
				fig = new Complex();
				break;
			case TRIANGLE:
				fig = new Triangle();
				break;
			case BUNCH:
				fig = new Bunch();
				break;
			default:
				throw new RuntimeException();
		}
		
		fig.idx = idx;
		
		line = reader.readLine().trim();
		parts = line.split(" = ");
		fig.weightAbs = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		//fig.frame = Boolean.parseBoolean(parts[1]);
		fig.frame = Double.parseDouble(parts[1]);
		
		//todo: remove if statement when all data is new
		line = reader.readLine().trim();
		parts = line.split(" = ");
		if (parts[0].trim().equalsIgnoreCase("degradation")) {
			fig.degradation = Double.parseDouble(parts[1]);
			
			//todo: remove this hotfix:
			fig.degradation = Math.max(1.0, fig.degradation);
			
			line = reader.readLine().trim();
			parts = line.split(" = ");
		}
		fig.deformation = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		fig.isSymmetric = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		fig.isRotating = Double.parseDouble(parts[1]);
		Vector rels = new Vector();				
		do {
			line = reader.readLine().trim();
			if (line.equalsIgnoreCase("RELATION begin")) {
				rels.add(Relation.readData_01(reader));
			} else if (line.equalsIgnoreCase("LAYERS begin")) {
				boolean layer = true;
				do {
					String line2 = reader.readLine().trim();
					if (line2.equalsIgnoreCase("LAYERS end")) {
						layer = false;
					} else {
						parts = line2.split(" = ");
						fig.layers.add(new Double(parts[1]));
					}
				} while (layer);
			}
		} while (line.equalsIgnoreCase("LAYERS begin") || line.equalsIgnoreCase("RELATION begin"));
		
		fig.relatives = new Relation[rels.size()+1];
		for(int i = 0; i < rels.size(); i++) {
			Relation rel = (Relation) rels.get(i);
			fig.relatives[rel.f2] = rel;
		}		
		
		fig.readSpecificData_01(reader);
		
		return fig;
	}
	
	public static Figure readData_02(BufferedReader reader) throws IOException {
		Figure fig;
		String line;
		String field;
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		int idx = Integer.parseInt(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		int type = Integer.parseInt(field);
		
		switch (type) {
			case CIRCLE:
				fig = new Circle();
				break;
			case SQUARE:
				fig = new Square();
				break;
			case CONVEX:
				fig = new Convex();
				break;
			case COMPLEX:
				fig = new Complex();
				break;
			case TRIANGLE:
				fig = new Triangle();
				break;
			case BUNCH:
				fig = new Bunch();
				break;
			default:
				throw new RuntimeException();
		}
		
		fig.idx = idx;
		
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		fig.weightAbs = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		fig.frame = Double.parseDouble(field);
		line = reader.readLine().trim();
		if (line.indexOf("degradation") != -1) {
			field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
			fig.degradation = Double.parseDouble(field);
			
			//todo: remove this hotfix:
			fig.degradation = Math.max(1.0, fig.degradation);
			
			line = reader.readLine().trim();
		}
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		fig.deformation = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		fig.isSymmetric = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		fig.isRotating = Double.parseDouble(field);
		Vector rels = new Vector();				
		do {
			line = reader.readLine().trim();
			if (line.equalsIgnoreCase("<RELATION>")) {
				rels.add(Relation.readData_02(reader));
			} else if (line.equalsIgnoreCase("<LAYERS>")) {
				boolean layer = true;
				do {
					String line2 = reader.readLine().trim();
					if (line2.equalsIgnoreCase("</LAYERS>")) {
						layer = false;
					} else {
						field = line2.substring(line2.indexOf('>') + 1, line2.lastIndexOf('<'));
						fig.layers.add(new Double(field));
					}
				} while (layer);
			}
		} while (line.equalsIgnoreCase("<LAYERS>") || line.equalsIgnoreCase("<RELATION>"));
		
		fig.relatives = new Relation[rels.size()+1];
		for(int i = 0; i < rels.size(); i++) {
			Relation rel = (Relation) rels.get(i);
			fig.relatives[rel.f2] = rel;
		}		
		
		fig.readSpecificData_02(reader);
		
		return fig;
	}

	public static Figure readData_03(Element element, int size) {
		Figure fig;
		String type = element.getAttributes().getNamedItem("type").getNodeValue();
		if (type.equalsIgnoreCase("triangle")) {
			fig = new Triangle();
		} else if (type.equalsIgnoreCase("rectangle")) {
			fig = new Square();
		} else if(type.equalsIgnoreCase("ellipse")) {
			fig = new Circle();
		} else if(type.equalsIgnoreCase("convex")) {
			fig = new Convex();
		} else if(type.equalsIgnoreCase("complex")) {
			fig = new Complex();
		} else {
			throw new RuntimeException();
		}
		fig.relatives = new Relation[size];
		fig.idx = Integer.parseInt(element.getAttributes().getNamedItem("id").getNodeValue());
		
		NodeList childNodes = element.getChildNodes();
		int nodeCount = 0;
		int stepCount = 0;
		while (stepCount <= 7) {
			Node node = childNodes.item(nodeCount);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element) node;
				switch (stepCount) {
				case 0:
					fig.readSpecificData_03(e);
					break;
				case 1:
					if (e.getTagName().equalsIgnoreCase("layers")) {
						StringTokenizer tokens = new StringTokenizer(e.getTextContent());
						while (tokens.hasMoreTokens()) {
							fig.layers.add(new Double(Double.parseDouble(tokens.nextToken())));
						}
						stepCount--;
					} else {
						fig.weightAbs = Double.parseDouble(e.getTextContent());
					}
					break;
				case 2:
					fig.frame = Double.parseDouble(e.getTextContent());
					break;
				case 3:
					fig.degradation = Double.parseDouble(e.getTextContent());
					break;
				case 4:
					fig.deformation = Double.parseDouble(e.getTextContent());
					break;
				case 5:
					fig.isSymmetric = Double.parseDouble(e.getTextContent());
					break;
				case 6:
					fig.isRotating = Double.parseDouble(e.getTextContent());
					break;
				case 7:
					if (e.getTagName().equalsIgnoreCase("relation")) {
						Relation rel = Relation.readData_03(e);
						fig.relatives[rel.f2] = rel;
						stepCount--;
					} else {
						// nop (shape)
					}
					break;
				}
				stepCount++;
			}
			nodeCount++;
		}
		
		return fig;
	}
	
	public static String toString(Polyline pl) {
		StringBuffer buff = new StringBuffer();
		if (pl.isClosed()) {
			buff.append("pg:");
		} else {
			buff.append("pl:");
		}
		Iterator it = pl.getPointIterator();
		while (it.hasNext()) {
			Point p = (Point) it.next();
			buff.append(p.x);
			buff.append(',');
			buff.append(p.y);
			buff.append(';');
		}
		return buff.toString();
	}
	
	public static String toXml(Polyline pl, boolean convex, boolean profi) {
		StringBuffer buff = new StringBuffer();
		String tag = null;
		String par = null;
		if (profi) {
			tag = "pl";
			par = (pl.isClosed()) ? " cl=\"i\"" : " cl=\"n\"";
		}
		else if (convex) {
			tag = "ch";
			par = "";
		} else {
			tag = (pl.isClosed()) ? "pg" : "pl";
			par = "";
		}
		buff.append("<");
		buff.append(tag);
		buff.append(par);
		buff.append(">");
		Iterator it = pl.getPointIterator();
		while (it.hasNext()) {
			Point p = (Point) it.next();
			buff.append(p.x);
			buff.append(' ');
			buff.append(p.y);
			buff.append(' ');
		}		
		buff.append("</");
		buff.append(tag);
		buff.append(">");
		return buff.toString();
	}
		
	public static Polyline toPolyLine(String str) {
		StringTokenizer tokens = new StringTokenizer(str, ":,;");
		String s = tokens.nextToken();
		Polyline pl = new Polyline(s.equalsIgnoreCase("pg"));
		while(tokens.countTokens() >= 2) {
			double x = Double.parseDouble(tokens.nextToken());
			double y = Double.parseDouble(tokens.nextToken());
			pl.addPoint(new Point(x, y));
		}		
		return pl;
	}
	
	public static Polyline readXml(String str) {
		StringTokenizer tokens = new StringTokenizer(str, "<> ");
		String s = tokens.nextToken();
		Polyline pl = new Polyline(s.equalsIgnoreCase("pg") || s.equalsIgnoreCase("ch"));
		while(tokens.countTokens() >= 2) {
			double x = Double.parseDouble(tokens.nextToken());
			double y = Double.parseDouble(tokens.nextToken());
			pl.addPoint(new Point(x, y));
		}		
		return pl;
	}
	
	public static void relate(Figure f1, Figure f2, boolean quality) {
		
		//if (f1 == f2) return;
			
		double simTra = 0.0;
		double simRot = 0.0;
		double simMir = 0.0;	
		
		if (f1 != f2) {
			if (f1.type < Figure.CLASSES || f2.type < Figure.CLASSES) {
				simTra = simRot = simMir = f1.vote(f2);
			} else {
				if (quality) {
					ShapeSet s1 = f1.getShapeSet();
					ShapeSet s2 = f2.getShapeSet();
					simTra = StandardComparator.compare(s1, s2, false, false, true, true, false, false, false).value[0];
					simRot = StandardComparator.compare(s1, s2, false, true, true, true, false, false, false).value[0];
					simMir = StandardComparator.compare(s1, s2, false, true, true, true, true, false, false).value[0];
				} else {
					simTra = simRot = simMir = 0.0;
				}
			}
		} else {
			simTra = 1.0;
			simRot = 1.0;
			simMir = f1.isSymmetric;
		}
		
		double d1 = f1.getSize();
		double d2 = f2.getSize();
		double ratio = d1 / d2;
		double distance = f1.getCenter().distance(f2.getCenter());
		
		Relation r12 = new Relation(f1, f2);
		Relation r21 = new Relation(f2, f1);
		
		r12.partner = r21;
		r21.partner = r12;
		
		if (2*distance > d1 + d2) {
			//r12.place = Relation.PLACE_NEIGHBOUR;
			//r21.place = Relation.PLACE_NEIGHBOUR;
			r12.weightAbs = 1.5;
			r21.weightAbs = 1.5;
		} else if (distance < Math.max(d1,d2) / 8) {
			//r12.place = Relation.PLACE_CENTER;
			//r21.place = Relation.PLACE_CENTER;
			r12.weightAbs = 2;
			r21.weightAbs = 2;
		} else {
			if (ratio > 2.0) {
				//r12.place = Relation.PLACE_INSIDE;
				//r21.place = Relation.PLACE_DEFAULT;
				r12.weightAbs = 1.5;
				r12.weightAbs = 1.0;
			} else {
				//r12.place = Relation.PLACE_DEFAULT;
				//r21.place = Relation.PLACE_DEFAULT;
				r12.weightAbs = 1.0;
				r12.weightAbs = 1.0;
			}
		}
		
		r12.centerDistAbs = distance;
		r21.centerDistAbs = distance;

		r12.centerDistRel = distance / d1 * 2;
		r21.centerDistRel = distance / d2 * 2;		
		
		//TODO: relations higher, weights lower for same...
		if (simTra > 0.95) {
			r12.rel = Relation.REL_SAME;
			r21.rel = Relation.REL_SAME;
			r12.weightAbs *= 0.75;
			r21.weightAbs *= 0.75;
		} else if (simRot > 0.95) {
			r12.rel = Relation.REL_ROTATED;
			r21.rel = Relation.REL_ROTATED;
			r12.weightAbs *= 1.5;
			r21.weightAbs *= 1.5;
		} else if (simMir > 0.95) {
			r12.rel = Relation.REL_MIRROR;
			r21.rel = Relation.REL_MIRROR;
			r12.weightAbs *= 1.5;
			r21.weightAbs *= 1.5;
		} else {
			r12.rel = Relation.REL_DEFAULT;
			r21.rel = Relation.REL_DEFAULT;
			r12.weightAbs *= 1;
			r21.weightAbs *= 1;
		}
		r12.ratio = ratio;
		r21.ratio = 1.0 / ratio;
		
		r12.weightAbs *= Math.sqrt(f1.weightAbs/f1.degradation * f2.weightAbs/f2.degradation);
		r21.weightAbs *= Math.sqrt(f1.weightAbs/f1.degradation * f2.weightAbs/f2.degradation);
		
		f1.relatives[f2.idx] = r12;
		f2.relatives[f1.idx] = r21;
	}

	public static Figure[] createStandardFigures() {
		Figure[] figures = new Figure[CLASSES];
		figures[0] = Triangle.createStandardFigure();
		figures[CIRCLE] = Circle.createStandardFigure();
		figures[SQUARE] = Square.createStandardFigure();
		figures[TRIANGLE] = Triangle.createStandardFigure();
		return figures;
	}
	
	public static class FigureInfo {
		public Figure figure;
		public boolean simple;
		
		public FigureInfo() {
			figure = null;
			simple = false;
		}
		
		public FigureInfo(Figure figure, boolean simple) {
			this.figure = figure;
			this.simple = simple;
		}
	}
	
	public static FigureInfo getFigure(Polyline poly, boolean recurseConvex, boolean recurseComplex) {
		FigureInfo fi = new FigureInfo();
		
		double diam = poly.getDiameter();
		ShapeSet set = new ShapeSet(poly);
				
		boolean found = false;

		Figure currentFigure;
		double currentWeight = 0;
		
		// squares...
		Figure[] squares = StandardFinder.findShape(set, Figure.square);
		
		for (int i = 0; i < squares.length; i++) {
			if (squares[i].weightAbs > Figure.ID_THRESHOLD[Figure.SQUARE]) {
				currentFigure = squares[i];
				double w = currentFigure.represented * currentFigure.represented * currentFigure.getLength();
				if (w > currentWeight) {
					fi.figure = currentFigure;
					currentWeight = w;
					found = true;
				}				
			}
		}
		
		// circles...
		Figure[] circles = StandardFinder.findShape(set, Figure.circle);
		for (int i = 0; i < circles.length; i++) {
			if (circles[i].weightAbs > Figure.ID_THRESHOLD[Figure.CIRCLE]) {
				currentFigure = circles[i];
				double w = currentFigure.represented * currentFigure.represented * currentFigure.getLength();
				if (w > currentWeight) {
					fi.figure = currentFigure;
					currentWeight = w;
					found = true;
				}
			}
		}
		
		// triangles...
		Figure[] triangles = StandardFinder.findShape(set, Figure.triangle);
		for (int i = 0; i < triangles.length; i++) {
			if (triangles[i].weightAbs > Figure.ID_THRESHOLD[Figure.TRIANGLE]) {
				currentFigure = triangles[i];
				double w = currentFigure.represented * currentFigure.represented * currentFigure.getLength();
				if (w > currentWeight) {
					fi.figure = currentFigure;
					currentWeight = w;
					found = true;
				}
			}
		}
		
		// convex...
		currentFigure = Convex.getConvexHull(set);
		double chRep = StandardComparator.representedBy(currentFigure.getShapeSet(), set, false);
		currentFigure.represented = chRep;
		if (chRep > 0.90 && (fi.figure == null || fi.figure.getLength() * 1.5 < currentFigure.getLength())) {
			if (recurseConvex) {
				FigureInfo chFigure = getFigure(currentFigure.getShapeSet().getPolyline(0), false, false);
				if (!(chFigure.figure instanceof Convex)) {
					currentFigure = chFigure.figure;
				}
			}
			fi.figure = currentFigure;
			found = true;
		}
		
		// simple?...
		if (fi.figure != null) {
			double setRep = StandardComparator.representedBy(set, fi.figure.getShapeSet(), false);
			fi.simple = setRep > 0.90;
			if (!fi.simple && recurseComplex) {
				ShapeSet f = fi.figure.getShapeSet();
				ShapeSet r = new ShapeSet();
				r.add(poly);
				r = StandardComparator.getLeftOver(r, f, true);
				if (r.getNumberOfPolylines() > 0) {
					int k = r.getNumberOfPolylines();
					double pitySum = fi.figure.getSalience() * fi.figure.pithiness() * fi.figure.pithiness() * chRep;
					double weightSum = fi.figure.getSalience() * fi.figure.pithiness();
					Figure[] figs = new Figure[k]; 
					Iterator it = r.getShapeIterator();
					for (int i = 0; i < k; i++) {
						ShapeSet nset = new ShapeSet();
						nset.add((Polyline) it.next());
						figs[i] = (new Complex(nset));
						pitySum += figs[i].getSalience() * figs[i].pithiness() * figs[i].pithiness();
						weightSum += figs[i].getSalience() * figs[i].pithiness();
					}
					Arrays.sort(figs, new FigureWeightComparator(FigureWeightComparator.DESCENDING, true));
					double weightSum2 = fi.figure.getSalience() * fi.figure.pithiness();
					int kk = 0;
				    while (weightSum2 < weightSum * 0.90 && kk < figs.length) {
						weightSum2 += figs[kk].getSalience() * figs[kk].pithiness();
						kk++;
					}
				    kk += 1;
					double pithiness = pitySum / weightSum / (Math.log(kk)/2.5+1);
					ShapeSet all = new ShapeSet();
					all.add(poly);
					Figure fig = new Complex(all);
					double pithinessAll = fig.pithiness();
					if (pithinessAll > pithiness) {
						found = false;
					}
				}
			}
		}
		
		if (!found) {
			ShapeSet pset = new ShapeSet();
			pset.add(poly);
			fi.figure = new Complex(pset);
			fi.simple = false;
		}		
		return fi;
	}
	
	public int idx;
		
	public int type;
	public double weightAbs;
	public double weightRel;
	public double frame;
	public double degradation;
	
	public double deformation;
	public double isSymmetric;
	public double isRotating;
	
	public Vector layers;
		
	public boolean matched;
	public double represented;
	
	public double boost;
			
	//public Polygon outline;
	
	public Relation[] relatives;
		
	protected Figure(int type) {
		idx = -1;
		this.type = type;
		weightAbs = 1.0;
		weightRel = 1.0;
		frame = 0.0;
		degradation = 1.0;
		deformation = 1.0;
		relatives = null;
		layers = new Vector();
		represented = 1.0;
	}
	
	public Figure(Figure other) {
		this.idx = other.idx;
		this.type = other.type;
		this.weightAbs = other.weightAbs;
		this.frame = other.frame;
		this.degradation = other.degradation;
		this.deformation = other.deformation;
		relatives = null;
		layers = new Vector();
		for (int i = 0; i < other.layers.size(); i++) {
			layers.add(new Double(((Double) other.layers.get(i)).doubleValue()));
		}
		this.represented = other.represented;
	}
	
	public void setNumberOfRelatives(int n) {
		relatives = new Relation[n];
	}
	
	public abstract void addSpecificData_01(Writer writer, String prefix) throws IOException;
	public abstract void addSpecificData_02(Writer writer, String prefix) throws IOException;
	public abstract void addSpecificData_03(Writer writer, String prefix) throws IOException;
	public abstract void readSpecificData_01(BufferedReader reader) throws IOException;
	public abstract void readSpecificData_02(BufferedReader reader) throws IOException;
	public abstract void readSpecificData_03(Element element);

	public abstract double getSize();
	public abstract double getSalience();
	public abstract ShapeSet getShapeSet();
	public abstract Point getCenter();
	public abstract boolean layering(Figure fig);
	public abstract double maybeFrame();
	
	public abstract double pithiness();
	
	public void writeData_01(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("FIGURE begin\n");
		buff.append(pprefix); buff.append("idx = "); buff.append(idx); buff.append("\n");
		
		buff.append(pprefix); buff.append("type = "); buff.append(type); buff.append("\n");
		buff.append(pprefix); buff.append("weight = "); buff.append(weightAbs); buff.append("\n");
		buff.append(pprefix); buff.append("frame = "); buff.append(frame); buff.append("\n");
		buff.append(pprefix); buff.append("degradation = "); buff.append(degradation); buff.append("\n");
		
		buff.append(pprefix); buff.append("deformation = "); buff.append(deformation); buff.append("\n");
		buff.append(pprefix); buff.append("isSymmetric = "); buff.append(isSymmetric); buff.append("\n");
		buff.append(pprefix); buff.append("isRotating = "); buff.append(isRotating); buff.append("\n");
		
		if (layers.size() > 0) {
			buff.append(pprefix); buff.append("LAYERS begin\n");
			for (int i = 0; i < layers.size(); i++) {
				buff.append(pprefix); buff.append("\t");
				buff.append("layer "); buff.append(i+1);buff.append(" = ");
				buff.append(layers.get(i)); buff.append("\n");
			}
			buff.append(pprefix); buff.append("LAYERS end\n");
		}
		
		writer.write(buff.toString());
		buff.setLength(0);
		
		for (int i = 0; i < relatives.length; i++) {
			if (i != idx) {
				relatives[i].writeData_01(writer, pprefix);
			}
		}
		
		addSpecificData_01(writer, pprefix);
		
		buff.append(prefix); buff.append("FIGURE end\n");
		writer.write(buff.toString());
	}
	

	public void writeData_02(Writer writer, String prefix) throws IOException {
		String pprefix = prefix;
		StringBuffer buff = new StringBuffer();
		if (!prefix.equalsIgnoreCase("")) {
			pprefix = prefix + "\t";
			buff.append(prefix); buff.append("<FIGURE>\n");
		}
		buff.append(pprefix); buff.append("<idx>"); buff.append(idx); buff.append("</idx>\n");
		
		buff.append(pprefix); buff.append("<type>"); buff.append(type); buff.append("</type>\n");
		buff.append(pprefix); buff.append("<weight>"); buff.append(weightAbs); buff.append("</weight>\n");
		buff.append(pprefix); buff.append("<frame>"); buff.append(frame); buff.append("</frame>\n");
		buff.append(pprefix); buff.append("<degradation>"); buff.append(degradation); buff.append("</degradation>\n");		
		buff.append(pprefix); buff.append("<deformation>"); buff.append(deformation); buff.append("</deformation>\n");
		buff.append(pprefix); buff.append("<isSymmetric>"); buff.append(isSymmetric); buff.append("</isSymmetric>\n");
		buff.append(pprefix); buff.append("<isRotating>"); buff.append(isRotating); buff.append("</isRotating>\n");
		
		if (layers.size() > 0) {
			buff.append(pprefix); buff.append("<LAYERS>\n");
			for (int i = 0; i < layers.size(); i++) {
				buff.append(pprefix); buff.append("\t");
				buff.append("<layer>"); buff.append(layers.get(i)); buff.append("</layer>\n");
			}
			buff.append(pprefix); buff.append("</LAYERS>\n");
		}
		
		writer.write(buff.toString());
		buff.setLength(0);
		
		for (int i = 0; i < relatives.length; i++) {
			if (i != idx) {
				relatives[i].writeData_02(writer, pprefix);
			}
		}
		
		addSpecificData_02(writer, pprefix);
		if (!prefix.equalsIgnoreCase("")) {
			buff.append(prefix); buff.append("</FIGURE>\n");
		}
		writer.write(buff.toString());
	}
	
	public void writeData_03(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix);
		buff.append("<figure");
		buff.append(" id=\"");
		buff.append(idx);
		buff.append("\"");
		buff.append(" type=\"");
		buff.append(getTypeName());
		buff.append("\"");		
		buff.append(">\n");
		writer.write(buff.toString());
		
		addSpecificData_03(writer, pprefix);
		
		buff.setLength(0);
		if (layers.size() > 0) {
			buff.append(pprefix);
			buff.append("<layers>");
			for (int i = 0; i < layers.size(); i++) {
				buff.append(layers.get(i));
				buff.append(' ');
			}
			buff.deleteCharAt(buff.length()-1);
			buff.append("</layers>\n");
		}

		buff.append(pprefix); buff.append("<weight>"); buff.append(weightAbs); buff.append("</weight>\n");
		buff.append(pprefix); buff.append("<frame>"); buff.append(frame); buff.append("</frame>\n");
		buff.append(pprefix); buff.append("<lowering>"); buff.append(degradation); buff.append("</lowering>\n");
		buff.append(pprefix); buff.append("<deformation>"); buff.append(deformation); buff.append("</deformation>\n");
		buff.append(pprefix); buff.append("<isSymmetric>"); buff.append(isSymmetric); buff.append("</isSymmetric>\n");
		buff.append(pprefix); buff.append("<isRotating>"); buff.append(isRotating); buff.append("</isRotating>\n");
		writer.write(buff.toString());
		
		for (int i = 0; i < relatives.length; i++) {
			if (i != idx) {
				relatives[i].writeData_03(writer, pprefix);
			}
		}
		
		getShapeSet().writeData_03(writer, pprefix);
		
		buff.setLength(0);
		buff.append(prefix); buff.append("</figure>\n");		
		writer.write(buff.toString());
	}
	
	public double vote(Figure other) {
		if (this.type < Figure.CLASSES && other.type < Figure.CLASSES) {
			double sv = voteStandard(other);
			int kMax = Math.max(this.layers.size(), other.layers.size()) + 1;
			int kMin = Math.min(this.layers.size(), other.layers.size()) + 1;
			double kDR  = (kMax - kMin) / (double) kMax;
			sv *= (1-kDR*kDR*0.4);
			sv += (1.0 - sv) *Math.min(0.5, 0.2*(kMin-1));
			return sv;
		}
		
		double minDef = Math.min(this.deformation, other.deformation);
		double maxDef = Math.min(this.deformation, other.deformation);
		double defRatio = minDef / maxDef;
		
		switch (this.type) {
				case CIRCLE:
				case SQUARE:
				case TRIANGLE: switch (other.type) {
						case CONVEX:	return ((Convex) other).standardVotes[this.type] * (1-(1-defRatio)*(1-defRatio));
						case BUNCH:	    return ((Bunch) other).standardVotes[this.type];
						case COMPLEX:	return ((Complex) other).standardVotes[this.type];
						default: throw new RuntimeException();
						}
				case CONVEX: switch (other.type) {
						case CIRCLE:
						case SQUARE:
						case TRIANGLE:	return ((Convex) this).standardVotes[other.type] * (1-(1-defRatio)*(1-defRatio));
						case CONVEX:
						case BUNCH:
						case COMPLEX:	return this.compare(other);
						default: throw new RuntimeException();
						}
				case BUNCH: switch (other.type) {
						case CIRCLE:
						case SQUARE:
						case TRIANGLE:	return ((Bunch) this).standardVotes[other.type] * (1-(1-defRatio)*(1-defRatio));
						case CONVEX:
						case BUNCH:
						case COMPLEX:	return this.compare(other);
						default: throw new RuntimeException();
						}
				case COMPLEX: switch (other.type) {
						case CIRCLE:
						case SQUARE:
						case TRIANGLE:	return ((Complex) this).standardVotes[other.type];
						case CONVEX:
						case BUNCH:
						case COMPLEX:	return this.compare(other);
						default: throw new RuntimeException();
						}
				default: throw new RuntimeException();
		}
	}
	
	public double voteStandard(Figure other) {
		double defMax = Math.max(this.deformation, other.deformation);
		double defMin = Math.min(this.deformation, other.deformation);
		double ratio = defMin / defMax;
		
		//TODO: convex
		switch (this.type) {
				case CIRCLE: switch (other.type) {
						case CIRCLE:	return 1.0 - (1-ratio)*(1-ratio);
						case SQUARE:	return 0.65;
						case TRIANGLE:	return 0.45;
						default: throw new RuntimeException();
						}
				case SQUARE: switch (other.type) {
						case CIRCLE:	return 0.65;
						case SQUARE:	return 1.0 - (1-ratio)*(1-ratio)*0.75;
						case TRIANGLE:	return 0.45;
						default: throw new RuntimeException();
						}
				case TRIANGLE: switch (other.type) {
						case CIRCLE:	return 0.45;
						case SQUARE:	return 0.45;
						case TRIANGLE:	return 1.0;
						default: throw new RuntimeException();
						}
				default: throw new RuntimeException();
		}
	}
	
	public abstract double getLength(); 
	
	public double compare(Figure other) {
		double plainSimilarity = StandardComparator.compare(this.getShapeSet(), other.getShapeSet());
		if (Math.min(this.isRotating, other.isRotating) > 0.85) {
			plainSimilarity += (1.0 - plainSimilarity) / 4;
		} else if (Math.abs(this.isRotating - other.isRotating) > 0.4) {
			plainSimilarity *= 0.85;
		}
		if (Math.min(this.isSymmetric, other.isSymmetric) > 0.85) {
			plainSimilarity += (1.0 - plainSimilarity) / 4;
		} else if (Math.abs(this.isSymmetric - other.isSymmetric) > 0.4) {
			plainSimilarity *= 0.85;
		}
		return plainSimilarity;
	}
	
	public double estimateWeight() {
		return Math.sqrt(getSize());
	}
		
	public double getRelation(Figure other) {		
		//TODO!!!
		return StandardComparator.compare(this.getShapeSet(), other.getShapeSet());
	}
	
	public double isRotating() {
			switch (type) {
			case Figure.CIRCLE:
				return 1.0;
			case Figure.SQUARE:
				return 1.0;
			case Figure.TRIANGLE:
				double dmin = Double.MAX_VALUE;
				double dmax = Double.MIN_VALUE;
				for (int i = 0; i < 3; i++) {
					double d = ((Triangle) this).getLengthSq(i);
					dmin = Math.min(dmin, d);
					dmax = Math.max(dmax, d);
				}
				return Math.sqrt(dmin / dmax);
			case Figure.CONVEX:
				return StandardComparator.isRotating(this.getShapeSet());
			case Figure.BUNCH:
				return StandardComparator.isRotating(this.getShapeSet());
			case Figure.COMPLEX:
				Complex complex = (Complex) this;
				ShapeSet set = new ShapeSet();
				set.add(complex.ch);
				return StandardComparator.isRotating(set);
				//return StandardComparator.isRotating(this.getShapeSet());
			default:
				throw new RuntimeException("default..."); 
		}
	}

	public double isSymmetric() {
		switch (type) {
			case Figure.CIRCLE:
				return 1.0;
			case Figure.SQUARE:
				return 1.0;
			case Figure.TRIANGLE:
				double dmin = Double.MAX_VALUE;
				double dmax = Double.MIN_VALUE;
				for (int i = 0; i < 3; i++) {
					double d = ((Triangle) this).getLengthSq(i);
					dmin = Math.min(dmin, d);
					dmax = Math.max(dmax, d);
				}
				return Math.sqrt(dmin / dmax);
			case Figure.CONVEX:
				return StandardComparator.isSymmetric(this.getShapeSet());
			case Figure.BUNCH:
				return StandardComparator.isSymmetric(this.getShapeSet());
			case Figure.COMPLEX:
				Complex complex = (Complex) this;
				ShapeSet set = new ShapeSet();
				set.add(complex.ch);
				return StandardComparator.isSymmetric(set);
				//TODO: replace
				//return StandardComparator.isSymmetric(this.getShapeSet());
			default:
				throw new RuntimeException("default..."); 
		}
	}
		
	public abstract boolean isCopy(Figure other);
	
	public abstract double sqDist(Figure other);
	
	public abstract void transfor(Transformation t);
	
	//public double dist(Figure other) {
	//	return Math.sqrt(sqDist(other));
	//}
	
	public abstract Figure copy();
	
	public double weightFactor() {
		return 1.0;
	}
	
	public String getClassName() {
		String name = getClass().getName(); 
		return name.substring(name.lastIndexOf('.') + 1);
	}
	
	// implementing WeightedCluster.Item
	private double itemWeight;
	
	public void setItemWeight(double w) {
		itemWeight = w;
	}
	
	public double getItemWeight() {
		return itemWeight;
	}
	
	public double itemDist(WeightedClustering.Item other) {
		return Math.sqrt(itemDistSQ(other));
	}
	
	public double itemDistSQ(WeightedClustering.Item other) {
		return sqDist((Figure) other);
	}
	
	public WeightedClustering.Item itemCopy() {
		return copy();
	}
	
	private boolean mark;
	
	public void mark(boolean m) {
		mark = m;
	}
	
	public boolean isMarked() {
		return mark;
	}
	
	
	abstract public String getTypeName();
}
