/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.awt.Color;
import java.awt.Graphics;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import profi.matching.Figure.FigureInfo;
import profi.model.Box;
import profi.model.DouglasPeucker;
import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.segmentation.RegionImage;

public class Image {

	public static class Matchings {
		private int min;
		private int max;
		private int i1;
		private int i2;
		private boolean first;
		private boolean[] c;
		private Permutations perms;
		
		public void init(int n1, int n2) {
			if (n1 <= n2) {
				min = n1;
				max = n2;
				i1 = 1;
				i2 = 0;
			} else {
				min = n2;
				max = n1;
				i1 = 0;
				i2 = 1;
			}
			perms = new Permutations();
			perms.init(min);
			first = true;
			c = new boolean[max];
			for (int i = 0; i < max; i++) {
				c[i] = i < min;
			}
		}
		
		public boolean hasNext() {
			boolean lastSubset = true;
			for (int i = max-min; i < max; i++) {
				lastSubset &= c[i];
			}			
			return first || !lastSubset || perms.hasNext();
		}
		
		public int[][] next() {
			first = false;
			int[][] ret = new int[min][2];
			
			if (!perms.hasNext()) {
				int idx = max - 1;
				int cs = 0;
				boolean gap = false;
				while (!(gap && c[idx])) {
					if (c[idx]) {
						cs++;
					} else {
						gap = true;
					}
					idx--;
				}
				c[idx] = false;
				c[idx+1] = true;
				for (int i = idx+2; i < max; i++) {
					c[i] = i <= idx+1+cs;
				}				
				perms.reset();
			}			
			int[] p = perms.next();
			int idx = -1;
			for (int i = 0; i < min; i++) {
				do {
					idx++;
				} while (!c[idx]);
				ret[i][i1] = idx;
				ret[i][i2] = p[i];
			}			
			return ret;
		}
		
	}
	
	public static class Permutations {
				
		boolean first;
		int n;		
		int[] p;
		
		public void init(int n) {
			this.n = n;
			first = true;
			p = new int[n];
			for (int i = 0; i < n; i++) {
				p[i] = i;
			}
		}
		
		public void reset() {
			first = true;
			for (int i = 0; i < n; i++) {
				p[i] = i;
			}
		}
		
		private void swap(int i, int j) {
			int tmp = p[i];
			p[i] = p[j];
			p[j] = tmp;
		}
		
		public boolean hasNext() {
			for (int i = 0; i < n; i++) {
				if (p[i] != n-1-i) {
					return true;
				}
			}
			return first;
		}
		
		int[] next() {
			if (first) {
				first = false;
				return p;
			}
			int i = n - 1;			
			while (p[i-1] >= p[i]) {
			 i = i-1;
			}
			int j = n;			
			while (p[j-1] <= p[i-1]) {
			 j = j-1;
			}
			swap(i-1, j-1);    
			i++;
			j = n;
			while (i < j) {
			  swap(i-1, j-1);
			  i++;
			  j--;
			}
			return p;
		}	
	}
	
	public static class Cluster {
		public Polyline polyLine;
		public Cluster c1, c2;		
		public Box boundingBox;
		public Polyline convexHull;
		
		public double diameter;
		
		public Cluster() {
			polyLine = null;
			c1 = null;
			c2 = null;
			boundingBox = null;
			convexHull = null;
			diameter = 0.0;
		}
		
		public Cluster(Polyline polyLine) {
			this.polyLine = polyLine;
			c1 = null;
			c2 = null;
			boundingBox = polyLine.getBoundingBox();
			ShapeSet s = new ShapeSet(); s.add(polyLine);
			convexHull = Convex.convexHull(s);
			diameter = convexHull.getDiameter();
		}
		
		public Cluster(Cluster c1, Cluster c2) {
			this.polyLine = null;
			this.c1 = c1;
			this.c2 = c2;
			boundingBox = new Box(
					Math.min(c1.boundingBox.getMinX(), c2.boundingBox.getMinX()),
					Math.min(c1.boundingBox.getMinY(), c2.boundingBox.getMinY()),
					Math.max(c1.boundingBox.getMaxX(), c2.boundingBox.getMaxX()),
					Math.max(c1.boundingBox.getMaxY(), c2.boundingBox.getMaxY()));
			ShapeSet s = new ShapeSet(); s.add(c1.convexHull);s.add(c2.convexHull);
			convexHull = Convex.convexHull(s);
			diameter = convexHull.getDiameter();
		}
		
		public ShapeSet getShape() {
			ShapeSet s = new ShapeSet();
			getShape(s);
			return s;
		}
		
		public void getShape(ShapeSet s) {
			if (polyLine != null) {
				s.add(polyLine);
			}
			if (c1 != null) {
				c1.getShape(s);
			}
			if (c2 != null) {
				c2.getShape(s);
			}
		}
	}
	
	public static class Graph {
		
		public static class Pair implements Comparable {
			public double weightedDegree;
			public int idx;
			
			public Pair(int idx, double weightedDegree) {
				this.idx = idx;
				this.weightedDegree = weightedDegree;
			}
			
			public int compareTo(Object o) {
				Pair p = (Pair) o;
				if (this.weightedDegree < p.weightedDegree) return -1;
				if (this.weightedDegree > p.weightedDegree) return 1;
				return 0;
			}			
		}
		
		public double[] weight;
		public double[] weight_initial;
		public boolean[][] edge;
		public BitSet[] edge_;
		public int oc1;
		public int oc2;
		public int on;
		public int rc1;
		public int rc2;
		public int rn;
		public int n;
		public int[] is;
		
		/**
		 * An empty assosiation graph for 2 components consisting of n1 and n2 figures.
		 * n1 and n2 should not exeed the value of 10.
		 * @param n1
		 * @param n2
		 */
		public Graph(int n1, int n2) {
			this.oc1 = n1;
			this.oc2 = n2;
			this.on = n1 * n2;
			this.rc1 = n1 * n1;
			this.rc2 = n2 * n2;
			this.rn = rc1 * rc2;
			this.n = on + rn;
			this.is = new int[Math.min(n1, n2)];
			for (int i = 0; i < is.length; i++) {
				is[i] = -1;
			}
			weight = new double[n];
			edge = new boolean[n][n];
			edge_ = new BitSet[n];
			for (int i = 0; i < n; i++) {
				edge_[i] = new BitSet(n);
				for (int j = 0; j < n; j++) {
					edge[i][j] = false;
				}	
			}			
		}
		
		public void setWeight(int o1, int o2, double w, double s) {
			weight[o1*oc2 + o2] = w*s;
		}
		
		public void setWeight(int c1o1, int c1o2, int c2o1, int c2o2, double w, double s) {
			int r1 = c1o1*oc1 + c1o2;
			int r2 = c2o1*oc2 + c2o2;
			weight[on + r1*rc2 + r2] = w*s;
		}
		
		public void setEdge(int fromC1, int fromC2, int toC1, int toC2, boolean b) {
			int nf = fromC1*oc2 + fromC2;
			int nt = toC1*oc2 + toC2;
			edge_[nf].set(nt, b);
			edge_[nt].set(nf, b);
			edge[nf][nt] = edge[nt][nf] = b;
		}
		
		public void setEdge(int fromC1, int fromC2, int toC1, int toC2) {
			setEdge(fromC1, fromC2, toC1, toC2, true);	
		}

		public void setEdge(
				int fromC1o1, int fromC1o2, int fromC2o1, int fromC2o2,
				int toC1o1, int toC1o2, int toC2o1, int toC2o2, boolean b) {
			int rfC1 = fromC1o1*oc1 + fromC1o2;
			int rfC2 = fromC2o1*oc2 + fromC2o2;
			int rtC1 = toC1o1*oc1 + toC1o2;
			int rtC2 = toC2o1*oc2 + toC2o2;
			
			int nf = on + rfC1*rc2 + rfC2;
			int nt = on + rtC1*rc2 + rtC2;
			edge_[nf].set(nt, b);
			edge_[nt].set(nf, b);
			edge[nf][nt] = edge[nt][nf] = b;
		}

		public void setEdge(
				int fromC1o1, int fromC1o2, int fromC2o1, int fromC2o2,
				int toC1o1, int toC1o2, int toC2o1, int toC2o2) {
			setEdge(fromC1o1, fromC1o2, fromC2o1, fromC2o2, toC1o1, toC1o2, toC2o1, toC2o2, true);
		}

		public void setEdge(
				int fromC1o, int fromC2o,
				int toC1o1, int toC1o2, int toC2o1, int toC2o2, boolean b) {
			int rt1 = toC1o1*oc1 + toC1o2;
			int rt2 = toC2o1*oc2 + toC2o2;
			
			int nf = fromC1o*oc2 + fromC2o;
			int nt = on + rt1*rc2 + rt2;
			edge_[nf].set(nt, b);
			edge_[nt].set(nf, b);
			edge[nf][nt] = edge[nt][nf] = b;
		}
		
		public void setEdge(
				int fromC1o, int fromC2o,
				int toC1o1, int toC1o2, int toC2o1, int toC2o2) {
			setEdge(fromC1o, fromC2o, toC1o1, toC1o2, toC2o1, toC2o2, true);
		}
		
		public double getEdgeRatio() {
			int edges = 0;
			for (int i = 0; i < n; i++) {
				for (int j = 0; j < n; j++) {
					//if (edge[i][j]) {
					if (edge_[i].get(j)) {
						edges++;
					}
				}	
			}
			return ((double) edges)/n/n;
		}
		

		public double[] getMWIS() {
			return weightedDegreeHeuristic();
		}
		
		private double[] weightedDegreeHeuristic() {
			double[] ret = {0, 0}; 
			int n = weight.length;
			double w1 = 0.0;
			double w2 = 0.0;
			Pair[] pairs = new Pair[n];
			boolean[] forbidden1 = new boolean[n];
			boolean[] forbidden2 = new boolean[n];
			for (int i = 0; i < n; i++) {
				double weightedDegree = 0.0;
				if (weight[i] == 0.0) {
					weightedDegree = Double.MAX_VALUE;
				} else {
					for (int j = 0; j < n; j++) {
						//if (edge[i][j]) {
						if (edge_[i].get(j)) {
							weightedDegree += weight[j] / weight[i];
						}
					}
				}
				pairs[i] = new Pair(i, weightedDegree);
				forbidden1[i] = false;
				forbidden2[i] = false;
			}			
			Arrays.sort(pairs);
			int objectCount = 0;
			for (int i = 0; i < n; i++) {
				int idx = pairs[i].idx;
				if (!forbidden1[idx]) {
					w1 += weight[idx];
					if (idx < on) {
						is[objectCount] = idx;
						objectCount++;
					}
					for (int e = 0; e < n; e++) {
						//forbidden[e] |= edge[idx][e];
						forbidden1[e] |= edge_[idx].get(e);
					}
				}
			}
			
			int[] is2 = new int [is.length];
			if (pairs.length > 1) {
				Pair tmp = pairs[0];
				pairs[0] = pairs[1];
				pairs[1] = tmp;
				objectCount = 0;
				for (int i = 0; i < n; i++) {
					int idx = pairs[i].idx;
					if (!forbidden2[idx]) {
						w2 += weight[idx];
						if (idx < on) {
							is2[objectCount] = idx;
							objectCount++;
						}
						for (int e = 0; e < n; e++) {
							//forbidden[e] |= edge[idx][e];
							forbidden2[e] |= edge_[idx].get(e);
						}
					}
				}
			}
			ret[0] = w1;
			ret[1] = w1;
			if (w2 > w1) {
				ret[1] = w2;
				is = is2;
			}			
			return ret;
		}
	}

	public static double epsRatio = 0.005;
	public static double relevanceThreshold = 0.05;
	public static int complexityThreshold = 1000;
	
	private boolean hasData;
	public String name;
	public ShapeSet original;
	public Vector figures;
	private Box boundingBox;
	public ShapeSet rest;
	
	private int width;
	private int height;
	
	//public boolean somethingChanged;
	
	public Image() {
		hasData = false;
		//somethingChanged = false;
	}
	
	public static class Result {
		
		public static class Pair {
			public int firstIdx;
			public int secondIdx;
			public double weight;
			public double sim;
						
			public Pair(int fi, int si, double s, double w) {
				firstIdx = fi;
				secondIdx = si;
				weight = w;
				sim = s;
			}
			
			public Pair() {
				firstIdx = -1;
				secondIdx = -1;
			}
		}
		
		public static final int ccm = 0;
		public static final int cpm = 1;
		public static final int ppm = 2;
		
		// images
		public String name1;
		public String name2;		
				
		boolean[] hasInfo;
		
		// figures
		public double[] figSim;
		public Pair[][] pairs;
		
		// transformation
		public double[] tSim;
		public SimilarityTransformation[] t;
		public ShapeSet[] firstSet;
		public ShapeSet[] secondSet;
		
		public double[][] tSimsPP;
		public SimilarityTransformation[][] tsPP; 
				
		public Result() {
			name1 = null;
			name2 = null;
			hasInfo = new boolean[3]; hasInfo[0] = false; hasInfo[1] = false;  hasInfo[2] = false;
			
			figSim = new double[2];
			pairs = new Pair[2][0];
			
			tSim = new double[2];
			tSimsPP = new double[0][0];
			tsPP = new SimilarityTransformation[0][0];
			t = new SimilarityTransformation[2];
			firstSet = new ShapeSet[2];
			secondSet = new ShapeSet[2];			
		}
		
		public double sim(int i) {
			if (hasInfo[i])
				return Math.max(figSim[i], tSim[i]);
			else
				return 0.0;
		}
		
		public double figSim(int i) {
			if (hasInfo[i])
				return figSim[i];
			else
				return 0.0;
		}

		public double tSim(int i) {
			if (hasInfo[i])
				return tSim[i];
			else
				return 0.0;
		}
		
		public boolean hasInfo(int i) {
			return hasInfo[i];
		}

		public static Result readData(File file) {
			try {
				BufferedReader reader = new BufferedReader(new FileReader(file));
				String line = reader.readLine();
				reader.close();
				if (line.contains("comparison description version 1.0")) {
					return readData_10(file);
				} else if (line.contains("comparison description version 1.2")) {
					return readData_12(file);
				} else {
					return readData_11(file);
				}
			} catch (Exception ex) {
				ex.printStackTrace();
			}
			return null;
		}
		
		public static Result readData_10(File file) {
			Result res = new Result();			
			try {
				Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
				Element resNode = document.getDocumentElement();
					
				NodeList childNodes = resNode.getChildNodes();
				
				res.name1 = childNodes.item(1).getTextContent();
				res.name2 = childNodes.item(3).getTextContent();
				
				Node tNode = childNodes.item(9);	
				{
					NodeList c2hildNodes = tNode.getChildNodes();
					res.tSim = new double[2];
					res.t = new SimilarityTransformation[2];
					res.firstSet = new ShapeSet[2];
					res.secondSet = new ShapeSet[2];
					{
						res.tSim[Result.ccm] = Double.parseDouble(c2hildNodes.item(3).getTextContent());
						NodeList c3hildNodes = c2hildNodes.item(7).getChildNodes();
						res.t[Result.ccm] = SimilarityTransformation.readData_03((Element) c3hildNodes.item(1));
						res.firstSet[Result.ccm] = ShapeSet.readData_03((Element) (c3hildNodes.item(3).getChildNodes().item(1)));
						res.secondSet[Result.ccm] = ShapeSet.readData_03((Element) (c3hildNodes.item(5).getChildNodes().item(1)));
					}
					{
						res.tSim[Result.cpm] = Double.parseDouble(c2hildNodes.item(11).getTextContent());
						NodeList c3hildNodes = c2hildNodes.item(15).getChildNodes();
						res.t[Result.cpm] = SimilarityTransformation.readData_03((Element) c3hildNodes.item(1));
						res.firstSet[Result.cpm] = ShapeSet.readData_03((Element) (c3hildNodes.item(3).getChildNodes().item(1)));
						res.secondSet[Result.cpm] = ShapeSet.readData_03((Element) (c3hildNodes.item(5).getChildNodes().item(1)));
					}
				}
				
				Node fNode = childNodes.item(11);
				{
					NodeList c2hildNodes = fNode.getChildNodes();
					res.figSim = new double[2];
					res.pairs = new Pair[2][];
					{
						res.figSim[Result.ccm] = Double.parseDouble(c2hildNodes.item(3).getTextContent());
						StringTokenizer mString = new StringTokenizer(c2hildNodes.item(7).getChildNodes().item(1).getTextContent());
						res.pairs[Result.ccm] = new Pair[mString.countTokens() / 4];
						for (int i = 0; i < res.pairs[Result.ccm].length; i++) {
							res.pairs[Result.ccm][i] = new Pair(
									Integer.parseInt(mString.nextToken()), Integer.parseInt(mString.nextToken()),
									Double.parseDouble(mString.nextToken()), Double.parseDouble(mString.nextToken()));
						}						
					}
					{
						res.figSim[Result.cpm] = Double.parseDouble(c2hildNodes.item(11).getTextContent());
						StringTokenizer mString = new StringTokenizer(c2hildNodes.item(15).getChildNodes().item(1).getTextContent());
						res.pairs[Result.cpm] = new Pair[mString.countTokens() / 4];
						for (int i = 0; i < res.pairs[Result.cpm].length; i++) {
							res.pairs[Result.cpm][i] = new Pair(
									Integer.parseInt(mString.nextToken()), Integer.parseInt(mString.nextToken()),
									Double.parseDouble(mString.nextToken()), Double.parseDouble(mString.nextToken()));
						}						
					}					
				}
				res.hasInfo = new boolean[3];
				res.hasInfo[0] = true;
				res.hasInfo[1] = true;
				res.hasInfo[2] = false;
			} catch (Exception ex) {
				ex.printStackTrace();
				return null;
			}
			return res;
		}

		public static Result readData_11(File file) {
			Result res = new Result();			
			try {
				Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
				Element resNode = document.getDocumentElement();
					
				NodeList childNodes = resNode.getChildNodes();
				
				res.name1 = childNodes.item(1).getTextContent();
				res.name2 = childNodes.item(3).getTextContent();

				
				res.tSim = new double[2];
				res.t = new SimilarityTransformation[2];
				res.firstSet = new ShapeSet[2];
				res.secondSet = new ShapeSet[2];
								
				Node tNodeCC = childNodes.item(9);				
				{
					NodeList c2hildNodesCC = tNodeCC.getChildNodes();
					res.tSim[Result.ccm] = Double.parseDouble(c2hildNodesCC.item(3).getTextContent());
					NodeList c3hildNodes = c2hildNodesCC.item(7).getChildNodes();
					res.t[Result.ccm] = SimilarityTransformation.readData_03((Element) c3hildNodes.item(1));
					res.firstSet[Result.ccm] = ShapeSet.readData_03((Element) (c3hildNodes.item(3).getChildNodes().item(1)));
					res.secondSet[Result.ccm] = ShapeSet.readData_03((Element) (c3hildNodes.item(5).getChildNodes().item(1)));					
				}
				Node tNodeCP = childNodes.item(11);
				{
					NodeList c2hildNodesCP = tNodeCP.getChildNodes();
					res.tSim[Result.cpm] = Double.parseDouble(c2hildNodesCP.item(3).getTextContent());
					NodeList c3hildNodes = c2hildNodesCP.item(7).getChildNodes();
					res.t[Result.cpm] = SimilarityTransformation.readData_03((Element) c3hildNodes.item(1));
					res.firstSet[Result.cpm] = ShapeSet.readData_03((Element) (c3hildNodes.item(3).getChildNodes().item(1)));
					res.secondSet[Result.cpm] = ShapeSet.readData_03((Element) (c3hildNodes.item(5).getChildNodes().item(1)));
				}

				res.figSim = new double[2];
				res.pairs = new Pair[2][];
				
				Node fNodeCC = childNodes.item(13);
				{
					NodeList c2hildNodesCC = fNodeCC.getChildNodes();
					res.figSim[Result.ccm] = Double.parseDouble(c2hildNodesCC.item(3).getTextContent());
					StringTokenizer mString = new StringTokenizer(c2hildNodesCC.item(7).getChildNodes().item(1).getTextContent());
					res.pairs[Result.ccm] = new Pair[mString.countTokens() / 4];
					for (int i = 0; i < res.pairs[Result.ccm].length; i++) {
						res.pairs[Result.ccm][i] = new Pair(
								Integer.parseInt(mString.nextToken()), Integer.parseInt(mString.nextToken()),
								Double.parseDouble(mString.nextToken()), Double.parseDouble(mString.nextToken()));
					}
				}
				
				Node fNodeCP = childNodes.item(15);
				{
					NodeList c2hildNodesCP = fNodeCP.getChildNodes();
					res.figSim[Result.cpm] = Double.parseDouble(c2hildNodesCP.item(3).getTextContent());
					StringTokenizer mString = new StringTokenizer(c2hildNodesCP.item(7).getChildNodes().item(1).getTextContent());
					res.pairs[Result.cpm] = new Pair[mString.countTokens() / 4];
					for (int i = 0; i < res.pairs[Result.cpm].length; i++) {
						res.pairs[Result.cpm][i] = new Pair(
								Integer.parseInt(mString.nextToken()), Integer.parseInt(mString.nextToken()),
								Double.parseDouble(mString.nextToken()), Double.parseDouble(mString.nextToken()));
					}					
				}
				res.hasInfo = new boolean[3];
				res.hasInfo[0] = true;
				res.hasInfo[1] = true;
				res.hasInfo[2] = false;
			} catch (Exception ex) {
				ex.printStackTrace();
				return null;
			}
			return res;
		}

		public static Result readData_12(File file) {
			Result res = new Result();			
			try {
				Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
				Element resNode = document.getDocumentElement();
					
				NodeList childNodes = resNode.getChildNodes();
				
				res.name1 = childNodes.item(1).getTextContent();
				res.name2 = childNodes.item(3).getTextContent();
				
				res.tSim = new double[2];
				res.t = new SimilarityTransformation[2];
				res.firstSet = new ShapeSet[2];
				res.secondSet = new ShapeSet[2];
								
				Node tNodeCC = childNodes.item(9);				
				{
					NodeList c2hildNodesCC = tNodeCC.getChildNodes();
					res.tSim[Result.ccm] = Double.parseDouble(c2hildNodesCC.item(3).getTextContent());
					NodeList c3hildNodes = c2hildNodesCC.item(7).getChildNodes();
					res.t[Result.ccm] = SimilarityTransformation.readData_03((Element) c3hildNodes.item(1));
					res.firstSet[Result.ccm] = ShapeSet.readData_03((Element) (c3hildNodes.item(3).getChildNodes().item(1)));
					res.secondSet[Result.ccm] = ShapeSet.readData_03((Element) (c3hildNodes.item(5).getChildNodes().item(1)));					
				}
				Node tNodeCP = childNodes.item(11);
				{
					NodeList c2hildNodesCP = tNodeCP.getChildNodes();
					res.tSim[Result.cpm] = Double.parseDouble(c2hildNodesCP.item(3).getTextContent());
					NodeList c3hildNodes = c2hildNodesCP.item(7).getChildNodes();
					res.t[Result.cpm] = SimilarityTransformation.readData_03((Element) c3hildNodes.item(1));
					res.firstSet[Result.cpm] = ShapeSet.readData_03((Element) (c3hildNodes.item(3).getChildNodes().item(1)));
					res.secondSet[Result.cpm] = ShapeSet.readData_03((Element) (c3hildNodes.item(5).getChildNodes().item(1)));
				}
				Node tNodePP = childNodes.item(13);
				{
					NodeList c2hildNodesPP = tNodePP.getChildNodes();
					res.tSim[Result.cpm] = Double.parseDouble(c2hildNodesPP.item(3).getTextContent());
					NodeList c3hildNodes = c2hildNodesPP.item(7).getChildNodes();
					StringTokenizer tokens = new StringTokenizer(c3hildNodes.item(1).getTextContent());
					int n = Integer.parseInt(tokens.nextToken());
					int m = Integer.parseInt(tokens.nextToken());
					res.tSimsPP = new double[n][m];
					res.tsPP = new SimilarityTransformation[n][m];
					for (int i = 0; i < n; i++) {
						for (int j = 0; j < m; j++) {
							NodeList c4hildNodes = c3hildNodes.item(3+2*(i*m+j)).getChildNodes();
							tokens = new StringTokenizer(((Text) c4hildNodes.item(0)).getWholeText());
							int ii = Integer.parseInt(tokens.nextToken());
							int jj = Integer.parseInt(tokens.nextToken());
							res.tSimsPP[ii][jj] = Double.parseDouble(tokens.nextToken());
							res.tsPP[ii][jj] = SimilarityTransformation.readData_03((Element) c4hildNodes.item(1));
						}
					}
				}

				res.figSim = new double[2];
				res.pairs = new Pair[2][];
				
				Node fNodeCC = childNodes.item(15);
				{
					NodeList c2hildNodesCC = fNodeCC.getChildNodes();
					res.figSim[Result.ccm] = Double.parseDouble(c2hildNodesCC.item(3).getTextContent());
					StringTokenizer mString = new StringTokenizer(c2hildNodesCC.item(7).getChildNodes().item(1).getTextContent());
					res.pairs[Result.ccm] = new Pair[mString.countTokens() / 4];
					for (int i = 0; i < res.pairs[Result.ccm].length; i++) {
						res.pairs[Result.ccm][i] = new Pair(
								Integer.parseInt(mString.nextToken()), Integer.parseInt(mString.nextToken()),
								Double.parseDouble(mString.nextToken()), Double.parseDouble(mString.nextToken()));
					}
				}
				
				Node fNodeCP = childNodes.item(17);
				{
					NodeList c2hildNodesCP = fNodeCP.getChildNodes();
					res.figSim[Result.cpm] = Double.parseDouble(c2hildNodesCP.item(3).getTextContent());
					StringTokenizer mString = new StringTokenizer(c2hildNodesCP.item(7).getChildNodes().item(1).getTextContent());
					res.pairs[Result.cpm] = new Pair[mString.countTokens() / 4];
					for (int i = 0; i < res.pairs[Result.cpm].length; i++) {
						res.pairs[Result.cpm][i] = new Pair(
								Integer.parseInt(mString.nextToken()), Integer.parseInt(mString.nextToken()),
								Double.parseDouble(mString.nextToken()), Double.parseDouble(mString.nextToken()));
					}					
				}
				res.hasInfo = new boolean[3];
				res.hasInfo[0] = true;
				res.hasInfo[1] = true;
				res.hasInfo[2] = true;
			} catch (Exception ex) {
				//ex.printStackTrace();
				System.err.println("Exception: " + ex.getMessage());
				return null;
			}
			return res;
		}

		public void writeData(Writer writer) throws IOException {
			StringBuffer buff = new StringBuffer();
			buff.append("<!-- comparison description version 1.3 -->\n");
			buff.append("<comparison>\n");
			buff.append("\t<image no=\"1\">"); buff.append(name1); buff.append("</image>\n");
			buff.append("\t<image no=\"2\">"); buff.append(name2); buff.append("</image>\n");
			if (hasInfo[ccm]) {				
				buff.append("\t<ccm>"); buff.append(Math.max(figSim[ccm], tSim[ccm])); buff.append("</ccm>\n");
			}
			if (hasInfo[cpm]) {				
				buff.append("\t<cpm>"); buff.append(Math.max(figSim[cpm], tSim[cpm])); buff.append("</cpm>\n");
			}
			for (int type = ccm; type <= cpm; type++) {
				if (hasInfo[type]) {
					buff.append("\t<result_transformation>\n");
					if (type == ccm) {
						buff.append("\t\t<algorithm type=\"complete-complete\" name=\"probabilistic alignment\" version=\"j1.0\"></algorithm>\n");
					} else {
						buff.append("\t\t<algorithm type=\"complete-partial\" name=\"probabilistic alignment\" version=\"j1.0\"></algorithm>\n");
					}
					buff.append("\t\t<similarity>"); buff.append(tSim[type]); buff.append("</similarity>\n");
					buff.append("\t\t<dissimilarity>"); buff.append(1.0-tSim[type]); buff.append("</dissimilarity>\n");
					buff.append("\t\t<info>\n");
					buff.append(t[type].toXML("\t\t\t")); 
					buff.append("\t\t\t<parts no=\"1\">\n");
					buff.append(firstSet[type].toXML("\t\t\t\t"));
					buff.append("\t\t\t</parts>\n");
					buff.append("\t\t\t<parts no=\"2\">\n");
					buff.append(secondSet[type].toXML("\t\t\t\t"));
					buff.append("\t\t\t</parts>\n");
					buff.append("\t\t</info>\n");
					buff.append("\t</result_transformation>\n");
				}
				if (type == cpm && hasInfo[ppm]) {
					buff.append("\t<result_transformation>\n");
					buff.append("\t\t<algorithm type=\"partial-partial\" name=\"probabilistic alignment\" version=\"j1.0\"></algorithm>\n");
					buff.append("\t\t<similarity>"); buff.append(0.0); buff.append("</similarity>\n");
					buff.append("\t\t<dissimilarity>"); buff.append(0.0); buff.append("</dissimilarity>\n");
					buff.append("\t\t<info>\n");
					int n = tSimsPP.length; 
					int m = tSimsPP[0].length;
					buff.append("\t\t\t<sizes>"); buff.append(n); buff.append(" "); buff.append(m); buff.append("</sizes>\n");
					for (int i = 0; i < n; i++) {
						for (int j = 0; j < m; j++) {
							buff.append("\t\t\t<match>\n");
							buff.append("\t\t\t\t"); buff.append(i);
							buff.append(" "); buff.append(j);
							buff.append(" "); buff.append(tSimsPP[i][j]);
							buff.append("\n");
							buff.append(tsPP[i][j].toXML("\t\t\t\t"));
							buff.append("\t\t\t</match>\n");
						}
					}					
					buff.append("\t\t</info>\n");
					buff.append("\t</result_transformation>\n");
				}
			}

			for (int type = ccm; type <= cpm; type++) {
				if (hasInfo[type]) {
					buff.append("\t<result_primitives>\n");
					if (type == ccm) {
						buff.append("\t\t<algorithm type=\"complete-complete\" name=\"exhaustive\" version=\"j1.0\"></algorithm>\n");
					} else {
						buff.append("\t\t<algorithm type=\"complete-partial\" name=\"exhaustive\" version=\"j1.0\"></algorithm>\n");
					}
					buff.append("\t\t<similarity>"); buff.append(figSim[type]); buff.append("</similarity>\n");
					buff.append("\t\t<dissimilarity>"); buff.append(1.0-figSim[type]); buff.append("</dissimilarity>\n");
					buff.append("\t\t<info>\n");
					buff.append("\t\t\t<matching>\n");
					for (int i = 0; i < pairs[type].length; i++) {
						buff.append("\t\t\t\t");
						buff.append(pairs[type][i].firstIdx);
						buff.append(" ");
						buff.append(pairs[type][i].secondIdx);
						buff.append(" ");
						buff.append(pairs[type][i].sim);
						buff.append(" ");
						buff.append(pairs[type][i].weight);
						buff.append("\n");
					}
					buff.append("\t\t\t</matching>\n");
					buff.append("\t\t</info>\n");
					buff.append("\t</result_primitives>\n");			
				}
			}
			buff.append("</comparison>\n");
			writer.write(buff.toString());			
		}		
	}
	
	public Result getSimilarity(Image other, boolean complete, boolean partial) {
		
		long t1 = System.currentTimeMillis();
		
		Result result = new Result();
		result.name1 = this.name;
		result.name2 = other.name;
		if (!(hasData && other.hasData && (complete || partial))) {
			return result;
		}
		
		result.hasInfo[Result.ccm] = complete;
		result.hasInfo[Result.cpm] = partial;
		
		double sMaxC = 0.0;
		double sMaxP = 0.0;
		
		double fr = Math.sqrt((this.getFigureRelevance() * other.getFigureRelevance()));
		double rr = Math.sqrt((this.getRelationRelevance() * other.getRelationRelevance()));
		
		int ff = this.figures.size() * other.figures.size();		
		boolean figFromT = (ff == 1) && this.figures.get(0) instanceof Complex && other.figures.get(0) instanceof Complex;
			
		
		if (ff < 100 && !figFromT) {
			double[][] fSims = getFigureSimilarities(other);
			double[][] rSims = getRelationSimilarities(other);
			
			for (int i = 0; i < this.figures.size(); i++) {
				Figure f1 = (Figure) this.figures.get(i);
				f1.boost = 1.0;
			}
			for (int i = 0; i < other.figures.size(); i++) {
				Figure f2 = (Figure) other.figures.get(i);
				f2.boost = 1.0;
			}
			
			this.normalizeWeights(fr / (fr+rr), rr / (fr+rr), other.figures.size());
			other.normalizeWeights(fr / (fr+rr), rr / (fr+rr), this.figures.size());
						
			Matchings ma = new Matchings();
			int tn = this.figures.size();
			int on = other.figures.size();
			ma.init(tn, on);
			double bestValueC = -Double.MAX_VALUE;
			int[][] bestMatchC = null;
			double bestValueP = -Double.MAX_VALUE;
			int[][] bestMatchP = null;
			while (ma.hasNext()) {
				int[][] m = ma.next();
				double sumC = 0.0;
				double sumP = 0.0;
				for (int i = 0; i < m.length; i++) {
					Figure tf1 = (Figure) this.figures.get(m[i][0]);
					Figure of1 = (Figure) other.figures.get(m[i][1]);
					sumC += fSims[m[i][0]][m[i][1]] * (tf1.weightRel + of1.weightRel) / 2;
					sumP += fSims[m[i][0]][m[i][1]] * tf1.weightRel;
					for (int j = 0; j < i; j++) {
						Figure tf2 = (Figure) this.figures.get(m[j][0]);
						Figure of2 = (Figure) other.figures.get(m[j][1]);
						sumC += rSims[m[i][0]*tn+m[j][0]][m[i][1]*on+m[j][1]] * (tf1.relatives[m[j][0]].weightRel + of1.relatives[m[j][1]].weightRel) / 2;
						sumC += rSims[m[j][0]*tn+m[i][0]][m[j][1]*on+m[i][1]] * (tf2.relatives[m[i][0]].weightRel + of2.relatives[m[i][1]].weightRel) / 2;
						sumP += rSims[m[i][0]*tn+m[j][0]][m[i][1]*on+m[j][1]] * tf1.relatives[m[j][0]].weightRel;
						sumP += rSims[m[j][0]*tn+m[i][0]][m[j][1]*on+m[i][1]] * tf2.relatives[m[i][0]].weightRel;
					}
				}
				if (sumC > bestValueC) {
					bestValueC = sumC;
					bestMatchC = m;
				}
				if (sumP > bestValueP) {
					bestValueP = sumP;
					bestMatchP = m;
				}
			}
			result.figSim[Result.ccm] = bestValueC;
			int size = bestMatchC.length;
			result.pairs[Result.ccm] = new Result.Pair[size];
			for (int i = 0; i < size; i++) {
				result.pairs[Result.ccm][i] = new Result.Pair(
						bestMatchC[i][0],
						bestMatchC[i][1],
						fSims[bestMatchC[i][0]][bestMatchC[i][1]],
						(((Figure) this.figures.get(bestMatchC[i][0])).weightRel + ((Figure) other.figures.get(bestMatchC[i][1])).weightRel) / 2);
			}
			sMaxC = bestValueC;
			
			result.figSim[Result.cpm] = bestValueP;
			size = bestMatchP.length;
			result.pairs[Result.cpm] = new Result.Pair[size];
			for (int i = 0; i < size; i++) {
				result.pairs[Result.cpm][i] = new Result.Pair(
						bestMatchP[i][0],
						bestMatchP[i][1],
						fSims[bestMatchP[i][0]][bestMatchP[i][1]],
						((Figure) this.figures.get(bestMatchP[i][0])).weightRel	);
			}
			sMaxP = bestValueP;				
		}
		
		ShapeSet s1f = this.getShapeSet();
		ShapeSet s1o = this.original;
		boolean f1 = this.hasFrames();
		
		ShapeSet s2f = other.getShapeSet();
		ShapeSet s2o = other.original;
		boolean f2 = other.hasFrames();
				
		StandardComparator.Result r = StandardComparator.compare(s1o, s2o, true);
		result.tSim[Result.ccm] = r.value[StandardComparator.Result.complete];
		result.t[Result.ccm] = r.transformation[StandardComparator.Result.complete];
		result.firstSet[Result.ccm] = r.set1[StandardComparator.Result.complete];
		result.secondSet[Result.ccm] = r.set2[StandardComparator.Result.complete];
		sMaxC = Math.max(sMaxC, result.tSim[Result.ccm]);
		result.tSim[Result.cpm] = r.value[StandardComparator.Result.partial];
		result.t[Result.cpm] = r.transformation[StandardComparator.Result.partial];
		result.firstSet[Result.cpm] = r.set1[StandardComparator.Result.partial];
		result.secondSet[Result.cpm] = r.set2[StandardComparator.Result.partial];
		sMaxP = Math.max(sMaxP, result.tSim[Result.cpm]);
		
		if (complete) {			
			if (f1 && sMaxC < 0.99) {
				ShapeSet s1n = this.getShapeSet(true);
				double fr1 = this.getFrameRatio();
				r = StandardComparator.compare(s1n, s2f, true);
				double sNew = r.value[StandardComparator.Result.complete] * (1.0 - Math.pow(1.0 - (fr1+1.0)/2, 2));
				if (sNew > sMaxC) {
					result.tSim[Result.ccm] = sNew;
					result.t[Result.ccm] = r.transformation[StandardComparator.Result.complete];
					result.firstSet[Result.ccm] = r.set1[StandardComparator.Result.complete];
					result.secondSet[Result.ccm] = r.set2[StandardComparator.Result.complete];
					sMaxC = sNew;
				}
			}
			if (f2 && sMaxC < 0.99) {
				ShapeSet s2n = other.getShapeSet(true);
				double fr2 = other.getFrameRatio();
				r = StandardComparator.compare(s1f, s2n, true);
				double sNew = r.value[StandardComparator.Result.complete] * (1.0 - Math.pow(1.0 - (1.0+fr2)/2, 2));
				if (sNew > sMaxC) {
					result.tSim[Result.ccm] = sNew;
					result.t[Result.ccm] = r.transformation[StandardComparator.Result.complete];
					result.firstSet[Result.ccm] = r.set1[StandardComparator.Result.complete];
					result.secondSet[Result.ccm] = r.set2[StandardComparator.Result.complete];
					sMaxC = sNew;
				}
			}
			if (f1 && f2 && sMaxC < 0.99) {
				ShapeSet s1n = this.getShapeSet(true);
				double fr1 = this.getFrameRatio();
				ShapeSet s2n = other.getShapeSet(true);
				double fr2 = other.getFrameRatio();
				r = StandardComparator.compare(s1n, s2n, true);
				double sNew = r.value[StandardComparator.Result.complete] * (1.0 - Math.pow(1.0 - (fr1+fr2)/2, 2));
				if (sNew > sMaxC) {
					result.tSim[Result.ccm] = sNew;
					result.t[Result.ccm] = r.transformation[StandardComparator.Result.complete];
					result.firstSet[Result.ccm] = r.set1[StandardComparator.Result.complete];
					result.secondSet[Result.ccm] = r.set2[StandardComparator.Result.complete];
					sMaxC = sNew;
				}
			}
		}
		if (figFromT) { // derive values for figures
			result.figSim[Result.ccm] = sMaxC;
			result.pairs[Result.ccm] = new Result.Pair[1];
			result.pairs[Result.ccm][0] = new Result.Pair(0, 0, sMaxC, 1.0);
			result.figSim[Result.cpm] = sMaxC;
			result.pairs[Result.cpm] = new Result.Pair[1];
			result.pairs[Result.cpm][0] = new Result.Pair(0, 0, sMaxC, 1.0);
		}

		long t2 = System.currentTimeMillis();
		
		//System.out.println("times: " + (t2-t1));
		
		return result;
	}

	public Result getSimilarity2(Image other, boolean ccm, boolean cpm, boolean ppm) {

		long t1 = System.currentTimeMillis();
		
		Result result = new Result();
		result.name1 = this.name;
		result.name2 = other.name;
		if (!(hasData && other.hasData && (ccm || cpm || ppm))) {
			return result;
		}
		
		result.hasInfo[Result.ccm] = ccm;
		result.hasInfo[Result.cpm] = cpm;
		result.hasInfo[Result.ppm] = ppm;

		ShapeSet s1f = this.getShapeSet();
		ShapeSet s1o = this.original;
		boolean hf1 = this.hasFrames();
		
		ShapeSet s2f = other.getShapeSet();
		ShapeSet s2o = other.original;
		boolean hf2 = other.hasFrames();
		
		if (ccm || cpm) {

			double sMaxC = 0.0;
			double sMaxP = 0.0;
			
			double fr = Math.sqrt((this.getFigureRelevance() * other.getFigureRelevance()));
			double rr = Math.sqrt((this.getRelationRelevance() * other.getRelationRelevance()));
			
			int ff = this.figures.size() * other.figures.size();		
			boolean figFromT = (ff == 1) && this.figures.get(0) instanceof Complex && other.figures.get(0) instanceof Complex;
				
			
			if (!figFromT) {

				for (int i = 0; i < this.figures.size(); i++) {
					Figure f1 = (Figure) this.figures.get(i);
					f1.boost = 1.0;
				}
				for (int i = 0; i < other.figures.size(); i++) {
					Figure f2 = (Figure) other.figures.get(i);
					f2.boost = 1.0;
				}
				
				this.normalizeWeights(fr / (fr+rr), rr / (fr+rr), other.figures.size());
				other.normalizeWeights(fr / (fr+rr), rr / (fr+rr), this.figures.size());
				
				int sizeAbs = Math.min(this.figures.size(), other.figures.size());
				result.pairs[Result.ccm] = new Result.Pair[sizeAbs];
				result.figSim[Result.ccm] = 0.0;
	
				result.pairs[Result.cpm] = new Result.Pair[sizeAbs];
				result.figSim[Result.cpm]  = 0.0;
				int steplength = 7;
				for (int step = 0; step*steplength < sizeAbs; step++) {
					int start = step*steplength;
					int end = (step+1)*steplength-1;
					Figure[] thisFigures = this.getFigures(start, end);
					Figure[] otherFigures = other.getFigures(start, end);
					
					double[][] fSims = getFigureSimilarities(thisFigures, otherFigures);
					double[][] rSims = getRelationSimilarities(thisFigures, otherFigures);
					
					Matchings ma = new Matchings();
					int tn = thisFigures.length;
					int on = otherFigures.length;
					ma.init(tn, on);
					double bestValueC = -Double.MAX_VALUE;
					int[][] bestMatchC = null;
					double bestValueP = -Double.MAX_VALUE;
					int[][] bestMatchP = null;
					while (ma.hasNext()) {
						int[][] m = ma.next();
						double sumC = 0.0;
						double sumP = 0.0;
						for (int i = 0; i < m.length; i++) {
							Figure tf1 = (Figure) thisFigures[m[i][0]];
							int ti = tf1.idx;
							Figure of1 = (Figure) otherFigures[m[i][1]];
							int oi = of1.idx;
							sumC += fSims[m[i][0]][m[i][1]] * (tf1.weightRel + of1.weightRel) / 2;
							sumP += fSims[m[i][0]][m[i][1]] * tf1.weightRel;
							for (int j = 0; j < i; j++) {
								Figure tf2 = (Figure) thisFigures[m[j][0]];
								int tj = tf2.idx;
								Figure of2 = (Figure) otherFigures[m[j][1]];
								int oj = of2.idx;
								sumC += rSims[m[i][0]*tn+m[j][0]][m[i][1]*on+m[j][1]] * (tf1.relatives[tj].weightRel + of1.relatives[oj].weightRel) / 2;
								sumC += rSims[m[j][0]*tn+m[i][0]][m[j][1]*on+m[i][1]] * (tf2.relatives[ti].weightRel + of2.relatives[oi].weightRel) / 2;
								sumP += rSims[m[i][0]*tn+m[j][0]][m[i][1]*on+m[j][1]] * tf1.relatives[tj].weightRel;
								sumP += rSims[m[j][0]*tn+m[i][0]][m[j][1]*on+m[i][1]] * tf2.relatives[ti].weightRel;
							}
						}
						if (sumC > bestValueC) {
							bestValueC = sumC;
							bestMatchC = m;
						}
						if (sumP > bestValueP) {
							bestValueP = sumP;
							bestMatchP = m;
						}
					}
					int size = bestMatchC.length;
					for (int i = 0; i < size; i++) {
						int f11 = thisFigures[bestMatchC[i][0]].idx;
						int f21 = otherFigures[bestMatchC[i][1]].idx;
						result.pairs[Result.ccm][step*steplength + i] = new Result.Pair(
								f11,
								f21,
								fSims[bestMatchC[i][0]][bestMatchC[i][1]],
								(thisFigures[bestMatchC[i][0]].weightRel + otherFigures[bestMatchC[i][1]].weightRel) / 2);
						for (int j = 0; j < start; j++) {
							int f12 = result.pairs[Result.ccm][j].firstIdx;
							int f22 = result.pairs[Result.ccm][j].secondIdx;
							Relation r11 = ((Figure) this.figures.get(f11)).relatives[f12];
							Relation r12 = ((Figure) this.figures.get(f12)).relatives[f11];
							Relation r21 = ((Figure) other.figures.get(f21)).relatives[f22];
							Relation r22 = ((Figure) other.figures.get(f22)).relatives[f21];
							
							bestValueC += r11.vote(r21) * (r11.weightRel + r21.weightRel)/2;
							bestValueC += r12.vote(r22) * (r12.weightRel + r22.weightRel)/2;
						}
					}
					result.figSim[Result.ccm] += bestValueC;
					sMaxC += bestValueC;
	
					size = bestMatchP.length;
					for (int i = 0; i < size; i++) {
						int f11 = thisFigures[bestMatchP[i][0]].idx;
						int f21 = otherFigures[bestMatchP[i][1]].idx;
						result.pairs[Result.cpm][step*steplength + i] = new Result.Pair(
								f11,
								f21,
								fSims[bestMatchP[i][0]][bestMatchP[i][1]],
								thisFigures[bestMatchP[i][0]].weightRel);
						for (int j = 0; j < start; j++) {
							int f12 = result.pairs[Result.cpm][j].firstIdx;
							int f22 = result.pairs[Result.cpm][j].secondIdx;
							Relation r11 = ((Figure) this.figures.get(f11)).relatives[f12];
							Relation r12 = ((Figure) this.figures.get(f12)).relatives[f11];
							Relation r21 = ((Figure) other.figures.get(f21)).relatives[f22];
							Relation r22 = ((Figure) other.figures.get(f22)).relatives[f21];
							
							bestValueP += r11.vote(r21) * r11.weightRel;
							bestValueP += r12.vote(r22) * r12.weightRel;
						}
					}
					result.figSim[Result.cpm] += bestValueP;
					sMaxP += bestValueP;		
				}					
			}
						
			StandardComparator.Result r = StandardComparator.compare(s1o, s2o, true);
			result.tSim[Result.ccm] = r.value[StandardComparator.Result.complete];
			result.t[Result.ccm] = r.transformation[StandardComparator.Result.complete];
			result.firstSet[Result.ccm] = r.set1[StandardComparator.Result.complete];
			result.secondSet[Result.ccm] = r.set2[StandardComparator.Result.complete];
			sMaxC = Math.max(sMaxC, result.tSim[Result.ccm]);
			result.tSim[Result.cpm] = r.value[StandardComparator.Result.partial];
			result.t[Result.cpm] = r.transformation[StandardComparator.Result.partial];
			result.firstSet[Result.cpm] = r.set1[StandardComparator.Result.partial];
			result.secondSet[Result.cpm] = r.set2[StandardComparator.Result.partial];
			sMaxP = Math.max(sMaxP, result.tSim[Result.cpm]);
			
			if (ccm) {
				if (hf1 && sMaxC < 0.99) {
					ShapeSet s1n = this.getShapeSet(true);
					double fr1 = this.getFrameRatio();
					r = StandardComparator.compare(s1n, s2f, true);
					double sNew = r.value[StandardComparator.Result.complete] * (1.0 - Math.pow(1.0 - (fr1+1.0)/2, 2));
					if (sNew > sMaxC) {
						result.tSim[Result.ccm] = sNew;
						result.t[Result.ccm] = r.transformation[StandardComparator.Result.complete];
						result.firstSet[Result.ccm] = r.set1[StandardComparator.Result.complete];
						result.secondSet[Result.ccm] = r.set2[StandardComparator.Result.complete];
						sMaxC = sNew;
					}
				}
				if (hf2 && sMaxC < 0.99) {
					ShapeSet s2n = other.getShapeSet(true);
					double fr2 = other.getFrameRatio();
					r = StandardComparator.compare(s1f, s2n, true);
					double sNew = r.value[StandardComparator.Result.complete] * (1.0 - Math.pow(1.0 - (1.0+fr2)/2, 2));
					if (sNew > sMaxC) {
						result.tSim[Result.ccm] = sNew;
						result.t[Result.ccm] = r.transformation[StandardComparator.Result.complete];
						result.firstSet[Result.ccm] = r.set1[StandardComparator.Result.complete];
						result.secondSet[Result.ccm] = r.set2[StandardComparator.Result.complete];
						sMaxC = sNew;
					}
				}
				if (hf1 && hf2 && sMaxC < 0.99) {
					ShapeSet s1n = this.getShapeSet(true);
					double fr1 = this.getFrameRatio();
					ShapeSet s2n = other.getShapeSet(true);
					double fr2 = other.getFrameRatio();
					r = StandardComparator.compare(s1n, s2n, true);
					double sNew = r.value[StandardComparator.Result.complete] * (1.0 - Math.pow(1.0 - (fr1+fr2)/2, 2));
					if (sNew > sMaxC) {
						result.tSim[Result.ccm] = sNew;
						result.t[Result.ccm] = r.transformation[StandardComparator.Result.complete];
						result.firstSet[Result.ccm] = r.set1[StandardComparator.Result.complete];
						result.secondSet[Result.ccm] = r.set2[StandardComparator.Result.complete];
						sMaxC = sNew;
					}
				}
			}
			if (figFromT) { // derive values for figures
				result.figSim[Result.ccm] = sMaxC;
				result.pairs[Result.ccm] = new Result.Pair[1];
				result.pairs[Result.ccm][0] = new Result.Pair(0, 0, sMaxC, 1.0);
				result.figSim[Result.cpm] = sMaxC;
				result.pairs[Result.cpm] = new Result.Pair[1];
				result.pairs[Result.cpm][0] = new Result.Pair(0, 0, sMaxC, 1.0);
			}
		}
		
		if (ppm) {
			int n = s1o.getNumberOfPolylines();
			ShapeSet[] s1ss = new ShapeSet[n];
			for (int i = 0; i < n; i++) {
				s1ss[i] = new ShapeSet(s1o.getPolyline(i));
			}
			int m = s2o.getNumberOfPolylines();
			ShapeSet[] s2ss = new ShapeSet[m];
			for (int i = 0; i < m; i++) {
				s2ss[i] = new ShapeSet(s2o.getPolyline(i));
			}
			result.tSimsPP = new double[n][m];
			result.tsPP =  new SimilarityTransformation[n][m];
			for (int i = 0; i < n; i++) {
				for (int j = 0; j < m; j++) {
					StandardComparator.Result rpp = StandardComparator.compare(s1ss[i], s2ss[j], false);
					result.tSimsPP[i][j] = rpp.value[StandardComparator.Result.partial];
					result.tsPP[i][j] = rpp.transformation[StandardComparator.Result.partial];
				}
			}
		}

		long t2 = System.currentTimeMillis();
		
		//System.out.println("times: " + (t2-t1));
		
		return result;
	}

	public double getFigureRelevance() {
		if (this.figures.size() == 1) {
			return 1.0;
		}
		//todo...
		//figureWeightFactor depends on recognizeability
		//return 1.0;
		return 0.75;
	}
	
	public double getRelationRelevance() {
		if (this.figures.size() == 1) {
			return 0.0;
		}
		//todo...
		//figureWeightFactor depends on recognizeability
		//return 0.0;
		return 0.25;
	}

	public boolean isDominatedByComplex() {
		double sumF = 0.0;
		double sumC = 0.0;
		for (int i = 0; i < figures.size(); i++) {
			Figure f = (Figure) figures.get(i);
			sumF += f.weightAbs;
			if (f instanceof Complex) {
				sumC += f.weightAbs;
			}
		}			
		return sumC/sumF > 0.5;			
	}

	public void normalizeWeights(double fig, double rel, int k) {
		if (Math.abs(1.0 - (fig + rel)) > 1e-3) System.out.println("Normalizing to " + (fig + rel));
		double sumFig = 0.0;
		double sumRel = 0.0;
		LinkedList allRels = new LinkedList();
		//...fig weightfactors...
		int n = figures.size();
		for(int i = 0; i < n; i++) {
			Figure figure = (Figure) figures.get(i);
			sumFig += figure.weightAbs/Math.sqrt(figure.degradation)*figure.boost;
			for (int j = 0; j < figure.relatives.length; j++) {
				Relation relation = figure.relatives[j]; 
				if (relation != null) {
					sumRel += relation.weightAbs;
					allRels.add(new Double(relation.weightAbs));
				}
			}
		}
		if (sumRel < 1e-6) {
			if (sumFig < 1e-6) {
				for(int i = 0; i < n; i++) {
					Figure figure = (Figure) figures.get(i);
					System.out.println("figureWeight == 0!");
					figure.weightRel = 0.0;
					for (int j = 0; j < figure.relatives.length; j++) {
						Relation relation = figure.relatives[j]; 
						if (relation != null) {
							relation.weightRel = 0.0;
						}
					}
				}
				return;					
			} else {
				fig = 1.0;
				rel = 0.0;
				sumRel = 1.0;
			}					
		}
		int k1 = this.figures.size();
		if (k1 > k && k > 1) {
			/*			
			Collections.sort(allRels);
			Collections.reverse(allRels);
			Iterator relsIt = allRels.iterator();
			double pSumRels = 0.0;
			for (int i = 0; i < k*(k-1) && relsIt.hasNext(); i++) {
				pSumRels += ((Double) relsIt.next()).doubleValue();
			}
			rel *= sumRel / pSumRels;
			*/
			double maxRels = getMaxRels2(k);
			rel *= sumRel / maxRels;
		}
		for(int i = 0; i < n; i++) {
			Figure figure = (Figure) figures.get(i);
			figure.weightRel = figure.weightAbs / Math.sqrt(figure.degradation) * figure.boost * fig / sumFig;
			for (int j = 0; j < figure.relatives.length; j++) {
				Relation relation = figure.relatives[j]; 
				if (relation != null) {
					relation.weightRel = relation.weightAbs * rel / sumRel;
				}
			}
		}			
	}
	
	double getMaxRels2(int k) {
		int n = figures.size();
		double[] sumWeight = new double[n];
		for (int i = 0; i < n; i++) {
			Figure fig = (Figure) figures.get(i);
			double[] relWeight = new double[n];
			for (int j = 0; j < n; j++) {
				relWeight[j] = (i==j) ? 0.0 : fig.relatives[j].weightAbs;
			}
			Arrays.sort(relWeight);
			sumWeight[i] = 0.0;
			for (int j = 1; j < k; j++) {
				sumWeight[i] += relWeight[n-j];
			}
		}
		double sum = 0.0;
		Arrays.sort(sumWeight);
		for (int i = 0; i < k; i++) {
			sum += sumWeight[n-1-i];
		}
		return sum;
	}
	
	double getMaxRels_(int k) {
		int kt = this.figures.size();
		int idx[] = new int[k];
		for (int i = 0; i < k-1; i++) {
			idx[i] = i;
		}
		idx[k-1] = k-2;
		double maxSum = 0.0;
		while (idx[0] < kt - k) {

			//indexes
			int lastIdx = k-1;
			for(int i = 0; i < k; i++) {
				lastIdx = k-1-i;
				if (idx[lastIdx] < kt-i-1) {
					idx[lastIdx]++;
					break;
				}
			}
			for (int i = lastIdx + 1; i < k; i++) {
				idx[i] = idx[lastIdx] + i - lastIdx;
			}
				
			//sum
			double sum = 0.0;
			for (int i = 0; i < k; i++) {
				Figure f1 = (Figure) figures.get(idx[i]);
				for (int j = i+1; j < k; j++) {
					Figure f2 = (Figure) figures.get(idx[j]);
					sum += f1.relatives[idx[j]].weightAbs;
					sum += f2.relatives[idx[i]].weightAbs;
				}
			}
			maxSum = Math.max(maxSum, sum);
		}
				
		return maxSum;
	}
	
	public double[][] getFigureSimilarities(Image other) {
		int tn = this.figures.size();
		int on = other.figures.size();
		double[][] sims = new double[tn][on];
		for (int i = 0; i < tn; i++) {
			Figure tf = (Figure) this.figures.get(i);
			for (int j = 0; j < on; j++) {
				Figure of = (Figure) other.figures.get(j);
				sims[i][j] = tf.vote(of);
			}
		}
		return sims;
	}
	
	public static double[][] getFigureSimilarities(Figure[] figs1, Figure[] figs2) {
		int tn = figs1.length;
		int on = figs2.length;
		double[][] sims = new double[tn][on];
		for (int i = 0; i < tn; i++) {
			Figure tf = figs1[i];
			for (int j = 0; j < on; j++) {
				Figure of = figs2[j];
				sims[i][j] = tf.vote(of);
			}
		}
		return sims;
	}

	public static double[][] getRelationSimilarities(Figure[] figs1, Figure[] figs2) {
		int tn = figs1.length;
		int on = figs2.length;
		double[][] sims = new double[tn*tn][on*on];
		for (int i1 = 0; i1 < tn; i1++) {
			Figure tf = figs1[i1];
			for (int i2 = 0; i2 < tn; i2++) {
				Relation tr = tf.relatives[figs1[i2].idx];
				for (int j1 = 0; j1 < on; j1++) {
					Figure of = figs2[j1];
					for (int j2 = 0; j2 < on; j2++) {
						Relation or = of.relatives[figs2[j2].idx];
						sims[i1*tn+i2][j1*on+j2] = (tr == null || or == null || i1 == i2 || j1 == j2) ? 0.0 : tr.vote(or);
					}
				}
			}
		}
		return sims;
	}
	
	public double[][] getRelationSimilarities(Image other) {
		int tn = this.figures.size();
		int on = other.figures.size();
		double[][] sims = new double[tn*tn][on*on];
		for (int i1 = 0; i1 < tn; i1++) {
			Figure tf = (Figure) this.figures.get(i1);
			for (int i2 = 0; i2 < tn; i2++) {
				Relation tr = tf.relatives[i2];
				for (int j1 = 0; j1 < on; j1++) {
					Figure of = (Figure) other.figures.get(j1);
					for (int j2 = 0; j2 < on; j2++) {
						Relation or = of.relatives[j2];
						sims[i1*tn+i2][j1*on+j2] = (tr == null || or == null || i1 == i2 || j1 == j2) ? 0.0 : tr.vote(or);
					}
				}
			}
		}
		return sims;
	}
	
	
	public Graph getGraph(Image other, double[][] fSims, double[][] rSims) {	
		int n1 = this.figures.size();
		int n2 = other.figures.size();
		Graph g = new Graph(n1, n2);
		
		// figures
		for (int i = 0; i < n1; i++) {
			Figure f1 = (Figure) this.figures.get(i);
			for (int j = 0; j < n2; j++) {
				// weights
				Figure f2 = (Figure) other.figures.get(j);
				double v = fSims[i][j];
				double w = (f1.weightRel + f2.weightRel) / 2;
				g.setWeight(i, j, w, v);
				// edges
				for (int k = 0; k < n2; k++) {
					if (k == j) continue;
					g.setEdge(i, j, i, k);
				}
				for (int k = 0; k < n1; k++) {
					if (k == i) continue;
					g.setEdge(i, j, k, j);
				}
			}
		}
		
		// relations
		for (int i1 = 0; i1 < n1; i1++) {
			Figure f11 = (Figure) this.figures.get(i1);
			for (int i2 = 0; i2 < n1; i2++) {
				Relation r1 = f11.relatives[i2];
				for (int j1 = 0; j1 < n2; j1++) {
					Figure f21 = (Figure) other.figures.get(j1);
					for (int j2 = 0; j2 < n2; j2++) {
						Relation r2 = f21.relatives[j2];
						// weights
						if (r1 == null || r2 == null) {
							g.setWeight(i1, i2, j1, j2, 0.0, 0.0);
						} else {
							double v = rSims[i1*n1+i2][j1*n2+j2];
							double w = (r1.weightRel + r2.weightRel) / 2;
							g.setWeight(i1, i2, j1, j2, w, v);
						}
						// edges to figures
						for (int k = 0; k < n2; k++) {
							if (k != j1) g.setEdge(i1, k, i1, i2, j1, j2);
							if (k != j2) g.setEdge(i2, k, i1, i2, j1, j2);
						}
						for (int k = 0; k < n1; k++) {
							if (k != i1) g.setEdge(k, j1, i1, i2, j1, j2);
							if (k != i2) g.setEdge(k, j2, i1, i2, j1, j2);
						}
						
						// edges to relations
						for (int k1 = 0; k1 < n2; k1++) {
							for (int k2 = 0; k2 < n2; k2++) {
								if (k1 == j1 && k2 == j2) continue;
								g.setEdge(i1, i2, j1, j2, i1, i2, k1, k2);
								//g.setEdge(i1, i2, j1, j2, i2, i1, k2, k1);
							}
						}
						for (int k1 = 0; k1 < n1; k1++) {
							for (int k2 = 0; k2 < n1; k2++) {
								if (k1 == i1 && k2 == i2) continue;
								g.setEdge(i1, i2, j1, j2, k1, k2, j1, j2);
								//g.setEdge(i1, i2, j1, j2, k2, k1, j2, j1);
							}
						}
					}
				}					
			}				
		}
		return g;
	}
	
	/*
	private void findComponents(ShapeSet shapeSet) {
		System.err.print("fc0 ");
		int k = shapeSet.getSize();
		if (k == 0) {
			hasData = false;
			return;
		}
		hasData = true;

		//replace by convexHull...
		//Polygon ch = Convex.convexHull(shapeSet);
		//shapeSet.setPlList(new LinkedList());
		//shapeSet.add(ch);
		
		boundingBox = shapeSet.getBoundingBox();
		double gd = boundingBox.getDiag();
		double gd2 = 0.0;
		double gm = boundingBox.getMaxDim();
		double eps = boundingBox.getMaxDim()/100;		
		
		Iterator it = shapeSet.getShapeIterator();
		while (it.hasNext())  {
			PolyLine polyLine = (PolyLine) it.next();
			Rectangle bb = polyLine.getBoundingBox();
			if (bb.getMaxDim() < gm * relevanceThreshold && polyLine instanceof Polygon) {
				it.remove();
			}
		}		
		System.err.print("fc1.");
		//shapeSet.reduce(500);
		shapeSet.reduce(1000);
		System.err.print(".");		
		shapeSet.connect(eps);
		System.err.print(".");		
		shapeSet.consolidate();
		System.err.print(".");
		//shapeSet.reduce(50);
		shapeSet.reduce(200);
		
		System.err.print("fc2 ");
		
		k = shapeSet.getSize();
		
		Vector clusters = new Vector(k);
		for (int i = 0; i < k; i++) {
			Cluster cluster = new Cluster();
			cluster.polyLine = (PolyLine) shapeSet.getShape(i);
			cluster.boundingBox = cluster.polyLine.getBoundingBox();
			clusters.add(i, cluster);
			gd2 = Math.max(gd2, cluster.boundingBox.getDiag());
		}

		// simplification
		for (int i = 0; i < k; i++) {
			Cluster cluster = (Cluster) clusters.get(i);
			double d = cluster.boundingBox.getDiag();
			
			// TODO... edge detection job?!
			if (d < gd2 * relevanceThreshold) {
				clusters.remove(i);
				i--;
				k--;
				continue;
			}
			
			double tol = d * epsRatio;
			PolygonSimplification ps = new DouglasPeucker();
			PolyLine pn = ps.simplify(cluster.polyLine, tol);
			cluster.polyLine.setPoints(pn.getPoints());
			cluster.polyLine.setCentreOfGravity(pn.getCenter());
			cluster.polyLine.updateLength();			
		}
		
		// TODO ...
		if (k == 0) {
			System.out.println("deleted everything!!!");
			hasData = false;
			return;
		}

		System.err.print("fc3 ");
		
		for (int i = 0; i < clusters.size(); i++) {
			Cluster cluster = (Cluster) clusters.get(i);
			clusters.set(i, new Cluster(cluster.polyLine));
		}
		
		for (int i = 0; i < clusters.size(); i++) {
			for (int j = i + 1; j < clusters.size(); j++) {
				boolean merge = false;
				Cluster c1 = (Cluster) clusters.get(i);
				Cluster c2 = (Cluster) clusters.get(j);
				Cluster cdu = new Cluster(c1, c2);
				merge = c1.convexHull.hasPointsInside(c2.convexHull) || c2.convexHull.hasPointsInside(c1.convexHull);
				if (!merge) {
					merge = c1.convexHull.getMinBoundaryDistance(c2.convexHull) < gd * 0.01;
				}
				if (merge) {
					
					//just for debugging ->
					if (cdu.boundingBox.getMaxDim() > 900) {
						boolean b1 = c1.convexHull.hasPointsInside(c2.convexHull);
						boolean b2 = c2.convexHull.hasPointsInside(c1.convexHull);
						boolean b3 = c1.convexHull.getMinBoundaryDistance(c2.convexHull) < gd * 0.01;
						boolean b11 = c1.convexHull.hasPointsInside(c2.convexHull);
						boolean b21 = c2.convexHull.hasPointsInside(c1.convexHull);
						boolean b31 = c1.convexHull.getMinBoundaryDistance(c2.convexHull) < gd * 0.01;
						if (b1) {
							
						}
					}
					// <- just for debugging
					
					clusters.set(i, new Cluster(c1, c2));
					clusters.remove(j);
					i = -1;
					break;
				}
			}
		}
		System.err.print("c:"+clusters.size());
		for (int i = 0; i < clusters.size(); i++) {
			for (int j = i + 1; j < clusters.size(); j++) {
				boolean merge = false;
				Cluster c1 = (Cluster) clusters.get(i);
				Cluster c2 = (Cluster) clusters.get(j);
				Cluster c = new Cluster(c1, c2);				
				merge = 
					//c.area < (c1.area + c2.area) * 1.01 || 
					//c.diameter < (c1.diameter + c2.diameter) * 0.75;
					c.area < (c1.area + c2.area) * 1.5 || //original
					c.diameter < (c1.diameter + c2.diameter) * 1.0;
				if (merge) {
					clusters.set(i, c);
					clusters.remove(j);
					i = -1;
					break;
				}
			}
		}		

		System.err.println("fc4");
		
		// TODO: Components
		components = new Component[clusters.size()];
		for (int i = 0; i < clusters.size(); i++) {
			Cluster c = (Cluster) clusters.get(i);
			components[i] = new Component();
			components[i].idx = i;
			components[i].original = c.getShape();
			components[i].bb = components[i].original.getBoundingBox();			
		}		
		System.out.println("components: " + components.length);
	}
	
	private void findFigures() {
				
		LinkedList allFigures = new LinkedList();		
		StandardComparator.Result[] results;
		Figure[] currentFigures; 
		
		for (int comp = 0; comp < components.length; comp++) {
			
			allFigures.clear();
						
			//*
			// squares...
			results = StandardComparator.findShape(components[comp].original, Figure.square.getShapeSet(), false);
			currentFigures = new Figure[results.length];
			for (int i = 0; i < results.length; i++) {
				if (results[i].value < Figure.ID_THRESHOLD[Figure.SQUARE]) {
					currentFigures[i] = null;
				} else {
					currentFigures[i] = Figure.square.copy();
					((Square) currentFigures[i]).transfor(results[i].transformation);
				}
			}
			for (int i = 0; i < currentFigures.length; i++) {
				if (currentFigures[i] == null) continue;
				for(int j = i+1; j < currentFigures.length; j++) {
					if (currentFigures[j] == null) continue;
					if (currentFigures[j].isCopy(currentFigures[i])) {
						currentFigures[j] = null;
					}
				}
				allFigures.addLast(currentFigures[i]);
			}
			
			//TODO:
			
			allFigures.clear();
			Square sq = new Square();
			sq.center = new Point(50,50);
			sq.vector = new Point(50,50);
			allFigures.add(sq);
			
			// 
			
			//
			// circles
			Figure[] circles = StandardFinder.findShape(components[comp].original, Figure.circle, false);
			for (int i = 0; i < circles.length; i++) {
				if (circles[i].weightAbs < Figure.ID_THRESHOLD[Figure.CIRCLE]) {
					circles[i] = null;
				}
			}
			for (int i = 0; i < circles.length; i++) {
				if (circles[i] == null) continue;
				for(int j = i+1; j < circles.length; j++) {
					if (circles[j] == null) continue;
					if (circles[j].isCopy(circles[i])) {
						circles[j] = null;
					}
				}
				allFigures.addLast(circles[i]);
			}
			// 
					
			//
			Figure[] triangles = StandardFinder.findShape(components[comp].original, Figure.triangle, false);
			for (int i = 0; i < triangles.length; i++) {
				if (triangles[i].weightAbs < Figure.ID_THRESHOLD[Figure.TRIANGLE]) {
					triangles[i] = null;
				}
			}
			for (int i = 0; i < triangles.length; i++) {
				if (triangles[i] == null) continue;
				for(int j = i+1; j < triangles.length; j++) {
					if (triangles[j] == null) continue;
					if (triangles[j].isCopy(triangles[i])) {
						triangles[j] = null;
					}
				}
				allFigures.addLast(triangles[i]);
			}
			// 
			
			Figure ch = Convex.getConvexHull(components[comp].original);
			
			double chLength = ch.getLength();
						
			components[comp].figures.ensureCapacity(allFigures.size() + 2);
			components[comp].figures.addAll(allFigures);
			Collections.sort(components[comp].figures, new Figure.FigureLengthComparator(Figure.FigureLengthComparator.DESCENDING));
			int size = 0;
			for (; size < components[comp].figures.size(); size++) {
				Figure fig = (Figure) components[comp].figures.get(size);
				if (fig.getLength() < chLength * 0.25) break;
			}
			components[comp].figures.setSize(Math.min(3, size));			
			
			double chRep = StandardComparator.representedBy(ch.getShapeSet(), components[comp].original);
			System.out.println(chRep);
			if (chRep > 0.90) {
				
				if (chRep > 0.99) ch.frame = 1.0;
				
				boolean chFound = false;
				for (int i = 0; i < components[comp].figures.size(); i++) {
					Figure f = (Figure) components[comp].figures.get(i);
					if (Figure.similarity(f, ch) > 0.90) {
						chFound = true;
						if (chRep > 0.99) f.frame = 1.0;
						//ch = f; // todo: not needed if figures kept
					}
				}
				if (!chFound) {
					components[comp].figures.add(ch);
				}
				//components[comp].figures.clear();
				//components[comp].figures.add(ch);
			} else {
				//components[comp].figures.clear();
			}
			
			ShapeSet found = new ShapeSet();
			for (int i = 0; i < components[comp].figures.size(); i++) {
				Figure fig = (Figure) components[comp].figures.get(i);
				fig.idx = i;
				found.add(fig.getShapeSet());
			}
			
			WeightedShapeSet rest_ = new WeightedShapeSet(components[comp].original); 
			WeightedShapeSet found_ = new WeightedShapeSet(found);
			AverageWeightedDistance.computeCorrelations(rest_, components[comp].original.getBoundingBox(), found_, true);
						
			ShapeSet rest = rest_.split(0.90, 0.5, false)[0];
			
			//Convex chRest = Convex.getConvexHull(rest);
			//if (chRest.getArea() < 0.9 * chArea) {
	
				if (rest.getSize() > 0) {
					double eps = Math.max(1.0, rest.getBoundingBox().getMaxDim()/100);
					rest.connect(eps);
					rest.consolidate();
					components[comp].rest = new ShapeSet();
					int k = rest.getSize();
					double gd = Math.min(components[comp].bb.getDiag(), components[comp].figures.size());
					for (int i = 0; i < k; i++) {
						PolyLine pl = (PolyLine) rest.getShape(i);
						double d = pl.getBoundingBox().getDiag();
						
						// TODO... edge detection job?!
						if (d >= gd * relevanceThreshold) {
							components[comp].rest.add(pl);
						}
					}
				}
				
				// TODO... maybe new Clustering for Rest
				if (components[comp].rest.getSize() > 0) {
					Complex com = new Complex(components[comp].rest);
					com.idx = components[comp].figures.size();
					components[comp].figures.add(com);
				}
			//} else {
			//	Complex com = new Complex(components[comp].original);
			//	com.idx = components[comp].figures.size();
			//	components[comp].figures.add(com);
			//	components[comp].rest = components[comp].original;
			//}
			
			for (int i = 0; i < components[comp].figures.size(); i++) {
				Figure f = (Figure) components[comp].figures.get(i);
				double l = f.getLength();
				l /= 1.0 + f.frame;
				f.weightAbs = l;
			}
		}
		{
			 
			//TODO!!!!!
			Vector figs = components[0].figures;
			figs.setSize(0);
			Triangle t = new Triangle((Triangle) Triangle.standardFigures[3]);
			t.p3 = new Point(654, 513);
			t.p2 = new Point(80, 511);
			t.p1 = new Point(359, 23);
			ShapeSet s = t.getShapeSet();
			Polygon pg = new Polygon();
			pg.addPoint(new Point(654, 513));
			pg.addPoint(new Point( 80, 511));
			pg.addPoint(new Point(359,  23));
			//ShapeSet s = new ShapeSet();
			//s.add(pg);
			//s.apply(new Similarity(0.7, 0.7, 100, 100));
			Figure fig = new Complex(s);
			fig.idx = 0;
			fig.weightAbs = 1.0;
			figs.add(fig);
			//654.0,513.0;80.0,511.0;187.0,318.0;359.0,23.0;488.0,229.0
			// /			
		}
		// standard votes
		for (int c = 0; c < components.length; c++) {
			Vector figs = components[c].figures;
			for (int i = 0; i < figs.size(); i++) {
				Figure o = (Figure) figs.get(i);
				if (o instanceof Convex) {
					((Convex) o).computeStandardVotes();
				}
				if (o instanceof Complex) {
					Complex co = (Complex) o;
					co.computeStandardVotes();			
				}
			}
		}
	}
	*/
	
	public void relateFigures(boolean quality) {
		for (int f = 0; f < figures.size(); f++) {
			Figure fig = (Figure) figures.get(f);
			fig.isRotating = fig.isRotating();
			fig.isSymmetric = fig.isSymmetric();
			fig.setNumberOfRelatives(figures.size());
		}
		for (int f1 = 0; f1 < figures.size(); f1++) {
			Figure fig1 = (Figure) figures.get(f1);
			for (int f2 = 0; f2 < figures.size(); f2++) {
				Figure fig2 = (Figure) figures.get(f2);
				if (f1 == f2) continue;
				Figure.relate(fig1, fig2, quality);					
			}
		}
	}
	
	public void weightFigures() {
		for (int f = 0; f < figures.size(); f++) {
			Figure fig = (Figure) figures.get(f);
			fig.weightAbs = Math.sqrt(fig.getSize());
			fig.degradation  = 1.0 + fig.frame * (fig.degradation - 1.0);
		}
	}
	
	/*
	public void defineStructure(ShapeSet shapeSet) {
		
		if (shapeSet.getSize() == 0) {
			defineNull();
			relateFigures();
			return;
		}
		
		findComponents(shapeSet);
		findFigures();
		relateFigures();
				
		
		for (int i = components.length-1; i >= 0; i--) {
			for (int j = i-1; j >= 0; j--) {
				double s = components[i].similarity(components[j]); 
				if (s > components[i].equality) {
					components[i].equality = s;
					components[i].other = j;
				}
			}
		}		
	}
	*/
	
	public static final int ONE_COMPLEX = 0;
	public static final int NO_CLASSIFICATION = 1;
	public static final int CLASSIFICATION = 2;
	public static final int REARRANGE = 3;
	
	public void defineStructure(RegionImage regionImage, int level) {
		this.width = regionImage.getWidth();
		this.height = regionImage.getHeight();
		switch (level) {
			case ONE_COMPLEX: {
				if (!regionImage.isSimplified()) {
					regionImage.simplifyBoundaries(Math.min(width,height));
				}
				this.name = regionImage.name;
				this.original = new ShapeSet();
				regionImage.getShapes(this.original);
				ensureComplexity(complexityThreshold, this.original);
				this.boundingBox = this.original.getBoundingBox();
				this.figures = new Vector();
				Complex com = new Complex(this.original);
				com.idx = 0;
				//com.computeStandardVotes();
				figures.add(com);						
				weightFigures();
				relateFigures(false);
				hasData = true;		
				break;
			}
			case NO_CLASSIFICATION: {
				if (!regionImage.isSimplified()) {
					regionImage.simplifyBoundaries(Math.min(width,height));
				}
				this.name = regionImage.name;
				this.original = new ShapeSet();
				regionImage.getShapes(this.original);
				ensureComplexity(complexityThreshold, this.original);
				this.boundingBox = this.original.getBoundingBox();
				this.figures = new Vector();
				Iterator poly = this.original.getShapeIterator();
				int i = 0;
				while (poly.hasNext()) {
					Polyline p = (Polyline) poly.next();
					ShapeSet set = new ShapeSet();
					set.add(p);
					Complex com = new Complex(set);
					com.idx = i;
					//com.computeStandardVotes();
					figures.add(com);
					i++;
				}						
				weightFigures();
				relateFigures(false);
				hasData = true;		
				break;
			}
			case CLASSIFICATION: {
				if (!regionImage.isSimplified()) {
					regionImage.simplifyBoundaries(Math.min(width,height));
				}
				this.name = regionImage.name;
				this.original = new ShapeSet();
				regionImage.getShapes(this.original);
				ensureComplexity(complexityThreshold, this.original);
				this.boundingBox = this.original.getBoundingBox();
				this.figures = new Vector();
				
				int n = this.original.getNumberOfPolylines();
				double[] sizes = new double[n];
				Iterator poly = this.original.getShapeIterator();
				for (int i = 0; i < n; i++) {
					sizes[i] = ((Polyline) poly.next()).getBoundingBox().getDiag();
				}
				for (int i = 0; i < n; i++) {
					int maxIdx = 0;
					for (int j = 0; j < n; j++) {
						if (sizes[j] > sizes[maxIdx]) {
							maxIdx = j;
						}
					}
					sizes[maxIdx] = -1.0;
					Polyline p = this.original.getPolyline(maxIdx);
					FigureInfo figInfo = Figure.getFigure(p, true, true);
					Figure fig = figInfo.figure;
					if (fig.type <= Figure.CLASSES) {
						ShapeSet lo = StandardComparator.getLeftOver(new ShapeSet(p), fig.getShapeSet(), true);
						if (lo.getLength() > p.getLength() / 10) {
							fig = new Complex(new ShapeSet(p));
						}
					}
					figures.add(fig);					
				}
				
				// standard votes			
				for (int i = 0; i < this.figures.size(); i++) {
					Figure o = (Figure) this.figures.get(i);
					o.idx = i;
					if (o instanceof Convex) {
						((Convex) o).computeStandardVotes();
					}
					if (o instanceof Bunch) {
						((Bunch) o).computeStandardVotes();			
					}
					if (o instanceof Complex) {
						((Complex) o).computeStandardVotes();			
					}
				}
						
				weightFigures();
				relateFigures(true);
				checkFrames();
				findDoublicates();
				hasData = true;
				break;
			}
			case REARRANGE: {
				defineStructure(regionImage);
				break;
			}
		}		
	}

	public void defineStructure(ShapeSet set, int level) {
		switch (level) {
			case ONE_COMPLEX: {
				this.original = new ShapeSet(set);
				ensureComplexity(complexityThreshold, this.original);
				this.boundingBox = this.original.getBoundingBox();
				this.figures = new Vector();
				Complex com = new Complex(this.original);
				com.idx = 0;
				//com.computeStandardVotes();
				figures.add(com);						
				weightFigures();
				relateFigures(false);
				hasData = true;		
				break;
			}
			case NO_CLASSIFICATION: {
				this.original = new ShapeSet(set);
				ensureComplexity(complexityThreshold, this.original);
				this.boundingBox = this.original.getBoundingBox();
				this.figures = new Vector();
				Iterator poly = this.original.getShapeIterator();
				int i = 0;
				while (poly.hasNext()) {
					Polyline p = (Polyline) poly.next();
					ShapeSet s = new ShapeSet();
					s.add(p);
					Complex com = new Complex(s);
					com.idx = i;
					//com.computeStandardVotes();
					figures.add(com);
					i++;
				}						
				weightFigures();
				relateFigures(false);
				hasData = true;		
				break;
			}
		}		
	}
	
	////////////////// parts of the former 
	////////////////// public void defineStructure(RegionImage regionImage)
	////////////////// begin
	
	public static LinkedList findFigures(ShapeSet set) { 

		LinkedList allFigures = new LinkedList();		
		StandardComparator.Result[] results;
		Figure[] currentFigures;
		Vector figures = new Vector();
		
		// squares...
		results = StandardComparator.findShape(set, Figure.square.getShapeSet());
		currentFigures = new Figure[results.length];
		for (int i = 0; i < results.length; i++) {
			if (results[i].value[0] < Figure.ID_THRESHOLD[Figure.SQUARE]) {
				currentFigures[i] = null;
			} else {
				currentFigures[i] = Figure.square.copy();
				((Square) currentFigures[i]).transfor(results[i].transformation[0]);
			}
		}
		for (int i = 0; i < currentFigures.length; i++) {
			if (currentFigures[i] == null) continue;
			for(int j = i+1; j < currentFigures.length; j++) {
				if (currentFigures[j] == null) continue;
				if (currentFigures[j].isCopy(currentFigures[i])) {
					currentFigures[j] = null;
				}
			}
			allFigures.addLast(currentFigures[i]);
		}
		
		// circles
		Figure[] circles = StandardFinder.findShape(set, Figure.circle);
		for (int i = 0; i < circles.length; i++) {
			if (circles[i].weightAbs < Figure.ID_THRESHOLD[Figure.CIRCLE]) {
				circles[i] = null;
			}
		}
		for (int i = 0; i < circles.length; i++) {
			if (circles[i] == null) continue;
			for(int j = i+1; j < circles.length; j++) {
				if (circles[j] == null) continue;
				if (circles[j].isCopy(circles[i])) {
					circles[j] = null;
				}
			}
			allFigures.addLast(circles[i]);
		}
		
		// triangles
		Figure[] triangles = StandardFinder.findShape(set, Figure.triangle);
		for (int i = 0; i < triangles.length; i++) {
			if (triangles[i].weightAbs < Figure.ID_THRESHOLD[Figure.TRIANGLE]) {
				triangles[i] = null;
			}
		}
		for (int i = 0; i < triangles.length; i++) {
			if (triangles[i] == null) continue;
			for(int j = i+1; j < triangles.length; j++) {
				if (triangles[j] == null) continue;
				if (triangles[j].isCopy(triangles[i])) {
					triangles[j] = null;
				}
			}
			allFigures.addLast(triangles[i]);
		}			
						
		Collections.sort(allFigures, new Figure.FigureLengthComparator(Figure.FigureLengthComparator.DESCENDING));
		return allFigures;
	}
	
	////////////////// parts of the former 
	////////////////// public void defineStructure(RegionImage regionImage)
	////////////////// end

	public void defineStructure(RegionImage regionImage) {
		
		if (!regionImage.isSimplified()) {
			regionImage.simplifyBoundaries(regionImage.diameter);
		}
		
		this.name = regionImage.name;
		this.width = regionImage.getWidth();
		this.height = regionImage.getHeight();
		this.original = new ShapeSet();
		regionImage.getShapes(this.original);
		ensureComplexity(complexityThreshold, this.original);
		this.boundingBox = this.original.getBoundingBox();
		this.figures = new Vector();
		
		// wanted maximal number of figures to be found
		int totalFigureLimit = 10;

		// wanted maximal number of figures per cluster to be found
		int clusterFigureLimit = totalFigureLimit;
		
		// first part of tree is still just a list
		boolean isSplit = false;
				
		// current node of tree
		RegionImage img = regionImage;
		
		// parts of trees to be processed one after another
		Vector clusters = new Vector();
		
		// first cluster containing all children - may be split at some level
		clusters.add(regionImage.children);
		
		// likelihood of finding a frame
		double frame = 1.0;
		
		// size of the whole image
		double diam = Convex.convexHull(this.original).getDiameter();		
		
		// processing the clusters (at the beginning there is only one, but that may produce more by being split
		for (int c = 0; c < clusters.size(); c++) {
			
			if (c == 1) {// split has just taken place and now clusters.size stays fixed
				clusterFigureLimit = (totalFigureLimit - figures.size()) / (clusters.size()-1);
			}
			
			// number of figures found for this cluster
			int cfCount = 0;
			
			// likelihood of finding a frame for this cluster
			double clFrame = frame;
			
			// whether no more frames are allowed 
			boolean stopFrame = false;
			
			// parts of contours not represented by figures
			ShapeSet rest = new ShapeSet();
			
			// all children for this cluster
			Vector children = (Vector) clusters.get(c);
			
			// still a list instead of a tree
			boolean isSimple = children.size() == 1;
			
			// only simple convex figures
			boolean bSimple = true;
			
			// reference size for cluster
			double clDiam = diam;
			if (isSimple) {
				clDiam = ((RegionImage) children.get(0)).boundary.getDiameter();
			}
			
			// process the "list" part of the tree
			while (isSimple) {
				// there is only one
				img = (RegionImage) children.get(0);	
				boolean hasPartsInside = false;
				FigureInfo figInfo = Figure.getFigure(img.boundary, true, true);
				Figure fig = figInfo.figure;
				if (fig.type <= Figure.CLASSES) { // one of the convex shapes
					// left over (the additional parts inside)
					ShapeSet lo = new ShapeSet();
					if (!figInfo.simple) { // there are additional parts
						bSimple = false;
						stopFrame = true;
						// determining the left over parts...
						ShapeSet set = new ShapeSet();
						set.add(img.boundary);
						lo = StandardComparator.getLeftOver(set, fig.getShapeSet(), true);
						hasPartsInside = lo.getNumberOfPoints() > 0;
						rest.add(lo);
					}
					// still a frame?
					clFrame *= fig.maybeFrame();
					if (fig.getSize() < 0.75 * clDiam) {
						clFrame *= fig.getSize() / clDiam;
					}
					
					// if the contained parts are small this figure gets degraded...
					double chD = 0.0;
					if (img.children.size() == 0 && !hasPartsInside) {// frames are not empty
						clFrame = 0.0;
					} else {
						for (int i = 0; i < img.children.size(); i++) {
							RegionImage ri = (RegionImage) img.children.get(i);
							ri.getShapes(lo);
						}
						lo = StandardComparator.getLeftOver(lo, fig.getShapeSet(), true);
						if (lo.getNumberOfPoints() == 0) {
							clFrame = 0.0;// frames are not empty
						} else {
							Polyline pg = Convex.convexHull(lo);
							chD = 	pg.getDiameter();
						}
					}
					fig.frame = clFrame;
					if (stopFrame) {
						clFrame = 0.0;
					}
					chD = Math.min(chD, fig.getSize());
					chD = Math.max(chD, fig.getSize() / 9);
					double ratio = chD/fig.getSize();
					fig.degradation = 2.0/Math.sqrt(ratio) * fig.frame + (1-fig.frame);
					//fig.degradation = 1.0 / (1.0 - ratio*fig.frame);
					
					cfCount += addFigure(fig, figures);
				} else { // no convex figure found
					rest.add(img.boundary);
				}
				
				// preparing for the next level in the tree
				children = img.children;
				isSimple = children.size() == 1;
			} // not simple any more (>=1 boundaries)
			if (!isSplit) { // splitting comes next
				// likelihood of being frame is inherited from the first "list" part
				frame = clFrame;
			}
			if (children.size() > 0 && bSimple) {
				// 3 possibilities for the children:
				// - every child for its own
				// - convex hull + interior
				// - all together in a Complex
				// which one is best? ...				
				
				// every part for its own...
				int k = children.size();
				Figure[] kFigs = new Figure[k];
				ShapeSet[] kSets = new ShapeSet[k];
				double[] kps = new double[k];
				double[] kpn = new double[k];
				ShapeSet chAll = new ShapeSet();
				double sizesSum = 0.0;
				double pithinessSum = 0.0;
				for (int i = 0; i < children.size(); i++) {
					RegionImage child = (RegionImage) children.get(i);
					Polyline p = child.boundary;
					kSets[i] = new ShapeSet();
					kSets[i].add(p);
					kFigs[i] = Figure.getFigure(p, true, true).figure;
					kpn[i] = kps[i] = kFigs[i].pithiness();
					for (int j = 0; j < i; j++) {
						double sim = StandardComparator.compare(kSets[i], kSets[j], true, true, false, true, false, false, false).value[0];
						if (sim > 0.95) {
							kpn[i] = (2*kps[i] + 3) / 5;
							if (kpn[j] == kps[j]) {
								kpn[j] = (2*kps[j] + 3) / 5;
							} else {
								kpn[j] = (2*kpn[j] + 1) / 3;
							}
						}
					}
					child.getShapes(chAll);
				}
				for (int i = 0; i < children.size(); i++) {
					sizesSum += kFigs[i].getSalience() * kpn[i];
					pithinessSum += kFigs[i].getSalience() * kpn[i] * kpn[i] * Math.sqrt(kFigs[i].represented*StandardComparator.representedBy(kSets[i], kFigs[i].getShapeSet(), false));
					kFigs[i].weightRel = kpn[i];
				}
				Arrays.sort(kFigs, new Figure.FigureWeightComparator(Figure.FigureWeightComparator.DESCENDING, true));
				double weightSum2 = 0.0;
				int kk = 0;
			    while (weightSum2 < sizesSum * 0.85 && kk < kFigs.length) {
					weightSum2 += kFigs[kk].getSalience() * kFigs[kk].weightRel;
					kk++;
				}
				double pithiness = pithinessSum / sizesSum / (Math.log(kk)/5+1);

				// ch + rest...
				Polyline ch = Convex.convexHull(chAll);
				Figure chFig = Figure.getFigure(ch, false, false).figure;
				double chR = StandardComparator.representedBy(chFig.getShapeSet(), chAll, false);
				chFig.represented = chR;
				double chPithiness = chFig.pithiness() * chR;
				boolean chBetter = pithiness < chPithiness;
				
				// test if ch really better...
				if (chBetter) {
					ShapeSet chRest = StandardComparator.getLeftOver(chAll, chFig.getShapeSet(), true);
					if (chRest.getNumberOfPoints() > 0) {					
	
						LinkedList chAllFigures = findFigures(chRest);
						Vector chFigures = new Vector();
						int chCount = 1;
						for (int f = 0; chCount < clusterFigureLimit && f < chAllFigures.size(); f++) {
							Figure newFigure = (Figure) chAllFigures.get(f);
							if (newFigure.getLength() > clDiam*1.4*2*0.1) {
								cfCount += addFigure(newFigure, chFigures);
							}				
						}
						
						ShapeSet found = new ShapeSet();
						for (int i = 0; i < chFigures.size(); i++) {
							found.add(((Figure) chFigures.get(i)).getShapeSet());
						}
						
						chRest = StandardComparator.getLeftOver(chRest, found, false);
						
						if (chRest.getNumberOfPoints() > 0) {
							ensureComplexity(500 / clusters.size(), chRest);
							LinkedList cls = cluster(chRest, clDiam, clusterFigureLimit - chCount);
							Iterator it = cls.iterator();
							while (it.hasNext()) {
								chFigures.add(new Complex((ShapeSet) it.next()));	
							}
						}
						
						int ck = chFigures.size() + 1;
						Figure[] ckFigs = new Figure[ck];
						ckFigs[0] = chFig;
						for (int i = 1; i < ck; i++) {
							ckFigs[i] = (Figure) chFigures.get(i-1);
						}
						double[] ckps = new double[ck];
						double[] ckpn = new double[ck];
						for (int i = 0; i < ck; i++) {
							ckpn[i] = ckps[i] = ckFigs[i].pithiness();
							for (int j = 0; j < i; j++) {
								double sim = StandardComparator.compare(ckFigs[i].getShapeSet(), ckFigs[j].getShapeSet(), true, true, false, true, false, false, false).value[0];
								if (sim > 0.95) {
									ckpn[i] = (2*ckps[i] + 3) / 5;
									if (ckpn[j] == ckps[j]) {
										ckpn[j] = (2*ckps[j] + 3) / 5;										
									} else {
										ckpn[j] = (2 * ckpn[j] + 1) / 3;
									}
								}
							}
						}
						double csizesSum = 0.0;
						double cpithinessSum = 0.0;
						for (int i = 0; i < ckFigs.length; i++) {
							csizesSum += ckFigs[i].getSalience() * ckpn[i];
							cpithinessSum += ckFigs[i].getSalience() * ckpn[i] * ckpn[i] * ckFigs[i].represented;
							ckFigs[i].weightRel = ckpn[i];
						}
						Arrays.sort(ckFigs, new Figure.FigureWeightComparator(Figure.FigureWeightComparator.DESCENDING, true));
						double cweightSum2 = 0.0;
						int ckk = 0;
					    while (cweightSum2 < csizesSum * 0.90 && ckk < ckFigs.length) {
							cweightSum2 += ckFigs[ckk].getSalience() * ckFigs[ckk].weightRel;
							ckk++;
						}
						chPithiness = cpithinessSum / csizesSum / (Math.log(ckk)/5+1);
						chBetter = pithiness < chPithiness;
					}
				}
				
				// all together...
				Complex com = new Complex(chAll);
				double cop = com.pithiness();
				
				//System.out.println("p-all: " + pithiness);
				//System.out.println("p-ch : " + chPithiness);
				//System.out.println("p-com: " + cop);
								
				if (cop > chPithiness && cop > pithiness && cop > 0.7) {
					figures.add(com);					
				} else if (chBetter && chFig.represented > 0.75 && chPithiness > 0.7) {
					clFrame *= chFig.maybeFrame();
					if (chFig.getSize() < 0.75 * clDiam) {
						clFrame *= chFig.getSize() / clDiam;
					}
					chFig.frame = clFrame;
					chFig.degradation = 2;
					cfCount += addFigure(chFig, figures);
					rest.add(StandardComparator.getLeftOver(chAll, chFig.getShapeSet(), true));					
				} else if (pithiness > 0.5 && !isSplit) {
					int cln = kFigs.length;
					double[] clvs = new double[cln];
					double clsum = 0.0;
					for (int i = 0; i < cln; i++) {
						clvs[i] = kFigs[i].getSalience() * kFigs[i].weightRel;
						clsum += clvs[i];
					}
					int cladded = 0;
					while (cladded < cln) {
						double clmax = -0.5;
						int clidx = -1;
						for (int i = 0; i < cln; i++) {
							if (clvs[i] > clmax) {
								clmax = clvs[i];
								clidx = i;
							}
						}
						if (clmax > clsum / 5 || cladded == cln -1) {
							Vector cluster = new Vector();
							cluster.add(children.get(clidx));
							clusters.add(cluster);
							isSplit = true;
							clvs[clidx] = -1.0;
							cladded++;
						} else {
							break;
						}
					}
					if (cladded < cln) {
						//somethingChanged = true;
						for (int i = 0; i < cln; i++) {
							if (clvs[i] > -0.5) {
								((RegionImage) children.get(i)).getShapes(rest);
							}
						}
					}					
				} else {
					rest.add(chAll);
				}				
			} else if (children.size() > 0) { // children size > 0 && !bsimple
				for (int i = 0; i < children.size(); i++) {
					RegionImage child = (RegionImage) children.get(i);
					child.getShapes(rest);
				}
			}
			
			if (rest.getNumberOfPoints() > 0) { // rest not empty			

				// find figures in the rest...
				LinkedList allFigures = findFigures(rest);
				for (int f = 0; cfCount < clusterFigureLimit && f < allFigures.size(); f++) {
					Figure newFigure = (Figure) allFigures.get(f);
					if (newFigure.getLength() > clDiam*1.4*2*0.1) {
						cfCount += addFigure(newFigure, figures);
					}				
				}				
				ShapeSet found = new ShapeSet();
				for (int i = 0; i < figures.size(); i++) {
					found.add(((Figure) figures.get(i)).getShapeSet());
				}
				rest = StandardComparator.getLeftOver(rest, found, true);
				
				// 2nd round of finding figures in the rest...
				if (rest.getNumberOfPoints() > 0) {
										
					allFigures = findFigures(rest);					
					for (int f = 0; cfCount < clusterFigureLimit && f < allFigures.size(); f++) {
						Figure newFigure = (Figure) allFigures.get(f);
						if (newFigure.getLength() > clDiam*1.4*2*0.1) {
							cfCount += addFigure(newFigure, figures);
						}				
					}					
					found = new ShapeSet();
					for (int i = 0; i < figures.size(); i++) {
						found.add(((Figure) figures.get(i)).getShapeSet());
					}
					rest = StandardComparator.getLeftOver(rest, found, false);
					
					// the rest that connot be classified as figures...
					if (rest.getNumberOfPoints() > 0) {
						ensureComplexity(500 / clusters.size(), rest);
						
						// 2 possibilities:
						// - every cluster (here cluster really means a bunch of close parts)
						// - every piece for its own
						
						// clustering the rest...
						LinkedList cls = cluster(rest, clDiam, clusterFigureLimit - cfCount);
						Iterator it = cls.iterator();
						Vector restFigsClustered = new Vector();
						while (it.hasNext()) {
							restFigsClustered.add(new Complex((ShapeSet) it.next()));	
						}
						boolean clustered = true;
						
						// every rest piece for its own...
						Vector restFigsAlone = new Vector();
						it = rest.getShapeIterator();
						while (it.hasNext()) {
							ShapeSet set = new ShapeSet();
							set.add((Polyline) it.next());
							restFigsAlone.add(new Complex(set));	
						}
						if (restFigsAlone.size() < 5) {
							double alonePithiness = 0.0;
							{
								int k = restFigsAlone.size();						
								double[] kps = new double[k];
								double[] kpn = new double[k];
								double sizesSum = 0.0;
								double pithinessSum = 0.0;
								for (int i = 0; i < restFigsAlone.size(); i++) {
									Figure f1 = (Figure) restFigsAlone.get(i);
									kpn[i] = kps[i] = f1.pithiness();
									for (int j = 0; j < i; j++) {
										Figure f2 = (Figure) restFigsAlone.get(j);
										double sim = StandardComparator.compare(f1.getShapeSet(), f2.getShapeSet(), true, true, false, true, false, false, false).value[0];
										if (sim > 0.95) {
											kpn[i] = (2*kps[i] + 3) / 5;
											if (kpn[j] == kps[j]) {
												kpn[j] = (2*kps[j] + 3) / 5;
											} else {
												kpn[j] = (2*kpn[j] + 1) / 3;
											}
										}
									}
								}
								for (int i = 0; i < restFigsAlone.size(); i++) {
									Figure f1 = (Figure) restFigsAlone.get(i);
									sizesSum += f1.getSalience() * kpn[i];
									pithinessSum += f1.getSalience() * kpn[i] * kpn[i];
								}
								alonePithiness = pithinessSum / sizesSum / (Math.log(k)/5+1);
							}
							double clusterPithiness = 1.0;
							{
								int k = restFigsClustered.size();						
								double[] kps = new double[k];
								double[] kpn = new double[k];
								double sizesSum = 0.0;
								double pithinessSum = 0.0;
								for (int i = 0; i < restFigsClustered.size(); i++) {
									Figure f1 = (Figure) restFigsClustered.get(i);
									kpn[i] = kps[i] = f1.pithiness();
									for (int j = 0; j < i; j++) {
										Figure f2 = (Figure) restFigsClustered.get(j);
										double sim = StandardComparator.compare(f1.getShapeSet(), f2.getShapeSet(), true, true, false, true, false, false, false).value[0];
										if (sim > 0.95) {
											kpn[i] = (2*kps[i] + 3) / 5;
											if (kpn[j] == kps[j]) {
												kpn[j] = (2*kps[j] + 3) / 5;
											} else {
												kpn[j] = (2*kpn[j] + 1) / 3;
											}
										}
									}
								}
								for (int i = 0; i < restFigsClustered.size(); i++) {
									Figure f1 = (Figure) restFigsClustered.get(i);
									sizesSum += f1.getSalience() * kpn[i];
									pithinessSum += f1.getSalience() * kpn[i] * kpn[i];
								}
								clusterPithiness = pithinessSum / sizesSum / (Math.log(k)/5+1);
							}
							clustered = clusterPithiness > alonePithiness;
						}	
						if (clustered) {
							for (int i = 0; i < restFigsClustered.size(); i++) {
								addFigure((Figure) restFigsClustered.get(i), figures);
							}
						} else {
							for (int i = 0; i < restFigsAlone.size(); i++) {
								addFigure((Figure) restFigsAlone.get(i), figures);
							}
						}					
					} // rest still not empty -> Complex
				} // 2nd round for rest (finding figures in the rest)
			} // rest not empty (finding figures in the rest)
		} // for all clusters 
				
		// standard votes			
		for (int i = 0; i < this.figures.size(); i++) {
			Figure o = (Figure) this.figures.get(i);
			o.idx = i;
			if (o instanceof Convex) {
				((Convex) o).computeStandardVotes();
			}
			if (o instanceof Bunch) {
				((Bunch) o).computeStandardVotes();			
			}
			if (o instanceof Complex) {
				((Complex) o).computeStandardVotes();			
			}
		}
				
		weightFigures();
		relateFigures(true);
		checkFrames();
		findDoublicates();
		hasData = true;		
	}
	
	private void checkFrames() {

		int n = figures.size();
		
		for (int i = 0; i < n; i++) {
			Figure f1 = (Figure) figures.get(i);
			if (f1.frame < Figure.FRAME_THRESHOLD) continue;
			Convex ch1 = Convex.getConvexHull(f1.getShapeSet());
			ch1.inflate(0.05);
			int figsInside = 0;
			for (int j = 0; j < n; j++) {
				if (j == i) continue;
				Relation rel = f1.relatives[j];
				if (rel.centerDistRel > 1.0 || rel.ratio < 0.9) continue;
				Figure f2 = (Figure) figures.get(j);
				Polyline ch2 = Convex.convexHull(f2.getShapeSet());
				if (ch1.arePointsInside(ch2)) {
					if (f2.type <= Figure.CLASSES) {
						figsInside++;
					} else {
						figsInside += 2;
					}
				}
			}
			if (figsInside <= 1) {
				f1.frame = 0.0;
			}
		}
		
		boolean hasNonFrames = false;
		for (int i = 0; i < n; i++) {
			Figure f = (Figure) figures.get(i);
			hasNonFrames |= f.frame < Figure.FRAME_THRESHOLD;
		}
		if (!hasNonFrames) {
			System.err.println("only frames - how comes?!");
			Figure f = (Figure) figures.get(n-1);
			f.frame = 0.0;
		}
	}
	
	private void findDoublicates() {
		if (figures.size() <= 1) {
			return;
		}
		Vector clusters = cluster(figures);
		int n = clusters.size();		
		for (int i = 0; i < n; i++) {
			
			// doublicate clusters...
			Vector c1 = (Vector) clusters.get(i);
			ShapeSet s1 = new ShapeSet();
			for (int ii = 0; ii < c1.size(); ii++) {
				s1.add(((Figure) c1.get(ii)).getShapeSet());
			}
			for (int j = i+1; j < n; j++) {
				Vector c2 = (Vector) clusters.get(j);
				ShapeSet s2 = new ShapeSet();
				for (int jj = 0; jj < c2.size(); jj++) {
					s2.add(((Figure) c2.get(jj)).getShapeSet());
				}
				double sim = StandardComparator.compare(s1, s2);
				if (sim > 0.95) {
					for (int f = 0; f < c2.size(); f++) {
						Figure fig = (Figure) c2.get(f);
						fig.weightAbs *= 0.25;
						for (int rel = 0; rel < fig.relatives.length; rel++) {
							Relation r12 = fig.relatives[rel];
							if (r12 == null) continue;
							Relation r21 = r12.partner;
							r12.weightAbs *= Math.sqrt(0.25);
							r21.weightAbs *= Math.sqrt(0.25);
						}
					}
				}
			}
			
			// doublicate figures...
			for (int f1 = 0; f1 < c1.size(); f1++) {
				Figure fig1 = (Figure) c1.get(f1);
				for (int f2 = f1+1; f2 < c1.size(); f2++) {
					Figure fig2 = (Figure) c1.get(f2);
					Relation r12 = fig1.relatives[fig2.idx];
					Relation r21 = fig2.relatives[fig1.idx];
					if (
							r12.ratio < 0.75 ||
							r21.ratio < 0.75 ||
							r12.centerDistRel < 0.5 ||
							r21.centerDistRel < 0.5 ||
							r12.centerDistRel > 12 && r21.centerDistRel > 12) {
						continue;
					}
					double sim = StandardComparator.compare(fig1.getShapeSet(), fig2.getShapeSet(), false, true, false, true, false, false, false).value[0];
					if (sim > 0.95) {
						fig2.weightAbs *= 0.25;
						for (int rel = 0; rel < fig2.relatives.length; rel++) {
							Relation rr12 = fig2.relatives[rel];
							if (rr12 == null) continue;
							Relation rr21 = rr12.partner;
							rr12.weightAbs *= Math.sqrt(0.25);
							rr21.weightAbs *= Math.sqrt(0.25);
						}
					}
				}
			}
		}		
	}
	
	private Vector cluster(Vector figures) {
		double diam = this.boundingBox.getDiag();		
		int n = figures.size();
		Figure[]   allFigs  = new Figure[n];
		Polyline[]  allCh    = new Polyline[n];
		double[] diams      = new double[n];
		double[][] maxDists = new double[n][n];
		double[][] minDists = new double[n][n];
		boolean[] enqueued  = new boolean[n];
		for (int i = 0; i < n; i++) {
			allFigs[i] = (Figure) figures.get(i);
			allCh[i] = Convex.convexHull(allFigs[i].getShapeSet());
			diams[i] = allCh[i].getDiameter();
			enqueued[i] = false;
			for (int j = 0; j < i; j++) {
				double dmax = RegionImage.distanceFarthest(allCh[i], allCh[j]);
				double dmin = RegionImage.distanceNearest(allCh[i], allCh[j]);
				maxDists[i][j] = maxDists[j][i] = dmax;
				minDists[i][j] = minDists[j][i] = dmin;
			}
		}
		Vector components = new Vector();
		Vector queue = new Vector();
		for (int i = 0; i < n; i++) {
			if (enqueued[i]) {
				continue;
			}
			Vector component = new Vector();
			components.add(component);
			component.add(allFigs[i]);
			queue.add(new Integer(i));
			enqueued[i] = true;
			while (queue.size() > 0) {
				int current = ((Integer) queue.lastElement()).intValue();
				queue.remove(queue.size()-1);
				for (int next = 0; next < n; next++) {
					if (
							!enqueued[next] && (
							maxDists[current][next] - diams[current] - diams[next] < 0.0 ||
							minDists[current][next] < Math.max(diam / 20, (diams[current] + diams[next]) / 5)
							)) {
						component.add(allFigs[next]);
						queue.add(new Integer(next));
						enqueued[next] = true;
					}
				}
			}
		}
		return components;
	}
	
	public static LinkedList cluster(ShapeSet set, double diam, int maxN) {
		LinkedList list = new LinkedList();
		
		if (maxN <= 1 || set.getNumberOfPolylines() <= 1) {
			list.add(set);
			return list;
		}
		
		// clustering nearby components
		int n = set.getNumberOfPolylines();
		Polyline[] allLines = new Polyline[n];
		Polyline[] allCh    = new Polyline[n];
		double[] diams      = new double[n];
		double[][] maxDists = new double[n][n];
		double[][] minDists = new double[n][n];
		boolean[] enqueued  = new boolean[n];
		Iterator it = set.getShapeIterator();
		for (int i = 0; i < n; i++) {
			allLines[i] = (Polyline) it.next();
			ShapeSet s = new ShapeSet();
			s.add(allLines[i]);
			allCh[i] = Convex.convexHull(s);
			diams[i] = allCh[i].getDiameter();
			enqueued[i] = false;
			for (int j = 0; j < i; j++) {
				double dmax = RegionImage.distanceFarthest(allCh[i], allCh[j]);
				double dmin = RegionImage.distanceNearest(allCh[i], allCh[j]);
				maxDists[i][j] = maxDists[j][i] = dmax;
				minDists[i][j] = minDists[j][i] = dmin;
			}
		}
		Vector components = new Vector();
		Vector queue = new Vector();
		for (int i = 0; i < n; i++) {
			if (enqueued[i]) {
				continue;
			}
			Vector component = new Vector();
			components.add(component);
			component.add(allLines[i]);
			queue.add(new Integer(i));
			enqueued[i] = true;
			while (queue.size() > 0) {
				int current = ((Integer) queue.lastElement()).intValue();
				queue.remove(queue.size()-1);
				for (int next = 0; next < n; next++) {
					if (
							!enqueued[next] && (
							maxDists[current][next] - diams[current] - diams[next] < 0.0 ||
							minDists[current][next] < Math.max(diam / 20, (diams[current] + diams[next]) / 5)
							)) {
						component.add(allLines[next]);
						queue.add(new Integer(next));
						enqueued[next] = true;
					}
				}
			}
		}
		
		// clustering until number is sufficiently small
		while (components.size() > maxN) {
			Vector c1 = null;
			Vector c2 = null;
			double minDist = Double.MAX_VALUE;
			for (int i = 0; i < components.size(); i++) {
				Vector v1 = (Vector) components.get(i);
				for (int j = 0; j < i; j++) {
					Vector v2 = (Vector) components.get(j);
					double d = Double.MAX_VALUE;
					for (int k = 0; k < v1.size(); k++) {
						Polyline pl1 = (Polyline) v1.get(k);
						for (int l = 0; l < v2.size(); l++) {
							Polyline pl2 = (Polyline) v2.get(l);
							d = Math.min(d, RegionImage.distanceNearest(pl1, pl2));
						}
					}
					if (d < minDist) {						
						minDist = d;
						c1 = v1;
						c2 = v2;
					}
				}
			}
			c1.addAll(c2);
			components.remove(c2);
		}
		for (int i = 0; i < components.size(); i++) {
			ShapeSet s = new ShapeSet();
			Vector v = (Vector) components.get(i);
			for (int j = 0; j < v.size(); j++) {
				s.add((Polyline) v.get(j));
			}
			list.add(s);
		}		
		return list;
	}
	
	private int addFigure(Figure fig, Vector figs) {
		for (int i = 0; i < figs.size(); i++) {
			Figure oldFig = (Figure) figs.get(i);
			if (oldFig.equals(fig)) {
				oldFig.frame = fig.frame;
				return 0;
			} else if (oldFig.layering(fig)) {
				oldFig.frame = fig.frame;
				return 0;
			}
		}
		if (figs.size() != 0) {
			ShapeSet all = new ShapeSet();
			double salienceSum = 0.0;
			for (int i = 0; i < figs.size(); i++) {
				Figure oldFig = (Figure) figs.get(i);
				all.add(oldFig.getShapeSet());
				salienceSum += oldFig.getSalience();
			}
			double figRep = StandardComparator.representedBy(fig.getShapeSet(), all, true);
			double salience = fig.getSalience();
			if (figRep > 0.75 || salience < salienceSum / 50) {
				return 0;
			}
		}
		figs.add(fig);
		return 1;
	}
	
	public void defineNull() {
		System.out.println("Image defined as null");
		this.hasData = true;
		this.boundingBox = new Box(0,0,1,1);
		Figure f = Figure.circle;
		f.weightAbs = 0.0;
		f.idx = 0;
		this.figures.add(f);
		this.original = f.getShapeSet();
	}
	
	public static Image readSeg(BufferedReader reader) throws IOException {
		Image img = new Image();
		img.figures = new Vector();
		ShapeSet set = new ShapeSet();
		while (reader.ready()) {
			String line = reader.readLine().trim();
			Polyline poly = null;
			if (line.endsWith(")): 0")) {
				poly = new Polyline(false);
			} else if (line.endsWith(")): 1")) {
				poly = new Polyline(true);
			} else {
				continue;
			}
			StringTokenizer tokens = new StringTokenizer(line, " \t:()", false);
			tokens.nextToken();
			while (tokens.countTokens() > 2) {
				String xs = tokens.nextToken();
				String ys = tokens.nextToken();
				double x = Integer.parseInt(xs);
				double y = Integer.parseInt(ys);
				poly.addPoint(new Point(x, y));
			}
			set.add(poly);
		}
		img.original = set;
		Square sq = Square.createSquare(set.getBoundingBox());
		img.boundingBox = sq.getShapeSet().getBoundingBox();
		img.figures.add(sq);
		sq.idx = 0;
		
		img.weightFigures();
		img.relateFigures(true);
		img.hasData = true;		
		
		return img;
	}
	
	
	public static Image readPath(File file, boolean view) throws IOException {
		Image img = new Image();
		img.figures = new Vector();
		img.original = new ShapeSet();
		try {
			Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
			Element doc = document.getDocumentElement();
			
			// src-img
			NodeList srcs = doc.getElementsByTagName("src-img");
			if (srcs.getLength() > 0) {
				img.name = srcs.item(0).getTextContent();
				img.width = Integer.parseInt(srcs.item(0).getAttributes().getNamedItem("width").getNodeValue());
				img.height = Integer.parseInt(srcs.item(0).getAttributes().getNamedItem("height").getNodeValue());
			} else {
				img.name = "";
				img.width = -1;
				img.height = -1;
			}

			// shape
			HashMap map = new HashMap();
			NodeList shapes = doc.getElementsByTagName("shape");
			for (int i = 0; i < shapes.getLength(); i++) {
				ShapeSet shapeSet = new ShapeSet();
				Node shape = shapes.item(i);
				int id = Integer.parseInt(shape.getAttributes().getNamedItem("id").getNodeValue());
				NodeList children = shape.getChildNodes();
				for (int j = 0; j < children.getLength(); j++) {
					Node child = children.item(j);
					if (child.getNodeType() == Node.ELEMENT_NODE) {
						if (((Element) child).getTagName().equals("pl")) {
							String closed = child.getAttributes().getNamedItem("cl").getNodeValue();
							Polyline pl = new Polyline(!closed.equalsIgnoreCase("n"));
							StringTokenizer data = new StringTokenizer(child.getTextContent());
							int minSize = closed.equalsIgnoreCase("d") ? 4 : 2;
							while (data.countTokens() >= minSize) {
								pl.addPoint(new Point(Double.parseDouble(data.nextToken()), Double.parseDouble(data.nextToken())));
							}
							pl = DouglasPeucker.simplify(pl, 1.0);
							shapeSet.add(pl);
							img.original.add(pl);
						} else if (((Element) child).getTagName().equals("pca")) {
							String closed = child.getAttributes().getNamedItem("cl").getNodeValue();
							Polyline pl = new Polyline(!closed.equalsIgnoreCase("n"));
							StringTokenizer data = new StringTokenizer(child.getTextContent());
							int minSize = closed.equalsIgnoreCase("d") ? 6 : 3;
							pl.addPoint(new Point(Double.parseDouble(data.nextToken()), Double.parseDouble(data.nextToken())));
							while (data.countTokens() >= minSize) {
								pl.addPoint(new Point(Double.parseDouble(data.nextToken()), Double.parseDouble(data.nextToken())));
								data.nextToken(); // discard curvature
							}
							pl = DouglasPeucker.simplify(pl, 1.0);
							shapeSet.add(pl);
							img.original.add(pl);
						}
					}
				}
				map.put(new Integer(id), shapeSet);
			}

			// views
			if (view) {
				NodeList views = doc.getElementsByTagName("view");
				if (views.getLength() > 0) {
					Node bestView = views.item(0);
					// todo: find best view...
					StringTokenizer data = new StringTokenizer(bestView.getTextContent());
					while (data.hasMoreTokens()) {
						int idx = Integer.parseInt(data.nextToken());
						Complex c = new Complex((ShapeSet) map.get(new Integer(idx)));
						img.figures.add(c);
					}				
				}
			} else {
				Iterator it = map.values().iterator();
				while (it.hasNext()) {
					Complex c = new Complex((ShapeSet) it.next());
					img.figures.add(c);
				}
			} 
		} catch (Exception ex) {			
		}
		int n = img.figures.size();
		for (int i = 0; i < n; i++) {
			Complex c = (Complex) img.figures.get(i);
			c.idx = i;
			c.relatives = new Relation[n];
			for (int j = 0; j < n; j++) {
				c.relatives[j] = new Relation(i, j);
				c.relatives[j].rel = Relation.REL_DEFAULT;
			}
		}
		img.boundingBox = img.original.getBoundingBox();
		if (img.width == -1) {
			img.width = (int) img.boundingBox.getMaxX(); 
		}
		if (img.height == -1) {
			img.height = (int) img.boundingBox.getMaxY(); 
		}
		
		img.hasData = true;
		return img;
	}

	public static ShapeSet Path2ShapeSet(File file, boolean view) throws IOException {
		ShapeSet set = new ShapeSet();
		try {
			Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
			Element doc = document.getDocumentElement();
			
			// src-img
			Node src = doc.getElementsByTagName("src-img").item(0);
			int height = -1;
			int width = -1;
			String name = "";
			Node widthNode = src.getAttributes().getNamedItem("width");
			if (widthNode != null) {
				width = Integer.parseInt(widthNode.getNodeValue());
			}
			Node heightNode = src.getAttributes().getNamedItem("height");
			if (heightNode != null) {
				height = Integer.parseInt(heightNode.getNodeValue());
			}
			name = src.getTextContent();
			int sepIdx = Math.max(name.lastIndexOf('/'), name.lastIndexOf('\\'));
			name = name.substring(sepIdx+1);
			
			// shape
			HashMap map = new HashMap();
			NodeList shapes = doc.getElementsByTagName("shape");
			for (int i = 0; i < shapes.getLength(); i++) {
				ShapeSet shapeSet = new ShapeSet();
				Node shape = shapes.item(i);
				int id = Integer.parseInt(shape.getAttributes().getNamedItem("id").getNodeValue());
				NodeList children = shape.getChildNodes();
				for (int j = 0; j < children.getLength(); j++) {
					Node child = children.item(j);
					if (child.getNodeType() == Node.ELEMENT_NODE) {
						if (((Element) child).getTagName().equals("pl")) {
							String closed = child.getAttributes().getNamedItem("cl").getNodeValue();
							Polyline pl = new Polyline(!closed.equalsIgnoreCase("n"));
							StringTokenizer data = new StringTokenizer(child.getTextContent());
							int minSize = closed.equalsIgnoreCase("d") ? 4 : 2;
							while (data.countTokens() >= minSize) {
								pl.addPoint(new Point(Double.parseDouble(data.nextToken()), Double.parseDouble(data.nextToken())));
							}
							//pl = DouglasPeucker.simplify(pl, 1.0);
							shapeSet.add(pl);
						} else if (((Element) child).getTagName().equals("pca")) {
							String closed = child.getAttributes().getNamedItem("cl").getNodeValue();
							Polyline pl = new Polyline(!closed.equalsIgnoreCase("n"));
							StringTokenizer data = new StringTokenizer(child.getTextContent());
							int minSize = closed.equalsIgnoreCase("d") ? 6 : 3;
							pl.addPoint(new Point(Double.parseDouble(data.nextToken()), Double.parseDouble(data.nextToken())));
							while (data.countTokens() >= minSize) {
								pl.addPoint(new Point(Double.parseDouble(data.nextToken()), Double.parseDouble(data.nextToken())));
								data.nextToken(); // discard curvature
							}
							//pl = DouglasPeucker.simplify(pl, 1.0);
							shapeSet.add(pl);
						}
					}
				}
				map.put(new Integer(id), shapeSet);
			}

			// views
			if (view) {
				NodeList views = doc.getElementsByTagName("view");
				if (views.getLength() > 0) {
					Node bestView = views.item(0);
					// todo: find best view...
					StringTokenizer data = new StringTokenizer(bestView.getTextContent());
					while (data.hasMoreTokens()) {
						int idx = Integer.parseInt(data.nextToken());
						set.add((ShapeSet) map.get(new Integer(idx)));
					}				
				}
			} else {
				Iterator it = map.values().iterator();
				while (it.hasNext()) {
					set.add((ShapeSet) it.next());
				}
			} 
		} catch (Exception ex) {			
		}
		return set;
	}
	
	public static Image readData(BufferedReader reader) throws IOException {
		String line = reader.readLine().trim();
		Image img = null;
		if (line.equalsIgnoreCase("image description version 01")) {
			img = readData_01(reader);
			img.hasData = true;
		} else if (line.equalsIgnoreCase("<!-- image description version 02 -->")) {
			img = readData_02(reader);
			img.hasData = true;
		}
		return img;
	}
	
	public static Image readData_01(BufferedReader reader) throws IOException {
		Image img = new Image();
		img.figures = new Vector();
		reader.readLine(); // "IMAGE BEGIN"
		String line;
		String[] parts;
		line = reader.readLine().trim();
		parts = line.split(" = ");
		double x1 = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		double y1 = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		double x2 = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		double y2 = Double.parseDouble(parts[1]);
		img.boundingBox = new Box(x1, y1, x2, y2);
		do {
			line = reader.readLine().trim();
			if (line.equalsIgnoreCase("FIGURE begin")) {
				img.figures.add(Figure.readData_01(reader));
			}
			if (line.equalsIgnoreCase("ORIGINAL begin")) {
				img.original = readOriginal_01(reader);
			}		
		} while (!line.equalsIgnoreCase("IMAGE end"));
		img.hasData = true;
		return img;
	}	

	public static Image readData_02(BufferedReader reader) throws IOException {
		Image img = new Image();
		img.figures = new Vector();
		String line;
		String field;
		do {
			line = reader.readLine().trim(); // "<IMAGE...>"
		} while (line.startsWith("<!--"));
		img.name = line.substring(line.indexOf('"') + 1, line.lastIndexOf('"'));
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		double x1 = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		double y1 = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		double x2 = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		double y2 = Double.parseDouble(field);
		img.boundingBox = new Box(x1, y1, x2, y2);
		do {
			line = reader.readLine().trim();
			if (line.equalsIgnoreCase("<FIGURE>")) {
				img.figures.add(Figure.readData_02(reader));
			}
			if (line.equalsIgnoreCase("<ORIGINAL>")) {
				img.original = readOriginal_02(reader);
			}		
		} while (!line.equalsIgnoreCase("</IMAGE>"));
		img.hasData = true;
		return img;
	}

	public static Image readData_03(File file) throws IOException {
		Image img = new Image();
		img.figures = new Vector();
		
		try {
			Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
			Element imgNode = document.getDocumentElement();
			
			// image
			img.name = imgNode.getAttributes().getNamedItem("name").getNodeValue();
			
			int size = 0;
			Node figuresNode = imgNode.getAttributes().getNamedItem("figures");
			if (figuresNode != null) {
				size = Integer.parseInt(figuresNode.getNodeValue());
			}
			Node sizeNode = imgNode.getAttributes().getNamedItem("size");
			if (sizeNode != null) {
				size = Integer.parseInt(sizeNode.getNodeValue());
			}
			
			img.width= -1;
			Node widthNode = imgNode.getAttributes().getNamedItem("width");
			if (widthNode != null) {
				img.width = Integer.parseInt(widthNode.getNodeValue());
			}
			img.height = -1;
			Node heightNode = imgNode.getAttributes().getNamedItem("height");
			if (heightNode != null) {
				img.height = Integer.parseInt(heightNode.getNodeValue());
			}
			
			String version = imgNode.getAttributes().getNamedItem("version").getNodeValue();
			
			NodeList childNodes = imgNode.getChildNodes();
			
			// figures
			int figCount = 0;
			int nodeCount = 0;
			while (figCount < size) {
				Node node = childNodes.item(nodeCount);
				if (node.getNodeType() == Node.ELEMENT_NODE) {
					Element e = (Element) node;
					if (e.getTagName().equals("figure")) {
						img.figures.add(Figure.readData_03(e, size));
						figCount++;
					}
				}
				nodeCount++;				
			}
			
			// original
			while (nodeCount < childNodes.getLength()) {
				Node node = childNodes.item(nodeCount);
				if (node.getNodeType() == Node.ELEMENT_NODE) {
					Element e = (Element) node;
					if (e.getTagName().equals("shape")) {
						img.original = ShapeSet.readData_03(e);
						break;
					}
				}
				nodeCount++;				
			}
		} catch (Exception ex) {
			//ex.printStackTrace();
			System.err.println("Exception: " + ex.getMessage());
			return img;
		}
		if (img.original == null) {
			img.original = new ShapeSet();
		}
		img.boundingBox = img.original.getBoundingBox();
		if (img.width == -1) {
			img.width = (int) img.boundingBox.getMaxX(); 
		}
		if (img.height == -1) {
			img.height = (int) img.boundingBox.getMaxY(); 
		}
		
		img.hasData = true;
		return img;
	}
	
	
	public static ShapeSet readOriginal_01(BufferedReader reader) throws IOException {
		ShapeSet set = new ShapeSet();		
		String line;
		String[] parts;
		boolean end = false;
		do {
			line = reader.readLine().trim();
			end = line.equalsIgnoreCase("ORIGINAL end");
			if (!end) {
				parts = line.split(" = ");
				set.add(Figure.toPolyLine(parts[1]));
			}
		} while (!end);
		set.updateLength();			
		return set;
	}

	public static ShapeSet readOriginal_02(BufferedReader reader) throws IOException {
		ShapeSet set = new ShapeSet();		
		String line;
		boolean end = false;
		do {
			line = reader.readLine().trim();
			end = line.equalsIgnoreCase("</ORIGINAL>");
			if (!end) {
				set.add(Figure.readXml(line));
			}
		} while (!end);
		set.updateLength();			
		return set;
	}
	
	public void writeData(Writer writer, int fn) throws IOException {
		if(!hasData) return;
		writeData_03(writer, fn);
	}
		
	public void writeData_01(Writer writer) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append("image description version 01\n");
		buff.append("IMAGE begin\n");
		writer.write(buff.toString());
		buff.setLength(0);
		String prefix = "\t";
		String pprefix = "\t\t";
		buff.append(prefix); buff.append("bb.x1 = "); buff.append(boundingBox.getMinX()); buff.append("\n");
		buff.append(prefix); buff.append("bb.y1 = "); buff.append(boundingBox.getMinY()); buff.append("\n");
		buff.append(prefix); buff.append("bb.x2 = "); buff.append(boundingBox.getMaxX()); buff.append("\n");
		buff.append(prefix); buff.append("bb.y2 = "); buff.append(boundingBox.getMaxY()); buff.append("\n");
		writer.write(buff.toString());
		buff.setLength(0);

		for (int i = 0; i < figures.size(); i++) {
			Figure fig = (Figure) figures.elementAt(i);
			fig.writeData_01(writer, prefix);
		}
		buff.append(prefix); buff.append("ORIGINAL begin\n");
		Iterator it = original.getShapeIterator();
		while (it.hasNext()) {
			Polyline pl = (Polyline) it.next();		
			buff.append(pprefix); buff.append("polyline = "); buff.append(Figure.toString(pl)); buff.append("\n");
		}
		buff.append(prefix); buff.append("ORIGINAL end\n");
		buff.append("IMAGE end");
		writer.write(buff.toString());
	}

	public void writeData_02(Writer writer, int fn) throws IOException {
		StringBuffer buff = new StringBuffer();
		String prefix = "\t";
		String pprefix = "\t\t";
		if (fn == -1) {
			buff.append("<!-- image description version 02 -->\n");
			buff.append("<IMAGE name = \"");
			buff.append(name);
			buff.append("\">\n");
			buff.append(prefix); buff.append("<bb.x1>"); buff.append(boundingBox.getMinX()); buff.append("</bb.x1>\n");
			buff.append(prefix); buff.append("<bb.y1>"); buff.append(boundingBox.getMinY()); buff.append("</bb.y1>\n");
			buff.append(prefix); buff.append("<bb.x2>"); buff.append(boundingBox.getMaxX()); buff.append("</bb.x2>\n");
			buff.append(prefix); buff.append("<bb.y2>"); buff.append(boundingBox.getMaxY()); buff.append("</bb.y2>\n");
			writer.write(buff.toString());
			buff.setLength(0);
		}
		
		for (int i = 0; i < figures.size(); i++) {
			if (fn != -1 && fn != i) continue;
			Figure fig = (Figure) figures.elementAt(i);
			fig.writeData_02(writer, (fn == -1) ? prefix : "");
		}
		
		if (fn == -1) {
			buff.append(prefix); buff.append("<ORIGINAL>\n");
			Iterator it = original.getShapeIterator();
			while (it.hasNext()) {
				Polyline pl = (Polyline) it.next();
				buff.append(pprefix);
				buff.append(Figure.toXml(pl, false, false));
				buff.append("\n");
			}
			buff.append(prefix); buff.append("</ORIGINAL>\n");
			buff.append("</IMAGE>\n");
			writer.write(buff.toString());
		}
	}

	public void writeData_03(Writer writer, int fn) throws IOException {
		StringBuffer buff = new StringBuffer();
		String prefix = "\t";
		if (fn == -1) {
			buff.append("<!-- image description version 3.4 -->\n");
			buff.append("<image");
			buff.append(" name=\"");
			buff.append(name);
			buff.append("\"");
			buff.append(" width=\"");
			buff.append(width);
			buff.append("\"");
			buff.append(" height=\"");
			buff.append(height);
			buff.append("\"");
			buff.append(" figures=\"");
			buff.append(figures.size());
			buff.append("\"");
			buff.append(" version=\"");
			buff.append("v3.4");
			buff.append("\"");
			buff.append(">\n");
			writer.write(buff.toString());
		}
		
		for (int i = 0; i < figures.size(); i++) {
			if (fn != -1 && fn != i) continue;
			Figure fig = (Figure) figures.elementAt(i);
			fig.writeData_03(writer, (fn == -1) ? prefix : "");
		}
		
		if (fn == -1) {
			original.writeData_03(writer, prefix);
			buff.setLength(0);
			buff.append("</image>\n");
			writer.write(buff.toString());
		}
		writer.flush();
	}

	public void writeImagesData(Writer writer) throws IOException {
		if(!hasData) return;
		
		Box bb = boundingBox;
		int l = (int) (bb.getMaxDim() * 1.1);
		Point c = bb.getCenter();
		int x0 = (int) c.x - l / 2;
		int y0 = (int) c.y - l / 2;
		
		// squares...
		writer.write("x: (" + (x0) + " " + (-y0) + " " + (x0+l) + " " + (-y0) + " " + (x0+l) + " " + (-y0-l) + " " + (x0) + " " + (-y0-l) + "): 1\n");
		writer.write("x: (" + (x0+l) + " " + (-y0) + " " + (x0+l+l) + " " + (-y0) + " " + (x0+l+l) + " " + (-y0-l) + " " + (x0+l) + " " + (-y0-l) + "): 1\n");
		writer.write("x: (" + (x0+l) + " " + (-y0-l) + " " + (x0+l+l) + " " + (-y0-l) + " " + (x0+l+l) + " " + (-y0-l-l) + " " + (x0+l) + " " + (-y0-l-l) + "): 1\n");
		writer.write("x: (" + (x0) + " " + (-y0-l) + " " + (x0+l) + " " + (-y0-l) + " " + (x0+l) + " " + (-y0-l-l) + " " + (x0) + " " + (-y0-l-l) + "): 1\n");
		
		
		Iterator plit;
		
		// original...
		plit = original.getShapeIterator();
		while (plit.hasNext()) {
			Polyline pl = (Polyline) plit.next();
			writer.write("o: ( ");
			Iterator pit = 	pl.getPoints().iterator();
			while (pit.hasNext()) {
				Point p = (Point) pit.next();
				writer.write(" " + p.x + " " + (-p.y));
			}
			writer.write("): " + (pl.isClosed() ? "1" : "0") + "\n");			
		}
		
		// figures...
		for (int i = 0; i < figures.size(); i++) {
			Figure fig = (Figure) figures.get(i);
			if (fig instanceof Complex) continue;
			plit = fig.getShapeSet().getShapeIterator();
			while (plit.hasNext()) {
				Polyline pl = (Polyline) plit.next();
				writer.write("f: ( ");
				Iterator pit = 	pl.getPoints().iterator();
				while (pit.hasNext()) {
					Point p = (Point) pit.next();
					writer.write(" " + p.x + " " + (-p.y-l));
				}
				writer.write("): " + (pl.isClosed() ? "1" : "0") + "\n");			
			}
		}
		
		// rest...
		plit = rest.getShapeIterator();
		while (plit.hasNext()) {
			Polyline pl = (Polyline) plit.next();
			writer.write("o: ( ");
			Iterator pit = 	pl.getPoints().iterator();
			while (pit.hasNext()) {
				Point p = (Point) pit.next();
				writer.write(" " + (p.x+l) + " " + (-p.y-l));
			}
			writer.write("): " + (pl.isClosed() ? "1" : "0") + "\n");			
		}		
	}		

	public boolean hasFrames() {
		boolean frames = false;
		for (int i = 0; i < figures.size(); i++) {
			Figure fig = (Figure) figures.get(i);
			frames |= fig.frame > Figure.FRAME_THRESHOLD;
		}
		return frames;
	}
	
	public ShapeSet getShapeSet() {
		ShapeSet s = new ShapeSet();
		for (int j = 0; j < figures.size(); j++) {
			Figure f = (Figure) figures.get(j);
			s.add(f.getShapeSet());
		}
		return s;
	}
	
	public ShapeSet getShapeSet(boolean noFrames) {
		ShapeSet s = new ShapeSet();
		for (int j = 0; j < figures.size(); j++) {
			Figure f = (Figure) figures.get(j);
			if (noFrames && f.frame > Figure.FRAME_THRESHOLD) {
				continue;
			}
			s.add(f.getShapeSet());
		}
		return s;
	}
	
	public Box getBoundingBox() {
		return boundingBox;
	}
	
	public ShapeSet getOriginal() {
		return this.original;
	}
	
	public static Image dummyImage() {
		Image i = new Image();
		
		Figure f0 = Figure.circle;
		f0.idx = 0;
		i.figures.add(f0);
		Figure f1 = Figure.square;
		f1.idx = 1;
		i.figures.add(f1);
		Figure f2 = Figure.triangle;
		f2.idx = 2;
		i.figures.add(f2);
		i.hasData = true;
		return i;
	}
	
	/*
	public void computeStandardVotes_() {
		// the standard votes
		for (int c = 0; c < components.length; c++) {
			Vector figs = components[c].figures;
			for (int i = 0; i < figs.size(); i++) {
				Figure o = (Figure) figs.get(i);
				if (o instanceof Convex) {
					((Convex) o).computeStandardVotes();
				}
				if (o instanceof Complex) {
					Complex co = (Complex) o;
					co.computeStandardVotes();			
				}
			}
		}
		*/
		/*// the relations
		for (int c = 0; c < components.length; c++) {
			Vector figs = components[c].figures;
			for (int i = 0; i < figs.size(); i++) {
				Figure ft = (Figure) figs.get(i);
				for (int r = 0; r < ft.relatives.length; r++) {
					Relation rel = ft.relatives[r];
					if (rel != null) {
						Figure fo = (Figure) figs.get(rel.f2);
						rel.weightAbs *= Math.min(ft.weightAbs, fo.weightAbs);
					}
				}
			}
		}*/
		
		/*
		// updating inter component relations - now using standard votes
		for (int i = components.length-1; i >= 0; i--) {
			components[i].equality = 0.0;
			components[i].other = -1;
			for (int j = i-1; j >= 0; j--) {
				double s = components[i].similarity(components[j]); 
				if (s > components[i].equality) {
					components[i].equality = s;
					components[i].other = j;
				}
			}
		}
		*/
	//}
	
	public boolean hasData() {
		if (!hasData) return false;
		if (figures.size() == 0) return false;
		return true;		
	}

	// just for temporary implementations of testing routines
	public void test() {
	
		Result r = new Result();
		r.name1 = "n1";
		r.name2 = "n2";
		
		r.hasInfo[Result.ccm] = true;
		r.hasInfo[Result.cpm] = true;
		
		r.t[Result.ccm] = new SimilarityTransformation();
		r.firstSet[Result.ccm] = new ShapeSet(new Polyline(true));
		r.secondSet[Result.ccm] = new ShapeSet(new Polyline(false));
		
		r.pairs[Result.ccm] = new Result.Pair[3];
		r.pairs[Result.ccm][0] = new Result.Pair(0, 1, 0.3, 1);
		r.pairs[Result.ccm][1] = new Result.Pair(1, 0, 0.7, 0.5);
		r.pairs[Result.ccm][2] = new Result.Pair();
		
		r.t[Result.cpm] = new SimilarityTransformation();
		r.firstSet[Result.cpm] = new ShapeSet(new Polyline(false));
		r.secondSet[Result.cpm] = new ShapeSet(new Polyline(true));
		
		r.pairs[Result.cpm] = new Result.Pair[3];
		r.pairs[Result.cpm][0] = new Result.Pair(0, 2, 0.3, 1);
		r.pairs[Result.cpm][1] = new Result.Pair(2, 0, 0.7, 0.5);
		r.pairs[Result.cpm][2] = new Result.Pair();
		
		
		try {
	        FileWriter out = new FileWriter("/home/datsche/scholz/result_000.xml", false);
	        r.writeData(out);
	        out.flush();
	        out.close();
	    } catch (IOException e) {
	    }
		
		/*
		Matchings matchings = new Matchings();
		
		System.out.println(10 + " " + 9);
		long t1 = System.currentTimeMillis();
		matchings.init(10, 9);		
		long m1 = 0;
		while(matchings.hasNext()) {
			int[][] m = matchings.next();
			m1 += m[0][0];
		}
		long t2 = System.currentTimeMillis();
		System.out.println((((double)(t2-t1))/1000 + m1*0.0));

		System.out.println(10 + " " + 10);

		long t3 = System.currentTimeMillis();
		matchings.init(10, 10);
		long m2 = 0;
		while(matchings.hasNext()) {
			int[][] m = matchings.next();
			m2 += m[0][0];
		}
		long t4 = System.currentTimeMillis();
		System.out.println((((double)(t4-t3))/1000 + m2*0.0));
		
		
		///*
		Image otherImage = new Image();
		for (int i = 0; i < 3; i++) {
			Figure f1 = Figure.circle.copy();
			f1.weightAbs = 100;
			otherImage.figures.add(f1);
		}
		otherImage.relateFigures();
		otherImage.hasData = true;
		// /
		
		figures = new Vector();
		int k = 4;
		Figure[] f = new Figure[k];
		for (int i = 0; i < k; i++) {
			f[i] = Figure.circle.copy();
			f[i].idx = i;
			f[i].weightAbs = 1;
			if (i == 1) {
				f[i].weightAbs = 100;
			}
			figures.add(f[i]);
		}
		relateFigures(true);

		for (int i = 0; i < k; i++) {
			f[i].relatives[i] = null;
		}
		hasData = true;
		
		normalizeWeights(0.0, 1.0, 3);
		*/
	}
		
	public static void ensureComplexity(int max, ShapeSet set) {
		
		int k = 1;
		double diag = set.getBoundingBox().getDiag();
		while (set.getNumberOfPoints() > max && k < 10) {
			double eps = diag * epsRatio * k;
			k++;
			ListIterator plIt = set.getShapeIterator();
			boolean keptOne = false;
			while (plIt.hasNext()) {
				Polyline pl = (Polyline) plIt.next();
				double d = pl.getBoundingBox().getDiag();
				if (d < eps * 4 && keptOne) {
					plIt.remove();
				} else {
					keptOne = true;
					Polyline pn = DouglasPeucker.simplify(pl, eps);
					plIt.set(pn);								
				}							
			}
		}
	}
		
	public String[] getFigureList() {
		int n = figures.size();
		String[] figs = new String[n];
		for (int i = 0; i < n; i++) {
			figs[i] = ((Figure) figures.get(i)).getClassName();
		}
		return figs;
	}
	
	public Figure[] getFigures(int from, int to) {
		if (from > this.figures.size()) return new Figure[0];
		int n = this.figures.size();
		int m = to - from + 1;
		int k = Math.min(n - from, m);
		int t = Math.min(n, to+1);
		double[] weights = new double[n];
		for (int i = 0; i < this.figures.size(); i++) {
			weights[i] = ((Figure) this.figures.get(i)).weightAbs;
		}				
		Figure[] figs = new Figure[k];
		for (int i = 0; i < t; i++) {
			int maxIdx = 0;
			for (int j = 0; j < n; j++) {
				if (weights[j] > weights[maxIdx]) {
					maxIdx = j;
				}
			}
			if (i >= from) {
				figs[i - from] = (Figure) this.figures.get(maxIdx);
			}
			weights[maxIdx] = -1.0;
		}		
		return figs;
	}
	
	public boolean ok() {
		boolean hasNonFrames = false;
		for (int i = 0; i < figures.size(); i++) {
			Figure f = (Figure) figures.get(i);
			hasNonFrames |= f.frame < Figure.FRAME_THRESHOLD;
		}
		return hasNonFrames;
	}
	
	public double getFrameRatio() {
		double l1 = 0.0;
		double l2 = 0.0;
		for (int i = 0; i < figures.size(); i++) {
			Figure fig = (Figure) figures.get(i);
			double l = fig.getLength();
			if (fig.frame >= Figure.FRAME_THRESHOLD) {
				l2 += l / fig.degradation;
			}
			else {
				l1 += l;
				l2 += l;
			}
		}		
		return l1/l2;
	}
	
	public void draw(Graphics gr, int w, int h, int fn, boolean colored, boolean fit) {
		if (fn == -1) {
			gr.setColor(Color.white);
			gr.fillRect(0, 0, w, h);
		}
		
		Box bb = this.getShapeSet().getBoundingBox();
		double scale = fit ? Math.min((w-10) / bb.getWidth(), (h-10) / bb.getHeight()) : 1.0;
		int dx = fit ? (int) (-bb.getMinX() * scale) + 5 : 0;
		int dy = fit ? (int) (-bb.getMinY() * scale) + 5 : 0;
		gr.translate(dx, dy);
		if (fn == -1) {
			ShapeSet set = this.getOriginal();
			gr.setColor(Color.lightGray);
			set.draw(gr, scale);
		}
		gr.setColor(Color.black);
		int i = 0;
		for (int f = 0; f < figures.size(); f++) {
			if (fn != -1 && f != fn) continue;
			if (colored) {
				if (fn != -1) {
					gr.setColor(Color.red);
				} else {
					int il = i%8;
					int ih = i/8+1;
					int r = (il%2 == 0) ? 0 : 255/ih;
					int g = ((il/2)%2 == 0) ? 0 : 255/ih;
					int b = (il/4 == 0) ? 0 : 255/ih;
					Color col = new Color(r, g, b);
					gr.setColor(col);
					i++;
					if (i%8==7) i += 2;
				}
			}
			Figure fig = (Figure) figures.get(f);
			ShapeSet set = fig.getShapeSet();
			set.draw(gr, scale);
		}			
	}
}
