/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;

import profi.matching.WeightedClustering.ItemComparator;
import profi.model.Box;
import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;

public class ProbabilisticAlignment {

	public static final int TRANSLATION = 0;
	public static final int TRANSLATION_SCALE = 1;
	public static final int TRANSLATION_ROTATION = 2;
	public static final int TRANSLATION_ROTATION_SCALE = 3;
	public static final int AFFINE = 4;
	public static final int AFFINE_DISTORTION_PENALIZED = 5;
			
	
	private static abstract class Calculator {

		protected static final double DIST_APPROX_FACTOR = (1+Math.sqrt(0.5))/2;
		
		public class Sample {
			public Point p1, p2;
			public double w1;
			public double w2;
			
			public Sample(Point p1, Point p2, double w) {
				this.p1 = p1;
				this.p2 = p2;
				this.w1 = w;
			}
		}
		protected boolean flip;
		
		protected Sample lastSample;
		protected LinkedList points;

		protected double xmin, xminBest;
		protected double xmax, xmaxBest;
		protected double ymin, yminBest;
		protected double ymax, ymaxBest;

		protected int plIdx1;
		protected int plIdx2;
		protected int[] pIdxs1;
		protected int[] pIdxs2;
		protected int pIdx1, pIdx1Best;
		protected int pIdx2, pIdx2Best;
		
		protected int dir;
		
		protected Calculator(boolean flip) {
			this.flip = flip;
			this.points = new LinkedList();
			pIdxs1 = new int[2];
			pIdxs2 = new int[2];
		}
		
		public abstract void reset(double scale);
		
		public void consolidate(boolean turn) {				
			pIdxs1[dir] = pIdx1Best;
			pIdxs2[dir] = pIdx2Best;
			if (turn) {
				this.pIdx1 = pIdx1Best = pIdxs1[1];
				this.pIdx2 = pIdx2Best = pIdxs2[1];	
			} else {
				dir = 1;
			}
		}
		
		public void addFirst(Point p1, Point p2, int plIdx1, int plIdx2, int pIdx1, int pIdx2) {
			this.plIdx1 = plIdx1;
			this.plIdx2 = plIdx2;
			dir = 0;
			this.pIdx1 = pIdx1Best = pIdxs1[1] = pIdx1;
			this.pIdx2 = pIdx2Best = pIdxs2[1] = pIdx2;			
			
			this.lastSample = new Sample(p1, p2, 0.0);
			this.points.add(lastSample);
			
			xmin = xminBest = p2.x;
			xmax = xmaxBest = p2.x;
			ymin = yminBest = p2.y;
			ymax = ymaxBest = p2.y;
		}
		public abstract void add(Point p1, Point p2, double length, int pIdx1, int pIdx2);
		public abstract boolean computeTransform();
		public abstract double getWeightBound();
		public abstract Transformation getTransform();
		
		protected WeightedClustering.Interval[] getInterval() {
			WeightedClustering.Interval[] interval = new WeightedClustering.Interval[2];
			interval[0] = new WeightedClustering.Interval(plIdx1, Math.min(pIdxs1[0], pIdxs1[1]), Math.max(pIdxs1[0], pIdxs1[1]));
			interval[1] = new WeightedClustering.Interval(plIdx2, Math.min(pIdxs2[0], pIdxs2[1]), Math.max(pIdxs2[0], pIdxs2[1]));
			return interval;
		}		
	}

	private static class TranslationCalculator extends Calculator {

		public double abortionThreshold;
		
		private double scale;
		
		private double sumX;
		private double sumY;
		private double sumWeight;
		
		private double sumXBest;
		private double sumYBest;
		private double sumWeightBest;
		
		private int n;
		private int nBest;
				
		private SimilarityTransformation mapBest;
		private double restBest;
		
				
		public TranslationCalculator(boolean flip) {
			super(flip);
			abortionThreshold = 0.2;
		}
		
		public void reset(double scale) {
			this.scale = scale;
			this.sumX = 0.0;
			this.sumY = 0.0;
			this.sumWeight = 0.0;
			this.points.clear();
			this.mapBest = null;

			sumXBest = 0.0;
			sumYBest = 0.0;
			sumWeightBest = 0.0;
						
			restBest = 0.0;

			xmin = xminBest = Double.MAX_VALUE;
			xmax = xmaxBest = -Double.MAX_VALUE;
			ymin = yminBest = Double.MAX_VALUE;
			ymax = ymaxBest = -Double.MAX_VALUE;
						
			this.n = 0;	
			nBest = 0;		
		}

		public void consolidate(boolean turn) {

			super.consolidate(turn);
			if (n == 0) return;
			
			sumX = sumXBest;
			sumY = sumYBest;
			sumWeight = sumWeightBest;

			xmin = xminBest;
			xmax = xmaxBest;
			ymin = yminBest;
			ymax = ymaxBest;
			
			n = nBest;
			while (this.points.size() > n + 1) {
				// todo: want to cut the list!
				this.points.removeLast();
			}
			((Sample) this.points.getLast()).w2 = 0.0;
			if (turn) {
				this.lastSample = (Sample) this.points.getFirst();
				this.lastSample.w1 = this.lastSample.w2;
				this.lastSample.w2 = 0.0;
			}
		}

		public void add(Point p1, Point p2, double length, int pIdx1, int pIdx2) {
			if (length <= 1e-6) return;

			this.pIdx1 = pIdx1;
			this.pIdx2 = pIdx2;
			if (flip) {
				this.sumX += (p2.x + p1.x*this.scale + this.lastSample.p2.x + this.lastSample.p1.x*this.scale) * length/2;
			} else {
				this.sumX += (p2.x - p1.x*this.scale + this.lastSample.p2.x - this.lastSample.p1.x*this.scale) * length/2;
			}
			this.sumY += (p2.y - p1.y*this.scale + this.lastSample.p2.y - this.lastSample.p1.y*this.scale) * length/2;
			this.sumWeight += length;
			
			if (p2.x < xmin) xmin = p2.x;
			else if (p2.x > xmax) xmax = p2.x;
			if (p2.y < ymin) ymin = p2.y;
			else if (p2.y > ymax) ymax = p2.y;
			
			this.lastSample.w2 = length/2;
			this.lastSample = new Sample(p1, p2, length/2);
			this.points.add(this.lastSample);
			n++;
		}

		public boolean computeTransform() {
			
			if (n==0) return true;
			
			double dx = this.sumX / this.sumWeight;
			double dy = this.sumY / this.sumWeight;
			
			double xError = 0.0;
			if (flip) {
				xError = Math.abs(-lastSample.p1.x * scale + dx - lastSample.p2.x);
			} else {
				xError = Math.abs(lastSample.p1.x * scale + dx - lastSample.p2.x);
			}
		    double yError = Math.abs(lastSample.p1.y * scale + dy - lastSample.p2.y);
		    
		    double bb = (xmax-xmin) + (ymax-ymin);

		    double sqError = xError * xError + yError * yError;
		    double tolerance = bb * abortionThreshold;
		    double sqTolerance = tolerance * tolerance;

//	System.out.println(sumWeight + " " + bb + " " + Math.sqrt(sqError));
		    
		    if (sqTolerance < sqError) {
		    	return false;
		    }

		    double approxError = (xError + yError) * DIST_APPROX_FACTOR;
		    //double rest = (sqTolerance - sqError) / (tolerance + approxError);
		    double rest = tolerance - 2*approxError + (sqError / tolerance);
		    		    		    		    
		    if (rest < restBest) {
		    	return true;
		    }
		    
		    pIdx1Best = pIdx1;
		    pIdx2Best = pIdx2;
		    
		    nBest = n;
		    sumXBest = sumX;
		    sumYBest = sumY;
		    sumWeightBest = sumWeight;
		    restBest = rest;
		    xminBest = xmin;
		    xmaxBest = xmax;
		    yminBest = ymin;
		    ymaxBest = ymax;
		    mapBest = new SimilarityTransformation(flip, scale, 0, dx, dy);
		    return true;
		}

		public double getWeightBound() {
			return sumWeight;
		}
		
		public Transformation getTransform() {
			if (mapBest == null) {
				SimilarityTransformation ret = new SimilarityTransformation(flip, 1.0, 0.0, 0.0, 0.0);
				ret.setItemSource(new Integer(1));
				return ret;
			}
			
			double dx = mapBest.getDx();
			double dy = mapBest.getDy();
			
			//TODO: weight Transform;		    
		    double sumSqE = 0;
		    Iterator it = points.iterator();
		    while (it.hasNext()) {
		    	Sample s = (Sample) it.next();
		    	sumSqE += (s.w1 + s.w2) * (
		    			(flip ? (-s.p1.x*scale + dx - s.p2.x) * (-s.p1.x*scale + dx - s.p2.x) : (s.p1.x*scale + dx - s.p2.x) * (s.p1.x*scale + dx - s.p2.x)) +
			    		(s.p1.y*scale + dy - s.p2.y) * (s.p1.y*scale + dy - s.p2.y));
		    }
		    
		    sumSqE /= sumWeight;

		    double bb = (xmax-xmin)*(xmax-xmin) + (ymax-ymin)*(ymax-ymin);
		    double e=Math.sqrt(sumSqE / bb);
//			    System.out.println("e: " + e);
		    
		    //System.out.println("n : " + n + " e   : " + e + " e/l : " + e/sumLength);
		    //System.out.println("n : " + n + " e   : " + Math.sqrt(error) + " e/l : " + Math.sqrt(error)/sumLength);
		    
		    //TODO: find best coefficients!
		    this.mapBest.weight = sumWeight / (1 + 5*e);

		    if (Double.isNaN(mapBest.weight)) {
		    	System.out.println("");
		    }
		    mapBest.setItemSource(getInterval());
			return this.mapBest;
		}
	}
	
	private class RigidCalculator extends Calculator {

		public double abortionThreshold;
		
		double scale;
				
		private double sumxsw, sumysw;
		private double sumxtw, sumytw;
		private double sumxtysw, sumytxsw;
		private double sumxtxsw, sumytysw;
		
		private double sumxswBest, sumyswBest;
		private double sumxtwBest, sumytwBest;
		private double sumxtyswBest, sumytxswBest;
		private double sumxtxswBest, sumytyswBest;
		
		private double sumWeight, sumWeightBest;

		int n;
		private int nBest;
		private double restBest;
		private SimilarityTransformation mapBest;
		
		public RigidCalculator(boolean flip) {
			super(flip);
			abortionThreshold = 0.2;
		}
		
		public void reset(double scale) {
			n = 0;
			this.scale = scale;
			sumxsw = 0.0; sumysw = 0.0;
			sumxtw = 0.0; sumytw = 0.0;
			sumxtysw = 0.0; sumytxsw = 0.0;
			sumxtxsw = 0.0; sumytysw = 0.0;
			this.sumWeight= 0.0;
			this.restBest = 0.0;
			this.nBest = 0;
			sumxswBest = 0.0; sumyswBest = 0.0;
			sumxtwBest = 0.0; sumytwBest = 0.0;
			sumxtyswBest = 0.0; sumytxswBest = 0.0;
			sumxtxswBest = 0.0; sumytyswBest = 0.0;
			this.sumWeightBest= 0.0;
			points.clear();
			mapBest = null;

			xmin = xminBest = 1e12;
			xmax = xmaxBest = -1e12;
			ymin = yminBest = 1e12;
			ymax = ymaxBest = -1e12;
		}
				
		public void consolidate(boolean turn) {

			super.consolidate(turn);
			if (n == 0) return;
			
			n = nBest;
			sumxsw = sumxswBest; sumysw = sumyswBest;
			sumxtw = sumxtwBest; sumytw = sumytwBest;
			sumxtysw = sumxtyswBest; sumytxsw = sumytxswBest;
			sumxtxsw = sumxtxswBest; sumytysw = sumytyswBest;
			this.sumWeight = this.sumWeightBest;

			xmin = xminBest;
			xmax = xmaxBest;
			ymin = yminBest;
			ymax = ymaxBest;
			
			((Sample) this.points.getLast()).w2 = 0.0;
			if (turn) {
				this.lastSample = (Sample) this.points.getFirst();
				this.lastSample.w1 = this.lastSample.w2;
				this.lastSample.w2 = 0.0;
			}
		}
		
		public void add(Point p1, Point p2, double length, int pIdx1, int pIdx2) {
			if (length <= 1e-6) return;

			this.pIdx1 = pIdx1;
			this.pIdx2 = pIdx2;
			
			double xs = flip ? -p1.x * scale : p1.x * scale;
			double ys = p1.y * scale;
			double xt = p2.x;
			double yt = p2.y;
			double xsl = flip ? -lastSample.p1.x * scale : lastSample.p1.x * scale;
			double ysl = lastSample.p1.y * scale;
			double xtl = lastSample.p2.x;
			double ytl = lastSample.p2.y;
			sumxsw += (xs+xsl)/2*length;
			sumysw += (ys+ysl)/2*length;
			sumxtw += (xt+xtl)/2*length;
			sumytw += (yt+ytl)/2*length;
			sumxtysw += (xt*ys+xtl*ysl)/2*length;
			sumytxsw += (yt*xs+ytl*xsl)/2*length;
			sumxtxsw += (xt*xs+xtl*xsl)/2*length;
			sumytysw += (yt*ys+ytl*ysl)/2*length;
			this.sumWeight += length;

			if (p2.x < xmin) xmin = p2.x;
			else if (p2.x > xmax) xmax = p2.x;
			if (p2.y < ymin) ymin = p2.y;
			else if (p2.y > ymax) ymax = p2.y;
						
			this.lastSample.w2 = length/2;
			this.lastSample = new Sample(p1, p2, length/2);
			this.points.add(this.lastSample);
			n++;
		}
		
		public boolean computeTransform() {
			if (n < 1) {
				mapBest = null;
				return true;
			}
			double xsm = sumxsw / sumWeight;
			double ysm = sumysw / sumWeight;
			double xtm = sumxtw / sumWeight;
			double ytm = sumytw / sumWeight;

			double p =
				sumxtysw - sumxtw*ysm -
				sumytxsw + sumytw*xsm;
			double q =
				sumxtxsw - sumxtw*xsm +
				sumytysw - sumytw*ysm;
			
			double l = Math.sqrt(p*p + q*q);
			if (l < 1e-6) {
				mapBest = null;
				return true;				
			}
			q /= l;
			p /= l;
			double a = q * scale;
			double b = p * scale;

		    double dx  = xtm - (q*xsm  + p*ysm);
		    double dy  = ytm - (-p*xsm  + q*ysm);
		    
		    //consistency...
		    double lx = lastSample.p1.x * (flip ? -1 : 1);
		    double ly = lastSample.p1.y;
		    double tx =  a * lx + b * ly + dx;
		    double ty = -b * lx + a * ly + dy;
		    double xError = Math.abs(tx - lastSample.p2.x);
		    double yError = Math.abs(ty - lastSample.p2.y);
		    double sqError = xError * xError + yError * yError;

		    double bb = (xmax - xmin) + (ymax - ymin);

		    double tolerance = bb * abortionThreshold;
		    
		    double sqTolerance = tolerance * tolerance;

//	System.out.println(sumWeight + " " + bb + " " + Math.sqrt(sqError));
		    
		    if (sqTolerance < sqError) {
		    	return false;
		    }

		    double approxError = (xError + yError) * DIST_APPROX_FACTOR;
		    //double rest = (sqTolerance - sqError) / (tolerance + approxError);
		    double rest = tolerance - 2*approxError + (sqError / tolerance);
		    		    
		    if (rest < restBest) {
		    	return true;
		    }

			this.pIdx1Best = pIdx1;
			this.pIdx2Best = pIdx2;
			
		    nBest = n;
			sumxswBest = sumxsw;
			sumyswBest = sumysw;
			sumxtwBest = sumxtw;
			sumytwBest = sumytw;
			sumxtyswBest = sumxtysw;
			sumytxswBest = sumytxsw;
			sumxtxswBest = sumxtxsw;
			sumytyswBest = sumytysw;
		    sumWeightBest = sumWeight;

		    xminBest = xmin;
		    xmaxBest = xmax;
		    yminBest = ymin;
		    ymaxBest = ymax;
		    restBest = rest;
		    mapBest = new SimilarityTransformation(flip, q*scale, p*scale, dx, dy);

		    if (Double.isNaN(q)) {
		    	System.out.println("");
		    }
		    return true;		    
		}
		
		public double getWeightBound() {
			return sumWeight;
		}
		
		public Transformation getTransform() {
			if (mapBest == null) {			
				SimilarityTransformation ret = new SimilarityTransformation(flip, 1.0, 0.0, 0.0, 0.0);
				ret.setItemSource(new Integer(1));
				return ret;
			}
			
			//TODO: weight Transform;		    
		    double sumSqE = 0;
		    Iterator it = points.iterator();
		    while (it.hasNext()) {
		    	Sample s = (Sample) it.next();
			    double tx = mapBest.getX(s.p1);
			    double ty = mapBest.getY(s.p1);
			    double xError = Math.abs(tx - s.p2.x);
			    double yError = Math.abs(ty - s.p2.y);
			    sumSqE += (s.w1 + s.w2) * (xError*xError + yError*yError);
		    }
		    
		    sumSqE /= sumWeightBest;

		    double bb = (xmax-xmin)*(xmax-xmin) + (ymax-ymin)*(ymax-ymin);
		    double e=Math.sqrt(sumSqE / bb);
		    
		    //TODO: find best coefficients!
		    this.mapBest.weight = sumWeightBest / (1 + 5*e);
		    if (Double.isNaN(mapBest.weight)) {
		    	System.out.println("");
		    }
		    mapBest.setItemSource(getInterval());
			return this.mapBest;
		}
	}

	public static class SimilarityCalculator extends Calculator {

		public double abortionThreshold;
		
		double scale;
				
		private double sumxsw, sumysw;
		private double sumxtw, sumytw;
		private double sumxsxsw, sumysysw;	
		private double sumxtysw, sumytxsw;
		private double sumxtxsw, sumytysw;
		
		private double sumxswBest, sumyswBest;
		private double sumxtwBest, sumytwBest;
		private double sumxsxswBest, sumysyswBest;
		private double sumxtyswBest, sumytxswBest;
		private double sumxtxswBest, sumytyswBest;
		
		private double sumWeight, sumWeightBest;

		int n;
		private int nBest;
		private double restBest;

		private double[][] m;
		private double[][] l;
		private double[][] u;
		private double[] b;
		private double[] x;
		private double[] xBest;
		private boolean hasBest;
				
		public SimilarityCalculator(boolean flip) {
			super(flip);
			abortionThreshold = 0.2;

			m = new double[4][4];
			l = new double[4][4];
			u = new double[4][4];
			b = new double[4];
			x = new double[4];
			xBest = new double [4];
		}
		
		public void reset(double scale) {
			n = 0;
			this.points.clear();
			this.lastSample = null;
			this.scale = scale;
			sumxsw = 0.0; sumysw = 0.0;
			sumxtw = 0.0; sumytw = 0.0;
			sumxsxsw = 0.0; sumysysw = 0.0;
			
			sumxtysw = 0.0; sumytxsw = 0.0;
			sumxtxsw = 0.0; sumytysw = 0.0;
			this.sumWeight= 0.0;
			this.restBest = -Double.MAX_VALUE;
			this.nBest = 0;
			sumxswBest = 0.0; sumyswBest = 0.0;
			sumxtwBest = 0.0; sumytwBest = 0.0;
			sumxsxswBest = 0.0; sumysyswBest = 0.0;
			sumxtyswBest = 0.0; sumytxswBest = 0.0;
			sumxtxswBest = 0.0; sumytyswBest = 0.0;
			this.sumWeightBest= 0.0;
			hasBest = false;

			xmin = xminBest = 1e12;
			xmax = xmaxBest = -1e12;
			ymin = yminBest = 1e12;
			ymax = ymaxBest = -1e12;
		}
				
		public void consolidate(boolean turn) {

			super.consolidate(turn);
			if (n == 0) return;
			
			n = nBest;
			sumxsw = sumxswBest; sumysw = sumyswBest;
			sumxtw = sumxtwBest; sumytw = sumytwBest;
			sumxsxsw = sumxsxswBest; sumysysw = sumysyswBest;
			sumxtysw = sumxtyswBest; sumytxsw = sumytxswBest;
			sumxtxsw = sumxtxswBest; sumytysw = sumytyswBest;
			this.sumWeight = this.sumWeightBest;

			xmin = xminBest;
			xmax = xmaxBest;
			ymin = yminBest;
			ymax = ymaxBest;
			
			while (this.points.size() > n + 1) {
				// todo: want to cut the list!
				this.points.removeLast();
			}
			((Sample) this.points.getLast()).w2 = 0.0;
			if (turn) {
				this.lastSample = (Sample) this.points.getFirst();
				this.lastSample.w1 = this.lastSample.w2;
				this.lastSample.w2 = 0.0;
			}
		}
		
		public void add(Point p1, Point p2, double length, int pIdx1, int pIdx2) {

			this.pIdx1 = pIdx1;
			this.pIdx2 = pIdx2;
			
			double xs = flip ? -p1.x * scale : p1.x * scale;
			double ys = p1.y * scale;
			double xt = p2.x;
			double yt = p2.y;
			double xsl = flip ? -lastSample.p1.x * scale : lastSample.p1.x * scale;
			double ysl = lastSample.p1.y * scale;
			double xtl = lastSample.p2.x;
			double ytl = lastSample.p2.y;
			sumxsw += (xs+xsl)/2*length;
			sumysw += (ys+ysl)/2*length;
			sumxtw += (xt+xtl)/2*length;
			sumytw += (yt+ytl)/2*length;
			sumxsxsw += (xs*xs+xsl*xsl)/2*length;
			sumysysw += (ys*ys+ysl*ysl)/2*length;
			sumxtysw += (xt*ys+xtl*ysl)/2*length;
			sumytxsw += (yt*xs+ytl*xsl)/2*length;
			sumxtxsw += (xt*xs+xtl*xsl)/2*length;
			sumytysw += (yt*ys+ytl*ysl)/2*length;
			this.sumWeight += length;

			if (p2.x < xmin) xmin = p2.x;
			else if (p2.x > xmax) xmax = p2.x;
			if (p2.y < ymin) ymin = p2.y;
			else if (p2.y > ymax) ymax = p2.y;
			
			this.lastSample.w2 = length/2;
			this.lastSample = new Sample(p1, p2, length/2);
			this.points.add(this.lastSample);
			n++;
		}
		
		public boolean computeTransform() {
			if (n < 1) {
				return true;
			}
			
			m[0][0] = m[1][1] = sumxsxsw + sumysysw;
			m[0][1] = m[1][0] = 0.0;
			m[0][2] = sumxsw;
			m[1][3] = -sumxsw;
			m[0][3] = m[1][2] = sumysw;
			
			m[2][0] = sumxsw;
			m[3][1] = -sumxsw;
			m[2][1] = m[3][0] = sumysw;
			m[2][2] = m[3][3] = sumWeight;
			m[2][3] = m[3][2] = 0.0;

			b[0] = sumxtxsw + sumytysw;
			b[1] = sumxtysw - sumytxsw;
			b[2] = sumxtw;
			b[3] = sumytw;
			

			decompose(m, l, u);
			solve(l, u, b, x);
			
			// too big differences in scale means bad match...
			double sqScale = (x[0]*x[0] + x[1]*x[1]);
						
			x[0] *= scale;
			x[1] *= scale;
			
		    //consistency...
		    double lx = lastSample.p1.x * (flip ? -1 : 1);
		    double ly = lastSample.p1.y;
		    double tx =  x[0] * lx + x[1] * ly + x[2];
		    double ty = -x[1] * lx + x[0] * ly + x[3];
		    double xError = Math.abs(tx - lastSample.p2.x);
		    double yError = Math.abs(ty - lastSample.p2.y);
		    double sqError = xError * xError + yError * yError;
		    
		    double bb = (xmax - xmin) + (ymax - ymin);
		    double tolerance = bb * abortionThreshold;
//			    double tolerance = sumWeight * abortionThreshold;
		    double sqTolerance = tolerance * tolerance;
		    
		    if (sqTolerance < sqError && restBest > -1.0) {
		    	return false;
		    }

		    double approxError = (xError + yError) * DIST_APPROX_FACTOR;
		    double rest = (sqTolerance - sqError) / (tolerance + approxError);
		    //double rest = tolerance - 2*approxError + (sqError / tolerance);
		    		    
		    if (rest < restBest) {
		    	return true;
		    }
		    this.pIdx1Best = pIdx1;
			this.pIdx2Best = pIdx2;
			
		    nBest = n;
			sumxswBest = sumxsw;
			sumyswBest = sumysw;
			sumxtwBest = sumxtw;
			sumytwBest = sumytw;
			sumxsxswBest = sumxsxsw;
			sumysyswBest = sumysysw;
			sumxtyswBest = sumxtysw;
			sumytxswBest = sumytxsw;
			sumxtxswBest = sumxtxsw;
			sumytyswBest = sumytysw;
		    sumWeightBest = sumWeight;
		    restBest = rest;
		    xBest[0] = x[0];
		    xBest[1] = x[1];
		    xBest[2] = x[2];
		    xBest[3] = x[3];
		    hasBest = true;

		    xminBest = xmin;
		    xmaxBest = xmax;
		    yminBest = ymin;
		    ymaxBest = ymax;
		    //mapBest = new WeightedSimilarity(q, p, scale, dx, dy, 0);
    	
		    if (sqScale < 0.25 || sqScale > 4) {
				return false;
			}
		    
		    return true;		    
		}

		//	 decompose
		private void decompose(double[][] M, double[][] L, double[][] U){
			
		    // Code: Cormen et al., page 756
		    int i, j, k;
		    for ( k = 0; k < M.length; ++k) {
		        U[ k][ k] = M[ k][ k];
		        for ( i = k+1; i < M.length; ++i) {
		            L[ i][ k] = M[ i][ k] / U[ k][ k];
		            U[ k][ i] = M[ k][ i];
		        }
		        for( i = k+1; i < M.length; ++i) {
		            for( j = k+1; j < M.length; ++j) {
		                M[ i][ j] = M[ i][ j] - L[ i][ k]*U[ k][ j];
		            }
		        }
		    }
		}

		// solve
		private void solve(double[][] L, double[][] U, double[] b, double[] x) {
			
		    // Code: Cormen et al., page 756
		    double[] y = new double[L.length];
		    int i, j;

		    // forward substitution
		    for ( i = 0; i < L.length; ++i) {
		        y[ i] = b[ i];
		        for ( j = 0; j < i; ++j) {
		            y[ i] -= L[ i][ j] * y[ j];
		        }
		    }

		    // back substitution
		    for ( i = L.length-1; i >= 0; --i) {
		        x[ i] = y[ i];
		        for ( j = i+1; j < L.length; ++j) {
		            x[ i] -= U[ i][ j] * x[ j];
		        }
		        x[i] /= U[ i][ i];
		    }
		}
		
		public double getWeightBound() {
			return sumWeight;
		}
		
		/*
		protected WeightedSimilarity getPreTransform() {
			if (n < 1) {
				return null;
			}
		
			m[0][0] = m[1][1] = sumxsxsw + sumysysw;
			m[0][1] = m[1][0] = 0.0;
			m[0][2] = sumxsw;
			m[1][3] = -sumxsw;
			m[0][3] = m[1][2] = sumysw;
			
			m[2][0] = sumxsw;
			m[3][1] = -sumxsw;
			m[2][1] = m[3][0] = sumysw;
			m[2][2] = m[3][3] = sumWeight;
			m[2][3] = m[3][2] = 0.0;
	
			b[0] = sumxtxsw + sumytysw;
			b[1] = sumxtysw - sumytxsw;
			b[2] = sumxtw;
			b[3] = sumytw;			
	
			decompose(m, l, u);
			solve(l, u, b, x);
			
			// too big differences in scale means bad match...
			double sqScale = (x[0]*x[0] + x[1]*x[1]);
			if (sqScale < 0.25 || sqScale > 4) {
				return null;
			}
					
			return new WeightedSimilarity(x[0], x[1], scale, x[2], x[3], 0.0);			
		}
		*/
		
		public Transformation getTransform() {
			if (!hasBest) {
				SimilarityTransformation ret = new SimilarityTransformation(flip, 1.0, 0.0, 0.0, 0.0);
				ret.setItemSource(new Integer(1));
				return ret;
			}
			
			double a  = xBest[0];
			double b  = xBest[1];
			double dx = xBest[2];
			double dy = xBest[3];
			
			//TODO: weight Transform;		    
		    double sumSqE = 0;
		    Iterator it = points.iterator();
		    while (it.hasNext()) {
		    	Sample s = (Sample) it.next();
			    double lx = flip ? -s.p1.x : s.p1.x;
			    double ly = s.p1.y;
			    double tx =  a * lx + b * ly + dx;
			    double ty = -b * lx + a * ly + dy;
			    double xError = Math.abs(tx - lastSample.p2.x);
			    double yError = Math.abs(ty - lastSample.p2.y);
			    sumSqE += (s.w1 + s.w2) * (xError*xError + yError*yError);
		    }

		    sumSqE /= sumWeightBest;

		    double bb = (xmax-xmin)*(xmax-xmin) + (ymax-ymin)*(ymax-ymin);
		    double e=Math.sqrt(sumSqE / bb);
		    SimilarityTransformation mapBest = new SimilarityTransformation(flip, a, b, dx, dy);
		    
		    //TODO: find best coefficients!
		    mapBest.weight = sumWeight / (1 + 5*e);
		    if (Double.isNaN(mapBest.weight)) {
		    	System.out.println("");
		    }
		    mapBest.setItemSource(getInterval());
			return mapBest;
		}
	}
	
	public static String TransformationType2String(int type) {		
		switch (type) {
		case TRANSLATION:
			return "translation";
		case TRANSLATION_SCALE:
			return "translation-scale";
		case TRANSLATION_ROTATION:
			return "translation-rotation";
		case TRANSLATION_ROTATION_SCALE:
			return "translation-rotation-scale";
		default:
			return "unknown";
		}		
	}
	
	private static double alignmentThreshold = 0.5; // sbs tests original
	//private static double alignmentThreshold = 0.0;
	public static double scaleThreshold = 2.0;
	public static double mapThreshold = 0.25;
	public static int mapCount = 20;
	
	private static int f_;
	private static int r_;
	private static int n_;
	
	/*
	public static SimilarityTransformation[] computeTransformations(
			ShapeSet start, ShapeSet goal, boolean flips, int transformationType, int samples, boolean forceRotation)  {
				
		ProbabilisticAlignment pa = new ProbabilisticAlignment(
				start, goal, transformationType);
		pa.forceRotation = forceRotation;
		LinkedList results = new LinkedList();
		for (f_ = 0; f_ < (flips ? 2 : 1); f_++) {
			pa.reset();
			pa.compute(samples);
			results.addAll(pa.getMaps(mapThreshold, mapCount));
			pa.flip(true);
		}
		Collections.sort(results, new ItemComparator(ItemComparator.DESCENDING));
		Iterator it = results.iterator();
		SimilarityTransformation[] ret = new SimilarityTransformation[Math.min(results.size(), mapCount)];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = (SimilarityTransformation) it.next();
		}
		return ret;
	}
	*/
	
	public static long t1, t2, t3, t4;
	
	
	public static SimilarityTransformation[] computeTransformations(
			ShapeSet start, ShapeSet goal, boolean flips, int transformationType, int rounds, int samples, boolean forceRotation)  {
				
		WeightedClustering results = new WeightedClustering(new SimilarityTransformation());
		results.weightThreshold = 0.0;
		results.setDiam(goal.getBoundingBox().getMaxDim());
		ProbabilisticAlignment pa = new ProbabilisticAlignment(
				start, goal, transformationType);
		pa.forceRotation = forceRotation;
		for (f_ = 0; f_ < (flips ? 2 : 1); f_++) {
			for (r_ = 0; r_ < rounds; r_++) {
				pa.reset();
				pa.compute(samples);
				Iterator it = pa.getMaps(mapThreshold, mapCount).iterator();
				int i = 0;
				while (it.hasNext()) {
					results.add((SimilarityTransformation) it.next());
					i++;
				}
			}			
			pa.flip(true);
		}
		results.performClustering();
		LinkedList ll = results.getItems(0.0, Integer.MAX_VALUE, true);
		Iterator it = ll.iterator();
		SimilarityTransformation[] ret = new SimilarityTransformation[ll.size()];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = (SimilarityTransformation) it.next();
		}
		return ret;
	}
	
	public static SimilarityTransformation[] dummyTransformation(
			ShapeSet start, ShapeSet goal, boolean flip, int transformationType, boolean forceRotation)  {
		SimilarityTransformation[] ret = new SimilarityTransformation[1];
		Box bb1 = start.getBoundingBox();
		Box bb2 = goal.getBoundingBox();
		Point c1 = bb1.getCenter();
		Point c2 = bb2.getCenter();
		double scale = (transformationType == TRANSLATION_SCALE || transformationType == TRANSLATION_ROTATION_SCALE) ? bb2.getDiag() / bb1.getDiag() : 1.0;
		double angle = (forceRotation) ? Math.PI / 2 : 0.0;
		SimilarityTransformation sim = SimilarityTransformation.explicit(flip, scale,angle, 0, 0);
		sim.applyTo(c1);
		ret[0] = new SimilarityTransformation(sim, SimilarityTransformation.translation(c2.x-c1.x, c2.y-c1.y));
		return ret;
	}
		
	
	public boolean finding;
	public boolean forceRotation;
	
	private int transformationType;
	private boolean useScale;
	private double[][] bestScales;
	
	private ShapeSet startShape;
	private ShapeSet goalShape;
	private Point[][] startPoints;
	private Point[][] goalPoints;
	private double[][] startDists;
	private double[][] goalDists;
	private double[][] startDistsN;
	private double[][] goalDistsN;
	
	private double nextStartDist;
	private double nextGoalDist;
	
	private int nGoalPoints;
	private int nStartPoints;
	
	private double startDiam;
	private double goalDiam;
	
	private double prescaling;
	private double minScale;
	private double maxScale;

	private int totalSteps;
	private int currentSteps;
	int scaleSteps;
	int nextScaleIdx;

	
	/**
	 * Diameter of the shapes.
	 */
	private double diam;

	private Calculator calc;
	private WeightedClustering clustering;
	private static Random randomGenerator = new Random();
	private static Random randomGenerator2 = new Random();

	
	public void flip(boolean flip) {
		switch (transformationType) {
		case TRANSLATION:
			calc = new TranslationCalculator(flip);
			this.useScale = false;
			break;
		case TRANSLATION_SCALE:
			calc = new TranslationCalculator(flip);
			this.useScale = true;
			break;
		case TRANSLATION_ROTATION:
			calc = new RigidCalculator(flip);
			this.useScale = false;
			break;
		case TRANSLATION_ROTATION_SCALE:
			calc = new SimilarityCalculator(flip);
			this.useScale = true;
			break;
		default:
			throw new RuntimeException("unknown transformation type");
		}
		
		Box rect1 = this.startShape.getBoundingBox();
				
		Point[] points = new Point[4];
		points[0] = new Point(rect1.getMinX(), rect1.getMinY());
		points[1] = new Point(rect1.getMaxX(), rect1.getMinY());
		points[2] = new Point(rect1.getMaxX(), rect1.getMaxY());
		points[3] = new Point(rect1.getMinX(), rect1.getMaxY());
		SimilarityTransformation.setPoints(points);
		clustering.setDiam(diam);
		
		Box rect2 = this.goalShape.getBoundingBox();
		
		Point cbb1 = rect1.getCenter();
		if (flip) {
			cbb1.x = -cbb1.x;
		}
		Point cbb2 = rect2.getCenter();
		
		SimilarityTransformation ref = new SimilarityTransformation(flip, 1.0, 0.0, cbb2.x-cbb1.x, cbb2.y-cbb1.y);
		double maxDist = (rect1.getMaxDim() + rect2.getMaxDim()) / 2.0 + rect1.getMaxDim()*maxScale; 
		clustering.setReference(ref, maxDist);
	}
	
	/**
	 * Creates an algorithm object.
	 * 
	 * @param startShape 
	 * @param goalShape
	 * @param distance
	 * @param useScale Whether scaling is allowed or not.
	 */
	public ProbabilisticAlignment(ShapeSet startShape, ShapeSet goalShape, int transformationType) {
		
		finding = false;
		forceRotation = false;
				
		minScale = 1.0 / scaleThreshold;
		maxScale = scaleThreshold;
		
		this.transformationType = transformationType;
						
		this.startShape = new ShapeSet(startShape);
		this.startDiam = this.startShape.getBoundingBox().getMaxDim();
		startPoints = createArrays(startShape);
		startDists = createDists(startPoints);
		startDistsN = createDistsN(startPoints, startShape.getLength());
		nStartPoints = 0;
		for (int i = 0; i < startPoints.length; i++) {
			nStartPoints += startPoints[i].length;
		}		
		
		this.goalShape = new ShapeSet(goalShape);
		this.goalDiam = this.goalShape.getBoundingBox().getMaxDim();
		goalPoints = createArrays(goalShape);
		goalDists = createDists(goalPoints);
		goalDistsN = createDistsN(goalPoints, goalShape.getLength());
		nGoalPoints = 0;
		for (int i = 0; i < goalPoints.length; i++) {
			nGoalPoints += goalPoints[i].length;
		}
		
		this.totalSteps = 0;
		this.currentSteps = 0;

		this.diam = goalShape.getBoundingBox().getMaxDim();
		clustering = new WeightedClustering(new SimilarityTransformation());
		flip(false);		
		if (this.useScale) {
			this.prescaling = this.goalDiam / this.startDiam;
			this.maxScale *= prescaling;
			this.minScale *= prescaling;
		} else {
			this.prescaling = 1.0;
		}
		bestScales = new double[1][2]; 
		bestScales[0][0] = this.prescaling;			 
		bestScales[0][1] = 1.0;
	}
		
	public static void setSeed(long seed) {
		randomGenerator.setSeed(seed);
		randomGenerator2.setSeed(seed/2+1);
	}
	
	private Point[][] createArrays(ShapeSet s) {		
		Point[][] points = new Point[s.getNumberOfPolylines()][];
		Iterator it = s.getShapeIterator();
		for (int i = 0; it.hasNext(); i++) {
			Polyline pl = (Polyline) it.next();
			points[i] = new Point[pl.getNumberOfPoints() + (pl.isClosed() ? 1 : 0)];
			Iterator pit = pl.getPointIterator();
			for (int j = 0; pit.hasNext(); j++) {
				points[i][j] = (Point) pit.next();
			}
			if (pl.isClosed()) {
				points[i][points[i].length-1] = points[i][0]; 
			}
		}
		return points;		
	}

	private double[][] createDists(Point[][] points) {
		double[][] tmp = new double[points.length][];
		for (int i = 0; i < points.length; i++) {
			tmp[i] = new double[points[i].length+3];
			tmp[i][0] = tmp[i][1] = tmp[i][tmp[i].length-1] = tmp[i][tmp[i].length-2] = 0.0;
			for (int j = 0; j < points[i].length-1; j++) {
				tmp[i][j+2] = points[i][j].distance(points[i][j+1]);
			}
		}
		return tmp;
	}
	
	private double[][] createDistsN(Point[][] points, double l) {
		double sumL = 0;
		double[][] tmp = new double[points.length][];
		for (int i = 0; i < points.length; i++) {
			tmp[i] = new double[points[i].length];
			tmp[i][0] = sumL/l;
			for (int j = 1; j < points[i].length; j++) {
				sumL += points[i][j-1].distance(points[i][j]);
				tmp[i][j] = sumL/l;
			}
		}
		tmp[tmp.length-1][tmp[tmp.length-1].length-1] = 1.0;
		return tmp;
	}
		
		
	public void compute(int newSteps) {
	
		this.totalSteps += newSteps;
		scaleSteps = Math.max(totalSteps/5, 1);
		nextScaleIdx = scaleSteps;
		t1 = System.currentTimeMillis();
		for (int i = 0; i < newSteps; i++) {
			doOneStep(currentSteps);			
			currentSteps++;
		}
		t2 = System.currentTimeMillis();
		clustering.performClustering();
		t3 = System.currentTimeMillis();
		//clustering.performClustering();
		t4 = System.currentTimeMillis();
		//System.out.println("times: " + (t2-t1) + " " + (t3-t2) + " " + (t4-t3));
	}
	
	protected void doOneStep(int stepIdx) {
		
		int f = f_;
		int r = r_;
		
		/*
		if (f == 0 && r == 3 && stepIdx == 256) {
			System.out.println();
		}
		*/

		double startSeed = randomGenerator.nextDouble();
		
		int plIdx1 = 0;
		Point[] startPoints = null;
		double[] startDists = null;
		int startFirstIdx = 0;
		for (int i = 0; i < this.startDistsN.length; i++) {
			if (startSeed < this.startDistsN[i][this.startDistsN[i].length-1]) {
				for (int j = 1; j < this.startDistsN[i].length; j++) {
					if (startSeed < this.startDistsN[i][j]) {
						startPoints = this.startPoints[i];
						startDists = this.startDists[i];
						plIdx1 = i;
						double d1 = startSeed - this.startDistsN[i][j-1];
						double d2 = this.startDistsN[i][j] - startSeed;
						if (d1 < d2) {
							startFirstIdx = j-1;
						} else {
							startFirstIdx = j;							
						}
						break;
					} else {
						continue;
					}
				}
				break;
			} else {
				// nop;
			}
		}
		
		
		double goalSeed = randomGenerator.nextDouble();
		int plIdx2 = 0;
		Point[] goalPoints = null;
		double[] goalDists = null;
		int goalFirstIdx = 0;
		for (int i = 0; i < this.goalDistsN.length; i++) {
			if (goalSeed < this.goalDistsN[i][this.goalDistsN[i].length-1]) {
				for (int j = 1; j < this.goalDistsN[i].length; j++) {
					if (goalSeed < this.goalDistsN[i][j]) {
						goalPoints = this.goalPoints[i];
						goalDists = this.goalDists[i];
						plIdx2 = i;
						double d1 = goalSeed - this.goalDistsN[i][j-1];
						double d2 = this.goalDistsN[i][j] - goalSeed;
						if (d1 < d2) {
							goalFirstIdx = j-1;
						} else {
							goalFirstIdx = j;							
						}
						break;
					} else {
						continue;
					}
				}
				break;
			} else {
				continue;
			}
		}

		
		double scale = 0.0;
		if (useScale) {

			while (scale == 0.0) {
				double startSeed2 = randomGenerator2.nextDouble();
				Point[] startPoints2 = null;
				int startFirstIdx2 = 0;
				for (int i = 0; i < this.startDistsN.length; i++) {
					if (startSeed2 < this.startDistsN[i][this.startDistsN[i].length-1]) {
						for (int j = 1; j < this.startDistsN[i].length; j++) {
							if (startSeed2 < this.startDistsN[i][j]) {
								startPoints2 = this.startPoints[i];
								double d1 = startSeed2 - this.startDistsN[i][j-1];
								double d2 = this.startDistsN[i][j] - startSeed2;
								if (d1 < d2) {
									startFirstIdx2 = j-1;
								} else {
									startFirstIdx2 = j;							
								}
								break;
							} else {
								continue;
							}
						}
						break;
					} else {
						continue;
					}
				}
				double goalSeed2 = randomGenerator2.nextDouble();
				Point[] goalPoints2 = null;
				int goalFirstIdx2 = 0;
				for (int i = 0; i < this.goalDistsN.length; i++) {
					if (goalSeed2 < this.goalDistsN[i][this.goalDistsN[i].length-1]) {
						for (int j = 1; j < this.goalDistsN[i].length; j++) {
							if (goalSeed2 < this.goalDistsN[i][j]) {
								goalPoints2 = this.goalPoints[i];
								double d1 = goalSeed2 - this.goalDistsN[i][j-1];
								double d2 = this.goalDistsN[i][j] - goalSeed2;
								if (d1 < d2) {
									goalFirstIdx2 = j-1;
								} else {
									goalFirstIdx2 = j;							
								}
								break;
							} else {
								continue;
							}
						}
						break;
					} else {
						continue;
					}
				}
				Point sourcePoint1 = startPoints[startFirstIdx];
				Point sourcePoint2 = startPoints2[startFirstIdx2];
				Point targetPoint1 = goalPoints[goalFirstIdx];
				Point targetPoint2 = goalPoints2[goalFirstIdx2];	
				
				double targetDistSQ = targetPoint1.distanceSQ(targetPoint2);
				double sourceDistSQ = sourcePoint1.distanceSQ(sourcePoint2)*prescaling*prescaling;
				if (targetDistSQ > sourceDistSQ*1e4 || targetDistSQ*1e4 < sourceDistSQ) {
					continue;
				}
				scale = Math.sqrt(targetDistSQ / sourceDistSQ)*prescaling;
			}			
			
			// scale based on scale of best map...
			if (false && stepIdx >= nextScaleIdx) {
				clustering.performClustering();
				LinkedList bs = clustering.getItems(0.25, 5, true);
				bestScales = new double[bs.size()][2];
				double weightSum = 0.0;
				Iterator it = bs.iterator();
				for (int i = 0; i < bestScales.length; i++) {
					SimilarityTransformation s = (SimilarityTransformation) it.next();
					bestScales[i][0] = Math.max(minScale, Math.min(maxScale, s.getScale()));
					bestScales[i][1] = s.weight;
					weightSum += s.weight;
				}
				if (bestScales.length == 0) {
					//System.out.println("l == 0!");
					bestScales = new double[1][2];
					bestScales[0][0] = prescaling;
					bestScales[0][1] = 1.0;
					weightSum = 1.0;
				}
				
				bestScales[0][1] = bestScales[0][1] / weightSum; 
				for (int i = 1; i < bestScales.length; i++) {
					bestScales[i][1] =  bestScales[i-1][1] + bestScales[i][1] / weightSum;
				}
				nextScaleIdx = stepIdx + scaleSteps;
			}
			double scaleChoose = randomGenerator.nextDouble();
			int scaleNumber = 0;
			for (int i = 0; i < bestScales.length; i++) {
				if (scaleChoose < bestScales[i][1]) {
					scaleNumber = i;
					break;
				}
			}
			if (finding) {
				do {
					scale = bestScales[scaleNumber][0] * Math.pow(2, randomGenerator.nextGaussian()*0.5);
				}
				while (scale < minScale / 2 || scale > maxScale);
			} else {
				do {					
					//scale = bestScales[scaleNumber][0] * Math.pow(2, randomGenerator.nextGaussian()*1.0);//sbs orig
					scale *= Math.pow(2, randomGenerator.nextGaussian()*0.1);//sbs exp
				} while (false && (scale < minScale / 2 || scale > maxScale * 2));
			}			
		} else {
			scale = 1.0;
		}
		
		for (int orientation = -1; orientation < 2; orientation+=2) {
			calc.reset(scale);
			calc.addFirst(startPoints[startFirstIdx], goalPoints[goalFirstIdx], plIdx1, plIdx2, startFirstIdx, goalFirstIdx);
											
			// going to both sides from the starting point...
			for (int dir = 1; dir > -2; dir -= 2) {
								
				// initialisation for the travel...
				int sourceCurrent = startFirstIdx;
				Point sourcePoint = startPoints[sourceCurrent];
				int targetCurrent = goalFirstIdx;
				Point targetPoint = goalPoints[targetCurrent];
				
				boolean inside = 
					sourceCurrent + dir*orientation >= 0 && sourceCurrent + dir*orientation < startPoints.length &&
					targetCurrent + dir >= 0 && targetCurrent + dir < goalPoints.length;
														
				nextStartDist = startDists[sourceCurrent+2+(dir*orientation-1)/2] * scale;
				nextGoalDist = goalDists[targetCurrent+2+(dir-1)/2];
				// looking for new points that yield better matches...
				while (inside) {
					int sourceNext = sourceCurrent + dir*orientation;
					int targetNext = targetCurrent + dir;
					//double sd = sourcePoint.distance(startPoints[sourceNext]) * scale;
					double sd = nextStartDist;
					//double td = targetPoint.distance(goalPoints[targetNext]);
					double td = nextGoalDist;
					double d = Math.min(sd, td);
					double l=0;
					double cos1 = 0.0;
					if (sourceCurrent + 2*dir*orientation >= 0 && sourceCurrent + 2*dir*orientation < startPoints.length) {
						Point p1 = startPoints[sourceCurrent];
						Point p2 = startPoints[sourceCurrent + dir*orientation];
						Point p3 = startPoints[sourceCurrent + 2*dir*orientation];
						cos1 = Point.innerProduct(p1, p2, p3) / startDists[sourceCurrent+2+(dir*orientation-1)/2] / startDists[sourceCurrent+2+(3*dir*orientation-1)/2] ;
					}
					double cos2 = 0.0;
					if (targetCurrent + 2*dir >= 0 && targetCurrent + 2*dir < goalPoints.length) {
						Point p1 = goalPoints[targetCurrent];
						Point p2 = goalPoints[targetCurrent + dir];
						Point p3 = goalPoints[targetCurrent + 2*dir];
						//double d1 = p1.distance(p2);
						//double d2 = p2.distance(p3);
						//int i1 = targetCurrent+2+(dir-1)/2;
						//int i2 = targetCurrent+2+(3*dir-1)/2;
						//double d1b = goalDists[i1];
						//double d2b = goalDists[i2];
						cos2 = Point.innerProduct(p1, p2, p3) / goalDists[targetCurrent+2+(dir-1)/2] / goalDists[targetCurrent+2+(3*dir-1)/2] ;
					}
					double at = alignmentThreshold * Math.min((1.0 - Math.max(cos1, cos2)), 0.015)/0.015;
					
					//System.out.println(cos1 + " " + Math.acos(cos1)/Math.PI*180 + " " + cos2 + " " + Math.acos(cos2)/Math.PI*180 + " " + at); 
					
					if (Math.abs(sd - td) <= d*at) {
						sourcePoint = startPoints[sourceNext];
						targetPoint = goalPoints[targetNext];
						sourceCurrent = sourceNext;
						targetCurrent = targetNext;
						nextStartDist = startDists[sourceCurrent+2+(dir*orientation-1)/2] * scale;
						nextGoalDist = goalDists[targetCurrent+2+(dir-1)/2];
						l = (sd + td) / 2;
					} else {
						// no corresponding pair found
						if (sd < td) {
							// create new target point						
							sourcePoint = startPoints[sourceNext];
							sourceCurrent = sourceNext;												
							targetPoint = targetPoint.intermediatePoint(sd/td, goalPoints[targetNext]);
							nextStartDist = startDists[sourceCurrent+2+(dir*orientation-1)/2];
							nextGoalDist = nextGoalDist - sd;
						} else {
							// create new source point						
							sourcePoint = sourcePoint.intermediatePoint(td/sd, startPoints[sourceNext]);
							targetPoint = goalPoints[targetNext];
							targetCurrent = targetNext;
							nextStartDist = nextStartDist - td;
							nextGoalDist = goalDists[targetCurrent+2+(dir-1)/2];
						}
						l = d;
					}
					if (l > 0) {
						if (finding) {
							calc.add(sourcePoint, targetPoint, l/Math.sqrt(scale), sourceNext, targetNext);
						} else {
							calc.add(sourcePoint, targetPoint, l, sourceNext, targetNext);
						}
					}
					
					inside = 
						sourceCurrent + dir*orientation >= 0 && sourceCurrent + dir*orientation < startPoints.length &&
						targetCurrent + dir >= 0 && targetCurrent + dir < goalPoints.length;
												
					// get the new map								
					boolean stillOk = calc.computeTransform();
					if (!stillOk) {
						break;
					}
				}
				calc.consolidate(dir == 1);
			} // end dir
			if (calc.getWeightBound() < clustering.minWeight) {
				continue;
			}
			SimilarityTransformation bestMap = (SimilarityTransformation) calc.getTransform();
			if (bestMap.getItemSource() == null) {
				System.out.println("null");
			}
				
	    	if (Double.isNaN(bestMap.weight)) {	    		
	    		System.err.println("nan1x");
	    		continue;
	    	}
	    	if (forceRotation && transformationType == TRANSLATION_ROTATION) {
	    		if (Math.abs(bestMap.getAngle()) < Math.PI / 6) {
	    			continue;
	    		}
	    	}
	    	
	    	/*
	    	if (Math.abs(bestMap.m[0][0]-0.5041) < 1e-2 && Math.abs(bestMap.m[0][2]+15.4809) < 1e-2 && Math.abs(bestMap.m[1][2]-16.93336) < 1e-2) {
	    		System.out.println(f + " " + r + " " + stepIdx);
	    	}
	    	*/
	    		
	    	clustering.add(bestMap);
		} // end orientation
	}

	public LinkedList getMaps(double f, int count) {
		return clustering.getItems(f, count, true);
	}
	
	public synchronized void reset() {
		clustering.reset();
		totalSteps = 0;
		currentSteps = 0;
		bestScales = new double[1][2]; 
		bestScales[0][0] = this.prescaling;			 
		bestScales[0][1] = 1.0;
		if (finding) {
			maxScale = this.prescaling * 1.5;
			minScale = this.prescaling / 10;
		} else {
			maxScale = this.prescaling * scaleThreshold;
			minScale = this.prescaling / scaleThreshold;
		}
	}
}
