/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Writer;
import java.util.StringTokenizer;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import profi.model.Box;
import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;

public class Square extends Figure {

	static Square createStandardFigure() {
		Square square = new Square();
		square.center = new Point(0.0, 0.0);
		square.radius = 100.0;
		square.angle = 0.0;
		square.ratio = 1.0;
		return square;
	}
	
	static Square createSquare(Box rec) {
		Square sq = new Square();
		sq.center = rec.getCenter();
		double h = rec.getHeight() * 1.1;
		double w = rec.getWidth() * 1.1;
		if (h > w) {
			sq.ratio = w / h;
			sq.angle = Math.PI / 2;
			sq.radius = h;
		} else {
			sq.ratio = h / w;
			sq.angle = 0;
			sq.radius = w;
		}
		return sq;
	}
	
	public Point center;
	public double radius;
	public double angle;
	// for rectangles
	public double ratio;
	
	public Square() {
		super(Figure.SQUARE);
	}

	public Square(Square other) {
		super(other);
		this.center = new Point(other.center);
		this.radius = other.radius;
		this.angle = other.angle;
		this.ratio = other.ratio;
	}

	public void addSpecificData_01(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("SQUARE\n");
		buff.append(prefix); buff.append("center.x = "); buff.append(center.x); buff.append("\n");
		buff.append(prefix); buff.append("center.y = "); buff.append(center.y); buff.append("\n");
		buff.append(prefix); buff.append("radius = "); buff.append(radius); buff.append("\n");
		buff.append(prefix); buff.append("angle = "); buff.append(angle); buff.append("\n");
		buff.append(prefix); buff.append("ratio = "); buff.append(ratio); buff.append("\n");
		writer.write(buff.toString());
	}

	public void addSpecificData_02(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<!-- SQUARE -->\n");
		buff.append(prefix); buff.append("<center.x>"); buff.append(center.x); buff.append("</center.x>\n");
		buff.append(prefix); buff.append("<center.y>"); buff.append(center.y); buff.append("</center.y>\n");
		buff.append(prefix); buff.append("<radius>"); buff.append(radius); buff.append("</radius>\n");
		buff.append(prefix); buff.append("<angle>"); buff.append(angle); buff.append("</angle>\n");
		buff.append(prefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		writer.write(buff.toString());
	}
	
	public void readSpecificData_01(BufferedReader reader) throws IOException {
		String line;
		String[] parts;
		center = new Point();
		
		line = reader.readLine().trim();
		parts = line.split(" = ");
		center.x = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		center.y = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		radius = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		angle = Double.parseDouble(parts[1]);
		line = reader.readLine().trim();
		parts = line.split(" = ");
		ratio = Double.parseDouble(parts[1]);
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("FIGURE end"));
	}
	
	public void readSpecificData_02(BufferedReader reader) throws IOException {
		String line;
		String field;
		center = new Point();
		
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		center.x = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		center.y = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		radius = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		angle = Double.parseDouble(field);
		line = reader.readLine().trim();
		field = line.substring(line.indexOf('>') + 1, line.lastIndexOf('<'));
		ratio = Double.parseDouble(field);
		do {
			line = reader.readLine().trim();
		} while (!line.equalsIgnoreCase("</FIGURE>"));
	}
	
	public ShapeSet getShapeSet() {
		ShapeSet set = new ShapeSet();
		for (int l = -1; l < layers.size(); l++) {
			double r = 1.0;
			if (l > -1) {
				r = ((Double) layers.get(l)).doubleValue();
			}
			r *= radius;
			double cos = Math.cos(angle);
			double sin = Math.sin(angle);
			Polyline poly = new Polyline(true);
			poly.addPoint(new Point(center.x + r*cos - r*ratio*sin, center.y + r*sin + r*ratio*cos));
			poly.addPoint(new Point(center.x - r*cos - r*ratio*sin, center.y - r*sin + r*ratio*cos));
			poly.addPoint(new Point(center.x - r*cos + r*ratio*sin, center.y - r*sin - r*ratio*cos));
			poly.addPoint(new Point(center.x + r*cos + r*ratio*sin, center.y + r*sin - r*ratio*cos));
			set.add(poly);
		}
		return set;
	}

	public double getSize() {
		return 2 * radius * Math.sqrt(1 + ratio*ratio);
	}

	public double getSalience() {
		return getSize() * (0.5 + ratio);
	}
	
	public double getLength() {
		return radius * 4 * (1+ratio);
	}

	public Point getCenter() {
		return center;
	}

	public Figure copy() {
		return new Square(this);
	}

	public boolean isCopy(Figure other) {
		if (other instanceof Square) return isTheSame((Square) other);
		else return false;
	}	

	public double sqDist(Figure o) {

		if (!(o instanceof Square)) {
			return Double.MAX_VALUE;
		}
		Square other = (Square) o;
		double l1 = this.radius;
		double l2 = other.radius;
		double minRadius = Math.min(l1, l2);
		double maxRadius = Math.max(l1, l2);
		double dr = maxRadius - minRadius;
		double centerDist = this.center.distance(other.center);
		double c1x = this.center.x;
		double c1y = this.center.y;
		double c2x = other.center.x;
		double c2y = other.center.y;
		
		double cos = Math.cos(this.angle);
		double sin = Math.sin(this.angle);
		double r = radius;		
		Point p11 = new Point(c1x + r*cos - r*ratio*sin, c1y + r*sin + r*ratio*cos);
		Point p12 = new Point(c1x - r*cos - r*ratio*sin, c1y - r*sin + r*ratio*cos);
		Point p13 = new Point(c1x - r*cos + r*ratio*sin, c1y - r*sin - r*ratio*cos);
		Point p14 = new Point(c1x + r*cos + r*ratio*sin, c1y + r*sin - r*ratio*cos);
		
		cos = Math.cos(other.angle);
		sin = Math.sin(other.angle);
		r = other.radius;
		Point p21 = new Point(c2x + r*cos - r*ratio*sin, c2y + r*sin + r*ratio*cos);
		Point p22 = new Point(c2x - r*cos - r*ratio*sin, c2y - r*sin + r*ratio*cos);
		Point p23 = new Point(c2x - r*cos + r*ratio*sin, c2y - r*sin - r*ratio*cos);
		Point p24 = new Point(c2x + r*cos + r*ratio*sin, c2y + r*sin - r*ratio*cos);
		
		double d11 = p11.distanceSQ(p21); 
		double d12 = p11.distanceSQ(p22);
		double d13 = p11.distanceSQ(p23);
		double d14 = p11.distanceSQ(p24);
		double d21 = p12.distanceSQ(p21);
		double d22 = p12.distanceSQ(p22);
		double d23 = p12.distanceSQ(p23);
		double d24 = p12.distanceSQ(p24);
		double d31 = p13.distanceSQ(p21);
		double d32 = p13.distanceSQ(p22);
		double d33 = p13.distanceSQ(p23);
		double d34 = p13.distanceSQ(p24);
		double d41 = p14.distanceSQ(p21);
		double d42 = p14.distanceSQ(p22);
		double d43 = p14.distanceSQ(p23);
		double d44 = p14.distanceSQ(p24);
		double d1 = Math.min(d11, Math.min(d12, Math.min(d13, d14)));
		double d2 = Math.min(d21, Math.min(d22, Math.min(d23, d24)));
		double d3 = Math.min(d31, Math.min(d32, Math.min(d33, d34)));
		double d4 = Math.min(d41, Math.min(d42, Math.min(d43, d44)));
		return Math.max(d1, Math.max(d2, Math.max(d3, d4)));
	}

	public boolean isTheSame(Square other) {
		double minRadius = Math.min(this.radius, other.radius);
		double maxRadius = Math.max(this.radius, other.radius);
		double dr = maxRadius - minRadius;
		double centerDist = this.center.distance(other.center);
		if (dr > maxRadius/5 || centerDist > maxRadius/5) {
			return false;
		} else {
			double maxSqDist = sqDist(other);
			return (maxSqDist < maxRadius * maxRadius / 5 / 5);
		}
	}

	public void transfor(Transformation t) {
		if (t instanceof SimilarityTransformation) {
			apply((SimilarityTransformation) t);
		} else {
			System.out.println("t not supported.");
		}		
	}
	
	public void apply(SimilarityTransformation sim) {
		sim.applyTo(this.center);
		this.radius *= sim.getScale();
		this.angle += sim.getAngle();
	}
	
	public double weightFactor() {
		return 2.0;
	}

	public boolean layering(Figure fig) {
		if (fig instanceof Square) {
			Square other = (Square) fig;	
			double centerDist = this.center.distance(other.center);
			if (centerDist > other.radius/3 || other.radius < this.radius * 0.5) {
				return false;
			}
			if (this.ratio / other.ratio > 1.2 || other.ratio / this.ratio > 1.2) {
				return false;
			}			
			int dAngle = (int) ((this.angle - other.angle) / Math.PI * 180);
			dAngle = dAngle % 180;
			dAngle += 180;
			dAngle = dAngle % 180;
			boolean angleOk = false;
			if (this.ratio > 0.9) {
				// nearly square -> 90
				dAngle = dAngle % 90;
				angleOk = dAngle < 5 || dAngle > 85;				
			} else {
				// rectangle
				angleOk = dAngle < 5 || dAngle > 175;
			}
			if (!angleOk) {
				return false;
			}
			double rNew = other.radius / this.radius * (1 + other.ratio / this.ratio) / 2;
			for (int l = -1; l < layers.size(); l++) {
				double r = 1.0;
				if (l > -1) {
					r = ((Double) layers.get(l)).doubleValue();
				}
				if(Math.abs(rNew-r) < 0.2) {
					return true;
				}
			}
			this.layers.add(new Double(rNew));
			return true;
			
		}
		return false;
	}
	
	public double maybeFrame() {
		return 1.0;
	}

	public double pithiness() {
		return 1-(1-ratio)*(1-ratio)*0.75;
	}

	public void addSpecificData_03(Writer writer, String prefix) throws IOException {
		String pprefix = prefix + "\t";
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<rectangle>\n");
		buff.append(pprefix); buff.append("<center>"); buff.append(center.x); buff.append(" "); buff.append(center.y); buff.append("</center>\n");
		buff.append(pprefix); buff.append("<radius>"); buff.append(radius); buff.append("</radius>\n");
		buff.append(pprefix); buff.append("<ratio>"); buff.append(ratio); buff.append("</ratio>\n");
		buff.append(pprefix); buff.append("<angle>"); buff.append(angle); buff.append("</angle>\n");
		buff.append(prefix); buff.append("</rectangle>\n");
		writer.write(buff.toString());		
	}

	public String getTypeName() {
		return "rectangle";
	}

	public void readSpecificData_03(Element element) {
		NodeList childNodes = element.getChildNodes();
		int nodeCount = 0;
		int stepCount = 0;
		while (stepCount <= 3) {
			Node node = childNodes.item(nodeCount);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element) node;
				switch (stepCount) {
				case 0:
					StringTokenizer tokens = new StringTokenizer(e.getTextContent());
					center = new Point(Double.parseDouble(tokens.nextToken()), Double.parseDouble(tokens.nextToken()));
					break;
				case 1:
					radius = Double.parseDouble(e.getTextContent());
					break;
				case 2:
					ratio = Double.parseDouble(e.getTextContent());
					break;
				case 3:
					angle = Double.parseDouble(e.getTextContent());
					break;
				}
				stepCount++;
			}
			nodeCount++;
		}
	}

	public Object getItemSource() {
		// TODO Auto-generated method stub
		return null;
	}

	public void setItemSource(Object source) {
		// TODO Auto-generated method stub
		
	}
}
