/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;

import profi.model.Box;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;

public class StandardComparator {

	public static class Result {
		
		public static final int complete = 0;
		public static final int partial = 1;
		
		public boolean[] hasInfo;		
		public double[] value;
		public SimilarityTransformation[] transformation;
		public ShapeSet[] set1;
		public ShapeSet[] set2;

		public Result() {
			hasInfo = new boolean[2];
			hasInfo[complete] = false;
			hasInfo[partial] = false;
			this.value = new double[2];
			this.transformation = new SimilarityTransformation[2];
			this.set1 = new ShapeSet[2];
			this.set2 = new ShapeSet[2];
		}
		
		public Result(int idx, double value, SimilarityTransformation transformation) {
			this.hasInfo = new boolean[2];
			this.hasInfo[idx] = true;
			this.value = new double[2];
			this.value[idx] = value;
			this.transformation = new SimilarityTransformation[2];
			this.transformation[idx] = transformation;
			this.set1 = new ShapeSet[2];
			this.set2 = new ShapeSet[2];
		}

		public static class ResultComparator implements Comparator {

			public static ResultComparator crc = new ResultComparator(complete);
			public static ResultComparator prc = new ResultComparator(partial);
			
			private int idx;
			
			public ResultComparator(int idx) {
				this.idx = idx;
			}
			
			public int compare(Object arg0, Object arg1) {
				Result r1 = (Result) arg0;
				Result r2 = (Result) arg1;
				if (r1.value[idx] < r2.value[idx]) return 1; // descending order! 
				if (r1.value[idx] > r2.value[idx]) return -1; // descending order! 
				return 0;
			}
			
		}		
	}
	
	public static int iterations = 3;
	public static double mapThreshold = 0.25;
	public static double findThreshold = 0.75;

		
	public static double compare(ShapeSet movable, ShapeSet fixed) {
		return compare(
				movable, fixed,
				true,	// flip
				true,	// rotate
				true,	// scale
				true,	// translate
				false,	// forceFlip
				false,	// forceRotation
				false).value[Result.complete];	// useSets
	}
	
	public static Result compare(ShapeSet movable, ShapeSet fixed, boolean useSets) {
		return compare(
				movable, fixed,
				true,	// flip
				true,	// rotate
				true,	// scale
				true,	// translate
				false,	// forceFlip
				false,	// forceRotation
				useSets);	// useSets
	}
	
	public static double isRotating(ShapeSet set) {
		return compare(
				set, set,
				false,	// flip
				true,	// rotate
				false,	// scale
				true,	// translate
				false,	// forceFlip
				true,	// forceRotation
				false).value[Result.complete];	// useSets
	}
	
	public static double isSymmetric(ShapeSet set) {
		return compare(
				set, set,
				false,	// flip
				true,	// rotate
				false,	// scale
				true,	// translate
				true,	// forceFlip
				false,	// forceRotation
				false).value[Result.complete];	// useSets
	}
	
	public static double isSame(ShapeSet set1, ShapeSet set2) {
		return compare(
				set1, set2,
				false,	// flip
				false,	// rotate
				false,	// scale
				false,	// translate
				false,	// forceFlip
				false,	// forceRotation
				false).value[Result.complete];	// useSets
	}
	
	public static Result compare(
			ShapeSet movable, ShapeSet fixed,
			boolean flip,
			boolean rotate,
			boolean scale,
			boolean translate,
			boolean forceFlip,
			boolean forceRotation,
			boolean useSets) {
		
		long t1 = System.currentTimeMillis();
		
		AverageWeightedDistance.finding = false;
		
		int size1 = movable.getNumberOfPoints();					
		int size2 = fixed.getNumberOfPoints();

		int tt = ProbabilisticAlignment.TRANSLATION;
		SimilarityTransformation[] maps = null;
		if (!translate) {
			maps = new SimilarityTransformation[1];
			maps[0] = SimilarityTransformation.identity;
		} else {
			if (forceFlip && !flip) {
				movable = new ShapeSet(movable);
				movable.apply(SimilarityTransformation.flipHorizontally);
			}
			if (rotate && !scale) tt = ProbabilisticAlignment.TRANSLATION_ROTATION;
			else if (rotate && scale)  tt = ProbabilisticAlignment.TRANSLATION_ROTATION_SCALE;
			else if (scale)  tt = ProbabilisticAlignment.TRANSLATION_SCALE;
			maps = ProbabilisticAlignment.computeTransformations(
					movable, fixed, flip, tt, 3, (int) ((scale ? 70 : 30) * Math.sqrt(size1*size2)), forceRotation);				
		}
		if (maps.length == 0) {
			maps = ProbabilisticAlignment.dummyTransformation(movable, fixed, flip, tt, forceRotation);
			maps[0].itemSource = new LinkedList[2][0];
		}
		
		long t2 = System.currentTimeMillis();

		Box bb1 = movable.getBoundingBox();
		AverageWeightedDistance.computeCorrelations(movable, bb1);
		
		Box bb2 = fixed.getBoundingBox();
		AverageWeightedDistance.computeCorrelations(fixed, bb2);	
		
		double bestSimC = -Double.MAX_VALUE;
		double bestSimP = -Double.MAX_VALUE;
		SimilarityTransformation bestTC = null;
		SimilarityTransformation bestTP = null;
		for (int i = 0; i < maps.length; i++) {
			
			ShapeSet moved = new ShapeSet(movable);
			moved.apply(maps[i]);
			bb1 = moved.getBoundingBox();
			
			AverageWeightedDistance.computeCorrelations(moved, bb1, fixed);
			AverageWeightedDistance.computeCorrelations(fixed, bb2, moved);
			double c1 = moved.getWeight();
			double c2 = fixed.getWeight();			
			double cc = (c1 * moved.getLength() + c2 * fixed.getLength()) /
				(moved.getLength() + fixed.getLength());
			if (cc > bestSimC) {
				bestSimC = cc;
				bestTC = (SimilarityTransformation) maps[i];
			}
			if (c1 > bestSimP) {
				bestSimP = c1;
				bestTP = (SimilarityTransformation) maps[i];
			}
		}

		long t3 = System.currentTimeMillis();
		
		//System.out.println("times: " + (t2-t1) + " " + (t3-t2));
		
		Result res = new Result();
		res.hasInfo[Result.complete] = true;
		res.value[Result.complete] = bestSimC;
		if (bestTC != null && bestTC.itemSource != null) {
			res.transformation[Result.complete] = bestTC;
			res.set1[Result.complete] = movable.getSubset(((LinkedList[][]) bestTC.itemSource)[0]);
			res.set2[Result.complete] = fixed.getSubset(((LinkedList[][]) bestTC.itemSource)[1]);
		} else {
			res.transformation[Result.complete] = SimilarityTransformation.identity;
			res.set1[Result.complete] = new ShapeSet();
			res.set2[Result.complete] = new ShapeSet();
		}
		
		res.hasInfo[Result.partial] = true;
		res.value[Result.partial] = bestSimP;
		if (bestTP != null && bestTP.itemSource != null) {
			res.transformation[Result.partial] = bestTP;
			res.set1[Result.partial] = movable.getSubset(((LinkedList[][]) bestTP.itemSource)[0]);
			res.set2[Result.partial] = fixed.getSubset(((LinkedList[][]) bestTP.itemSource)[1]);
		} else {
			res.transformation[Result.partial] = SimilarityTransformation.identity;
			res.set1[Result.partial] = new ShapeSet();
			res.set2[Result.partial] = new ShapeSet();
		}
		return res;
	}

	public static double representedBy(ShapeSet request, ShapeSet other, boolean otherBB) {
		AverageWeightedDistance.finding = true;
				
		Box bb = otherBB ? other.getBoundingBox() : request.getBoundingBox();
		AverageWeightedDistance.computeCorrelations(request, bb);
		request.updateWeightedLength();
		
		AverageWeightedDistance.computeCorrelations(request, bb, other);
		return request.getWeight();
	}

	public static ShapeSet getLeftOver(ShapeSet all, ShapeSet found, boolean sharp) {

		Box r1 = all.getBoundingBox();
				
		ShapeSet rest_ = new ShapeSet(all); 
		AverageWeightedDistance.computeCorrelations(rest_, r1, found);
					
		ShapeSet leftOver = rest_.split(sharp ? 0.95 : 0.90, sharp ? 0.75 : 0.5, false)[0];
		ShapeSet rest = new ShapeSet();	
		
		if (leftOver.getNumberOfPolylines() > 0) {
			double eps = Math.max(1.0, r1.getMaxDim()/100);
			leftOver.connect(eps);
			leftOver.consolidate(1e-3);
			int k = leftOver.getNumberOfPolylines();
			
			double gd = r1.getDiag();
			for (int i = 0; i < k; i++) {
				Polyline pl = (Polyline) leftOver.getPolyline(i);
				double d = pl.getBoundingBox().getDiag();
				if (d >= gd * 0.05) {
					rest.add(pl);
				}
			}
		}
		return rest;
	}

	public static Result[] findShape(ShapeSet fixed, ShapeSet movable) {
		AverageWeightedDistance.finding = true;
		
		LinkedList results = new LinkedList();
		ShapeSet s1_ = new ShapeSet(fixed);
		Box bb1 = fixed.getBoundingBox();
		//AverageWeightedDistance.computeCorrelations(s1_, bb1, s1_, false);
		//s1_.updateWeightedLength();						
		int size1 = s1_.getNumberOfPoints();
		
		ShapeSet s2_ = new ShapeSet(movable);
		Box bb2 = movable.getBoundingBox();
		AverageWeightedDistance.computeCorrelations(s2_, bb2);
		s2_.updateWeightedLength();
		int size2 = s2_.getNumberOfPoints();
	
		for (int side = 0; side < 2; side++) {
			ProbabilisticAlignment algo = new ProbabilisticAlignment(
					s2_, // movable
					fixed, // fixed
					ProbabilisticAlignment.TRANSLATION_ROTATION_SCALE
					);
			int samples = (int) (50 * Math.sqrt(size1*size2));
			algo.finding = true;
	
			for (int k = 0; k < iterations; k++) {
				
				algo.reset();
				algo.compute(samples);
				LinkedList maps = algo.getMaps(mapThreshold, 50);
				Iterator it = maps.iterator();
				while (it.hasNext()) {
					SimilarityTransformation sim = (SimilarityTransformation) it.next();
					ShapeSet moved = new ShapeSet(s2_);
					moved.apply(sim);
					bb2 = moved.getBoundingBox();
					
					AverageWeightedDistance.computeCorrelations(moved, bb2, s1_);
					double c_ = moved.getWeight();
					if (c_ > findThreshold) {
						results.add(new Result(0, c_, sim));
					}
				}
			
		
			}
			algo.flip(true);				
		}
		
		Result[] returns = new Result[results.size()];
		Iterator it = results.iterator();
		for (int i = 0; it.hasNext(); i++) {
			returns[i] = (Result) it.next();
		}
		Arrays.sort(returns, Result.ResultComparator.crc);
		
		return returns;
	}
	
}
