/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;

import profi.model.Box;
import profi.model.Polyline;
import profi.model.ShapeSet;
import profi.model.SimilarityTransformation;
import profi.model.Transformation;

public class StandardFinder {

	public static class Result implements Comparable {
		
		public double value;
		public Transformation transformation;

		public Result() {			
			this(0.0, null);
		}
		
		public Result(double value, Transformation transformation) {
			this.value = value;
			this.transformation = transformation;
		}

		public int compareTo(Object o) {
			Result other = (Result) o;
			if (this.value < other.value) return 1; // descending order! 
			if (this.value > other.value) return -1; // descending order! 
			return 0;
		}		
	}
	
	public static int iterations = 5;
	public static double mapThreshold = 0.25;
	public static double findThreshold = 0.75;
	public static double findTriThreshold = 0.85;
		
	public static Figure[] findShape(ShapeSet fixed, Figure figure) {
		AverageWeightedDistance.finding = true;
		LinkedList results = new LinkedList();
		
		Box bb2 = fixed.getBoundingBox();		
		int size2 = Math.max(20, fixed.getNumberOfPoints());
		
		Polyline pg = null;
		if (fixed.getNumberOfPolylines() == 1 && ((Polyline) fixed.getPlList().get(0)).isClosed()) {
			pg = (Polyline) fixed.getPlList().get(0);
		} else {
			pg = Convex.convexHull(fixed);
		}
		DeformationFinder pd = new DeformationFinder(pg);
				
		if (figure instanceof Circle) {
		
			Circle c = pd.getEllipse(mapThreshold);
			
			CircleFinder c_algo = new CircleFinder(fixed);
			int samples = fixed.getNumberOfPoints();
			c_algo.reset();
			c_algo.steps(samples);
			Figure[] circles = c_algo.getFigures(mapThreshold/4);
			
			for (int l = 0; l < circles.length && l < 50; l++) {
				
				ShapeSet moved = new ShapeSet(circles[l].getShapeSet());
				Box bb1 = moved.getBoundingBox();
				AverageWeightedDistance.computeCorrelations(moved, bb1);
				AverageWeightedDistance.computeCorrelations(moved, bb1, fixed);
				double c_ = moved.getWeight();
				if (c_ > findThreshold) {
					circles[l].represented = c_;
					circles[l].weightAbs = c_;
					results.add(circles[l]);
				}
			}
			Figure.FigureWeightComparator comp = new Figure.FigureWeightComparator(Figure.FigureWeightComparator.DESCENDING);
			Figure[] returns = new Figure[results.size() + ((c != null) ? 1 : 0)];
			Iterator it = results.iterator();
			for (int i = 0; it.hasNext(); i++) {
				returns[i] = (Figure) it.next();
			}
			if (c!= null) {
				returns[returns.length-1] = c;
			}
			Arrays.sort(returns, comp);
			return returns;
		} else 	if (figure instanceof Triangle) {
			
			TriangleFinder c_algo = new TriangleFinder(fixed);
			c_algo.findLines();
			LinkedList triangles = c_algo.getTriangles();
			Iterator trIt = triangles.iterator();
			for (int l = 0; trIt.hasNext() && l < 500; l++) {
				
				Triangle t = (Triangle) trIt.next();
				
				ShapeSet moved = new ShapeSet(t.getShapeSet());
				Box bb1 = moved.getBoundingBox();
				AverageWeightedDistance.computeCorrelations(moved, bb1);
				AverageWeightedDistance.computeCorrelations(moved, bb1, fixed);
				double c_ = moved.getWeight();
				if (c_ > findTriThreshold) {
					t.weightAbs = t.getLength() * c_;
					t.represented = c_;
					results.add(t);
				}
			}
			Figure.FigureWeightComparator comp = new Figure.FigureWeightComparator(Figure.FigureWeightComparator.DESCENDING);
			Figure[] returns = new Figure[results.size()];
			Iterator it = results.iterator();
			for (int i = 0; it.hasNext(); i++) {
				returns[i] = (Figure) it.next();
			}
			Arrays.sort(returns, comp);
			for (int i = 0; i < returns.length; i++) {
				returns[i].weightAbs = returns[i].represented;
			}
			return returns;
		} else if (figure instanceof Square) {
		
			Square sq = pd.getRectangle();			

			ShapeSet movable = figure.getShapeSet();
			ShapeSet s2_ = new ShapeSet(movable);
			Box bb1 = movable.getBoundingBox();
			AverageWeightedDistance.computeCorrelations(s2_, bb1);
			s2_.updateWeightedLength();
			int samples = 50 * fixed.getNumberOfPoints();
			
			ProbabilisticAlignment algo = new ProbabilisticAlignment(
					s2_, // movable
					fixed, // fixed
					ProbabilisticAlignment.TRANSLATION_ROTATION_SCALE
					);
	
			for (int k = 0; k < iterations; k++) {
				
				algo.reset();
				algo.compute(samples);
				LinkedList maps = algo.getMaps(mapThreshold, 50);
				Iterator it = maps.iterator(); 			
				
				while (it.hasNext()) {
					
					ShapeSet moved = new ShapeSet(s2_);
					SimilarityTransformation sim = (SimilarityTransformation) it.next(); 
					moved.apply(sim);
					bb1 = moved.getBoundingBox();
					AverageWeightedDistance.computeCorrelations(moved, bb1);
					AverageWeightedDistance.computeCorrelations(moved, bb1, fixed);
					//AverageWeightedDistance.computeCorrelations(s1_, bb1, moved, true);
					double c_ = moved.getWeight();
					if (c_ > findThreshold) {
						results.add(new Result(c_, sim));
					}
				}
			
		
			}

			Figure.FigureWeightComparator comp = new Figure.FigureWeightComparator(Figure.FigureWeightComparator.DESCENDING);
			Figure[] returns = new Figure[results.size() + ((sq != null) ? 1 : 0)];
			Iterator it = results.iterator();
			for (int i = 0; it.hasNext(); i++) {
				Result r = (Result) it.next();
				returns[i] = figure.copy();
				returns[i].transfor(r.transformation);
				returns[i].weightAbs =  r.value;
				returns[i].represented =  r.value;
			}
			if (sq!= null) {
				returns[returns.length-1] = sq;
			}
			Arrays.sort(returns, comp);
			return returns;
		} else {
			ShapeSet movable = figure.getShapeSet();
			ShapeSet s2_ = new ShapeSet(movable);
			Box bb1 = movable.getBoundingBox();
			AverageWeightedDistance.computeCorrelations(s2_, bb1);
			s2_.updateWeightedLength();
			int size1 = Math.max(20, s2_.getNumberOfPoints());

			ProbabilisticAlignment algo = new ProbabilisticAlignment(
					s2_, // movable
					fixed, // fixed
					ProbabilisticAlignment.TRANSLATION_ROTATION_SCALE
					);
			int samples = (int) (50 * Math.sqrt(size1*size2));
			
			for (int side = 0; side < 2; side++) {		
				for (int k = 0; k < iterations; k++) {
					
					algo.reset();
					algo.compute(samples);
					LinkedList maps = algo.getMaps(mapThreshold, 50);
					Iterator it = maps.iterator();
					while (it.hasNext()) {
						
						ShapeSet moved = new ShapeSet(s2_);
						SimilarityTransformation sim = (SimilarityTransformation) it.next(); 
						moved.apply(sim);
						bb1 = moved.getBoundingBox();
						AverageWeightedDistance.computeCorrelations(moved, bb1);
						AverageWeightedDistance.computeCorrelations(moved, bb1, fixed);
						//AverageWeightedDistance.computeCorrelations(s1_, bb1, moved, true);
						double c_ = moved.getWeight();
						if (c_ > findThreshold) {
							results.add(new Result(c_, sim));
						}
					}
				
			
				}
				algo.flip(true);
			}
			
			Figure.FigureWeightComparator comp = new Figure.FigureWeightComparator(Figure.FigureWeightComparator.DESCENDING);
			Figure[] returns = new Figure[results.size()];
			Iterator it = results.iterator();
			for (int i = 0; it.hasNext(); i++) {
				Result r = (Result) it.next();
				returns[i] = figure.copy();
				returns[i].transfor(r.transformation);
				returns[i].weightAbs = returns[i].getLength() * r.value;
				returns[i].represented = r.value;
			}
			Arrays.sort(returns, comp);
			for (int i = 0; it.hasNext(); i++) {
				returns[i].weightAbs = returns[i].represented;
			}
			return returns;
		}
	}
}
