/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Observable;
import java.util.Random;
import java.util.Vector;

import profi.model.Box;
import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;

public class TriangleFinder {

	
	public static final int TRANSLATION = 0;
	public static final int TRANSLATION_SCALE = 1;
	public static final int TRANSLATION_ROTATION = 2;
	public static final int TRANSLATION_ROTATION_SCALE = 3;
	public static final int AFFINE = 4;
	public static final int AFFINE_DISTORTION_PENALIZED = 5;
			
	private static class Segment {
		public Point p1;
		public Point p2;
		public double sql;
		
		public Segment(Point p1, Point p2) {
			this.p1 = p1;
			this.p2 = p2;
			sql = p1.distanceSQ(p2);
		}		
	}
	
	private static class Line {
		
		static class B {
			public double x;
			public boolean start;
			
			public B(double x, boolean start) {
				this.x = x;
				this.start = start;
			}
		}
		
		public Point p0;
		public Point v;
		
		public Vector points;
		
		public double totalLength;
		public double maxLength;
		
		public double xMin, xMax, yMin, yMax;
		public Vector intervals;
		
		public Line(Segment s) {
			p0 = s.p1;
			v = new Point(s.p2.x - s.p1.x, s.p2.y - s.p1.y);
			v.normalize();
			points = new Vector();
						
			this.totalLength = 0.0;
			intervals = new Vector();
			this.xMin = Double.MAX_VALUE;
			this.xMax = Double.MIN_VALUE;
			this.yMin = Double.MAX_VALUE;
			this.yMax = Double.MIN_VALUE;
			
			add(s);
		}
		
		public void add(Segment s) {
			
			if (intervals.size() == 0) {
				//System.out.print(".");
			} else {
				//System.out.print(":");
			}
			xMin = Math.min(xMin, Math.min(s.p1.x, s.p2.x));
			xMax = Math.max(xMax, Math.max(s.p1.x, s.p2.x));
			yMin = Math.min(yMin, Math.min(s.p1.y, s.p2.y));
			yMax = Math.max(yMax, Math.max(s.p1.y, s.p2.y));
			
			double v1 = (s.p1.x - p0.x)*v.x + (s.p1.y - p0.y)*v.y;
			double v2 = (s.p2.x - p0.x)*v.x + (s.p2.y - p0.y)*v.y;
			if (v2 < v1) {
				double tmp = v1;
				v1 = v2;
				v2 = tmp;
			}
			
			int idx;
			for (idx = 0; idx < intervals.size();) {
				B b = (B) intervals.get(idx);
				if (b.x < v1) {
					idx++;
				} else {
					break;
				}
			}
			if (idx >= intervals.size()) {
				intervals.add(new B(v1, true));
				intervals.add(new B(v2, false));
				totalLength += v2 - v1;
				return;
			}
			boolean searching = true;
			B b = (B) intervals.get(idx);
			if (b.start) {
				intervals.insertElementAt(new B(v1, true), idx);
				idx++;
			} else {
				if (b.x > v2) {
					searching = false;
				} else {
					//intervals.remove(idx);
				}
			}
			while (searching) {
				B next = (B) intervals.get(idx);
				if (next.x < v2) {
					intervals.remove(idx);
					if (idx >= intervals.size()) {
						intervals.add(new B(v2, false));
						searching = false;
					}
				} else {
					if (next.start) {
						intervals.insertElementAt(new B(v2, false), idx);
					} else {
						//nop
					}
					searching = false;
				}				
			}
			
			// update length:
			totalLength = 0.0;
			for (int i = 0; i < intervals.size(); i+=2) {
				totalLength += ((B) intervals.get(i+1)).x - ((B) intervals.get(i)).x; 
			}			
		}
		
		public boolean supports(Segment s, double dim, boolean sharp) {
			// angle < 5? means nearly parallel
			Point ov = new Point(s.p2.x - s.p1.x, s.p2.y - s.p1.y);
			ov.normalize();
			if (Math.abs(this.v.x * ov.x + this.v.y * ov.y) < 0.996 && sharp) return false;

			double v1 = (s.p1.x - p0.x)*v.x + (s.p1.y - p0.y)*v.y;
			double v2 = (s.p2.x - p0.x)*v.x + (s.p2.y - p0.y)*v.y;
			double dx1 = s.p1.x - (p0.x + v1 * v.x);
			double dy1 = s.p1.y - (p0.y + v1 * v.y);
			double dx2 = s.p2.x - (p0.x + v2 * v.x);
			double dy2 = s.p2.y - (p0.y + v2 * v.y);
			double d1 = Math.sqrt(dx1*dx1 + dy1*dy1);
			double d2 = Math.sqrt(dx2*dx2 + dy2*dy2);
			double d = Math.max(d1, d2);
			if (sharp) return d < dim / 15;
			
			double d1r = Math.abs(d1 / v1);
			double d2r = Math.abs(d2 / v2);
			double dr = Math.max(d1r, d2r);
			return d < dim / 15 || dr < 0.1; // ca tan(5.5) 
		}
		
		public boolean mayIntersect(Line other) {
			// angle < 20 -> triangle too pointed
			if (Math.abs(this.v.x * other.v.x + this.v.y * other.v.y) > 0.94) return false;
			
			double dim = Math.max(this.maxDim(), other.maxDim());
			double dx = 0.0;
			double dy = 0.0;
			if (this.xMax < other.xMin) dx = other.xMin - this.xMax;
			if (other.xMax < this.xMin) dx = this.xMin - other.xMax;
			if (this.yMax < other.yMin) dy = other.yMin - this.yMax;
			if (other.yMax < this.yMin) dy = this.yMin - other.yMax;
			return Math.sqrt(dx*dx + dy*dy) < dim * 0.25;
		}
		
		public Point intersection(Line other) {
			/*
			if (Math.abs(other.v.y - 1) > 1e3) {
				double lambda = (other.p0.x * other.v.y + p0.y - other.p0.y) / (other.v.y - 1);
				return new Point(p0.x + lambda * v.x, p0.y + lambda * v.y);
			} else {
				double lambda = (p0.x * v.y + other.p0.y - p0.y) / (v.y - 1);
				return new Point(other.p0.x + lambda * other.v.x, other.p0.y + lambda * other.v.y);
			}
			*/
			double lambda =
				(other.p0.x*other.v.y - this.p0.x*other.v.y + this.p0.y*other.v.x - other.p0.y*other.v.x) / 
				(other.v.y*this.v.x - this.v.y*other.v.x);
			return new Point(p0.x + lambda * v.x, p0.y + lambda * v.y);
		}
		
		public double getLength(Point p1, Point p2) {
			
			double v1 = (p1.x - p0.x)*v.x + (p1.y - p0.y)*v.y;
			double v2 = (p2.x - p0.x)*v.x + (p2.y - p0.y)*v.y;
			if (v2 < v1) {
				double tmp = v1;
				v1 = v2;
				v2 = tmp;
			}
			
			double l = 0.0;
			for (int i = 0; i < intervals.size(); i+=2) {
				double s = ((B) intervals.get(i)).x;
				double e = ((B) intervals.get(i+1)).x;
				if (s >= v2) {
					break;
				} else if (e <= v1) {
					continue;
				} else if (v1 < s) {
					if (v2 < e) {
						l += v2-s;
					} else {
						l += e-s;
					}
				} else {
					if (v2 < e) {
						l += v2-v1;
						break;
					} else {
						l+= e-v1;
					}
				}
			}
			return l;			
		}
		
		public double maxDim() {
			return Math.max(xMax - xMin, yMax - yMin);
		}
	}	

	private static class DirectionComparator implements Comparator {
		
		public int compare(Object o1, Object o2) {
			double x1 = ((Line) o1).v.x;
			double x2 = ((Line) o2).v.x;
			return x1 == x2 ? 0 :
				x1 < x2 ? -1 : 1;
		}		
	}

	private static class SegmentComparator implements Comparator {
		
		public int compare(Object o1, Object o2) {
			double x1 = ((Segment) o1).sql;
			double x2 = ((Segment) o2).sql;
			return x1 == x2 ? 0 :
				x1 < x2 ? 1 : -1;
		}		
	}
	
		
	/**
	 * Whether scaling is allowed or not.
	 * <code>true</code> means maps are similarities, <code>false</code> means maps are rigid motions.
	 */
	
	
	private ShapeSet startShape;
	int nStartPoints;
	Box bb;
	double dim;
	Segment[] startPoints;
	LinkedList linesList;
	Line[] lines;
		
	
	/**
	 * The clustering algorithm - as well as the clusters.
	 */
	private WeightedClustering mqlsClustering;
	
	/**
	 * Creates an algorithm object.
	 * 
	 * @param startShape 
	 * @param goalShape
	 * @param distance
	 * @param useScale Whether scaling is allowed or not.
	 */
	public TriangleFinder(ShapeSet startShape) {
		/*
		Segment s1 = new Segment(new Point(2,1), new Point(5,2));
		Segment s2 = new Segment(new Point(2,2), new Point(6, 3));
		Line l1 = new Line(s1);
		Line l2 = new Line(s2);
		Point p = l1.intersection(l2);
		System.out.println(p);
		...
		*/
		
						
		this.startShape = startShape;
		startPoints = createArrays(startShape);
		bb = startShape.getBoundingBox();
		dim = bb.getMaxDim();

		nStartPoints = startPoints.length;
		
		this.mqlsClustering = new WeightedClustering(Figure.triangle);
		
		double maxDim = startShape.getBoundingBox().getMaxDim();
		this.mqlsClustering.setDiam(maxDim / 20);
		this.mqlsClustering.setReference(Figure.triangle, maxDim);		
	}
		
	private Segment[] createArrays(ShapeSet lineSet) {
		LinkedList segs = new LinkedList();
		
		Polyline[] pls;
		
		List plList = lineSet.getPlList();
		pls = new Polyline[plList.size()];
		Iterator plIt = plList.iterator();
		for (int i = 0; plIt.hasNext(); i++) {
			pls[i] = (Polyline) plIt.next();
		}			
		
		for (int i = 0; i < pls.length; i++) {
			LinkedList pList = pls[i].getPoints();
			boolean pg = pls[i].isClosed();
			Iterator pIt = pList.iterator();
			Point start = (Point) pIt.next();
			for (int j = 0; pIt.hasNext(); j++) {
				Point end = (Point) pIt.next();
				Segment s = new Segment(start, end);
				if (s.sql > 0) segs.add(s);
				start = end;
			}
			if (pg) {
				Segment s = new Segment((Point) pList.getLast(), (Point) pList.getFirst());
				if (s.sql > 0) segs.add(s);
			}
		}
		Segment[] segments = new Segment[segs.size()];
		Iterator segIt = segs.iterator();
		for (int i = 0; segIt.hasNext(); i++) {
			segments[i] = (Segment) segIt.next();
		}
		return segments;		
	}
	
	public void findLines() {
		Arrays.sort(startPoints, new SegmentComparator());
		linesList = new LinkedList();
		for (int i = 0; i < startPoints.length; i++) {
			Segment s = startPoints[i];
			Iterator it = linesList.iterator();
			boolean found = false;
			while (it.hasNext()) {
				Line l = (Line) it.next();
				if (l.supports(s, dim, true)) {
					l.add(s);
					found = true;
				}					
			}
			if (!found) {
				Line l = new Line(s);
				linesList.add(l);
			}
		}		
		lines = new Line[linesList.size()];
		Iterator it = linesList.iterator();
		for (int i = 0; it.hasNext(); i++) {
			lines[i] = (Line) it.next();
		}
		for (int i = 0; i < startPoints.length; i++) {
			Segment s = startPoints[i];
			for (int j = 0; j < lines.length; j++) {
				if (lines[j].supports(s, dim, false)) {
					lines[j].add(s);
				}
			}
		}
	}
	
	public LinkedList getTriangles() {
		
		LinkedList triangles = new LinkedList();
		for (int i = 0; i < lines.length; i++) {
			for (int j = i + 1; j < lines.length; j++) {
				for (int k = j + 1; k < lines.length; k++) {
					if (lines[i].mayIntersect(lines[j]) && lines[j].mayIntersect(lines[k]) && lines[k].mayIntersect(lines[i])) {
						Point p1 = lines[i].intersection(lines[j]);
						Point p2 = lines[j].intersection(lines[k]);
						Point p3 = lines[k].intersection(lines[i]);
						if (
								(p1.x - bb.getMaxX() > dim / 2) ||
								(p2.x - bb.getMaxX() > dim / 2) ||
								(p3.x - bb.getMaxX() > dim / 2) ||
								(p1.y - bb.getMaxY() > dim / 2) ||
								(p2.y - bb.getMaxY() > dim / 2) ||
								(p3.y - bb.getMaxY() > dim / 2) ||
								(bb.getMinX() - p1.x > dim / 2) ||
								(bb.getMinX() - p2.x > dim / 2) ||
								(bb.getMinX() - p3.x > dim / 2) ||
								(bb.getMinY() - p1.y > dim / 2) ||
								(bb.getMinY() - p2.y > dim / 2) ||
								(bb.getMinY() - p3.y > dim / 2)) continue;
						Triangle t = new Triangle(p1, p2, p3);
						double l1 = lines[i].getLength(p1, p3);
						Point p13 = p1.intermediatePoint(0.5, p3);
						double l1a = lines[i].getLength(p1, p13);
						double l1b = lines[i].getLength(p3, p13);
						double l2 = lines[j].getLength(p1, p2);
						Point p12 = p1.intermediatePoint(0.5, p2);
						double l2a = lines[j].getLength(p1, p12);
						double l2b = lines[j].getLength(p2, p12);
						double l3 = lines[k].getLength(p2, p3);
						Point p23 = p2.intermediatePoint(0.5, p3);
						double l3a = lines[k].getLength(p2, p23);
						double l3b = lines[k].getLength(p3, p23);
						double l4 = l1+l2+l3;
						double d1 = p1.distance(p3);
						double d2 = p1.distance(p2);
						double d3 = p2.distance(p3);
						if (0.75 * t.getLength() > l4) continue;
						if (0.5 * d1 > l1) continue;
						if (0.5 * 0.5 * d1 > l1a) continue;
						if (0.5 * 0.5 * d1 > l1b) continue;
						if (0.5 * d2 > l2) continue;
						if (0.5 * 0.5 * d2 > l2a) continue;
						if (0.5 * 0.5 * d2 > l2b) continue;
						if (0.5 * d2 > l3) continue;
						if (0.5 * 0.5 * d3 > l3a) continue;
						if (0.5 * 0.5 * d3 > l3b) continue;
						triangles.add(t);
					}
				}
			}
		}
		return triangles;		
	}
	
	public LinkedList getQuads() {
		
		LinkedList quads = new LinkedList();
		for (int i = 0; i < lines.length; i++) {
			for (int j = i + 1; j < lines.length; j++) {
				for (int k = j + 1; k < lines.length; k++) {
					for (int l = k + 1; l < lines.length; l++) {
						
						//find Rhomboids...
						
						if (lines[i].mayIntersect(lines[j]) && lines[j].mayIntersect(lines[k]) && lines[k].mayIntersect(lines[i])) {
							Point p1 = lines[i].intersection(lines[j]);
							Point p2 = lines[j].intersection(lines[k]);
							Point p3 = lines[k].intersection(lines[l]);
							Point p4 = lines[l].intersection(lines[i]);
							if (
									(p1.x - bb.getMaxX() > dim / 2) ||
									(p2.x - bb.getMaxX() > dim / 2) ||
									(p3.x - bb.getMaxX() > dim / 2) ||
									(p4.x - bb.getMaxX() > dim / 2) ||
									(p1.y - bb.getMaxY() > dim / 2) ||
									(p2.y - bb.getMaxY() > dim / 2) ||
									(p3.y - bb.getMaxY() > dim / 2) ||
									(p4.y - bb.getMaxY() > dim / 2) ||
									(bb.getMinX() - p1.x > dim / 2) ||
									(bb.getMinX() - p2.x > dim / 2) ||
									(bb.getMinX() - p3.x > dim / 2) ||
									(bb.getMinX() - p4.x > dim / 2) ||
									(bb.getMinY() - p1.y > dim / 2) ||
									(bb.getMinY() - p2.y > dim / 2) ||
									(bb.getMinY() - p3.y > dim / 2) ||
									(bb.getMinY() - p4.y > dim / 2)) continue;
							Triangle t = new Triangle(p1, p2, p3);
							if (t.getLength() > (lines[i].getLength(p1, p4) + lines[j].getLength(p1, p2) + lines[k].getLength(p2, p3) + lines[l].getLength(p3, p4)) * 2) continue;
							quads.add(t);
						}
					}
				}
			}
		}
		return quads;		
	}	
}
