/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.matching;

import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;


public class WeightedClustering {

	public static interface Item {
		public double getItemWeight();
		public void setItemWeight(double w);
		public double itemDist(Item other);
		public double itemDistSQ(Item other);
		public Item itemCopy();
		public Object getItemSource();
		public void setItemSource(Object source);
		public void mark(boolean m);
		public boolean isMarked();
	}
	

	/**
	 * Comparator for the items.
	 * Compares based on the weight.
	 */
	public static class ItemComparator implements Comparator {
		
		/**
		 * Constant indicating ascending order.
		 */
		public static final int ASCENDING = 1;

		/**
		 * Constant indicating descending order.
		 */
		public static final int DESCENDING = -1;
		
		/**
		 * The ordering direction - ascending or descending.
		 */
		private int dir;
		
		/**
		 * Default orders ascending.
		 */
		public ItemComparator() {
			this(ASCENDING);
		}
		
		/**
		 * Comparator ordering ascending or descending.
		 * 
		 * @param dir The ordering direction - ascending or descending.
		 */
		public ItemComparator(int dir) {
			this.dir = dir;
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Object o1, Object o2) {
			double w1 = ((Item) o1).getItemWeight();
			double w2 = ((Item) o2).getItemWeight();
			return w1 == w2 ? 0 :
				w1 < w2 ? -1 * this.dir : 1 * this.dir;
		}		
	}
	
	public static class Interval {
		public int idx;
		public int start;
		public int end;
		public Interval(int idx, int start, int end) {
			this.idx = idx;
			this.start = start;
			this.end = end;
		}
	}
	
	public static class IntervalBound implements Comparable {
		public int idx;
		public int value;
		public boolean isStart;
		public int compareTo(Object arg0) {
			IntervalBound other = (IntervalBound) arg0;
			return (this.value*2 + (this.isStart ? 0 : 1) - other.value*2 + (other.isStart ? 0 : 1));
		}
		public IntervalBound(Interval interval, boolean isStart) {
			this.idx = interval.idx;
			this.value = isStart ? interval.start : interval.end;
			this.isStart = isStart;
		}
	}
	
	
	/**
	 * A cluster of weighted items.
	 */
	private class Cluster {

		LinkedList[][] mergeIntervals() {
			LinkedList[][] intervals = new LinkedList[2][];

			LinkedList intervalBounds[] = new LinkedList[2];
			intervalBounds[0] = new LinkedList();
			intervalBounds[1] = new LinkedList();
			int[] k = {0, 0};
			Iterator it = points2.iterator();
			while (it.hasNext()) {
				Item item = (Item) it.next(); 
				Object o = item.getItemSource();
				if (o instanceof Integer) continue;
				if (o instanceof LinkedList[][]) {
					LinkedList[][] oldIntervals = (LinkedList[][]) o;
					for (int side = 0; side < 2; side++) {
						k[side] = Math.max(k[side], oldIntervals[side].length);
						for (int i = 0; i < oldIntervals[side].length; i++) {
							Iterator oldIt = oldIntervals[side][i].iterator();
							while (oldIt.hasNext()) {
								Interval oldInt = (Interval) oldIt.next();
								intervalBounds[side].add(new IntervalBound(oldInt, true));
								intervalBounds[side].add(new IntervalBound(oldInt, false));
							}
						}
					}
				} else {				
					Interval[] interval = (Interval[]) o;
					if (interval[0].end != interval[0].start) {
						k[0] = Math.max(k[0], interval[0].idx+1);
						intervalBounds[0].add(new IntervalBound(interval[0], true));
						intervalBounds[0].add(new IntervalBound(interval[0], false));
					}
					if (interval[1].end != interval[1].start) {
						k[1] = Math.max(k[1], interval[1].idx+1);
						intervalBounds[1].add(new IntervalBound(interval[1], true));
						intervalBounds[1].add(new IntervalBound(interval[1], false));
					}
				}
			}
			
			for (int side = 0; side < 2; side++) {
				intervals[side] = new LinkedList[k[side]];
				int[] depth = new int[k[side]];
				for (int i = 0; i < k[side]; i++) {
					intervals[side][i] = new LinkedList();
					depth[i] = 0;
				}
				Collections.sort(intervalBounds[side]);
				it = intervalBounds[side].iterator();
				while (it.hasNext()) {
					IntervalBound bound = (IntervalBound) it.next();
					if (depth[bound.idx] == 0) {
						Interval interval = new Interval(bound.idx, bound.value, bound.value);
						intervals[side][bound.idx].add(interval);
					}
					depth[bound.idx] += bound.isStart ? 1 : -1;
					if (depth[bound.idx] == 0) {
						Interval interval = (Interval) intervals[side][bound.idx].getLast(); 
						interval.end = bound.value;
					}
				}
			}
			return intervals;
		}				
		public Node parent;
		
		/**
		 * The cluster's centroid.
		 * This happens to be the most heavy weighted point with additional weight.
		 */
		private Item centroid;
		
		private double weight;
		
		/**
		 * The points subsumed by this cluster.
		 */
		private LinkedList points;

		private LinkedList points2;
		
		public Cluster(Item s) {
			this.centroid = s.itemCopy();
			this.points = new LinkedList();
			this.points.add(s);
			this.points2 = new LinkedList();
			this.weight = 0.0;
		}
		
		/**
		 * The centroid.
		 * 
		 * @return The centroid.
		 */
		public Item getCentroid() {
			return this.centroid;
		}
		
		/**
		 * The cluster's weight.
		 * This happens to be the sum of the weights of the cluster's points.
		 * 
		 * @return The cluster's weight.
		 */
		public double weight() {
			return this.weight;
		}
		
		/**
		 * The cluster's points.
		 * 
		 * @return The cluster's points.
		 */
		public LinkedList getPoints() {
			return this.points;
		}
		
		/**
		 * Adds a point to this cluster.
		 * Since the centroid may change, some points may be dropped.
		 * 
		 * @param s The new point.
		 * @return All the dropped points.
		 */
		public void add(Item s) {
			this.points.add(s);
		}
		
		public void clearPoints() {
			points.clear();
			points2.clear();
			weight = 0.0;
		}
		
		public void clearPoints2() {
			points2.clear();
			weight = 0.0;
		}
				
		public void add2(Item s) {
			this.points2.add(s);
		}

		/**
		 * Recomputes the cluster's weight.
		 */
		private void updateWeight() {
			
			double weight = 0;
			Iterator it = points2.iterator();
			while (it.hasNext()) {
				Item current = (Item) it.next();
				weight += current.getItemWeight();
			}						
			this.weight = weight;
		}
	}

	/**
	 * Comparator for the clusters.
	 * Compares based on the weight.
	 */
	private static class ClusterComparator implements Comparator {
		
		/**
		 * Constant indicating ascending order.
		 */
		public static final int ASCENDING = 1;

		/**
		 * Constant indicating descending order.
		 */
		public static final int DESCENDING = -1;

		/**
		 * Constant indicating usage of cluster weight.
		 */
		public static final int  USE_CLUSTER_WEIGHT = -1;

		/**
		 * Constant indicating usage of centroid weight.
		 */
		public static final int  USE_CENTROID_WEIGHT = 1;
		
		/**
		 * The ordering direction - ascending or descending.
		 */
		private int dir;
		
		private int weight;
		
		/**
		 * Default orders ascending.
		 */
		public ClusterComparator() {
			this(ASCENDING, USE_CLUSTER_WEIGHT);
		}
		
		/**
		 * Comparator ordering ascending or descending.
		 * 
		 * @param dir The ordering direction - ascending or descending.
		 */
		public ClusterComparator(int dir, int weight) {
			this.dir = dir;
			this.weight = weight;
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Object o1, Object o2) {
			double w1 = 0.0;
			double w2 = 0.0;
			if (weight == USE_CLUSTER_WEIGHT) {
				w1 = ((Cluster) o1).weight();
				w2 = ((Cluster) o2).weight();
			} else {
				w1 = ((Cluster) o1).getCentroid().getItemWeight();
				w2 = ((Cluster) o2).getCentroid().getItemWeight();
			}
			return w1 == w2 ? 0 :
				w1 < w2 ? -1 * this.dir : 1 * this.dir;
		}		
	}

	private class Node {
		public Item center;
		public Cluster cluster;
		public Node parent;
		public Node insides;
		public Node outsides;
		public double sqRadius;
		public double outerThreshold;
		public double innerThreshold;
				
		private int children;
		
		public int countChildren() {
			int n = 1;
			if (insides != null) {
				n += insides.countChildren();
			}
			if (outsides != null) {
				n += outsides.countChildren();
			}
			return n;
		}
			
		public Node() {
			children = 0;
		}
				
		public void findNeighbours(Item s, LinkedList neighbours) {
			double sqDist = center.itemDistSQ(s);
			if (sqDist <= outerThreshold) {
				if (insides != null) {
					insides.findNeighbours(s, neighbours);
				}
				if (cluster != null && sqDist <= clusterSqRadius) {
					neighbours.add(cluster);
				}
			}
			if (sqDist > innerThreshold) {
				if (outsides != null) {
					outsides.findNeighbours(s, neighbours);
				}
			}
		}
		
		public boolean hasNeighbours(Item s) {
			boolean found = false;
			double sqDist = center.itemDistSQ(s);
			if (sqDist <= outerThreshold) {
				if (cluster != null && sqDist <= clusterSqRadius) {
					return true;
				}
				if (insides != null) {
					found |= insides.hasNeighbours(s);
				}
			}
			if (sqDist > innerThreshold) {
				if (outsides != null) {
					found |= outsides.hasNeighbours(s);
				}
			}
			return found;
		}
		
		public void add(Cluster c) {
			double sqDist = center.itemDistSQ(c.getCentroid());
			if (sqDist <= sqRadius) {
				if (insides != null) {
					insides.add(c);
				} else {
					children++;
					insides = new Node();
					insides.parent = this;
					insides.center = c.getCentroid();
					insides.cluster = c;
					c.parent = insides;
					insides.sqRadius = Math.max(sqRadius/4, clusterSqRadius);
					insides.innerThreshold = Math.max(insides.sqRadius - 2*Math.sqrt(sqRadius*clusterSqRadius) + clusterSqRadius, 0.0);
					insides.outerThreshold = insides.sqRadius + 2*Math.sqrt(sqRadius*clusterSqRadius) + clusterSqRadius;
				}
			} else {
				if (outsides != null) {
					outsides.add(c);
				} else {
					children++;
					outsides = new Node();
					outsides.parent = this;
					outsides.center = c.getCentroid();
					outsides.cluster = c;
					c.parent = outsides;
					outsides.sqRadius = sqRadius;
					outsides.innerThreshold = innerThreshold;
					outsides.outerThreshold = outerThreshold;
				}				
			}
		}		
	}


	public double weightThreshold;	

	// absolute values
	private double clusterRadius;
	private double clusterSqRadius;	
	
	/**
	 * The weight of the most heavyweighted point.
	 */
	private double maxWeight;
	public double minWeight;
	
	/**
	 * All the clusters.
	 */
	private LinkedList clusters;
	
	/**
	 * All points waiting to get clustered.
	 */
	private LinkedList newPoints;
	private LinkedList allPoints;
	
	private int n;
	
	private Node root;
	private double maxSqDist;
	private Item reference;
		
	private static ClusterComparator clustersComparator = new ClusterComparator(
			ClusterComparator.DESCENDING,
			ClusterComparator.USE_CLUSTER_WEIGHT);
	private static ClusterComparator centroidsComparator  = new ClusterComparator(
			ClusterComparator.DESCENDING,
			ClusterComparator.USE_CENTROID_WEIGHT);
	
	//private Item bestSim;
	
	/**
	 * Maximum distance of a point to the centroid.
	 * i.e. The cluster's maximum radius.
	 */
	public static double clusterRadiusRel = 0.05;
		
	/**
	 * Default has no clusters.
	 */
	public WeightedClustering(Item std) {
		
		this.clusters = new LinkedList();
		this.newPoints = new LinkedList();
		this.allPoints = new LinkedList();
		this.maxWeight = 0.0;
		this.minWeight = 0.0;
		this.n = 0;
		
		this.reference = null;
		this.maxSqDist = 1e12;
				
		root = new Node();
		root.center = std;
		root.insides = new Node();
		root.insides.center = std;
		weightThreshold = 1.0/500;
	}
	
	public void setReference(Item ref, double maxDist) {
		
		this.reference = ref;
		this.maxSqDist = maxDist * maxDist * 4;
	}
	
	/**
	 * Adds a weighted similarity.
	 * Clusters are not updated.
	 * 
	 * @param s The similarity.
	 */
	public void add(Item t) {
		if (t == null) {
			return;
		}
		n++;
		if (t.getItemWeight() <= minWeight) {
			return;
		}
		this.newPoints.add(t);
		this.allPoints.add(t);
		//this.bestSim = null;
		if (maxWeight < t.getItemWeight()) {
			maxWeight = t.getItemWeight();
			minWeight = maxWeight * weightThreshold;
		}
	}
	
	public void setDiam(double diam) {
		this.clusterRadius = clusterRadiusRel * diam;
		this.clusterSqRadius = this.clusterRadius * this.clusterRadius;
		root.insides.sqRadius = (diam*2)*(diam*2);
		root.insides.innerThreshold = root.insides.sqRadius - 2*(diam*2)*this.clusterRadius + this.clusterRadius*this.clusterRadius; 
		root.insides.outerThreshold = root.insides.sqRadius + 2*(diam*2)*this.clusterRadius + this.clusterRadius*this.clusterRadius; 
	}


	/*
	public void performClustering() {

		if (newPoints.size() == 0) {
			return;
		}

		System.out.println();
		System.out.println("new  : " + newPoints.size());
		System.out.println("all  : " + allPoints.size());
		System.out.println("nodes: " + root.countChildren());
		
		Collections.sort(newPoints, new ItemComparator(ItemComparator.DESCENDING));

		//if (true) return;
		
		int count = 0;
		Iterator clusterIt;
						
		// 1. Phase: building up the clusters...
		LinkedList neighbours = new LinkedList();		
		while (this.newPoints.size() > 0) {
			count++;
			Item rm = (Item) newPoints.removeFirst();

			if (rm.getItemWeight() <= maxWeight * weightThreshold) {
				newPoints.clear();
				break;
			}

			neighbours.clear();
			int[] counter = {0};
			root.insides.findNeighbours(rm, neighbours, counter);
			if (neighbours.size() == 0) {
				Cluster newCluster = new Cluster(rm);
				this.clusters.add(newCluster);
				root.insides.add(newCluster);
			} else {
				Collections.sort(neighbours, centroidsComparator);
				Cluster c = (Cluster) neighbours.getFirst();
				if (c.centroid.getItemWeight() >= rm.getItemWeight()) {
					c.add(rm);
				} else {
					Cluster newCluster = new Cluster(rm);
					this.clusters.add(newCluster);
					root.insides.add(newCluster);		
					Iterator nIt = neighbours.iterator();
					LinkedList points = new LinkedList();
					while (nIt.hasNext()) {
						c = (Cluster) nIt.next();
						points.addAll(c.getPoints());
						c.parent.removeCluster();
						this.clusters.remove(c);
					}
					this.newPoints.addAll(newCluster.addAll(points));
				}
			}		
		}

// *
		Iterator it1 = clusters.iterator();
		int i = 0;
		while (it1.hasNext()) {
			Cluster c1 = (Cluster) it1.next();
			Iterator it2 = clusters.listIterator();
			int j = 0;
			while (j < i) {
				Cluster c2 = (Cluster) it2.next();
				double d = c1.centroid.dist(c2.centroid, this.testingPoints);
				if (d <= this.clusterRadius) {
					System.out.println("Clustering: error found");					
				}				
				j++;
			}
		}
		
// * /
		
	
		// 2. Phase: filling the clusters...
		clusterIt = this.clusters.iterator();			
		while (clusterIt.hasNext()) {
			((Cluster) clusterIt.next()).clearPoints2();
		}
		
		//int nNodes = root.insides.countChildren();
		
		Iterator pointsIt;
		pointsIt = this.allPoints.iterator();
		while (pointsIt.hasNext()) {
			Item s = (Item) pointsIt.next();
			if (s.getItemWeight() <=  maxWeight * weightThreshold) {
				pointsIt.remove();
				continue;
			}
			neighbours.clear();
			int[] counter = {0};
			root.insides.findNeighbours(s, neighbours, counter);
			Iterator nIt = neighbours.iterator();
			while (nIt.hasNext()) {
				((Cluster) nIt.next()).add2(s);
			}
			//System.out.println("n:" + nNodes + " " + counter[0] + " " + neighbours.size());
		}

		Iterator it = this.clusters.iterator();
		while (it.hasNext()) {
			((Cluster) it.next()).updateWeight();
		}
		
		if (this.clusters.size() < 1) {
			// System.out.println("MacQueenLikeSimilarityClustering.performClustering: clusters < 1");
			this.bestSim = null;
			return;
		}
		
		Collections.sort(this.clusters, clustersComparator);
		
		{
			if (this.clusters.size() > 0) {
				double wMin = ((Cluster) this.clusters.getFirst()).weight() * weightThreshold;
				it = this.clusters.iterator();
				while (it.hasNext()) {
					Cluster c = ((Cluster) it.next());
					double w = c.weight();
					if (w < wMin) {
						it.remove();
						c.parent.removeCluster();
					} 
				}
			}
		}
		
		Cluster bestCluster = (Cluster) this.clusters.getFirst();
		this.bestSim = bestCluster.getCentroid();			
	}	
	*/

	
	public void performClusteringNaive() {
		Collections.sort(allPoints, new ItemComparator(ItemComparator.DESCENDING));
		
		this.clusters = new LinkedList();
		LinkedList points = new LinkedList();
		
		Iterator pIt = allPoints.iterator();
		while (pIt.hasNext()) {
			Item p = (Item) pIt.next();
			boolean found = false;
			Iterator cIt = clusters.iterator();
			while (cIt.hasNext()) {
				Cluster c = (Cluster) cIt.next();
				if (c.centroid.itemDistSQ(p) < clusterSqRadius) {
					found = true;
					break;
				}
			}
			if (found) {
				points.add(p);
			} else {
				Cluster newCluster = new Cluster(p);
				newCluster.add2(p);
				clusters.add(newCluster);
			}			
		}
		pIt = points.iterator();
		while (pIt.hasNext()) {
			Item p = (Item) pIt.next();
			Iterator cIt = clusters.iterator();
			while (cIt.hasNext()) {
				Cluster c = (Cluster) cIt.next();
				if (c.centroid.itemDistSQ(p) < clusterSqRadius) {
					c.add2(p);
				}
			}
		}

		Iterator it = this.clusters.iterator();
		while (it.hasNext()) {
			((Cluster) it.next()).updateWeight();
		}
		
		if (this.clusters.size() < 1) {
			//this.bestSim = null;
			return;
		}
		
		Collections.sort(this.clusters, clustersComparator);
				
		Cluster bestCluster = (Cluster) this.clusters.getFirst();
	}
	
	public void performClustering() {
		
		int n1 = 0;
		int n2 = 0;

		long t1 = System.currentTimeMillis();
		
		if (newPoints.size() == 0) {
			return;
		}

		root.insides.insides = null;
		root.insides.outsides = null;
		root.outsides = null;
		newPoints.clear();
		Collections.sort(allPoints, new ItemComparator(ItemComparator.DESCENDING));
		
		Iterator clusterIt;
		
		long t2 = System.currentTimeMillis();
		
		// 1. Phase: building up the clusters...
		LinkedList neighbours = new LinkedList();
		Iterator apIt = allPoints.iterator();
		while (apIt.hasNext()) {
					
			Item rm = (Item) apIt.next();
			rm.mark(false);

			if (rm.getItemWeight() <= maxWeight * weightThreshold) {
				break;
			}
			
			if (!root.insides.hasNeighbours(rm)) {
				Cluster newCluster = new Cluster(rm);
				newCluster.add2(rm);
				rm.mark(true);
				this.clusters.add(newCluster);
				root.insides.add(newCluster);
				n1++;
			} else {
				//nop
				n2++;
			}
		}
	
		long t3 = System.currentTimeMillis();
		
		// 2. Phase: filling the clusters...
				
		Iterator pointsIt;
		pointsIt = this.allPoints.iterator();
		while (pointsIt.hasNext()) {
			
			Item s = (Item) pointsIt.next();
			if (s.getItemWeight() <=  maxWeight * weightThreshold) {
				pointsIt.remove();
				continue;
			}
			if (s.isMarked()) continue;
			neighbours.clear();
			root.insides.findNeighbours(s, neighbours);
			Iterator nIt = neighbours.iterator();
			while (nIt.hasNext()) {
				((Cluster) nIt.next()).add2(s);
			}
		}

		Iterator it = this.clusters.iterator();
		while (it.hasNext()) {
			((Cluster) it.next()).updateWeight();
		}
		
		if (this.clusters.size() < 1) {
			//this.bestSim = null;
			return;
		}
		
		Collections.sort(this.clusters, clustersComparator);
				
		Cluster bestCluster = (Cluster) this.clusters.getFirst();
		//this.bestSim = bestCluster.getCentroid();		

		long t4 = System.currentTimeMillis();

		//System.out.println("times: " + (t2-t1) + " " + (t3-t2) + " " + (t4-t3));
		//System.out.println("maps: " + n1 + " " + n2);
	}	

	
	/**
	 * The maps that have <code>weight</code> at least <code>bestMap.weight * f</code>.
	 * 
	 * @param f The factor.
	 * @return The maps.
	 */
	public LinkedList getItems(double f, int count, boolean source) {
		LinkedList ret = new LinkedList();
		if (this.clusters.size() < 1) {
			return ret;
		}
		double wMax = ((Cluster) this.clusters.getFirst()).weight();
		Iterator it = this.clusters.iterator();
		int i = 0;
		while (it.hasNext() && i < count) {
			Cluster c = (Cluster) it.next(); 
			double w = c.weight();
			if (w >= wMax*f) {
				Item item = c.getCentroid().itemCopy();
				item.setItemWeight(c.weight);
				if (source) {
					item.setItemSource(c.mergeIntervals());
				}
				ret.add(item);
				i++;
			} else {
				break;
			}
		}
		return ret;
	}
	
	public LinkedList getMaps() {
		return allPoints;
	}
		
	public void reset() {
		//this.bestSim = null;
		this.clusters.clear();
		this.newPoints.clear();
		this.allPoints.clear();
		maxWeight = 0.0;
		minWeight = 0.0;
		n = 0;
		root.insides = new Node();
		root.insides.center = reference;
		root.insides.sqRadius = maxSqDist;
		root.insides.innerThreshold = root.insides.sqRadius - 2*Math.sqrt(maxSqDist)*clusterRadius + clusterSqRadius;
		root.insides.outerThreshold = root.insides.sqRadius + 2*Math.sqrt(maxSqDist)*clusterRadius + clusterSqRadius;
	}
	// getters and setters
	
	
	public double getClusterDiam() {
		return clusterRadius*2;
	}
			
	public int numberOfPoints() {
		return n;
	}

}
