/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.model;

import java.util.LinkedList;

import profi.matching.WeightedClustering;
import profi.matching.WeightedClustering.Item;

public class AffineTransformation implements Transformation, WeightedClustering.Item {

	private static Point[] points; 
	
	public static void setPoints(Point[] points) {
		AffineTransformation.points = points;
	}
	
	protected static void mult(double[][] m, double[][] m1, double[][] m2) {
		for (int r = 0; r < 3; r++) {
			for (int c = 0; c < 3; c++) {
				m[r][c] = 0.0;
				for (int i = 0; i < 3; i++) {
					m[r][c] += m1[r][i]*m2[i][c];
				}
			}
		}
	}
	
	public double[][] m;
	protected boolean lengthPreserving;
	
	public double weight;
	
	public AffineTransformation(String s) {
		m = new double[3][3];
		m[0][0] = m[1][1] = m[2][2] = 1.0;
		lengthPreserving = true;
	}

	public AffineTransformation(AffineTransformation other, String s) {
		m = new double[3][3];
		for (int r = 0; r < 3; r++) {
			for (int c = 0; c < 3; c++) {
				m[r][c] = other.m[r][c];
			}
		}
		this.lengthPreserving = other.lengthPreserving;
	}
	
	public AffineTransformation(AffineTransformation t1, AffineTransformation t2, String s) {
		m = new double[3][3];
		mult(this.m, t1.m, t2.m);
		this.lengthPreserving = t1.lengthPreserving && t2.lengthPreserving;
	}
	
	public AffineTransformation(double m00, double m01, double m10, double m11, double dx, double dy, String s) {
		m = new double[3][3];
		m[0][0] = m00; m[0][1] = m01; m[0][2] = dx;
		m[1][0] = m10; m[1][1] = m11; m[1][2] = dy;
		m[2][0] = 0.0; m[2][1] = 0.0; m[2][2] = 1.0;
		this.lengthPreserving = false;
	}

	public void applyTo(Point p) {
		double x = p.x;
		double y = p.y;
		p.x = x * m[0][0] + y * m[0][1] + m[0][2];
		p.y = x * m[1][0] + y * m[1][1] + m[1][2];
	}

	public double getX(Point p) {
		return p.x * m[0][0] + p.y * m[0][1] + m[0][2];
	}
	
	public double getY(Point p) {
		return p.x * m[1][0] + p.y * m[1][1] + m[1][2];
	}

	public boolean isLengthPreserving() {
		return lengthPreserving;
	}
	
	// implementing WeightedClustering.Item
	
	public double getItemWeight() {		
		return weight;
	}
	
	public void setItemWeight(double w) {		
		weight = w;
	}

	public double itemDist(Item other) {
		return Math.sqrt(itemDistSQ(other));
	}

	public Item itemCopy() {
		return new AffineTransformation(this, null);
	}

	public double itemDistSQ(Item other) {
		AffineTransformation t = (AffineTransformation) other;
		double d2max = 0.0;
		for (int i = 0; i < points.length; i++) {
			double dx = this.getX(points[i]) - t.getX(points[i]);
			double dy = this.getY(points[i]) - t.getY(points[i]);
			d2max = Math.max(d2max, dx*dx + dy*dy);
		}
		return d2max;
	}
	
	public Object itemSource;

	public Object getItemSource() {
		return itemSource;
	}

	public void setItemSource(Object source) {
		itemSource = source;
	}

	private boolean mark;
	
	public void mark(boolean m) {
		mark = m;
	}
	
	public boolean isMarked() {
		return mark;
	}
}
