/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.model;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;

public class Correlations {
	
	public static boolean finding;
	
	public static class Correlation {
				
		public double diameterRec;
		public double length;
		public double start;
		public double end;
		public boolean isPoint;
		public double slopeFactor;
		public double b, m;
		public double a0,a1,a2;
		public Correlation d;
		
		public Correlation() {			
		}
		
		public Correlation(boolean d) {
			a0 = 1.0; a1 = 0.0; a2 = 0.0;
			start = 0.0;
			end = 1.0;
		}
				
		public Correlation(Correlation c) {
			diameterRec = c.diameterRec;
			length = c.length;
			start = c.start;
			end = c.end;
			isPoint = c.isPoint;
			slopeFactor = c.slopeFactor;
			b = c.b; m = c.m;
			a0 = c.a0; a1 = c.a1; a2 = c.a2;
			d = c.d;
		}
		
		public boolean isConst() {
			return (a2 < 1.0e-6 && a1 < 1.0e-6);
		}
		
		public double getA0() {
			return a0;
		}
		
		public double getMax() {
			return Math.max(
					a0 + start * (a1  + a2*start),
					a0 + end * (a1  + a2*end));
		}
		
		public double getMin() {
			return Math.min(
					a0 + start * (a1  + a2*start),
					a0 + end * (a1  + a2*end));
		}
		
		/**
		 * pre: Correlation has to (really) cross y inside [start, end]! 
		 */
		public double getX(double y) {
			double _a0 = a0 - y;
			if (Math.abs(a2) < 1e-6) {
				return -_a0 / a1;
			}
			double p = a1/a2;
			double q = _a0/a2;
			double d = p * p / 4.0 - q;
			double sqrt = Math.sqrt(d);
			double i1 = -p/2 - sqrt;
			if (i1 >= start) {
				return i1;
			} else {
				return -p/2 + sqrt;
			}
		}

		public Correlation flip() {
			Correlation flipped = new Correlation();
			flipped.diameterRec = this.diameterRec;
			flipped.length = this.length;
			flipped.start = 1 - this.end;
			flipped.end = 1 - this.start;
			flipped.isPoint = this.isPoint;
			flipped.slopeFactor = this.slopeFactor;
			flipped.m = -this.m;
			flipped.b = this.b + this.m;
			flipped.a0 = this.a0 + this.a1 + this.a2;
			flipped.a1 = - this.a1 - 2*this.a2;
			flipped.a2 = this.a2;
			return flipped;
		}
		
		
		public static Correlation createDummy(double length) {
			Correlation dummy = new Correlation();
			dummy.diameterRec = 1.0;
			dummy.slopeFactor = 0.0;
			dummy.length = length;
			dummy.start = 0.0;
			dummy.end = 1.0;
			dummy.m = 0.0;
			dummy.b = 0.0;
			dummy.a0 = 0.0;
			dummy.a1 = 0.0;
			dummy.a2 = 0.0;
			return dummy;
		}
		
		public void update() {
			//a0 = slopeFactor - slopeFactor * diameterRec * 4 * b;
			//a1 = -slopeFactor * 4 * m * diameterRec;
			//a2 = 0.0;
			double f = finding ? 40 : 25;
			// TODO: only for finding Figures - 25
			a0 = slopeFactor - b*b*f*slopeFactor*diameterRec*diameterRec;
			a1 = -2*m*b*f*slopeFactor*diameterRec*diameterRec;
			a2 = -m*m*f*slopeFactor*diameterRec*diameterRec;
		}
				
		public double getLength() {
			return length * (end - start);
		}
		
		public double getLength(boolean useD) {
			double w = (Math.max(d.getValue(end), 1.0) + Math.max(d.getValue(start), 1.0)) * 0.5;
			return length * (end - start) / w;
		}
		
		public double getLength(double start, double end) {
			return length * (end - start);
		}
		
		public double getLength(double start, double end, boolean useD) {
			double w = (Math.max(d.getValue(end), 1.0) + Math.max(d.getValue(start), 1.0)) * 0.5;
			return length * (end - start) / w;
		}
		
		public double getIntegral() {
			double e = ((a2 / 3 * end + a1 / 2) * end + a0) * end;
			double s = ((a2 / 3 * start + a1 / 2) * start + a0) * start;			
			double v = (e-s)*length;
			//TODO: 0.0 not necessary if Correlations.merge fixed
			return Math.max(Math.min(v, length * (end - start)), 0.0);
		}

		public double getIntegral(boolean useD) {
			double w = (Math.max(d.getValue(end), 1.0) + Math.max(d.getValue(start), 1.0)) * 0.5;
			double e = ((a2 / 3 * end + a1 / 2) * end + a0) * end;
			double s = ((a2 / 3 * start + a1 / 2) * start + a0) * start;			
			double v = (e-s)*length/w;
			//TODO: 0.0 not necessary if Correlations.merge fixed
			return Math.max(Math.min(v, length * (end - start)), 0.0);
		}
		
		public double getIntegral(double start, double end) {
			double e = ((a2 / 3 * end + a1 / 2) * end + a0) * end;
			double s = ((a2 / 3 * start + a1 / 2) * start + a0) * start;			
			double v = (e-s)*length;
			//TODO: 0.0 not necessary if Correlations.merge fixed
			return Math.max(Math.min(v, length * (end - start)), 0.0);
		}
		
		public double getIntegral(double start, double end, boolean useD) {
			double w = (Math.max(d.getValue(end), 1.0) + Math.max(d.getValue(start), 1.0)) * 0.5;
			double e = ((a2 / 3 * end + a1 / 2) * end + a0) * end;
			double s = ((a2 / 3 * start + a1 / 2) * start + a0) * start;			
			double v = (e-s)*length/w;
			//TODO: 0.0 not necessary if Correlations.merge fixed
			return Math.max(Math.min(v, length * (end - start)), 0.0);
		}
		
		public double getIntegral(Correlation denominator) {
			double end_ = Math.min(this.end, denominator.end);
			double start_ = Math.max(this.start, denominator.start);
			if (end_ <= start_ + 1e-6) return 0.0;
			//TODO!!!
			double d = (Math.max(denominator.getValue(start_), 1.0) + Math.max(denominator.getValue(end_), 1.0)) / 2;
			double e = ((a2 / 3 * end_ + a1 / 2) * end_ + a0) * end_;
			double s = ((a2 / 3 * start_ + a1 / 2) * start_ + a0) * start_;			
			double v = (e-s)*length/d;
			if(e-s > 1.0) System.out.println("here...1"); 
			//TODO: 0.0 not necessary if Correlations.merge fixed
			return Math.max(v, 0.0);
		}
		
		public double getLength(Correlation denominator) {
			double end_ = Math.min(this.end, denominator.end);
			double start_ = Math.max(this.start, denominator.start);
			if (end_ <= start_ + 1e-6) return 0.0;
			//TODO!!!
			double d = (Math.max(denominator.getValue(start_), 1.0) + Math.max(denominator.getValue(end_), 1.0)) / 2;
			double v = (end_-start_)*length/d;
			//TODO: 0.0 not necessary if Correlations.merge fixed
			return Math.max(v, 0.0);
		}
		
		public double getWeightedLength() {
			if (end <= start + 1e-6) return 0.0;
			//TODO!!!
			double d = (Math.max(getValue(start), 1.0) + Math.max(getValue(end), 1.0)) / 2;
			double v = (end - start) * length/d;
			//double e = Math.log(end + b1)/(b2-b1) + Math.log(end + b2)/(b1-b2);
			//double s = Math.log(start + b1)/(b2-b1) + Math.log(start + b2)/(b1-b2);
			//double v = (e-s)*length;
			//TODO: 0.0 not necessary if Correlations.merge fixed
			return Math.max(v, 0.0);
		}
		
		public static double computeValue(double sf, double dr) {
			// TODO: only for finding Figures - 25
			
			return sf * (1.0 - (finding ? 40 : 25) * dr * dr);
		}
		
		public boolean goesOver(Correlation other, double l) {
			if (other == null) return true;
			double vThis = this.getValue(l);
			double vOther = other.getValue(l);
			if (vThis > vOther + 1e-6) {
				return true;
			} else if (vThis + 1e-6 < vOther) {
				return false;
			} else {
				return 2*this.a2*l + this.a1 > 2*other.a2*l + other.a1;
			}
		}
		
		public double getValue(double l) {
			return (a2*l + a1)*l + a0;
		}

		public double[] getIntersections(Correlation other, double min, double max) {
			double[] result;
			double d0 = this.a0 - other.a0;
			double d1 = this.a1 - other.a1;
			double d2 = this.a2 - other.a2;
			if (Math.abs(d2) > 1e-6) {
				double p = d1/d2;
				double q = d0/d2;
				double d = p * p / 4.0 - q;
				if (d > 1e-6) {
					result = new double[2];
					double sqrt = Math.sqrt(d);
					double i1 = -p/2 - sqrt;
					double i2 = -p/2 + sqrt;
					if (i1 > min && i2 < max) {
						result = new double[2];
						result[0] = i1;
						result[1] = i2;
					} else if (i1 > min && i1 < max) {
						result = new double[1];
						result[0] = i1;
					} else if (i2 > min && i2 < max) {
						result = new double[1];
						result[0] = i2;
					} else {
						result = new double[0];
					}
				} else if (d < -1e-6) {
					result = new double[0];
				} else {
					double i1 = -p/2;
					if(i1 > min && i1 < max) {
						result = new double[1];
						result[0] = i1;
					} else {
						result = new double[0];
					}
				}
			} else if (Math.abs(d1) > 1e-6) {
				double i1 = d0 / d1;
				if (i1 > min && i1 < max) {
					result = new double[1];
					result[0] = i1;
				} else {
					result = new double[0];
				}
			} else if (Math.abs(d0) > 1e-6) {
				result = new double[0];
			} else {
				result = new double[3];
			}			
			return result;
		}

		public double next(Correlation other, double min) {
			double max = Math.min(this.end, other.end);
			double d0 = this.a0 - other.a0;
			double d1 = this.a1 - other.a1;
			double d2 = this.a2 - other.a2;
			
			if (
					Math.min(other.getValue(min), other.getValue(max)) > Math.max(getValue(min), getValue(max)) ||
					Math.max(other.getValue(min), other.getValue(max)) < Math.min(getValue(min), getValue(max))) {
				return max;
			}
			
			if (Math.abs(d2) > 1e-6) {
				double p = d1/d2;
				double q = d0/d2;
				double d = p * p / 4.0 - q;
				if (d > 1e-6) {
					double sqrt = Math.sqrt(d);
					double i1 = -p/2 - sqrt;
					double i2 = -p/2 + sqrt;
					if (i1 > min && i2 < max) {
						return i1;
					} else if (i1 > min && i1 < max) {
						return i1;
					} else if (i2 > min && i2 < max) {
						return i2;
					} else {
						return max;
					}
				} else if (d < -1e-6) {
					return max;
				} else {
					double i1 = -p/2;
					if(i1 > min && i1 < max) {
						return i1;
					} else {
						return max;
					}
				}
			} else if (Math.abs(d1) > 1e-6) {
				double i1 = d0 / d1;
				if (i1 > min && i1 < max) {
					return i1;
				} else {
					return max;
				}
			} else if (Math.abs(d0) > 1e-6) {
				return max;
			} else {
				return max;
			}	
		}		
	}	

	public LinkedList correlations;
	public double minValue;
	public double maxValue;
	
	public Correlations() {
		correlations = new LinkedList();
	}
	
	public static Correlations createDummy(double length) {
		Correlations c = new Correlations();
		c.correlations.add(Correlation.createDummy(length));
		c.minValue = 0.0;
		c.maxValue = 0.0;
		return c;
	}
	
	public void getIntegral(double[] result) {			
		result[1] += ((Correlation) correlations.getFirst()).length;
		Iterator it = correlations.iterator();
		while (it.hasNext()) {
			result[0] += ((Correlation) it.next()).getIntegral();
		}
	}
	
	public double getIntegral(Correlations denominator) {
		double result = 0.0;
		double length = 0.0;
		Iterator nIt = this.correlations.iterator();
		Correlation nc = null;
		boolean nNext = true;
		Iterator dIt = denominator.correlations.iterator();
		Correlation dc = null;
		boolean dNext = true;
		while (nNext || dNext) {
			if (nNext && nIt.hasNext()) {
				nc = (Correlation) nIt.next();
			}
			if (dNext && dIt.hasNext()) {
				dc = (Correlation) dIt.next();
			}				
			result += nc.getIntegral(dc);
			length = nc.length;
			nNext = nc.end <= dc.end;
			dNext = dc.end <= nc.end;
		}

		if(result > length) System.out.println("here...3"); 
		return result;
	}
	
	public void repair() {
		if (correlations.size() == 0) {
			//System.out.println("Correlations.repair: empty!");
			return;
		}
		ListIterator it = correlations.listIterator();
		Correlation current = (Correlation) it.next();
		if (current.getValue(current.start) < 1.0 || current.getValue(current.end) < 1.0) {
			current.a0 = 1.0;
			current.a1 = 0.0;
			current.a2 = 0.0;
		}
		if (current.start > 0.0) {
			Correlation c = new Correlation(current);
			c.start = 0.0;
			c.end = current.start;
			c.a0 = 1.0;
			c.a1 = 0.0;
			c.a2 = 0.0;
			it.previous();
			it.add(c);
			it.next();
		}
		Correlation next = null;
		while (it.hasNext()) {
			next = (Correlation) it.next();
			if (current.end < next.start) {
				Correlation c = new Correlation(current);
				c.start = current.end;
				c.end = next.start;
				c.a0 = 1.0;
				c.a1 = 0.0;
				c.a2 = 0.0;
				it.previous();
				it.add(c);
				it.next();
			}				
			current = next;
			if (current.getValue(current.start) < 1.0 || current.getValue(current.end) < 1.0) {
				current.a0 = 1.0;
				current.a1 = 0.0;
				current.a2 = 0.0;
			}
		}			
		if (current.end != 1.0) {
			Correlation c = new Correlation(current);
			c.start = current.end;
			c.end = 1.0;
			c.a0 = 1.0;
			c.a1 = 0.0;
			c.a2 = 0.0;
			correlations.add(c);
		}
	}

	public void add(Correlations other) {
		if (other.maxValue <= 0.0) {
			return;
		}
		if (this.correlations.size() == 0) {
			if (other.correlations.size() == 0) {
				return;
			} else {
				Correlation d = Correlation.createDummy(((Correlation) other.correlations.getFirst()).length);
				this.correlations.add(d);
			}
		}
		//TODO!!!
		
		Correlation dummy = Correlation.createDummy(1.0);
		LinkedList newCorrelations = new LinkedList();
		
		ListIterator sumIt = this.correlations.listIterator();
		Correlation sumC = (Correlation) sumIt.next();
		Iterator otherIt = other.correlations.iterator();
		double s = 0.0;
		double e = 1.0;
		while (otherIt.hasNext()) {
			Correlation otherC = (Correlation) otherIt.next();
			double[] intersections = otherC.getIntersections(dummy, -1e6, 1e6);
							
			if (intersections.length == 0) {
				if (otherC.a0 < 0.0) {
					continue;
				} else {
					s = otherC.start;
					e = otherC.end;
				}
			} else if (intersections.length == 1) {
				if (otherC.a2 == 0) {
					// how can that be?
					//System.out.println("Correlations.add: a2 == 0!");
				} else {
					continue;
				}
			} else if (intersections.length == 2) {
				s = Math.max(otherC.start, intersections[0]);
				e = Math.min(otherC.end, intersections[1]);
			} else {
				continue;
			}
			
			if (s + 1e-6 >= e) {
				continue;
			}
			while (sumIt.hasNext() && sumC.end <= s) {
				if (sumC.end > sumC.start) {
					newCorrelations.add(sumC);
				}
				sumC = (Correlation) sumIt.next();
			}
			if (sumC.start < e) {
				Correlation leftC = new Correlation(sumC);
				leftC.end = s;
				if (leftC.end > leftC.start) {
					newCorrelations.add(leftC);
				}
				sumC.start = s;
			}
			while (sumC.end <= e) {
				sumC.a0 += otherC.a0;
				sumC.a1 += otherC.a1;
				sumC.a2 += otherC.a2;
				if (sumC.end > sumC.start) {
					newCorrelations.add(sumC);
				}
				if (sumIt.hasNext()) {
					sumC = (Correlation) sumIt.next();
				} else {
					break;
				}
			}
			if (sumC.start < e && sumC.end > e) {
				Correlation midC = new Correlation(sumC);
				midC.end = e;
				midC.a0 += otherC.a0;
				midC.a1 += otherC.a1;
				midC.a2 += otherC.a2;
				if (midC.end > midC.start) {
					newCorrelations.add(midC);
				}
				sumC.start = e;
			}
		}
		if (sumC.start < sumC.end && sumC.end > e) {
			newCorrelations.add(sumC);
		}
		while (sumIt.hasNext()) {
			sumC = (Correlation) sumIt.next();
			if (sumC.end > sumC.start) {
				newCorrelations.add(sumC);
			}
		}
		this.correlations = newCorrelations;
	}
	
	private static LinkedList newList = new LinkedList();
	
	public void merge(Correlation[] other) {
		
		/*
		LinkedList copy1 = new LinkedList();
		Iterator cIt = correlations.iterator();
		while (cIt.hasNext()) copy1.addLast(new Correlation((Correlation) cIt.next()));
		Correlation[] copy2 = new Correlation[other.length];
		for (int i = 0; i < other.length; i++) copy2[i] = (other[i] == null) ? null : new Correlation(other[i]);
		*/
		
		newList.clear();
		double min = 1.0;

		//try {
			
		ListIterator iter = this.correlations.listIterator();
		int i = 0;
		
		double event = 0.0;
		double nextEvent = 0.0;
		int eventType = 0;
		
		Correlation cer = null;
		Correlation cor = null;
		Correlation current = null;
					
		while (event < 1.0) {
			switch (eventType) {
			case 0: // both at end
				cer = (Correlation) iter.next();
				cor = other[i];
				i++;
				if (cer.goesOver(cor, event)) {
					newList.addLast(cer);
					current = cer;
				} else {
					newList.addLast(cor);
					current = cor;
				}
				break;
			case 1: // cer at end
				if (current == cer) {
					cer = (Correlation) iter.next();
					if (cer.goesOver(cor, event)) {
						newList.addLast(cer);
						current = cer;
					} else {
						cor.start = event;
						newList.addLast(cor);
						current = cor;
					}
				} else {
					if (!iter.hasNext())
						System.out.println("damn 1");
					cer = (Correlation) iter.next();
					if (cer.goesOver(cor, event)) {
						cor = new Correlation(cor);
						cor.start = event;
						current.end = event;
						newList.addLast(cer);
						current = cer;
					} else {
						// nop
					}
				}
				break;
			case 2: // cor at end
				if (current == cor) {
					cor = other[i];
					i++;
					if (cor.goesOver(cer, event)) {
						newList.addLast(cor);
						current = cor;
					} else {
						cer.start = event;
						newList.addLast(cer);
						current = cer;
					}
				} else {
					cor = other[i];
					i++;
					if (cor.goesOver(cer, event)) {
						cer = new Correlation(cer);
						cer.start = event;
						current.end = event;
						newList.addLast(cor);
						current = cor;
					} else {
						// nop
					}
				}
				break;
			case 3: // intersection
				if (cer == current) {
					if (cer.goesOver(cor, event)) {
						// just touching -> nop
					} else {
						cer = new Correlation(cer);
						cer.start = event;
						current.end = event;
						cor.start = event;
						newList.addLast(cor);
						current = cor;
					}
				} else {
					if (cor.goesOver(cer, event)) {
						// just touching -> nop
					} else {
						cor = new Correlation(cor);
						cor.start = event;
						current.end = event;
						cer.start = event;
						newList.addLast(cer);
						current = cer;
					}						
				}
				break;
			default:
				System.out.println("fool!");				
			}
			min = Math.min(min, current.getValue(event));
			try {
				nextEvent = cer.next(cor, event);
			} catch (NullPointerException ex) {
				for (int oi = 0; oi < other.length; oi++) {
					System.err.print("cor " + oi + ": ");
					System.err.println(other[oi].start + " " + other[oi].end);
				}					
			}
			if (nextEvent <= event) System.out.println("fool! 2");		
			if (nextEvent == cer.end && nextEvent == cor.end) {
				eventType = 0;
			} else if (nextEvent == cer.end) {
				eventType = 1;
			} else if (nextEvent == cor.end) {
				eventType = 2;
			} else {
				eventType = 3;
			}
			event = nextEvent;
		}
		min = Math.min(min, current.getValue(1.0));

		/*
		//if (((Correlation) newList.getLast()).end != 1.0) throw new Exception();
		
		Iterator testIt = newList.iterator();
		double end = 0;
		while (testIt.hasNext()) {
			Correlation c = (Correlation) testIt.next();
			if (end != c.start) {
				System.out.println(end + " ... " + c.start);
				throw new Exception();
			}
			end = c.end;
		}
		
		
		} catch (Exception ex) {

			System.out.println("found:");
			this.correlations = copy1;
			other = copy2;
			
			newList.clear();
			min = 1.0;
			
			ListIterator iter = this.correlations.listIterator();
			int i = 0;
			
			double event = 0.0;
			double nextEvent = 0.0;
			int eventType = 0;
			
			Correlation cer = null;
			Correlation cor = null;
			Correlation current = null;
						
			while (event < 1.0) {
				switch (eventType) {
				case 0: // both at end
					cer = (Correlation) iter.next();
					cor = other[i];
					i++;
					if (cer.goesOver(cor, event)) {
						newList.addLast(cer);
						current = cer;
					} else {
						newList.addLast(cor);
						current = cor;
					}
					break;
				case 1: // cer at end
					if (current == cer) {
						cer = (Correlation) iter.next();
						if (cer.goesOver(cor, event)) {
							newList.addLast(cer);
							current = cer;
						} else {
							cor.start = event;
							newList.addLast(cor);
							current = cor;
						}
					} else {
						if (!iter.hasNext())
							System.out.println("damn 1");
						cer = (Correlation) iter.next();
						if (cer.goesOver(cor, event)) {
							cor = new Correlation(cor);
							cor.start = event;
							current.end = event;
							newList.addLast(cer);
							current = cer;
						} else {
							// nop
						}
					}
					break;
				case 2: // cor at end
					if (current == cor) {
						cor = other[i];
						i++;
						if (cor.goesOver(cer, event)) {
							newList.addLast(cor);
							current = cor;
						} else {
							cer.start = event;
							newList.addLast(cer);
							current = cer;
						}
					} else {
						cor = other[i];
						i++;
						if (cor.goesOver(cer, event)) {
							cer = new Correlation(cer);
							cer.start = event;
							current.end = event;
							newList.addLast(cor);
							current = cor;
						} else {
							// nop
						}
					}
					break;
				case 3: // intersection
					if (cer == current) {
						if (cer.goesOver(cor, event)) {
							// just touching -> nop
						} else {
							cer = new Correlation(cer);
							cer.start = event;
							current.end = event;
							cor.start = event;
							newList.addLast(cor);
							current = cor;
						}
					} else {
						if (cor.goesOver(cer, event)) {
							// just touching -> nop
						} else {
							cor = new Correlation(cor);
							cor.start = event;
							current.end = event;
							cer.start = event;
							newList.addLast(cer);
							current = cer;
						}						
					}
					break;
				default:
					System.out.println("fool!");				
				}
				min = Math.min(min, current.getValue(event));
								
				nextEvent = cer.next(cor, event);
				if (nextEvent <= event) System.out.println("fool! 2");		
				if (nextEvent == cer.end && nextEvent == cor.end) {
					eventType = 0;
				} else if (nextEvent == cer.end) {
					eventType = 1;
				} else if (nextEvent == cor.end) {
					eventType = 2;
				} else {
					eventType = 3;
				}
				event = nextEvent;
			}
			min = Math.min(min, current.getValue(1.0));			
			
			
		}
		*/
		
		
		this.correlations.clear();
		this.correlations.addAll(newList);
		minValue = min;			
	}
	
	public void merge(Correlations other) {
		
		//FIXME: sometimes list gets mixed up!
		if (other.maxValue <= this.minValue) {
			return;
		} else if (this.maxValue <= other.minValue) {
			this.correlations = other.correlations;
		}
		
		
		LinkedList newCorrelations = new LinkedList();
					
		Iterator[] it = new Iterator[2];
		it[0] = this.correlations.iterator();
		it[1] = other.correlations.iterator();
		Correlation[] c = new Correlation[2];
		c[0] = new Correlation((Correlation) it[0].next());
		c[1] = new Correlation((Correlation) it[1].next());
		boolean[] next = new boolean[2];
		next[0] = false;
		next[1] = false;
		int high = c[0].goesOver(c[1], 0.0) ? 0 : 1;
		int low = (high + 1) %2;
		double v0= c[high].getValue(0.0);
		double newMax = v0;
		double newMin = v0;

		double start = 0.0;
		double end = Math.min(c[0].end, c[1].end);
		double[] intersections = c[0].getIntersections(c[1], start, end);
		int idx = (intersections.length != 3) ? 0 : 3;
		boolean go = true;
		
		while (go) {
			if(idx >= intersections.length) {
				boolean closed = false;
				if (end == c[high].end) {
					newCorrelations.add(c[high]);
					c[low].start = end;
					closed = true;
					double value = c[high].getValue(end);
					newMax = Math.max(newMax, value);
					newMin = Math.min(newMin, value);
					if (it[high].hasNext()) {
						c[high] = new Correlation((Correlation) it[high].next());
					} else {
						break;
					}
				}
				if (end == c[low].end) {
					if (it[low].hasNext()) {
						c[low] = new Correlation((Correlation) it[low].next());
					} else {
						if(!closed) {
							//System.err.println("Correlations.merge: List not Closed!!!");
							//System.err.println(c[high].end + " " + c[low].end);
							//System.err.println("");
						}
						break;
					}
				}
				start = Math.max(c[0].start, c[1].start);
				end = Math.min(c[0].end, c[1].end);
				intersections = c[0].getIntersections(c[1], start, end);
				idx = (intersections.length != 3) ? 0 : 3;
				if (c[high].goesOver(c[low], start)) {
					if (closed) {
						double value = c[high].getValue(start);
						c[high].start = start;
						newMax = Math.max(newMax, value);
						newMin = Math.min(newMin, value);
					}
				} else {
					high = (high + 1) %2;
					low = (low + 1) %2;
					c[high].start = start;
					double value = c[high].getValue(start);
					newMax = Math.max(newMax, value);
					newMin = Math.min(newMin, value);
				}
			} else {
				if(c[high].goesOver(c[low], intersections[idx])) {
					//nop
				} else {
					Correlation cNew = new Correlation(c[high]);
					cNew.end = intersections[idx];
					newCorrelations.add(cNew);
					
					high = (high + 1) %2;
					low = (low + 1) %2;
					double value = c[high].getValue(intersections[idx]);
					newMax = Math.max(newMax, value);
					newMin = Math.min(newMin, value);
					
					c[high].start = intersections[idx];				
				}
				idx++;
			}
		}
		
		this.correlations = newCorrelations;
		this.minValue = newMin;
		this.maxValue = newMax;
	}
}
