/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.model;

import java.util.Iterator;
import java.util.LinkedList;

/**
 * Douglas-Peucker Algorithm with slight modification.
 */
public class DouglasPeucker {

	public static Polyline simplify(Polyline p, double epsilon) {
		Point[] points = new Point[p.getNumberOfPoints() + (p.isClosed() ? 1 : 0)];
		Iterator it = p.getPointIterator();
		int i = 0;
		while (it.hasNext()) {
			points[i] = (Point) it.next();
			i++;
		}
		if (p.isClosed()) {
			points[points.length-1] = points[0];
		}
		LinkedList list = new LinkedList();
		doSimplify(points, list, 0, points.length-1, epsilon);
		if (!p.isClosed()) {
			list.add(points[points.length-1]);
		}
		Polyline ps = new Polyline(p.isClosed());
		ps.addPoints(list);
		return ps;
	}
	

	private static void doSimplify(Point[] points, LinkedList list, int i, int j, double epsilon) {
		Point left = points[i];
		if (j-i <= 1) {
			list.add(left);
			return;
		}
		Point right = points[j];
		double d = left.distanceSQ(right);
		double maxSqDist = 0.0;
		int maxIdx = i+1;
		if (d < epsilon*epsilon*1e-4) {
			for (int idx = i + 1; idx < j; idx++) {
				double sqDist = left.distanceSQ(points[idx]);
				if (sqDist > maxSqDist) {
					maxSqDist = sqDist;
					maxIdx = idx;
				}
			}
		} else {
			for (int idx = i + 1; idx < j; idx++) {
				double sqDist = sqDist(left, right, points[idx]);
				double ld = left.distanceSQ(points[idx]);
				double rd = right.distanceSQ(points[idx]);
				if (ld > d) sqDist = Math.max(sqDist, rd); // for not neglecting the endpoints
				if (rd > d) sqDist = Math.max(sqDist, ld);
				if (sqDist > maxSqDist) {
					maxSqDist = sqDist;
					maxIdx = idx;
				}
			}
		}
		if (maxSqDist > epsilon*epsilon) {
			doSimplify(points, list, i, maxIdx, epsilon);
			doSimplify(points, list, maxIdx, j, epsilon);
		} else {
			list.add(points[i]);
		}
	}
	
	public static Polyline simplifyCrossingFree(Polyline p, double epsilon) {
		Point[] points = new Point[p.getNumberOfPoints() + (p.isClosed() ? 1 : 0)];
		Iterator it = p.getPointIterator();
		int i = 0;
		while (it.hasNext()) {
			points[i] = (Point) it.next();
			i++;
		}
		if (p.isClosed()) {
			points[points.length-1] = points[0];
		}
		LinkedList list = new LinkedList();
		doSimplifyCrossingFree(points, list, 0, points.length-1, epsilon);
		if (!p.isClosed()) {
			list.add(points[points.length-1]);
		}
		Polyline ps = new Polyline(p.isClosed());
		ps.addPoints(list);
		return ps;
	}
	
	private static void doSimplifyCrossingFree(Point[] points, LinkedList list, int i, int j, double epsilon) {
	
		Point left = points[i];
		if (j-i <= 1) {
			list.add(left);
			return;
		}
		Point right = points[j];
		double d = left.distanceSQ(right);
		double maxSqDist = 0.0;
		int maxIdx = i+1;
		if (d < epsilon*epsilon*1e-4) {
			for (int idx = i + 1; idx < j; idx++) {
				double sqDist = left.distanceSQ(points[idx]);
				if (sqDist > maxSqDist) {
					maxSqDist = sqDist;
					maxIdx = idx;
				}
			}
		} else {
			for (int idx = i + 1; idx < j; idx++) {
				double sqDist = sqDist(left, right, points[idx]);
				double ld = left.distanceSQ(points[idx]);
				double rd = right.distanceSQ(points[idx]);
				if (ld > d) sqDist = Math.max(sqDist, rd); // for not neglecting the endpoints
				if (rd > d) sqDist = Math.max(sqDist, ld);
				if (sqDist > maxSqDist) {
					maxSqDist = sqDist;
					maxIdx = idx;
				}
			}
		}
		if (maxSqDist > epsilon*epsilon || crossing(i, j, points)) {
			doSimplifyCrossingFree(points, list, i, maxIdx, epsilon);
			doSimplifyCrossingFree(points, list, maxIdx, j, epsilon);
		} else {
			list.add(points[i]);
		}
	}
	
	private static boolean crossing(int start,  int end, Point[] points) {
		//if (true) return false;
		for (int i = 0; i < start - 1; i++) {
			if (
					Point.isLeft(points[start], points[end], points[i]) && 
					Point.isLeft(points[end], points[start], points[i+1]) && 
					Point.isLeft(points[i+1], points[i], points[start]) && 
					Point.isLeft(points[i], points[i+1], points[end])) return true;
			if (
					Point.isLeft(points[start], points[end], points[i+1]) && 
					Point.isLeft(points[end], points[start], points[i]) && 
					Point.isLeft(points[i], points[i+1], points[start]) && 
					Point.isLeft(points[i+1], points[i], points[end])) return true;
			if (
					Point.isLeft(points[start], points[end], points[i]) && 
					Point.isLeft(points[end], points[start], points[i]) &&
					(points[start].x < points[i].x && points[i].x < points[end].x || points[start].x > points[i].x && points[i].x > points[end].x) &&
					(points[start].y < points[i].y && points[i].y < points[end].y || points[start].y > points[i].y && points[i].y > points[end].y)) return true;
			
			if (
					Point.isLeft(points[start], points[end], points[i+1]) && 
					Point.isLeft(points[end], points[start], points[i+1]) &&
					(points[start].x < points[i+1].x && points[i+1].x < points[end].x || points[start].x > points[i+1].x && points[i+1].x > points[end].x) &&
					(points[start].y < points[i+1].y && points[i+1].y < points[end].y || points[start].y > points[i+1].y && points[i+1].y > points[end].y)) return true;
		}
		for (int i = end + 1; i < points.length - 1; i++) {
			if (
					Point.isLeft(points[start], points[end], points[i]) && 
					Point.isLeft(points[end], points[start], points[i+1]) && 
					Point.isLeft(points[i+1], points[i], points[start]) && 
					Point.isLeft(points[i], points[i+1], points[end])) return true;
			if (
					Point.isLeft(points[start], points[end], points[i+1]) && 
					Point.isLeft(points[end], points[start], points[i]) && 
					Point.isLeft(points[i], points[i+1], points[start]) && 
					Point.isLeft(points[i+1], points[i], points[end])) return true;
			if (
					Point.isLeft(points[start], points[end], points[i]) && 
					Point.isLeft(points[end], points[start], points[i]) &&
					(points[start].x < points[i].x && points[i].x < points[end].x || points[start].x > points[i].x && points[i].x > points[end].x) &&
					(points[start].y < points[i].y && points[i].y < points[end].y || points[start].y > points[i].y && points[i].y > points[end].y)) return true;
			
			if (
					Point.isLeft(points[start], points[end], points[i+1]) && 
					Point.isLeft(points[end], points[start], points[i+1]) &&
					(points[start].x < points[i+1].x && points[i+1].x < points[end].x || points[start].x > points[i+1].x && points[i+1].x > points[end].x) &&
					(points[start].y < points[i+1].y && points[i+1].y < points[end].y || points[start].y > points[i+1].y && points[i+1].y > points[end].y)) return true;
		}
		return false;
	}
	
	private static double sqDist(Point p1, Point p2, Point other) {
		double dx=p2.getX()-p1.getX();
		double dy=p2.getY()-p1.getY();
		double d = dy*(other.x - p1.x) - dx*(other.y - p1.y);
		return d*d / (dx*dx+dy*dy);
	}
}
