/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.model;

public class Point {

	public final static Point ORIGIN = new Point();

	public double x;

	public double y;

	public Point() {
		this(0.0, 0.0);
	}

	public Point(double x, double y) {
		this.x = x;
		this.y = y;
	}
	
	public boolean equals(Point other) {
		return this.x == other.x && this.y == other.y;
	}


	public Point(Point point) {
		this(point.x, point.y);
	}

	public void add(Point p) {
		x += p.x;
		y += p.y;
	}

	public void subtract(Point p) {
		x -= p.x;
		y -= p.y;
	}

	public void mult(double factor) {
		x *= factor;
		y *= factor;
	}

	public double getX() {
		return x;
	}

	public double getY() {
		return y;
	}

	public void setX(double x) {
		this.x = x;
	}

	public void setY(double y) {
		this.y = y;
	}	

	public static boolean isLeft(Point p1, Point p2, Point p) {
		return ( (p2.x-p1.x) * (p.y - p1.y) - (p.x - p1.x) * (p2.y-p1.y) ) > 0.0;
	}

	public void setX(float pF) {
		x = pF;
	}

	public void setY(float pF) {
		y = pF;
	}
		
	public void normalize() {
		double l = this.distance();
		this.x /= l;
		this.y /= l;
	}

	public double distance(Point other) {
		return Math.sqrt(
				(this.x - other.x)*(this.x - other.x) +
				(this.y - other.y)*(this.y - other.y));
	}
	
	public double distanceSQ(Point other) {
		return 
			(this.x - other.x)*(this.x - other.x) +
			(this.y - other.y)*(this.y - other.y);
	}

	public double distance() {
		return Math.sqrt(this.x * this.x + this.y * this.y);
	}

	public double distanceSQ() {
		return this.x * this.x + this.y * this.y;
	}
	
	public Point intermediatePoint(double offset, Point that) {
		if (offset < 0 || offset > 1)
			throw new IllegalArgumentException("offset = " + offset);
		if (offset == 0)
			return new Point(this);
		if (offset == 1)
			return new Point(that);
		return new Point((that.x - this.x) * offset + this.x, (that.y - this.y)
				* offset + this.y);
	}

	public String toString() {
		return "Point (" + x + ", " + y + ")";
	}
	
	public static double innerProduct(Point p1, Point p2, Point p3) {
		double dx1 = p2.x - p1.x;
		double dy1 = p2.y - p1.y;
		double dx2 = p3.x - p2.x;
		double dy2 = p3.y - p2.y;
		return dx1 * dx2 + dy1 * dy2;
	}
	
	public double innerProduct(Point other) {
		return this.x * other.x + this.y * other.y;
	}
	
	public void rotate(double angle) {
		double sin = Math.sin(angle);
		double cos = Math.cos(angle);
		double newX = x*cos - y*sin;
		y = x*sin + y*cos;
		x = newX;		
	}
	
}
