/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.model;

import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import org.w3c.dom.Element;

import profi.model.Correlations.Correlation;

public class Polyline {

private static final double EPS = 1e-6;
	
	private static final Correlation DEFAULT = new Correlation(true);
	
	public static class VertexIterator implements Iterator {
		
		Iterator it;
		Object first;
		boolean hasNext;
		
		public VertexIterator(Iterator it, Object first) {
			this.it = it;
			this.first = first;
			this.hasNext = true;
		}

		public boolean hasNext() {
			return hasNext;
		}

		public Object next() {
			if (it.hasNext()) {
				return it.next();
			} else if (hasNext) {
				hasNext = false;
				return first;
			} else {
				throw new NoSuchElementException();
			}
		}

		public void remove() {
			throw new UnsupportedOperationException();
		}		
	}
	
	public static class WeightedLine {
		public Point start;
		public Point end;
		public LinkedList externCorrelations;
		public LinkedList internCorrelations;
		public LinkedList mergedCorrelations;
		
		public WeightedLine() {
			start = null;
			end = null;
			externCorrelations = new LinkedList();
			internCorrelations = new LinkedList();
		}

		public WeightedLine(Point s, Point e) {
			start = s;
			end = e;
			externCorrelations = new LinkedList();
			internCorrelations = new LinkedList();
		}
		
		public WeightedLine(WeightedLine other, Point s, Point e) {
			this.start = s;
			this.end = e;
			externCorrelations = new LinkedList();
			Iterator it = other.externCorrelations.iterator();
			while (it.hasNext()) {
				this.externCorrelations.add(new Correlation((Correlation) it.next()));
			}
			internCorrelations = new LinkedList();
			it = other.internCorrelations.iterator();
			while (it.hasNext()) {
				this.internCorrelations.add(new Correlation((Correlation) it.next()));
			}
			if (other.mergedCorrelations != null) {
				mergedCorrelations = new LinkedList();
				it = other.mergedCorrelations.iterator();
				while (it.hasNext()) {
					this.mergedCorrelations.add(new Correlation((Correlation) it.next()));
				}
			}
		}
		
		public void mergeCorrelations() {
			mergedCorrelations = new LinkedList();
			
			Iterator nIt = this.externCorrelations.iterator();
			Correlation nc = null;
			boolean nNext = nIt.hasNext();
			if (!nNext) return;
			double nStart = 0.0;
			
			Iterator dIt = this.internCorrelations.iterator();
			Correlation dc = DEFAULT;
			boolean dNext = dIt.hasNext();
			
			while (nNext || dNext) {
				if (nNext) {
					if (!nIt.hasNext()) {
						return;
					}
					nc = (Correlation) nIt.next();
					nStart = nc.start;
				}
				if (dNext) {
					if (!dIt.hasNext()) {
						dc = DEFAULT;
					} else {
						dc = (Correlation) dIt.next();
					}
				}
				
				if (nc.end == dc.end) {
					Correlation m = new Correlation(nc);
					m.start = nStart;
					m.d = dc;
					mergedCorrelations.add(m);
					nNext = true;
					dNext = true;
				} else if (nc.end < dc.end) {
					Correlation m = new Correlation(nc);
					m.start = nStart;
					m.d = dc;
					mergedCorrelations.add(m);
					nNext = true;
					dNext = false;
				} else {
					Correlation m = new Correlation(nc);
					m.start = nStart;
					m.d = dc;
					mergedCorrelations.add(m);
					nNext = false;
					dNext = true;
					nStart = dc.end;
				}				
			}			
		}

		public double getWeightedLength() {
			double sum = 0.0;
			Iterator it = this.internCorrelations.iterator();
			while (it.hasNext()) {
				Correlation c = ((Correlation) it.next());
				double v = c.getWeightedLength();
				if (v > c.length * (c.end - c.start)) {
					//System.out.println("error 1!");
				}
				sum += v;
			}
			if (sum > start.distance(end)) {
				//System.out.println("error 2!: " + ((sum - start.distance(end))/sum)+ " " + start.distance(end));
			}
			return sum;
		}
		
		public void flip() {
			start.x = -start.x;
			end.x = -end.x;
			Point tmp = start;
			start = end;
			end = tmp;
			LinkedList ic = new LinkedList();
			Iterator it = internCorrelations.iterator();
			while (it.hasNext()) {
				Correlation c = (Correlation) it.next();				
				ic.addFirst(c.flip());
			}
			internCorrelations = ic;
			mergedCorrelations = null;
		}

		public void updateLength() {
			double l = start.distance(end);
			for (Iterator it = internCorrelations.iterator(); it.hasNext();) {
				((Correlation) it.next()).length = l;
			}
			mergedCorrelations = null;
		}
		
		public double getWeight(boolean useDenominator) {
			double sum = 0.0;
			Iterator nIt = this.externCorrelations.iterator();
			Correlation nc = null;
			boolean nNext = nIt.hasNext();
			
			if (useDenominator) {
				Iterator dIt = this.internCorrelations.iterator();
				Correlation dc = null;
				boolean dNext = dIt.hasNext();
				while (nNext || dNext) {
					if (nNext) {
						nc = (Correlation) nIt.next();
					}
					if (dNext) {
						dc = (Correlation) dIt.next();
					}
					if (dc == null) {
						if (nc == null) {
							return 0.0;
						}
						sum += nc.getIntegral();
						nNext = nIt.hasNext();
					} else {
						if (nc == null) {
							return 0.0;
						}
						sum += nc.getIntegral(dc);
						dNext = dc.end <= nc.end && dIt.hasNext();
						nNext = nc.end <= dc.end && nIt.hasNext();
					}
				}
			} else {
				while (nIt.hasNext()) {
					if (nNext) {
						nc = (Correlation) nIt.next();
						sum += nc.getIntegral();
					}
				}
			}
			return sum;
		}

		public void getWeight(double[] results, boolean useDenominator) {
			Iterator nIt = this.externCorrelations.iterator();
			Correlation nc = null;
			boolean nNext = nIt.hasNext();
			
			if (useDenominator) {
				Iterator dIt = this.internCorrelations.iterator();
				Correlation dc = null;
				boolean dNext = dIt.hasNext();
				while (nNext || dNext) {
					if (nNext) {
						nc = (Correlation) nIt.next();
						nNext = false;
					}
					if (dNext) {
						dc = (Correlation) dIt.next();
						dNext = false;
					}
					if (dc == null) {
						if (nc == null) {
							//nop
						} else {
							results[0] += nc.getIntegral();
							results[1] += (nc.end - nc.start) * nc.length;
							nNext = nIt.hasNext();
						}
					} else {
						if (nc == null) {
							// nop
						} else {
							results[0] += nc.getIntegral(dc);
							results[1] += nc.getLength(dc);
							dNext = dc.end <= nc.end && dIt.hasNext();
							nNext = nc.end <= dc.end && nIt.hasNext();
						}
					}
				}
			} else {
				while (nIt.hasNext()) {
					if (nNext) {
						nc = (Correlation) nIt.next();
						results[0] += nc.getIntegral();
						results[1] += (nc.end - nc.start) * nc.length;
					}
				}
			}
		}

		public void getWeights(double[] thresholds, double[][] results, boolean weighted) {
			int k = thresholds.length;
			//TODO: weighted			
			if (!weighted) {
				Iterator iter = this.externCorrelations.iterator();
				while (iter.hasNext()) {
					Correlation c = (Correlation) iter.next();
					double max = c.getMax();
					double min = c.getMin();
					double lastThreshold = 1.0;
					double lastX = 0.0;
					for (int i = 0; i < k; i++) {
						if (min > thresholds[i]) {
							if (max <= lastThreshold) {
								results[i][0] += c.getIntegral();
								results[i][2] += c.getLength();
							} else {
								double x = lastX;
								if (c.getValue(c.start) > c.getValue(c.end)) {
									results[i][0] += c.getIntegral(x, c.end);
									results[i][2] += c.getLength(x, c.end);
								} else {
									results[i][0] += c.getIntegral(c.start, x);
									results[i][2] += c.getLength(c.start, x);
								}
							}
							break;
						} else if (c.isConst() && c.getA0() >= thresholds[i] - EPS) {
							for(int j = i; j < thresholds.length && c.getA0() <= thresholds[j]; j++) {
								results[j][1] += c.getIntegral();
								results[j][3] += c.getLength();
							}
							break;
						} else if (max <= thresholds[i]) {
							continue;
						} else if (max > thresholds[i] && min <= thresholds[i]) {
							double x = c.getX(thresholds[i]);
							if (max > lastThreshold) {
								if (c.getValue(c.start) > c.getValue(c.end)) {
									results[i][0] += c.getIntegral(lastX, x);
									results[i][2] += c.getLength(lastX, x);
								} else {
									results[i][0] += c.getIntegral(x, lastX);
									results[i][2] += c.getLength(x, lastX);
								}
							} else {
								if (c.getValue(c.start) > c.getValue(c.end)) {
									results[i][0] += c.getIntegral(c.start, x);
									results[i][2] += c.getLength(c.start, x);
								} else {
									results[i][0] += c.getIntegral(x, c.end);
									results[i][2] += c.getLength(x, c.end);
								}
							}
							lastX = x;
						} else {
							//what else?!
							System.out.println("what else?!");
						}
					}
				}
			} else {
				Iterator iter = this.mergedCorrelations.iterator();
				while (iter.hasNext()) {
					Correlation c = (Correlation) iter.next();
					double max = c.getMax();
					double min = c.getMin();
					double lastThreshold = 1.0;
					double lastX = 0.0;
					for (int i = 0; i < k; i++) {
						if (min > thresholds[i]) {
							if (max <= lastThreshold) {
								results[i][0] += c.getIntegral(true);
								results[i][2] += c.getLength(true);
							} else {
								double x = lastX;
								if (c.getValue(c.start) > c.getValue(c.end)) {
									results[i][0] += c.getIntegral(x, c.end, true);
									results[i][2] += c.getLength(x, c.end, true);
								} else {
									results[i][0] += c.getIntegral(c.start, x, true);
									results[i][2] += c.getLength(c.start, x, true);
								}
							}
							break;
						} else if (c.isConst() && c.getA0() >= thresholds[i] - EPS) {
							for(int j = i; j < thresholds.length && c.getA0() <= thresholds[j]; j++) {
								results[j][1] += c.getIntegral(true);
								results[j][2] += c.getLength(true);
							}
							break;
						} else if (max <= thresholds[i]) {
							continue;
						} else if (max > thresholds[i] && min <= thresholds[i]) {
							double x = c.getX(thresholds[i]);
							if (max > lastThreshold) {
								if (c.getValue(c.start) > c.getValue(c.end)) {
									results[i][0] += c.getIntegral(lastX, x, true);
									results[i][2] += c.getLength(lastX, x, true);
								} else {
									results[i][0] += c.getIntegral(x, lastX, true);
									results[i][2] += c.getLength(x, lastX, true);
								}
							} else {
								if (c.getValue(c.start) > c.getValue(c.end)) {
									results[i][0] += c.getIntegral(c.start, x, true);
									results[i][2] += c.getLength(c.start, x, true);
								} else {
									results[i][0] += c.getIntegral(x, c.end, true);
									results[i][2] += c.getLength(x, c.end, true);
								}
							}
							lastX = x;
						} else {
							//what else?!
							System.out.println("what else?!");
						}
					}
				}				
			}
		}

		public void getLength(double threshold, double[] result, boolean weighted) {
			if (!weighted) {
				Iterator it = this.externCorrelations.iterator();
				while (it.hasNext()) {
					Correlation c = (Correlation) it.next();
					double max = c.getMax();
					double min = c.getMin();
					if (min > threshold) {
						result[0] += c.length * (c.end - c.start);
					} else if (c.isConst() && max >= threshold - EPS) {
						result[1] += c.length * (c.end - c.start);
					} else if (max < threshold) {
						//nop
					} else {
						double x = c.getX(threshold);
						if(x < c.start || x > c.end) {
							System.out.println("x out");
						}
						else if (c.getValue(c.start) > c.getValue(c.end)) {
							result[0] += c.length * (x - c.start);
						} else {
							result[0] += c.length * (c.end - x);
						}
					}
				}
			} else {
				Iterator it = this.mergedCorrelations.iterator();
				while (it.hasNext()) {
					Correlation c = (Correlation) it.next();
					double max = c.getMax();
					double min = c.getMin();
					if (min > threshold) {
						result[0] += c.getLength(true);
					} else if (c.isConst() && max >= threshold - EPS) {
						result[1] += c.getLength(true);
					} else if (max < threshold) {
						//nop
					} else {
						double x = Math.max(Math.min(c.getX(threshold), c.end), c.start);
						if (c.getValue(c.start) > c.getValue(c.end)) {
							result[0] += c.getLength(c.start, x, true);
						} else {
							result[0] += c.getLength(x, c.end, true);
						}
					}
				}				
			}
		}

		public void getLength(double[] hist, double[] thresholds, boolean weighted) {
			//TODO...
			
			Iterator it = this.externCorrelations.iterator();
			while (it.hasNext()) {
				Correlation c = (Correlation) it.next();
				double max = Math.min(1.0, Math.max(0.0, c.getMax()));
				double min = Math.min(1.0, Math.max(0.0, c.getMin()));
				int maxIdx = 0;
				int minIdx = maxIdx;
				
				if (min == 1.0 && max == 1.0) {
					minIdx = 1;
				} else if (min == 0.0 && max == 0.0) {
					maxIdx = thresholds.length - 2;
					minIdx = thresholds.length - 1;
				} else {				
					while (maxIdx + 1 < thresholds.length && thresholds[maxIdx+1] > max) {
						maxIdx++;
					}
					while (minIdx + 1 < thresholds.length && thresholds[minIdx] >= min) {
						minIdx++;
					}
				}
				int k = minIdx - maxIdx - 1;
				double[] ts = new double[k+2];
				if (c.getValue(c.start) < c.getValue(c.end)) {
					ts[0]   = 1.0;
					ts[k+1] = 0.0;
				} else {
					ts[0]   = 0.0;
					ts[k+1] = 1.0;
				}
				for (int i = 0; i < k; i++) {
					ts[i+1] = c.getX(thresholds[maxIdx + 1 + k]);
					hist[maxIdx + i] += Math.abs(ts[i+1] - ts[i]) * c.length;
				}
				hist[maxIdx + k] += Math.abs(ts[k+1] - ts[k]) * c.length;				
			}
		}	
		
		public double getRelativeWeight() {
			double length = start.distance(end);
			double sum = 0.0;
			Iterator it = this.externCorrelations.iterator();
			while (it.hasNext()) {
				sum += ((Correlation) it.next()).getIntegral();
			}
			return sum / length;
		}
	}
	
	public static Polyline readData_03(Element e) {
		boolean std = true;
		String closed = null;
		if (e.getTagName().equalsIgnoreCase("pl")) {
			closed = e.getAttributes().getNamedItem("cl").getNodeValue();
		} else if (e.getTagName().equalsIgnoreCase("convexHull")) {
			closed = "i";
		} else {
			std = false;
			closed = e.getAttributes().getNamedItem("cl").getNodeValue();
		}		
		int skip = closed.equalsIgnoreCase("d") ? (std ? 2 : 3) : 0; 
		Polyline pl = new Polyline(!closed.equalsIgnoreCase("n"));
		StringTokenizer tokens = new StringTokenizer(e.getTextContent());
		pl.addPoint(new Point(Double.parseDouble(tokens.nextToken()), Double.parseDouble(tokens.nextToken())));
		if (std) {
			while (tokens.countTokens() >= 2 + skip) {
				pl.addPoint(new Point(Double.parseDouble(tokens.nextToken()), Double.parseDouble(tokens.nextToken())));
			}
		} else {
			while (tokens.countTokens() >= 3 + skip) {
				pl.addPoint(new Point(Double.parseDouble(tokens.nextToken()), Double.parseDouble(tokens.nextToken())));
				tokens.nextToken();
			}
		}		
		return pl;
	}
	
	protected LinkedList lines;
	
	private LinkedList points;
	
	private boolean closed;
	
	private double length;	
	private boolean valid;
	
	private boolean weighted;
	
	private Iterator pointsIt;
	private boolean hasNext;
	
	public Polyline(boolean closed) {
		points = new LinkedList();
		this.closed = closed;
		length = 0;
		valid = true;
		weighted = false;
	}

	public Polyline(Polyline other) {
		points = new LinkedList();
		Iterator it = other.getPointIterator();
		while (it.hasNext()) {
			points.add(new Point((Point) it.next()));
		}
		this.closed = other.closed;
		length = other.length;
		valid = other.valid;
		if (other.lines != null) {
			this.lines = new LinkedList();
			it = other.lines.iterator();
			resetPointsIterator();
			Point s = next();
			while (it.hasNext()) {
				Point e = next();
				this.lines.add(new WeightedLine((WeightedLine) it.next(), s, e));
				s = e;
			}
		}
		weighted = other.weighted;
	}
	
	public void makeLines() {
		lines = new LinkedList();
		if (points.size() < 2) {
			return;
		}
		Iterator it = points.iterator();
		Point s = (Point) it.next();
		while (it.hasNext()) {
			Point e = (Point) it.next();
			lines.add(new WeightedLine(s, e));
			s = e;
		}
		if (closed) {
			lines.add(new WeightedLine(s, (Point) points.getFirst()));
		}
	}
	
	public LinkedList getLines() {
		return lines;
	}
	
	public void setClosed(boolean closed) {
		this.closed = closed;
		valid = false;
	}
	
	public double getWeight() {
		//todo...
		return this.getLength();
	}
	
	public boolean isClosed() {
		return closed;
	}
	
	private void resetPointsIterator() {
		pointsIt = points.iterator();
		hasNext = pointsIt.hasNext();
	}
	
	protected boolean hasNext() {
		return hasNext;
	}
	
	protected Point next() {
		if (closed) {
			if (pointsIt.hasNext()) {
				return (Point) pointsIt.next();
			} else if (hasNext) {
				hasNext = false;
				return (Point) points.getFirst();
			} else {
				throw new RuntimeException();
			}
		} else {
			Point p = (Point) pointsIt.next();
			hasNext = pointsIt.hasNext();
			return p;
		}
	}
	
	public void addPoint(Point p) {
		points.add(p);
		valid = false;
	}
	
	public void addPoints(Collection points) {
		Iterator pit = points.iterator();
		while (pit.hasNext()) {
			addPoint((Point) pit.next());				
		}
		valid = false;
	}
	
	public void apply(Transformation t) {
		for (Iterator i = points.iterator(); i.hasNext();) {
			t.applyTo((Point) i.next());
		}
		valid &= t.isLengthPreserving();
		if (!t.isLengthPreserving() && lines!= null) {
			for (Iterator i = lines.iterator(); i.hasNext();) {
				((WeightedLine) i.next()).updateLength();
			}
		}
	}
	
	public Iterator getPointIterator() {
		return points.iterator();
	}

	public Iterator getVertexIterator() {
		if (closed) {
			return new VertexIterator(points.iterator(), points.getFirst());
		} else {
			return points.iterator();
		}
	}
	
	public LinkedList getPoints() {
		return points;
	}
	
	public double getLength() {
		if (!valid) {
			updateLength();
		}
		return length;
	}
	
	public int getNumberOfPoints() {
		return points.size();
	}
	
	public void updateLength() {
		length = 0;
		if (points.size() >= 2) {
			Iterator iter = points.iterator();
			Point first = (Point) iter.next();
			Point a = first;
			Point b = null;
			while (iter.hasNext()) {
				b = (Point) iter.next();
				length += a.distance(b);
				a = b;
			}
			if (closed) {
				length += b.distance(first);
			}
		}
		valid = true;
	}
		
	public Point getPoint(int k) {
		Iterator it = points.iterator();
		for (int i = 0; i < k; i++, it.next());
		return (Point) it.next();
	}
	
	public int size() {
		return points.size();
	}		
	
	public Box getBoundingBox() {
		Iterator iter = points.iterator();
		if (!iter.hasNext()) {
			return null;
		}
		Point p = (Point) iter.next();
		double xMin = p.x;
		double xMax = p.x;
		double yMin = p.y;
		double yMax = p.y;
		while (iter.hasNext()) {
			p = (Point) iter.next();
			if (p.x < xMin)
				xMin = p.x;
			else if (p.x > xMax)
				xMax = p.x;
			if (p.y < yMin)
				yMin = p.y;
			else if (p.y > yMax)
				yMax = p.y;
		}
		return new Box(xMin, yMin, xMax, yMax);
	}
	
	public double getDiameter() {
		double d = 0.0;
		Iterator it1 = getPointIterator();
		for (int i = 0; it1.hasNext(); i++) {
			Point p1 = (Point) it1.next();
			Iterator it2 = getPointIterator();
			for (int j = 0; j < i; j++) {
				Point p2 = (Point) it2.next();
				d = Math.max(d, p1.distanceSQ(p2));
			}
		}
		return Math.sqrt(d);
	}
	
	public void consolidate(double eps) {
		if (this.points.size() < 1) return;
		double eps2 = eps*eps;
		Iterator pointsIt = getPointIterator();
		Point next = (Point) pointsIt.next();
		boolean changed = false;
		while (pointsIt.hasNext()) {
			Point current = (Point) pointsIt.next();
			if (next.distanceSQ(current) < eps2) {
				pointsIt.remove();
				changed = true;
			} else {
				next = current;
			}
		}
		if (((Point) points.getFirst()).distanceSQ((Point) points.getLast()) < eps2) {
			points.removeLast();
			changed = true;
		}
		valid &= !changed;		
	}
	
	public double getMinDist(Polyline other) {
		Point p11 = (Point) this.points.getFirst();
		Point p12 = (Point) this.points.getLast();
		Point p21 = (Point) other.points.getFirst();
		Point p22 = (Point) other.points.getLast();
		double dMin = Double.MAX_VALUE;
		double d11 = p11.distanceSQ(p21); dMin = Math.min(dMin, d11);
		double d12 = p11.distanceSQ(p22); dMin = Math.min(dMin, d12);
		double d21 = p12.distanceSQ(p21); dMin = Math.min(dMin, d21);
		double d22 = p12.distanceSQ(p22); dMin = Math.min(dMin, d22);
		return dMin;
	}	
	
	public void connect(Polyline other) {
		Point p11 = (Point) this.points.getFirst();
		Point p12 = (Point) this.points.getLast();
		Point p21 = (Point) other.points.getFirst();
		Point p22 = (Point) other.points.getLast();
		double dMin = Double.MAX_VALUE;
		double d11 = p11.distanceSQ(p21); dMin = Math.min(dMin, d11);
		double d12 = p11.distanceSQ(p22); dMin = Math.min(dMin, d12);
		double d21 = p12.distanceSQ(p21); dMin = Math.min(dMin, d21);
		double d22 = p12.distanceSQ(p22); dMin = Math.min(dMin, d22);
				
		LinkedList list = other.points;
		if (dMin == d11) {
			while (list.size() > 0) {
				points.addFirst(list.removeFirst());
			}
		} else if (dMin == d12) {
			while (list.size() > 0) {
				points.addFirst(list.removeLast());
			}
		} else if (dMin == d21) {
			while (list.size() > 0) {
				points.addLast(list.removeFirst());
			}
		} else {
			while (list.size() > 0) {
				points.addLast(list.removeLast());
			}
		}
		valid = false;
	}
	public LinkedList[] split(double upperThreshold, double lowerThreshold, boolean needMapped) {
		LinkedList[] newLines = new LinkedList[2];
		newLines[0] = new LinkedList(); // unmapped
		newLines[1] = new LinkedList(); // mapped
		
		if (!needMapped) {
			
			WeightedLine[] lines = new WeightedLine[this.lines.size()];
			Iterator liIt = this.lines.iterator();
			for (int i = 0; liIt.hasNext(); i++) {
				lines[i] = (WeightedLine) liIt.next();
			}			
			
			Correlation upper = Correlation.createDummy(1);
			upper.a0 = upperThreshold;
			Correlation lower = Correlation.createDummy(1);
			lower.a0 = lowerThreshold;
			
			int plStartIdx = 0;
			double plStartValue = 0.0;
			int plEndIdx = 0;
			double plEndValue = 1.0;
			double level = -1;
			
			int lineIdx = -1;
			boolean nextLine = lineIdx < lines.length - 1;
			boolean nextC = nextLine;
			boolean hasWork = nextLine || nextC;
			boolean endFound = false;
			Iterator cIt = null;
			Correlation c;
			
			while (hasWork) {
				if (nextLine) {
					lineIdx++;
					cIt = lines[lineIdx].externCorrelations.iterator();
					if (!cIt.hasNext()) {
						System.out.print('-');
					} else {
						//nop
					}
					nextC = cIt.hasNext();
				}
				if (nextC) {
					c = (Correlation) cIt.next();

					double s = c.getValue(c.start);
					double e = c.getValue(c.end);
					
					if (level == 0) {
						if (s >= upperThreshold) {
							plEndIdx = lineIdx - 1;
							plEndValue = 1.0;
							endFound = true;
							level = 2;
						} else if (s >= lowerThreshold && e >= upperThreshold) {
							plEndIdx = lineIdx - 1;
							plEndValue = 1.0;
							endFound = true;
							level = 2;
						} else if (s >= lowerThreshold && e >= lowerThreshold) {
							plEndIdx = lineIdx - 1;
							plEndValue = 1.0;
							level = 1.0;
						} else if (s >= lowerThreshold) {
							// nop
						} else if (e >= upperThreshold) {
							double[] ts = c.getIntersections(lower, c.start, c.end);
							if (ts.length == 1) {
								plEndIdx = lineIdx;
								plEndValue = ts[0];
							} else {
								plEndIdx = lineIdx;
								plEndValue = 1.0;
							}
							endFound = true;
							level = 2;
						} else if (e >= lowerThreshold) {
							double[] ts = c.getIntersections(lower, c.start, c.end);
							if (ts.length == 1) {
								plEndIdx = lineIdx;
								plEndValue = ts[0];
							} else {
								plEndIdx = lineIdx;
								plEndValue = 1.0;
							}
							level = 1.0;							
						} else {
							// nop
						}
					} else if (level == 1.0) {
						if (s >= upperThreshold) {
							endFound = true;
							level = 2;
						} else if (s >= lowerThreshold && e >= upperThreshold) {
							endFound = true;
							level = 2;
							plEndIdx = lineIdx-1;
							plEndValue = 1.0;
						} else if (s >= lowerThreshold && e >= lowerThreshold) {
							// nop
						} else if (s >= lowerThreshold) {
							level = 0;							
						} else if (e >= upperThreshold) {
							double[] ts = c.getIntersections(lower, c.start, c.end);
							if (ts.length == 1) {
								plEndIdx = lineIdx;
								plEndValue = ts[0];
							} else {
								plEndIdx = lineIdx;
								plEndValue = 1.0;
							}
							endFound = true;
							level = 2;
						} else if (e >= lowerThreshold) {
							double[] ts = c.getIntersections(lower, c.start, c.end);
							if (ts.length == 1) {
								plEndIdx = lineIdx;
								plEndValue = ts[0];
							} else {
								plEndIdx = lineIdx;
								plEndValue = 1.0;
							}
							level = 1.0;							
						} else {
							level = 0;
						}
					} else if (level == 1.1) {
						if (s >= upperThreshold) {
							level = 2;
						} else if (s >= lowerThreshold && e >= upperThreshold) {
							level = 2;
						} else if (s >= lowerThreshold && e >= lowerThreshold) {
							// nop
						} else if (s >= lowerThreshold) {
							double[] ts = c.getIntersections(lower, c.start, c.end);
							if (ts.length == 1) {
								plStartIdx = lineIdx;
								plStartValue = ts[0];
							} else {
								plStartIdx = lineIdx;
								plStartValue = 0.0;
							}
							level = 0;
						} else if (e >= upperThreshold) {
							level = 2;
						} else if (e >= lowerThreshold) {
							plStartIdx = lineIdx;
							plStartValue = 0.0;
							double[] ts = c.getIntersections(lower, c.start, c.end);
							if (ts.length == 1) {
								plEndIdx = lineIdx;
								plEndValue = ts[0];
							} else {
								plEndIdx = lineIdx;
								plEndValue = 1.0;
							}
							level = 1.0;							
						} else {
							plStartIdx = lineIdx;
							plStartValue = 0.0;
							level = 0;
						}
					} else if (level == 2) {
						if (e >= upperThreshold) {
							// nop
						} else if (e >= lowerThreshold) {
							level = 1.1;
						} else {
							double[] ts = c.getIntersections(lower, c.start, c.end);
							if (ts.length == 1) {
								plStartIdx = lineIdx;
								plStartValue = ts[0];
							} else {
								plStartIdx = lineIdx;
								plStartValue = 0.0;
							}
							level = 0;
						}
					} else if (level == -1) {
						if (e >= upperThreshold) {
							level = 2;
						} else if (s >= upperThreshold && e >= lowerThreshold) {
							level = 1.1;
						} else if (e >= lowerThreshold) {
							plStartIdx = lineIdx;
							plStartValue = 0.0;
							plEndIdx = lineIdx;
							plEndValue = 0.0;
							level = 1.0;
						} else if (s >= upperThreshold) {
							double[] ts = c.getIntersections(lower, c.start, c.end);
							if (ts.length == 1) {
								plStartIdx = lineIdx;
								plStartValue = ts[0];
							} else {
								plStartIdx = lineIdx;
								plStartValue = 0.0;
							}
							level = 0;							
						} else {
							plStartIdx = lineIdx;
							plStartValue = 0.0;
							level = 0;
						}
					} else {
						throw new RuntimeException();
					}					
				} else {
					if (level <= 1.0) {
						plEndIdx = lineIdx;
						plEndValue = 1.0;
						endFound = true;
					}
				}
				nextC = cIt.hasNext();
				nextLine = !nextC && lineIdx < lines.length - 1;
				hasWork = nextLine || nextC;
				
				if (!hasWork && level <= 1.0) {
					plEndIdx = lineIdx;
					plEndValue = 1.0;
					endFound = true;
				}
											
				if (endFound) {
					Polyline pl = new Polyline(false);
					if (plEndIdx >= 0) {
						pl.addPoint(lines[plStartIdx].start.intermediatePoint(plStartValue, lines[plStartIdx].end));
						for (int i = plStartIdx; i < plEndIdx; i++) {
							pl.addPoint(new Point(lines[i].end));
						}
						pl.addPoint(lines[plEndIdx].start.intermediatePoint(plEndValue, lines[plEndIdx].end));
						pl.updateLength();
						if (pl.length > 0.0) {
							newLines[0].add(pl);
						}
					}
					endFound = false;
				}
			}			
			return newLines;
		} else {
			throw new RuntimeException("not implemented");
		}				
	}
	
	//todo: replace by proper one
	public double getThickness() {
		int n = points.size();
		Point[] ps = new Point[n];
		Iterator it = points.iterator();
		for (int i = 0; it.hasNext(); i++) {
			ps[i] = new Point((Point) it.next());
		}
		SimilarityTransformation r = SimilarityTransformation.rotation(15.0/180*Math.PI);
		double thickness = Double.MAX_VALUE;
		for (int a = 0; a < 4; a++) {
			if (a!=0) {
				for (int i = 0; i < n; i++) {
					r.applyTo(ps[i]);
				}
			}
			double minX = Double.MAX_VALUE;
			double maxX = -Double.MAX_VALUE;
			double minY = Double.MAX_VALUE;
			double maxY = -Double.MAX_VALUE;
			for (int i = 0; i < n; i++) {
				minX = Math.min(minX, ps[i].x);
				maxX = Math.max(maxX, ps[i].x);
				minY = Math.min(minY, ps[i].y);
				maxY = Math.max(maxY, ps[i].y);
			}
			thickness = Math.min(Math.min(thickness, maxX-minX), maxY-minY);
		}
		return thickness;
	}
	
	public boolean inside(Point p) {
		int counter = 0;
		p = new Point(p.x+1e-6, p.y+1e-6);
		points.addLast(points.getFirst());
		Iterator it = points.iterator();
		Point n = new Point((Point) it.next());
		n.subtract(p);
		while (it.hasNext()) {
			Point c = n; 
			n = new Point((Point) it.next());
			n.subtract(p);
			if (n.x * n.y == 0.0) {
				throw new RuntimeException("Points not in general position:" + n.x + " " + n.y);
			}
			if (c.x * n.x > 0.0 && c.y * n.y > 0.0) continue;
			if (c.x < 0.0 && c.y > 0.0) {
				if (n.x > 0.0 && n.y > 0.0) {
					counter++;
				} else if (n.x > 0.0 && n.y < 0.0) {
					counter += Point.isLeft(c, n, p) ? -2 : 2; 
				} else if (n.x < 0.0 && n.y < 0.0) {
					counter--;
				}
			} else if (c.x > 0.0 && c.y > 0.0) {
				if (n.x > 0.0 && n.y < 0.0) {
					counter++;
				} else if (n.x < 0.0 && n.y < 0.0) {
					counter += Point.isLeft(c, n, p) ? -2 : 2; 
				} else if (n.x < 0.0 && n.y > 0.0) {
					counter--;
				}
			} else if (c.x > 0.0 && c.y < 0.0) {
				if (n.x < 0.0 && n.y < 0.0) {
					counter++;
				} else if (n.x < 0.0 && n.y > 0.0) {
					counter += Point.isLeft(c, n, p) ? -2 : 2; 
				} else if (n.x > 0.0 && n.y > 0.0) {
					counter--;
				}
			} else if (c.x < 0.0 && c.y < 0.0) {
				if (n.x < 0.0 && n.y > 0.0) {
					counter++;
				} else if (n.x > 0.0 && n.y > 0.0) {
					counter += Point.isLeft(c, n, p) ? -2 : 2; 
				} else if (n.x > 0.0 && n.y < 0.0) {
					counter--;
				}
			}
		}
		
		points.removeLast();
		return counter != 0; 
	}

	public void writeData_03(Writer writer, String prefix) throws IOException {
		writeData_03(writer, prefix, "pl");
	}

	public void writeData_03(Writer writer, String prefix, String tag) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix);
		buff.append("<");
		buff.append(tag);
		buff.append(" cl=\"");
		buff.append(isClosed() ? 'i' : 'n');
		buff.append("\">");
		Iterator it = points.iterator();
		boolean hasData = it.hasNext();
		while (it.hasNext()) {
			Point p = (Point) it.next();
			buff.append(p.x);
			buff.append(' ');
			buff.append(p.y);
			buff.append(' ');
		}
		if (hasData) {
			buff.deleteCharAt(buff.length()-1);
		}
		buff.append("</");
		buff.append(tag);
		buff.append(">\n");
		writer.write(buff.toString());
	}

	public String toXML(String prefix) {
		return toXML(prefix, "pl");
	}
	
	public String toXML(String prefix, String tag) {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix);
		buff.append("<");
		buff.append(tag);
		buff.append(" cl=\"");
		buff.append(isClosed() ? 'i' : 'n');
		buff.append("\">");
		Iterator it = points.iterator();
		boolean hasData = it.hasNext();
		while (it.hasNext()) {
			Point p = (Point) it.next();
			buff.append(p.x);
			buff.append(' ');
			buff.append(p.y);
			buff.append(' ');
		}
		if (hasData) {
			buff.deleteCharAt(buff.length()-1);
		}
		buff.append("</");
		buff.append(tag);
		buff.append(">\n");
		return buff.toString();
	}
	
}
