/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.model;

import java.awt.Graphics;
import java.io.IOException;
import java.io.Writer;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.StringTokenizer;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import profi.matching.Figure;
import profi.matching.WeightedClustering;
import profi.matching.WeightedClustering.Interval;
import profi.model.Polyline.WeightedLine;


public class ShapeSet {

	public static ShapeSet readData_03(Element element) {
		ShapeSet set = new ShapeSet();		
		NodeList childNodes = element.getChildNodes();
		int nodeCount = 0;
		while (nodeCount < childNodes.getLength()) {
			Node node = childNodes.item(nodeCount);
			if (node.getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element) node;
				if (e.getTagName().equals("pl")) {
					set.add(Polyline.readData_03(e));
				}
			}
			nodeCount++;				
		}
		return set;
	}
	
	private List plList;

	private double length;	
	private boolean hasLength;

	private boolean weighted;
	private double weightedLength;
	private boolean hasWeightedLength;

	public ShapeSet() {
		plList = new LinkedList();
		length = 0.0;
		hasLength = true;
		weightedLength = 0.0;
		hasWeightedLength = false;
		weighted = false;
	}

	public ShapeSet(Polyline p) {
		plList = new LinkedList();
		plList.add(p);
		hasLength = false;
		weightedLength = 0.0;
		hasWeightedLength = false;
		weighted = false;
	}

	public ShapeSet(ShapeSet other) {
		plList = new LinkedList();
		Iterator it = other.plList.iterator();
		while (it.hasNext()) {
			plList.add(new Polyline((Polyline) it.next()));
		}
		this.length = other.length;
		this.hasLength = other.hasLength;
		this.weightedLength = other.weightedLength;
		this.hasWeightedLength = other.hasWeightedLength;
		this.weighted = other.weighted;
	}
	
	public void makeLines() {
		Iterator it = plList.iterator();
		while(it.hasNext()) {
			((Polyline) it.next()).makeLines();
		}
		weighted = true;
	}
	
	public LinkedList getLines() {
		if (!weighted) makeLines();
		Iterator it = plList.iterator();
		LinkedList list = new LinkedList();
		while(it.hasNext()) {
			list.addAll(((Polyline) it.next()).getLines());
		}
		return list;
	}

	public double getLength() {
		if (!hasLength)	updateLength();
		return length;
	}
			
	public double getWeight() {
		double weight = 0.0;
		Iterator it = getLines().iterator();
		while(it.hasNext()) {
			weight += ((WeightedLine) it.next()).getWeight(true);
		}
		return weight / getWeightedLength();
	}

	public void updateLength() {
		length = 0;
		Iterator it = getShapeIterator();
		while (it.hasNext()) {
			Polyline p = (Polyline) it.next();
			length += p.getLength();
		}
	}
	
	public double getWeightedLength() {
		if (!hasWeightedLength) updateWeightedLength();
		return weightedLength;
	}
	
	public void updateWeightedLength() {
		weightedLength = 0.0;		
		Iterator it = getLines().iterator();
		while (it.hasNext()) {
			weightedLength += ((WeightedLine) it.next()).getWeightedLength();
		}
		hasWeightedLength = true;
	}
	
	public List getPlList() {
		return plList;
	}

	public void apply(Transformation t) {
		Iterator it = plList.iterator();
		while (it.hasNext()) {
			((Polyline) it.next()).apply(t);
		}
		hasLength &= t.isLengthPreserving();
		hasWeightedLength &= t.isLengthPreserving();
	}

	public void add(Polyline pl) {
		plList.add(pl);
		hasLength = false;
	}

	public void add(ShapeSet other) {
		Iterator it = other.getShapeIterator();
		while (it.hasNext()) {
			plList.add(it.next());
		}
		hasLength = false;
	}
	
	public void add(LinkedList list) {
		Iterator it = list.iterator();
		while (it.hasNext()) {
			plList.add(it.next());
		}
		hasLength = false;
	}

	public ListIterator getShapeIterator() {
		return plList.listIterator();
	}
	
	public Polyline getPolyline(int i) {
		return (Polyline) plList.get(i);
	}
	
	public Box getBoundingBox() {
		Iterator iter = getShapeIterator();
		if (!iter.hasNext()) {
			return null;
		}
		Box b = ((Polyline) iter.next()).getBoundingBox();
		while (iter.hasNext()) {
			b.add(((Polyline) iter.next()).getBoundingBox());
		}
		return b;
	}

	public int getNumberOfPolylines() {
		return plList.size();
	}

	public int getNumberOfPoints() {
		int n = 0;
		Iterator it = plList.iterator();
		while (it.hasNext()) {
			n += ((Polyline) it.next()).getNumberOfPoints();
		}
		return n;
	}
		
	public void consolidate(double eps) {
		Iterator shapeIt = getShapeIterator();
		while (shapeIt.hasNext()) {
			((Polyline) shapeIt.next()).consolidate(eps);
		}		
	}
	
	public void connect(double eps) {
		LinkedList pgs = new LinkedList();
		LinkedList pls = new LinkedList();
		Iterator it = getShapeIterator();
		while (it.hasNext()) {
			Polyline p = (Polyline) it.next();
			if (p.isClosed()) {
				pgs.add(p);
			} else {
				pls.add(p);
			}
		}
		
		boolean connected = false;
		do {
			connected = false;
			Polyline poly1 = null;
			Polyline poly2 = null;
			double minDist = Double.MAX_VALUE;
			Iterator it1 = pls.iterator();
			while (it1.hasNext()) {
				Polyline pl1 = (Polyline) it1.next();
				Iterator it2 = pls.iterator();
				while (it2.hasNext()) {
					Polyline pl2 = (Polyline) it2.next();
					if (pl1 == pl2) continue;
					double dist = pl1.getMinDist(pl2);
					if (dist < minDist) {
						minDist = dist;
						poly1 = pl1;
						poly2 = pl2;
					}					
				}
			}
			if (minDist < eps) {
				poly1.connect(poly2);
				pls.remove(poly2);
				connected = true;
			}
		} while (connected);
		this.plList.clear();
		this.plList.addAll(pgs);
		this.plList.addAll(pls);
		updateLength();
	}
	

	public ShapeSet[] split(double upperThreshold, double lowerThreshold, boolean needMapped) {
		ShapeSet[] returns = new ShapeSet[2];
		returns[0] = new ShapeSet();
		returns[1] = new ShapeSet();
		
		Iterator it = this.plList.iterator();
		while (it.hasNext()) {
			LinkedList[] newShapes = ((Polyline) it.next()).split(upperThreshold, lowerThreshold, needMapped);
			returns[0].add(newShapes[0]);
			returns[1].add(newShapes[1]);
		}
		return returns;
	}
	
	public ShapeSet getSubset(LinkedList[] intervals) {
		ShapeSet subset = new ShapeSet();
		for (int i = 0; i < intervals.length; i++) {
			Iterator it = intervals[i].iterator();
			Polyline polyline = this.getPolyline(i);
			while (it.hasNext()) {
				Polyline subline = new Polyline(false);
				Interval interval = (Interval) it.next();
				for (int j = interval.start; j < interval.end; j++) {
					subline.addPoint(new Point(polyline.getPoint(j)));
				}
				if (interval.end == polyline.getNumberOfPoints()) {
					subline.addPoint(new Point(polyline.getPoint(0)));
				} else {
					subline.addPoint(new Point(polyline.getPoint(interval.end)));
				}
				subset.add(subline);
			}
		}
		return subset;
	}
	
	private class PlComparator implements Comparator {

		public int compare(Object arg0, Object arg1) {
			double l1 = ((Polyline) arg0).getLength();
			double l2 = ((Polyline) arg1).getLength();
			if (l1 < l2) return 1;
			if (l1 > l2) return -1;
			return 0;
		}
		
	}

	public void writeData_03(Writer writer, String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix);
		buff.append("<shape>\n");
		writer.write(buff.toString());
		Iterator it = getShapeIterator();
		String pprefix = prefix + "\t";
		while (it.hasNext()) {
			Polyline pl = (Polyline) it.next();
			pl.writeData_03(writer, pprefix);
		}		
		buff.setLength(0);
		buff.append(prefix);
		buff.append("</shape>\n");
		writer.write(buff.toString());
	}

	public String toXML(String prefix) throws IOException {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<shape>\n");
		Iterator it = getShapeIterator();
		String pprefix = prefix + "\t";
		while (it.hasNext()) {
			Polyline pl = (Polyline) it.next();
			buff.append(pl.toXML(pprefix));
		}
		buff.append(prefix); buff.append("</shape>\n");
		return buff.toString();
	}

	public void draw(Graphics g, double scale) {
		Iterator plIt = plList.iterator();
		while (plIt.hasNext()) {
			Polyline o = (Polyline) plIt.next();
			Iterator it = o.getPoints().iterator();
			Point pFirst, p1, p2;
			pFirst = p1 = (Point) it.next();
			while (it.hasNext()) {
				p2 = (Point) it.next();
				g.drawLine((int) (p1.x*scale), (int) (p1.y*scale), (int) (p2.x*scale), (int) (p2.y*scale));
				p1 = p2;
			}
			if (o.isClosed()) {
				g.drawLine((int) (p1.x*scale), (int) (p1.y*scale),
						(int) (pFirst.x*scale), (int) (pFirst.y*scale));
			}			
		}
	}
}
