/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.model;

import java.util.StringTokenizer;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import profi.matching.WeightedClustering.Item;

public class SimilarityTransformation extends AffineTransformation {

	public static SimilarityTransformation flipHorizontally = new SimilarityTransformation(true, 1.0, 0.0, 0.0, 0.0);
	public static SimilarityTransformation identity = new SimilarityTransformation();
	
	public static SimilarityTransformation rotation(double angle) {
		double a = Math.cos(angle);
		double b = -Math.sin(angle);
		SimilarityTransformation sim = new SimilarityTransformation(false, a, b, 0.0, 0.0);
		sim.angle = angle;
		sim.scale = 1.0;
		sim.explicit = true;
		sim.lengthPreserving = true;
		return sim;
	}
	
	public static SimilarityTransformation scaling(double s) {
		SimilarityTransformation sim = new SimilarityTransformation(false, s, 0.0, 0.0, 0.0);
		sim.angle = 0.0;
		sim.scale = s;
		sim.explicit = true;
		return sim;
	}

	public static SimilarityTransformation translation(double dx, double dy) {
		SimilarityTransformation sim = new SimilarityTransformation(false, 1.0, 0.0, dx, dy);
		sim.angle = 0.0;
		sim.scale = 1.0;
		sim.explicit = true;
		sim.lengthPreserving = true;
		return sim;
	}

	public static SimilarityTransformation explicit(boolean flip, double scale, double angle, double dx, double dy) {
		double a = Math.cos(angle) * scale;
		double b = -Math.sin(angle) * scale;
		SimilarityTransformation sim = new SimilarityTransformation(flip, a, b, dx, dy);
		sim.angle = angle;
		sim.scale = scale;
		sim.explicit = true;
		sim.lengthPreserving = (scale == 1.0);
		return sim;
	}

	public static SimilarityTransformation readData_03(Element e) {
		NodeList childNodes = e.getChildNodes();
		StringTokenizer tokens = new StringTokenizer(childNodes.item(1).getTextContent());
		double dx = Double.parseDouble(tokens.nextToken());
		double dy = Double.parseDouble(tokens.nextToken());
		double rotation = Double.parseDouble(childNodes.item(3).getTextContent());
		double scale = Double.parseDouble(childNodes.item(5).getTextContent());
		boolean mirrored = childNodes.item(7).getTextContent().trim().equalsIgnoreCase("true");
		return explicit(mirrored, scale, rotation, dx, dy);
	}
	
	protected boolean flip;
	protected double scale;
	protected double angle;
	protected boolean explicit;
	
	public SimilarityTransformation() {
		super(null);
		flip = false;
		scale = 1.0;
		angle = 0.0;
		explicit = true;
	}

	public SimilarityTransformation(SimilarityTransformation other) {
		super(other, null);
		flip = other.flip;
		scale = other.scale;
		angle = other.angle;
		explicit = other.explicit;
	}

	public SimilarityTransformation(SimilarityTransformation t1, SimilarityTransformation t2) {
		super(t1, t2, null);
		flip = t1.flip ^ t2.flip;
		scale = t1.scale * t2.scale;
		angle = t1.angle + t2.angle;
		explicit = t1.explicit && t2.explicit;
	}
		
	public SimilarityTransformation(boolean flip, double a, double b, double dx, double dy) {
		super(null);
		this.flip = flip;
		explicit = false;
		lengthPreserving = false;
		m[0][0] = flip ? -a :  a; m[0][1] = b; m[0][2] = dx;
		m[1][0] = flip ?  b : -b; m[1][1] = a; m[1][2] = dy;
	}

	public Item copyItem() {
		return new SimilarityTransformation(this);
	}
	
	public boolean getFlip() {
		return flip;
	}
	
	public double getScale() {
		if (!explicit) {
			makeExplicit();
		}
		return scale;
	}
	
	public double getAngle() {
		if (!explicit) {
			makeExplicit();
		}
		return angle;
	}
	
	public double getDx() {
		return m[0][2];
	}
	
	public double getDy() {
		return m[1][2];
	}	
	
	protected void makeExplicit() {
		scale = Math.sqrt(m[0][0]*m[0][0] + m[0][1]*m[0][1]);
		lengthPreserving = Math.abs(scale-1.0) < 1e-4;
		double m00 = Math.min(1.0, Math.max(-1.0, m[1][1]/scale));
		angle = Math.acos(m00) * (-m[0][1] > 0.0 ? 1.0 : -1.0);
		explicit = true;
	}
	
	public String toXML(String prefix) {
		StringBuffer buff = new StringBuffer();
		buff.append(prefix); buff.append("<transformation type=\"similarity\">\n");
		buff.append(prefix); buff.append("\t<translation>"); buff.append(getDx()); buff.append(" "); buff.append(getDy()); buff.append("</translation>\n");
		buff.append(prefix); buff.append("\t<rotation>"); buff.append(getAngle()); buff.append("</rotation>\n");
		buff.append(prefix); buff.append("\t<scale>"); buff.append(getScale()); buff.append("</scale>\n");
		buff.append(prefix); buff.append("\t<mirrored>"); buff.append(getFlip()); buff.append("</mirrored>\n");
		buff.append(prefix); buff.append("</transformation>\n"); 
		return buff.toString();
	}

	public Item itemCopy() {
		return new SimilarityTransformation(this);
	}
}
