/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.segmentation;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import profi.matching.Convex;
import profi.matching.Figure;
import profi.model.Box;
import profi.model.DouglasPeucker;
import profi.model.Point;
import profi.model.Polyline;
import profi.model.ShapeSet;

public class RegionImage {

	public static class Bound implements Comparable {
		
		public int x;
		public boolean left;

		public Bound(int x, boolean left) {
			this.x = x;
			this.left = left;
		}

		public int compareTo(Object o) {

			Bound other = (Bound) o;
			if (this.x < other.x) return -1;
			if (this.x > other.x) return  1;
			if (!this.left && other.left) return -1;
			if (this.left && !other.left) return  1;
			return 0;
		}
	}

	private static class Segment {
		
		public int x;
		public int y;		
		public int turns;
		
		public Segment(int x, int y) {
			this.x = x;
			this.y = y;
			turns = 0;
		}
		
		public Segment(Segment s) {
			this.x = s.x;
			this.y = s.y;
			this.turns = s.turns;
		}
		
		public boolean equals(Segment other) {
			return (other.x == this.x && other.y == this.y && d(other.turns) == d(this.turns));
		}

		public int d(int turns) {
			if (turns >= 0) {
				return turns % 4;
			} else {
				return ((turns % 4) + 4) % 4;
			}
		}
		
		public void turnLeft() {
			turns++;
		}
		
		public void turnRight() {
			move();
			turns--;
			move();
		}

		public void move() {
			int d = d(turns);
			int dx = (d % 2 == 0) ? 0 : ((d == 1) ? +1 : -1);
			int dy = (d % 2 == 1) ? 0 : ((d == 0) ? -1 : +1);
			x += dx;
			y += dy;
		}
		
		public int rightX() {
			int d = d(turns);
			int dx = (d % 2 == 0) ? 0 : ((d == 1) ? +1 : -1);
			return x + dx;
		}
		
		public int rightY() {
			int d = d(turns);
			int dy = (d % 2 == 1) ? 0 : ((d == 0) ? -1 : +1);
			return y + dy;
		}
		
		public int rightRightX() {
			int dh = d(turns+1);
			int dx = (dh / 2 == 0) ? -1 : +1;
			return x + dx;
		}
		
		public int rightRightY() {
			int dh = d(turns);
			int dy = (dh / 2 == 0) ? -1 : 1;
			return y + dy;
		}
		
		public int getXCoordinate() {
			int dh = d(turns+1);
			int dx = (dh / 2 == 0) ? 0 : 1;
			return x + dx;
		}

		public int getYCoordinate() {
			int dh = d(turns);
			int dy = (dh / 2 == 0) ? 0 : 1;
			return y + dy;
		}
	}
	
	public static double rEps = 0.01;
	public static double rangeDiam = 0.25;
	public static double rangeThick = 3.0;
	public static double rangeDist = 2.0;
	public static double boundDist = 4.0;
		
	public static void test() {
		
		Polyline pg = new Polyline(true);
		for (int i = 0; i < 500; i++) {
			Point p = new Point(100 * Math.cos(2 * Math.PI / 500 * i), 100 * Math.sin(2 * Math.PI / 500 * i));
			pg.addPoint(p);
		}
		double eps = rEps * 200 * 1.4;
		pg = DouglasPeucker.simplify(pg, eps);
		
		System.out.println(eps + " " + pg.getPoints().size());
		
		/*
		PixelImage pImg = PixelImage.getPixelImage("/home/datsche/scholz/test.png");
		int h = pImg.getHeight();
		int w = pImg.getWidth();
		System.out.print("  ");
		for (int c = 0; c < w; c++) {
			System.out.print(c%10);
		}
		System.out.println();
		for (int r = h-1; r >= 0; r--) {
			System.out.print(r%10 + " ");
			for (int c = 0; c < w; c++) {
				int col = pImg.getBrightness(c, r);
				System.out.print(col == 0 ? "X" : ".");
			}
			System.out.println();
		}
		
		RegionImage rImg = new RegionImage(pImg);
		ShapeSet set = new ShapeSet();
		rImg.getShapes(set);
				
		System.out.println("test...");
		Iterator it = set.getShapeIterator();
		while (it.hasNext()) {
			Polygon p = (Polygon) it.next();
			System.out.println(Figure.toString(p));
		}
		System.out.println("...completed");
		*/
	}
	
	public static ShapeSet getShapeSet(String fileName) {
		RegionImage rImg = getRegionImage(fileName, true);
		ShapeSet set = new ShapeSet();
		rImg.getShapes(set);
		return set;
	}
	
	public static RegionImage getRegionImage(String fileName, boolean simplify) {
		PixelImage pImg = PixelImage.getPixelImage(fileName);
		RegionImage rImg = new RegionImage(pImg);
		if (simplify) {
			rImg.eliminateNoise(rImg.diameter, true);
			rImg.simplifyBoundaries(rImg.diameter);
		}
		return rImg;
	}
	
	public static RegionImage readData(File file) throws IOException {
		RegionImage img =  new RegionImage();
		try {
			Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
			Element doc = document.getDocumentElement();
			
			// src-img
			NodeList srcs = doc.getElementsByTagName("src-img");
			if (srcs.getLength() > 0) {
				img.name = srcs.item(0).getTextContent();
				img.width = Double.parseDouble(srcs.item(0).getAttributes().getNamedItem("width").getNodeValue());
				img.height = Double.parseDouble(srcs.item(0).getAttributes().getNamedItem("height").getNodeValue());
				img.originalWidth = (int) img.width;
				img.originalHeight = (int) img.height;
				img.diameter = Math.sqrt(img.width*img.width + img.height*img.height);
			} else {
				img.name = "";
			}
			
			// creation
			NodeList creations = doc.getElementsByTagName("creation");
			if (creations.getLength() > 0) {
				NodeList stages = creations.item(0).getChildNodes();
				int k = 0;
				for (int i = 0; i < stages.getLength(); i++) {
					if (stages.item(i).getNodeType() == Node.ELEMENT_NODE) {
						k++;
					}
				}				
				img.creation = new String[k];
				k = 0;
				for (int i = 0; i < stages.getLength(); i++) {
					if (stages.item(i).getNodeType() == Node.ELEMENT_NODE) {
						img.creation[k] = stages.item(i).getAttributes().getNamedItem("method").getNodeValue();
						k++;
					}
				}								
			}
			
			// shape
			NodeList shapes = doc.getElementsByTagName("shape");
			//int n = shapes.getLength();
			Vector imgs = new Vector();
			for (int i = 0; i < shapes.getLength(); i++) {
				Node shape = shapes.item(i);
				RegionImage imgNew = new RegionImage();
				imgs.add(imgNew);
				imgNew.idx = Integer.parseInt(shape.getAttributes().getNamedItem("id").getNodeValue());
				NodeList children = shape.getChildNodes();
				for (int j = 0; j < children.getLength(); j++) {
					Node child = children.item(j);
					if (child.getNodeType() == Node.ELEMENT_NODE) {
						if (((Element) child).getTagName().equals("pl")) {
							Polyline outline = new Polyline(child.getAttributes().getNamedItem("cl").getNodeValue().equals("i"));
							StringTokenizer data = new StringTokenizer(child.getTextContent());
							while (data.countTokens() >= 2) {
								outline.addPoint(new Point(Double.parseDouble(data.nextToken()), Double.parseDouble(data.nextToken())));
							}
							imgNew.boundary = outline;
							break;
						} else if (((Element) child).getTagName().equals("pca")) {
							Polyline outline = new Polyline(child.getAttributes().getNamedItem("cl").getNodeValue().equals("i"));
							StringTokenizer data = new StringTokenizer(child.getTextContent());
							outline.addPoint(new Point(Double.parseDouble(data.nextToken()), Double.parseDouble(data.nextToken())));
							while (data.countTokens() >= 3) {
								outline.addPoint(new Point(Double.parseDouble(data.nextToken()), Double.parseDouble(data.nextToken())));
								data.nextToken(); // discard curvature
							}
							imgNew.boundary = outline;
							break;
						}
					}
				}
			}

			// views
			NodeList views = doc.getElementsByTagName("view");
			RegionImage[] imgArr = null;
			if (views.getLength() > 0) {
				Node bestView = views.item(0);
				// todo: find best view...
				StringTokenizer data = new StringTokenizer(bestView.getTextContent());
				imgArr = new RegionImage[data.countTokens()];
				int m = 0;
				while (data.hasMoreTokens()) {
					int idx = Integer.parseInt(data.nextToken());
					for (int i = 0; i < imgs.size(); i++) {
						RegionImage imgNew = (RegionImage) imgs.get(i);
						if (imgNew.idx == idx) {
							imgArr[m] = imgNew;
							m++;
							break;
						}
					}
				}				
			} else {
				imgArr = new RegionImage[imgs.size()];
				for (int i = 0; i < imgs.size(); i++) {
					imgArr[i] = (RegionImage) imgs.get(i);
				}
			}
			
			// relations
			NodeList relations = doc.getElementsByTagName("shape-relations");
			if(relations.getLength() > 0) {
				Node relation = relations.item(0);
				String att = relation.getAttributes().getNamedItem("approx").getNodeValue();
				int quality = att.equals("full") ? 3 : att.equals("ch") ? 2 : 1;
				for (int i = 1; i < relations.getLength(); i++) {
					Node relationN = relations.item(0);
					String attN = relation.getAttributes().getNamedItem("cl").getNodeValue();
					int qualityN = att.equals("full") ? 3 : att.equals("ch") ? 2 : 1;
					if (qualityN > quality) {
						relation = relationN;
						quality = qualityN;
					}
				}
				NodeList cont = relation.getChildNodes();
				for (int j = 0; j < cont.getLength(); j++) {
					Node c = cont.item(j);
					if (c.getNodeType() == Node.ELEMENT_NODE) {
						if (((Element) c).getTagName().equals("contains")) {
							Vector[] contained = new Vector[imgArr.length];
							for (int k = 0; k < imgArr.length; k++) {
								contained[k] = new Vector();
							}
							StringTokenizer data = new StringTokenizer(c.getTextContent());
							while (data.countTokens() >= 2) {
								int outer0 = Integer.parseInt(data.nextToken());
								int outer = -1;
								for (int k = 0; k < imgArr.length; k++) {
									if (imgArr[k].idx == outer0) {
										outer = k;
										break;
									}
								}
								int inner0 = Integer.parseInt(data.nextToken());
								int inner = -1;
								for (int k = 0; k < imgArr.length; k++) {
									if (imgArr[k].idx == inner0) {
										inner = k;
										break;
									}
								}
								if (inner >= 0 && outer >= 0) {
									contained[inner].add(new Integer(outer));
								}
							}
							for (int k = 0; k < imgArr.length; k++) {
								if (contained[k].size() == 0) {
									imgArr[k].parent = img;
									img.children.add(imgArr[k]);
								} else {
									int bestIdx = 0;
									int bestSize = contained[((Integer) contained[k].get(0)).intValue()].size();
									for (int l = 1; l < contained[k].size(); l++) {
										int newSize = contained[((Integer) contained[k].get(l)).intValue()].size();
										if (newSize > bestSize) {
											bestSize = newSize;
											bestIdx = l;
										}
									}
									int p = ((Integer) contained[k].get(bestIdx)).intValue();
									imgArr[k].parent = imgArr[p];
									imgArr[p].children.add(imgArr[k]);
								}
							}
							break;
						}
					}
				}				
			} else {
				for (int i = 0; i < imgArr.length; i++) {
					imgArr[i].parent = img;
					img.children.add(imgArr[i]);
				}
			}			

			ShapeSet set = new ShapeSet();
			img.getShapes(set);
			Box bb = set.getBoundingBox();
			if (img.diameter == 0.0) {
				img.width = bb.getWidth();
				img.height = bb.getHeight();				
				img.diameter = Math.sqrt(img.width*img.width + img.height*img.height);
			}
			img.boundary = new Polyline(true);
			img.boundary.addPoint(new Point(bb.getMinX(), bb.getMinY()));
			img.boundary.addPoint(new Point(bb.getMaxX(), bb.getMinY()));
			img.boundary.addPoint(new Point(bb.getMaxX(), bb.getMaxY()));
			img.boundary.addPoint(new Point(bb.getMinX(), bb.getMaxY()));
			img.hasData = true;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}		
		return img;
	}

	public boolean hasData;
	private PixelImage img;
	public String name;
	private RegionImage parent;
	public Vector children;
	public Polyline boundary;
	public int idx;
	public double height;
	public double width;
	public int originalHeight;
	public int originalWidth;
	public String[] creation;
	
	//public Figure figure;
	public double diameter;
	private double thickness;
	private double area;
	private int color;
	private int tmpIdx; // just for temporary usage
			
	public RegionImage(PixelImage img) {
		this.img = img;
		this.name = img.name;
		parent = null;
		//figure = null;
		children = new Vector();
		boundary = new Polyline(true);
		boundary.addPoint(new Point(0, 0));
		boundary.addPoint(new Point(img.getWidth(), 0));
		boundary.addPoint(new Point(img.getWidth(), img.getHeight()));
		boundary.addPoint(new Point(0, img.getHeight()));
		this.height = img.getHeight();
		this.width = img.getWidth();
		this.originalHeight = img.getOriginalHeight();
		this.originalWidth = img.getOriginalWidth();
		diameter = Math.sqrt(img.getWidth()*img.getWidth() + img.getHeight()*img.getHeight());
		thickness = Math.min(img.getWidth(), img.getHeight());
		color = img.bg;

		detectBoundaries();
		
		String[] tmpArray = {"d-SegmentationNaive_001", "d-TextureNearestNeigbour_001", "d-SimplificationDouglasPeucker_001"};
		creation = tmpArray;
		hasData = true;
	}

	public RegionImage(RegionImage parent, Polyline boundary) {
		this.img = parent.img;
		this.parent = parent;
		//figure = null;
		children = new Vector();
		this.boundary = boundary;
		diameter = boundary.getDiameter();
		thickness = boundary.getThickness();
		color = PixelImage.getOther(parent.color);
		
		detectBoundaries();
		hasData = true;
	}
	
	private RegionImage() {
		this.parent = null;
		children = new Vector();
		diameter = 0.0;
		hasData = false;
		this.creation = new String[0];
	}
	
	public boolean isSimplified() {
		for (int i = 0; i < creation.length; i++) {
			if (creation[i].indexOf("d-Simplification") >=0) {
				return true;
			}
		}
		return false;
	}

	public void getShapes(ShapeSet set) {
		if (parent != null) {
			set.add(boundary);
		}
		for (int i = 0; i < children.size(); i++) {
			RegionImage img = (RegionImage) children.get(i);
			img.getShapes(set);
		}
	}
	
	public int getNumberOfShapes() {
		int n = 0;
		if (parent != null) {
			n++;
		}
		for (int i = 0; i < children.size(); i++) {
			RegionImage img = (RegionImage) children.get(i);
			n += img.getNumberOfShapes();
		}
		return n;
	}
	
	public int getWidth() {
		return originalWidth;
	}

	public int getHeight() {
		return originalHeight;
	}
	
	public int setIdx(int idx) {
		int n = 0;
		if (parent != null) {
			this.idx = idx;
		} else {
			this.idx = idx-1;
		}
		int maxIdx = this.idx;
		for (int i = 0; i < children.size(); i++) {
			RegionImage img = (RegionImage) children.get(i);
			maxIdx = img.setIdx(maxIdx+1);
		}
		return maxIdx;
	}
	
	public double getDiameter() {
		return diameter;
	}
		
	public void detectBoundaries() {
		Box bb = boundary.getBoundingBox();
		int minX = (int) bb.getMinX();
		int maxX = (int) bb.getMaxX();
		int minY = (int) bb.getMinY();
		int maxY = (int) bb.getMaxY();
		Vector[] bounds = new Vector[maxY-minY+1];
		for (int i = 0; i < bounds.length; i++) {
			bounds[i] = new Vector();
		}
		if (parent != null) {
			addBounds(bounds, minY, boundary, false);
		} else {
			for (int i = 0; i < bounds.length; i++) {
				bounds[i].add(new Bound(minX, true));
				bounds[i].add(new Bound(maxX, false));
			}
		}

		for (int y = maxY; y >= minY; y--) {
			for (int i = 0; i < bounds[y-minY].size(); i+=2) {
				int x1 = ((Bound) bounds[y-minY].get(i)).x;
				int x2 = ((Bound) bounds[y-minY].get(i+1)).x;
				area += x2 - x1 -1;
			}
		}
		
		for (int y = maxY; y >= minY; y--) {
			for (int i = 0; i < bounds[y-minY].size(); i+=2) {
				int x1 = ((Bound) bounds[y-minY].get(i)).x;
				int x2 = ((Bound) bounds[y-minY].get(i+1)).x;
				for (int x = x1+1; x < x2; x++) {
					if (getBrightness(x, y, bounds, minY) != color) {
						Polyline p = followContour(x, y, bounds, minY);
						addBounds(bounds, minY, p, true);
						RegionImage child = new RegionImage(this, p);
						children.add(child);
						break;
					}
				}
			}
		}	
	}
	
	public void eliminateNoise(double diam, boolean first) {
		int n = children.size();
		boolean[][] connected = new boolean[n][n];
		boolean[] enqueued = new boolean[n];
		double distance[][] = new double[n][n];
		double minDist[] = new double[n];
		for (int i = 0; i < n; i++) {
			enqueued[i] = false;
			minDist[i] = Double.MAX_VALUE;
			
		}
		for (int i = 0; i < n; i++) {			
			RegionImage ci = (RegionImage) children.get(i);
			ci.tmpIdx = i;
			for (int j = i+1; j < n; j++) {
				RegionImage cj = (RegionImage) children.get(j);	
				double dmin = Math.min(ci.diameter, cj.diameter);
				double dmax = Math.max(ci.diameter, cj.diameter);
				double tmin = Math.min(ci.thickness, cj.thickness);
				double tmax = Math.max(ci.thickness, cj.thickness);
				double ratio = dmax / dmin;
				double dij   = distanceNearest(ci.boundary, cj.boundary) * ratio;
				distance[i][j] = distance[j][i] = dij;
				minDist[i] = Math.min(minDist[i], dij);
				minDist[j] = Math.min(minDist[j], dij);				
			}
		}
		for (int i = 0; i < n; i++) {
			minDist[i] = Double.MAX_VALUE;
			RegionImage ci = (RegionImage) children.get(i);
			connected[i][i] = true;
			for (int j = i+1; j < n; j++) {
				RegionImage cj = (RegionImage) children.get(j);		
				double dmin = Math.min(ci.diameter, cj.diameter);
				double dmax = Math.max(ci.diameter, cj.diameter);
				double tmin = Math.min(ci.thickness, cj.thickness);
				double tmax = Math.max(ci.thickness, cj.thickness);
				double amin = Math.min(ci.diameter * ci.thickness, cj.diameter * cj.thickness);
				double amax = Math.max(ci.diameter * ci.thickness, cj.diameter * cj.thickness);
				double dratio = dmin / dmax;
				double aratio = amin / amax;
				connected[i][j] = connected[j][i] =	
					//dij < Math.max(Math.min(ci.diameter, cj.diameter) * rangeD,	Math.min(ci.thickness, cj.thickness) * rangeT);
					dratio > 0.3 && aratio > 0.2 &&
					distance[i][j] < 
					Math.min(
							Math.min(Math.min(minDist[i], minDist[j]) * rangeDist, dmin * boundDist),
							Math.max(dmin * rangeDiam, tmin * rangeThick)); 
			}
		}
		Vector components = new Vector();
		Vector queue = new Vector();
		for (int i = 0; i < n; i++) {
			if (enqueued[i]) {
				continue;
			}
			Vector component = new Vector();
			components.add(component);
			component.add(children.get(i));
			queue.add(new Integer(i));
			enqueued[i] = true;
			while (queue.size() > 0) {
				int current = ((Integer) queue.lastElement()).intValue();
				queue.remove(queue.size()-1);
				for (int next = 0; next < n; next++) {
					if (!enqueued[next] && connected[current][next]) {
						component.add(children.get(next));
						queue.add(new Integer(next));
						enqueued[next] = true;
					}
				}
			}
		}
		double maxD = 0.0;
		int k = components.size();
		if (false && k > 100 && first) {

			// merging of hatched areas
			for (int i = 0; i < components.size(); i++) {
				Vector component = (Vector) components.get(i);
				if (component != null && component.size() > 15) {
					for (int j = 0; j < component.size(); j++) {
						children.remove(component.get(j));									
					}
					children.add(merge(component, connected));
				}
			}
			
			eliminateNoise(diam, false);
			
			n = children.size();
			for (int i = 0; i < n; i++) {
				((RegionImage) children.get(i)).eliminateNoise(diam, true); 
			}
			return;
		} 
		double[] ds = new double[k];
		for (int i = 0; i < components.size(); i++) {
			Vector component = (Vector) components.get(i);
			ShapeSet set = new ShapeSet();
			for (int j = 0; j < component.size(); j++) {
				RegionImage ri = (RegionImage) component.get(j);
				set.add(ri.boundary);				
			}
			ds[i] = Convex.convexHull(set).getDiameter();
			if (ds[i] > maxD) {
				maxD = ds[i];
			}			
		}
		
		// noise
		double threshold = Math.max(maxD / 10, diam / 20);
		for (int i = 0; i < k; i++) {
			if (ds[i] < threshold) {
				Vector component = (Vector) components.get(i);
				for (int j = 0; j < component.size(); j++) {
					children.remove(component.get(j));									
				}
				components.set(i, null);
			}
		}
		
		// merging of hatched areas
		RegionImage[] mergeds = new RegionImage[k];
		for (int i = 0; i < components.size(); i++) {
			Vector component = (Vector) components.get(i);
			if (component != null && component.size() > 15) {
				mergeds[i] = merge(component, connected);
			}
		}
		
		// area
		double area = this.area;
		double childsArea = 0.0;
		for (int i = 0; i < components.size(); i++) {
			if (mergeds[i] != null) {
				childsArea += mergeds[i].area;
			} else if (components.get(i) != null) {
				Vector component = (Vector) components.get(i);
				for (int j = 0; j < component.size(); j++) {
					RegionImage img = (RegionImage) component.get(j);
					childsArea += img.area;
				}
			}
		}
		boolean keep = (parent == null) || (childsArea < area * 0.75);
		for (int i = 0; i < components.size(); i++) {
			Vector component = (Vector) components.get(i);
			if (component == null) continue;  // component was noise
			if (mergeds[i] == null) continue; // component consists of real figures
			if (keep) {
				if (component.size() > 10) {
					for (int j = 0; j < component.size(); j++) {
						children.remove(component.get(j));									
					}
				}
				children.add(mergeds[i]);
				mergeds[i].parent = this;
				
			} else {
				for (int j = 0; j < component.size(); j++) {
					children.remove(component.get(j));									
				}
			}			
		}

		n = children.size();
		for (int i = 0; i < n; i++) {
			((RegionImage) children.get(i)).eliminateNoise(diam, true); 
		}
	}
	
	public static double distanceFarthest(Polyline pg1, Polyline pg2) {
		double maxSqDist = 0.0;
		Iterator i1 = pg1.getPointIterator();
		while (i1.hasNext()) {
			Point p1 = (Point) i1.next();
			Iterator i2 = pg2.getPointIterator();
			while (i2.hasNext()) {
				Point p2 = (Point) i2.next();
				double sqDist = p1.distanceSQ(p2);
				if (sqDist > maxSqDist) {
					maxSqDist = sqDist;
				}
			}
		}
		return Math.sqrt(maxSqDist);
	}

	public static double distanceNearest(Polyline pg1, Polyline pg2) {
		double minSqDist = Double.MAX_VALUE;
		Iterator i1 = pg1.getPointIterator();
		while (i1.hasNext()) {
			Point p1 = (Point) i1.next();
			Iterator i2 = pg2.getPointIterator();
			while (i2.hasNext()) {
				Point p2 = (Point) i2.next();
				double sqDist = p1.distanceSQ(p2);
				if (sqDist < minSqDist) {
					minSqDist = sqDist;
				}
			}
		}
		return Math.sqrt(minSqDist);
	}

	public int getBrightness(int x, int y, Vector[] bounds, int minY) {
		int yr = y-minY;
		if (yr < 0 || yr >= bounds.length) {
			return this.color;
		}		
		
		boolean inside;
		Vector b = bounds[yr];
		int r = -1;
		do {
			r += 2;
			int xrb = ((Bound) b.get(r)).x;
			if (xrb > x) {
				break;
			}
		}
		while (r < b.size() - 1);
		inside = ((Bound) b.get(r-1)).x < x && ((Bound) b.get(r)).x > x;
		
		if (inside) {			
			return img.getBrightness(x, y);
		} else {
			return this.color;
		}
	}
	
	public void addBounds(
			Vector[] bounds, int minY,
			Polyline boundary,
			boolean isHole) {
		Iterator it = boundary.getPointIterator();
		Point first = (Point) it.next();
		Point current = first;
		Point next = null;
		boolean complete = false;
		while (!complete) {
			if (it.hasNext()) {
				next = (Point) it.next();
			} else {
				next = first;
				complete = true;
			}
			if (isHole) {
				if (next.y < current.y) {
					int nx = (int) next.x;
					int ny = (int) next.y;
					bounds[ny - minY].add(new Bound(nx, false));
				} else if (next.y > current.y) {
					int cx = (int) current.x;
					int cy = (int) current.y;
					bounds[cy - minY].add(new Bound(cx-1, true));
				}
			} else {
				if (next.y < current.y) {
					int nx = (int) next.x;
					int ny = (int) next.y;
					bounds[ny - minY].add(new Bound(nx-1, true));
				} else if (next.y > current.y) {
					int cx = (int) current.x;
					int cy = (int) current.y;
					bounds[cy - minY].add(new Bound(cx, false));
				}
			}
			current = next;
		}
		for (int i = 0; i < bounds.length; i++) {
			Collections.sort(bounds[i]);
		}
	}

	public void simplifyBoundaries(double d0) {
		Box bb = boundary.getBoundingBox();
		double dr = bb.getDiag() / d0;
		// 1.1 for the noisy pixels!
		double eps = Math.max(1.1, (-0.5*dr*dr*dr + 1.25*dr*dr + 0.25) * d0 * rEps);
		this.boundary = DouglasPeucker.simplify(this.boundary, eps);
		for (int i = 0; i < children.size(); i++) {
			RegionImage img = (RegionImage) children.get(i);
			img.simplifyBoundaries(d0);
		}
	}

	private Polyline followContour(int x0, int y0, Vector[] bounds, int minY) {
		Polyline p = new Polyline(true);		
		int fg = getBrightness(x0, y0, bounds, minY);
		int bg = PixelImage.getOther(fg);
		Segment startSeg = new Segment(x0, y0);		
		Segment seg = new Segment(startSeg);
		do {
			int crr = getBrightness(seg.rightRightX(), seg.rightRightY(), bounds, minY);
			if (crr != bg) {
				seg.turnRight();
			} else {
				int cr = getBrightness(seg.rightX(), seg.rightY(), bounds, minY);
				if (cr != bg) {
					seg.move();
				} else {
					seg.turnLeft();
				}
			}
			p.addPoint(new Point(seg.getXCoordinate(), seg.getYCoordinate()));			
		} while (!seg.equals(startSeg));
		return p;
	}
	
	private RegionImage merge(Vector component, boolean[][] connected) {
		RegionImage[] parts = new RegionImage[component.size()];
		double maxX = -Double.MAX_VALUE;
		double maxY = -Double.MAX_VALUE;
		for (int i = 0; i < parts.length; i++) {
			parts[i] = (RegionImage) component.get(i);
			Box bb = parts[i].boundary.getBoundingBox();
			maxX = Math.max(maxX, bb.getMaxX());
			maxY = Math.max(maxY, bb.getMaxY());
		}
		int w = (int) maxX+2;
		int h = (int) maxY+2;

		BufferedImage img = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
		Graphics2D g = img.createGraphics();
		g.setColor(this.color > 128 ? Color.white : Color.black);
		g.fillRect(0, 0, w, h);
		g.setColor(this.color > 128 ? Color.black : Color.white);
		
		for (int i = 0; i < parts.length; i++) {
			for (int j = i+1; j < parts.length; j++) {
				if (!connected[parts[i].tmpIdx][parts[j].tmpIdx]) continue;
				boolean found = false;
				for (int k = j; k < parts.length; k++) { // j because 2 cliques also wanted!
					if (connected[parts[i].tmpIdx][parts[k].tmpIdx] && connected[parts[j].tmpIdx][parts[k].tmpIdx]) {
						ShapeSet clique = new ShapeSet();
						clique.add(parts[i].boundary);
						clique.add(parts[j].boundary);
						clique.add(parts[k].boundary);
						Convex ch = Convex.getConvexHull(clique);
						Box bb = ch.getOutline().getBoundingBox();
						int fromX = Math.max(0, (int) bb.getMinX()-1);
						int toX   = Math.min(w, (int) bb.getMaxX()+1);
						int fromY = Math.max(0, (int) bb.getMinY()-1);
						int toY   = Math.min(h, (int) bb.getMaxY()+1);
						for (int r = fromY; r < toY; r++) {
							for (int c = fromX; c < toX; c++) {
								Point p = new Point(c + 0.5, r + 0.5);
								if (ch.isPointInside(p)) {
									g.drawLine(c,r,c,r);
								}
							}
						}
					}
				}
			}
		}
		PixelImage pix = new PixelImage(img);
		pix.bg = this.color;
		RegionImage merged = new RegionImage(pix);
		return (RegionImage) merged.children.get(0);
	}
		
	private boolean searchCliques(
			LinkedList cliques,
			LinkedList candidate,
			int last,
			RegionImage[] parts,
			boolean[][] connected) {
		boolean anyAugmentations = false;
		for (int i = last + 1; i < parts.length; i++) {
			RegionImage newI = parts[i];
			Iterator it = candidate.iterator();
			boolean add = true;
			while (it.hasNext()) {
				RegionImage oldI = (RegionImage) it.next();
				if (!connected[oldI.tmpIdx][newI.tmpIdx]) {
					add = false;
					break;
				}
			}
			if (add) {
				candidate.add(newI);
				boolean augmented = searchCliques(cliques, candidate, i, parts, connected);
				if (augmented) {
					anyAugmentations = true;
				} else {
					ShapeSet set = new ShapeSet();
					Iterator sIt = candidate.iterator();
					while (sIt.hasNext()) {
						set.add(((RegionImage) sIt.next()).boundary);
					}
					cliques.add(Convex.getConvexHull(set));
				}
				candidate.removeLast();
			}
		}
		return anyAugmentations;
	}
	
	public void writeData(Writer writer) throws IOException {
		if (!hasData) {
			return;
		}
		setIdx(0);
		writeData_01(writer);
	}	

	public void writeData_01(Writer writer) throws IOException {
		StringBuffer buff = new StringBuffer();
		String prefix = "\t";
		String pprefix = "\t\t";
		
		buff.append("<shapes xmlns=\"http://www.cs.uu.nl/profi/shapes-1\" v-maj=\"1\" v-min=\"1\">\n");
		buff.append("  <src-img width=\"");
		buff.append(     (originalWidth));
		buff.append(     "\" height=\"");
		buff.append(     (originalHeight));
		buff.append(     "\">");
		buff.append(     name);
		buff.append(     "</src-img>\n");
		buff.append("  <creation>\n");
		for (int i = 0; i < creation.length; i++) {
			buff.append("    <stage method=\"");
			buff.append(       creation[i]);
			buff.append(       "\"></stage>\n");
		}
		buff.append("  </creation>\n");
		appendShapes(buff);
		buff.append("  <shape-relations approx=\"full\">\n");
		buff.append("    <contains>");
		appendContains(buff);
		buff.append(     "</contains>\n");
		buff.append("  </shape-relations>\n");
		buff.append("  <view>");
		int n = getNumberOfShapes();
		for (int i = 0; i < n; i++) {
			buff.append(" "); buff.append(i);
		} 
		buff.append("</view>\n");
		buff.append("</shapes>\n");
		 
		writer.write(buff.toString());
	}

	public void appendContains(StringBuffer buff) {
		if (parent != null) {
			Vector allChildren = new Vector();
			allChildren.addAll(children);
			for (int i = 0; i < allChildren.size(); i++) {
				RegionImage img = (RegionImage) allChildren.get(i);
				allChildren.addAll(img.children);
				buff.append(" ");
				buff.append(this.idx);
				buff.append(" ");
				buff.append(img.idx);
			}	
		}
		for (int i = 0; i < children.size(); i++) {
			RegionImage img = (RegionImage) children.get(i);
			img.appendContains(buff);
		}
	}

	public void appendShapes(StringBuffer buff) {
		if (parent != null) {
			
			//<shape id="1" relevance="0.75">
			//   <pca cl="i"> 0 0 1.0 2 2 0 </pca>
			//  <apbb> 0 4 3 0 </apbb>
			// <properties area="12" pm="14.2" />
			//</shape>
				
			buff.append("  <shape id=\"");
			buff.append(     idx);
			buff.append(     "\">\n    ");
			buff.append(     Figure.toXml(boundary, false, true));
			buff.append(     "\n    <apbb>");
			Box bb = boundary.getBoundingBox();
			buff.append(bb.getMinX());
			buff.append(" ");
			buff.append(bb.getMinY());
			buff.append(" ");
			buff.append(bb.getMaxX());
			buff.append(" ");
			buff.append(bb.getMaxY());
			buff.append(     "</apbb>\n");
			buff.append("  </shape>\n");			
		}
		for (int i = 0; i < children.size(); i++) {
			RegionImage img = (RegionImage) children.get(i);
			img.appendShapes(buff);
		}
	}	
}
