/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.util;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;

import javax.swing.JComponent;

public class ImageCanvas extends JComponent {

	private BufferedImage original;
	private BufferedImage scaled;
	private int width;
	private int height;
	private boolean preserveEdges;
	
	public ImageCanvas(BufferedImage image, boolean preserveEdges) {			
		this.original = image;
		this.preserveEdges = preserveEdges;
		width = 0;
		height = 0;	
	}
	
	public ImageCanvas(boolean preserveEdges) {			
		this.preserveEdges = preserveEdges;
		width = 0;
		height = 0;	
	}
	
	public void setPreserveEdges(boolean preserveEdges) {
		this.preserveEdges = preserveEdges;
	}
	
	public void updateScaled() {
		width = 0;
	}
	
	public void setImage(BufferedImage image) {
		this.original = image;
		updateScaled();
		repaint();
	}
	
	public BufferedImage getImage() {
		return this.original;
	}
			
	public void paintComponent(Graphics g) {
		if (original == null) {
			return;
		}
		int imageHeight = original.getHeight(this);
		int imageWidth = original.getWidth(this);
		float scale = Math.min(this.getWidth() / (float) imageWidth, this.getHeight() / (float) imageHeight);
		int displayWidth = (int) (imageWidth * scale);
		int displayHeight = (int) (imageHeight * scale);
		
		if (displayWidth != width || displayHeight != height) {
			width = displayWidth;
			height = displayHeight;
			if (preserveEdges) {
				scaled = scale(original, width, height);
			}
		}
		
		int dx = (getWidth() - width) / 2; 
		int dy = (getHeight() - height) / 2;
		if (preserveEdges) {
			g.drawImage(
					this.scaled,
					dx, dy, dx + width, dy + height,
					0, 0, width, height,
					Color.WHITE, this);
		} else {
			g.drawImage(
					this.original,
					dx, dy, dx + width, dy + height,
					0, 0, imageWidth, imageHeight,
					Color.WHITE, this);
			
		}
	}
	
	private BufferedImage scale(BufferedImage original, int w, int h) {
		int margin = 2;
		int oldWidth = original.getWidth(this);
		int oldHeight = original.getHeight(this);
		if (oldWidth <=1 || oldHeight <=1) return original;
		if (w < 2*margin || h < 2*margin) return original;
		BufferedImage newImage = new BufferedImage(w, h, BufferedImage.TYPE_3BYTE_BGR);
		Graphics g = newImage.getGraphics();
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, w, h);
		
		ColorModel cm = original.getColorModel();

		double scale = Math.min((w-2*margin) / (double) oldWidth, (h-2*margin) / (double) oldHeight);
		
		for (int oldX = 0; oldX < oldWidth; oldX++) {
			int newX = Math.min(Math.max((int) (oldX * scale) + margin, 0), w-1);
			for (int oldY = 0; oldY < oldHeight; oldY++) {
				int newY = Math.min(Math.max((int) (oldY * scale) + margin, 0), h-1);
				int color = original.getRGB(oldX, oldY);
				if (color != -1) { //!white
					int oc = newImage.getRGB(newX, newY);
					int or = cm.getRed(oc);
					int og = cm.getGreen(oc);
					int ob = cm.getBlue(oc);
					int omin = Math.min(Math.min(or, og), ob);
					int osum = or + og + ob;
					int nr = cm.getRed(color);
					int ng = cm.getGreen(color);
					int nb = cm.getBlue(color);
					int nmin = Math.min(Math.min(nr, ng), nb);
					int nsum = nr + ng + nb;
					if (osum == 3*256 || nmin < omin || (nmin == omin && nsum > osum)) {
						newImage.setRGB(newX, newY, color);
					}
				}
			}
		}
		return newImage;
	}

}
