/*
    This file is part of a program to align vectorised images and
    compute their similarity.

    Copyright (C) 2009  Freie Universitt Berlin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact: Professor Helmut Alt at the Institute of Computer Science.
    <http://www.fu-berlin.de/>     <http://www.inf.fu-berlin.de/>
*/
package profi.util;


import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.util.Vector;

import javax.swing.JComponent;

import profi.model.ShapeSet;

public class ImageCanvas2 extends JComponent {

	private static class Layer {
		public ShapeSet shapeSet;
		public Color color;
		public boolean visible;
		public String description;
	}
	
	private static final int margin = 10;
	
	private BufferedImage buffer;
	private int width;
	private int height;

	private boolean marginVisible;
	
	private BufferedImage image;
	private Vector layers;
	
	
	public ImageCanvas2(int width, int height) {
		this.width = width;
		this.height = height;
		this.image = null;
		this.layers = new Vector();
		marginVisible = true;
		updateBuffer();
	}
	
	public ImageCanvas2(BufferedImage image) {
		this.width = image.getWidth();
		this.height = image.getHeight();
		this.image = image;
		this.layers = new Vector();
		marginVisible = true;
		updateBuffer();
	}
	
	public void updateBuffer() {
		int currentWidth = Math.max(this.getWidth(), 1);
		int currentHeight = Math.max(this.getHeight(), 1);
		buffer = new BufferedImage(currentWidth, currentHeight, BufferedImage.TYPE_3BYTE_BGR);		
		float scale = Math.min(
				(currentWidth-(2*margin)) / (float) width,
				(currentHeight-(2*margin)) / (float) height);
		Graphics g = buffer.getGraphics();
		g.setColor(Color.white);
		g.fillRect(0, 0, currentWidth, currentHeight);
		if (marginVisible) {
			g.setColor(Color.lightGray);
			g.drawRect(margin-2, margin-2, (int) (width*scale)+3, (int) (height*scale)+3);
		}
		g.translate(margin, margin);
		if (image != null) {
			g.drawImage(image, 0, 0, (int) (width*scale), (int) (height*scale), Color.white, null);
		}
		for (int i = 0; i < layers.size(); i++) {
			Layer layer = (Layer) layers.get(i);
			if (layer.visible) {
				g.setColor(layer.color);
				layer.shapeSet.draw(g, scale);
			}
		}
	}
	
	public void addLayer(ShapeSet shapeSet, Color color, String description) {
		Layer layer = new Layer();
		layer.shapeSet = shapeSet;
		layer.color = color;
		layer.visible = true;
		layer.description = description;
		
		layers.add(layer);
		updateBuffer();
		repaint();
	}
	
	public String[] getLayerDescriptions() {
		String[] descriptions = new String[layers.size()];
		for (int i = 0; i < layers.size(); i++) {
			descriptions[i] = ((Layer) layers.get(i)).description;
		}
		return descriptions;
	}
	
	public void setLayerColor(int i, Color color) {
		((Layer) layers.get(i)).color = color;
		updateBuffer();
		repaint();
	}
	
	public void deleteLayer(int i) {
		layers.remove(i);
		updateBuffer();
		repaint();
	}
	
	public void deleteLayer(ShapeSet shapeSet) {
		for (int i = 0; i < layers.size(); i++) {
			if (((Layer) layers.get(i)).shapeSet == shapeSet) {
				layers.remove(i);
				break;
			}
		}
		updateBuffer();
		repaint();
	}
	
	public void clearLayers() {
		layers.clear();
		updateBuffer();
		repaint();
	}
	
	public void paintComponent(Graphics g) {

		int oldWidth = buffer.getWidth();
		int oldHeight = buffer.getHeight();
		int currentWidth = Math.max(this.getWidth(), 1);
		int currentHeight = Math.max(this.getHeight(), 1);
		
		if (currentWidth != oldWidth || currentHeight != oldHeight) {
			updateBuffer();
		}
		
		g.drawImage(
				this.buffer,
				0, 0, currentWidth, currentHeight,
				Color.WHITE, null);			
		}	
	
	public void setImage(BufferedImage image) {
		this.width = image.getWidth();
		this.height = image.getHeight();
		this.image = image;
		updateBuffer();
		repaint();
	}
	
}
